// egr.cc: implementation of functions for reduction of points & component groups
//////////////////////////////////////////////////////////////////////////
//
// Copyright 1990-2005 John Cremona
// 
// This file is part of the mwrank package.
// 
// mwrank is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or (at your
// option) any later version.
// 
// mwrank is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with mwrank; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
// 
//////////////////////////////////////////////////////////////////////////
 
// NB This is experimental and needs more work

#include "points.h"
#include "egr.h"

//#define DEBUG_EGR

int HasGoodReduction(const Point& P, const bigint& p)
{
#ifdef DEBUG_EGR
  cout<<"Testing whether "<<P<<" has good reduction at "<<p<<endl;
#endif
  bigint Z=getZ(P);
  if(is_zero(Z)) return 1; // identity is nonsingular
  bigint X=getX(P);
  bigint Y=getY(P);
  Curvedata E = P.getcurve();
  if(is_zero(p)) // test whether P is not on the "egg"
    {
      if(getconncomp(E)==1) return 1;
      bigint b2, b4, b6, b8;
      E.getbi(b2,b4,b6,b8);
      bigint fd = 6*X*X+b2*X*Z+b4*Z*Z;
      if(sign(fd)<0) return 0;
      bigint fdd = 12*X+b2*Z;
      if(sign(fdd)<0) return 0;      // assumes Z>0
      return 1;
    }
  X=mod(X,p);  Y=mod(Y,p);  Z=mod(Z,p);
  if(is_zero(Z)) return 1; // identity is nonsingular
  bigint a1,a2,a3,a4,a6;
  E.getai(a1,a2,a3,a4,a6);
  if(ndiv(p,-3*X*X - 2*a2*X*Z + a1*Y*Z - a4*Z*Z)) return 1;
  if(ndiv(p,a1*X + 2*Y + a3*Z)) return 1;
  if(ndiv(p,-a2*X*X + a1*X*Y - 2*a4*X*Z + Y*Y + 2*a3*Y*Z - 3*a6*Z*Z)) return 1;
  return 0;
}

int HasGoodReduction(const Point& P, const vector<bigint>& plist, bigint& p0)
{
  for(unsigned int i=0; i<plist.size(); i++)
    {p0=plist[i]; if(!HasGoodReduction(P,p0)) return 0;}
  return 1;
}

//#undef DEBUG_EGR
//#define DEBUG_EGR

// Since ComponentGroups are very small there's no need for any fancy
// tricks here...
int OrderInComponentGroup(const Point& P, const bigint& p, int group_order)
{
#ifdef DEBUG_EGR
  cout<<"In OrderInComponentGroup() with P="<<P<<", group order = "<<group_order<<endl;
#endif
  Point Q=P; int j=1;
  while(j<=group_order)
    {
      if(group_order%j==0)
	{
#ifdef DEBUG_EGR
	  //	  cout<<"j="<<j<<endl;
#endif
	  if(HasGoodReduction(Q,p)) return j;
	}
      j+=1; Q+=P;
    }
  cout<<"Error in OrderInComponentGroup(): component group order is "<<group_order<<" but none of the first "<<group_order<<" multiples of "<<P<<" have good reduction!"<<endl;
  abort();
  return 0;  // error (cannot reach here)
}

int InSameComponent(const Point& P, const Point& Q, const bigint& p)
{
  if(P==Q) return 1;
  return HasGoodReduction(P-Q,p);
}

// Type Im, multiplicative reduction where component group is cyclic
// of order m, using Tate curve formula from Silverman
// The formulas here are copied from pheight() in heights.cc

long ImageInComponentGroup(const Point&P, const bigint& pr)
{
  if(HasGoodReduction(P,pr)) return 0;
  bigint a1,a2,a3,a4,a6;
  P.getcurve().getai(a1,a2,a3,a4,a6);
  bigint x=getX(P),  y=getY(P),  z=getZ(P);
  bigint zroot = gcd(x,z); // = cube root of z
  return val(pr, 2*y + a1*x + a3*z) - 3*val(pr,zroot);
}

// class Kodaira_code just holds an int which "codes" the type as follows:
// (this coding originally from R.G.E.Pinch)
//
// Im                 -> 10*m
// I*m                -> 10*m+1
// I, II, III, IV     -> 1, 2, 3, 4
// I*, II*. III*, IV* -> 5, 6, 7, 8
//

// Returns [m] for cyclic of order m, [2,2] for 2*2 (type I*m, m even only)

vector<int> ComponentGroup(const CurveRed& E, const bigint& p)
{
  vector<int> ans(1);
  if(p==0)
    {
      ans[0]=getconncomp(E); return ans;
    }
  int m, code=getKodaira_code(E,p).code;
  switch (code%10){
  case 0: // Type Im
    ans[0] = code/10;
    return ans;
  case 1: // Type Im*
    m = (code-1)/10;
    if(even(m)) {ans[0]=2; ans.push_back(2);}
    else ans[0]=4;
    return ans;
  case 2:  // Type II
    ans[0]=1;  return ans;
  case 3: // Type III
    ans[0]=2;  return ans;
  case 4: // Type IV
    ans[0]=3;  return ans;
  case 5: // Type IV*
    ans[0]=3;  return ans;
  case 6: // Type III*
    ans[0]=2;  return ans;
  case 7: // Type II*
    ans[0]=1;  return ans;
  default: // fail condition (error)
    ans[0]=0;  return ans;
  };
}

//#define DEBUG_EGR
// replaces the (independent) points with a new set which spans the
// subgroup of the original with good reduction at p, returning the
// index
int gr1prime(vector<Point>& Plist, const CurveRed& E, const bigint& p)
{
  int j,k,m,n,n0,n1,npts=Plist.size();
#ifdef DEBUG_EGR
  cout<<"in gr1prime with p="<<p<<endl;
  //  bigfloat reg0 = regulator(Plist);
  //  cout<<"regulator = "<<reg0<<endl;
  //  cout<<n<<" points"<<endl;
#endif
  if(npts==0) return 1;
  Point P0 = Plist[0], P1;
  vector<int> CG = ComponentGroup(E,p);
  long CGOrder=CG[0];
  if(CG.size()>1) CGOrder*=CG[1];
#ifdef DEBUG_EGR
  cout<<"Component group structure = "<<CG<<endl;
#endif
  if(CG.size()>1) 
    {
#ifdef DEBUG_EGR
      cout<<"Non-cyclic component group "<<endl;
#endif
      int logm=0;  // = no. of gens so far
      m = 1;       // = order so far = 2^logm
      Point Third; // will hold sum of first two gens

      for (k=0; k<npts; k++)
	{
	  Point Pk=Plist[k];
#ifdef DEBUG_EGR
	  cout<< "processing point#"<<k<<": "<<Pk<<endl;
#endif
	  j=OrderInComponentGroup(Pk,p,2);
#ifdef DEBUG_EGR
	  cout<< "image has order "<<j<<" in component group"<<endl;
#endif
	  if (2%j)
	    {
	      cout<< "Error:  order of "<<Pk
		  <<" in the component group is not 1 or 2!"<<endl;
	      abort();
	    }

	  if (j==1) continue; // good reduction, nothing to do

	  switch(logm)
	    {
	    case 0:  // no generators yet, we have the first generator
	      {
		if(k>0)	// Make this the first point in list
		  {
		    Plist[k]=Plist[0];
		    Plist[0]=Pk;
		  }
		logm+=1; m*=2;
		break;
	      } // end of case 0
	      
	    case 1: // one generator so far: do we have a second generator?
	      {
		if (InSameComponent(Pk,Plist[0],p)) // no
		  {
		    Plist[k]=Pk-Plist[0];
		  }
		else                                  // yes
		  {
		    if(k>1)	// Make this the second point in list
		      {
			Plist[k]=Plist[1];
			Plist[1]=Pk;
		      }
		    Third = Plist[0]+Plist[1];
		    logm+=1; m*=2;
		  }
		break;
	      } // end of case 1
	    
	    case 2: // we already have 2 generators
	      {
		if (InSameComponent(Pk,Plist[0],p))
		  {
		    Plist[k]=Pk-Plist[0];
		  }
		else
		  if (InSameComponent(Pk,Plist[1],p))
		    {
		      Plist[k]=Pk-Plist[1];
		    }
		  else
		    {
		      if (InSameComponent(Pk,Third,p))
			{ 
			  Plist[k]=Pk-Third;
			}
		      else
			{
			  cout<<"Problem in non-cyclic component group case!"<<endl;
			  abort();
			}
		      
		    }
	      } // end of case 2
	    } // end of switch on logm
	}  // end of loop on points
      if(logm>0) Plist[0]=2*Plist[0];
      if(logm>1) Plist[1]=2*Plist[1];
      return m;
    }  // end of cyclic switch

#ifdef DEBUG_EGR
  cout<<"Cyclic component group of order "<<CG[0]<<endl;
#endif

  if(CG[0]==1) return 1;
#ifdef DEBUG_EGR
	  cout<< "processing point #1: "<<flush;
#endif
  m=OrderInComponentGroup(P0,p,CGOrder); // will hold current index
#ifdef DEBUG_EGR
  cout<< "image has order "<<m<<" in component group"<<endl;
  if((getKodaira_code(E,p).code)%10==0) // Type Im
    {
      long n = ImageInComponentGroup(P0,p);
      long m2 = CGOrder/gcd(CGOrder,n);  // order of n mod CGOrder
      if(m==m2)
	cout<<"New method agrees"<<endl;
      else
	cout<<"New method gives n = "<<n<<", order = "<<m2<<endl;
    }

#endif
  for(j=1; j<npts; j++)
    {
      n0 = m;
      P0=Plist[0];  // NB this will change
      P1=Plist[j];
#ifdef DEBUG_EGR
      cout<< "processing point #"<<(j+1)<<": "<<flush;
#endif
      n1 = OrderInComponentGroup(P1,p,CGOrder);
      if(n1==1)
	{
#ifdef DEBUG_EGR
	  cout<< "point has good reduction, no action needed "<<endl;
#endif     
	  continue;
	}
#ifdef DEBUG_EGR
      cout<< "image has order "<<n1<<" in component group"<<endl;
#endif
      // swap points so n0>=n1 for convenience
      if(n1>n0) 
	{
#ifdef DEBUG_EGR
	  cout<<" swapping "<<P0<<" and "<<P1<<endl;
#endif
	  Point Q=P0; P0=P1; P1=Q; 
	  n=n0; n0=n1; n1=n; m=n0;
	}

      if((n0%n1)==0) // P1 is a multiple of P0 in CG...
	{
	  while((!HasGoodReduction(P1,p))) {P1=P1-P0;}
#ifdef DEBUG_EGR
	  cout<<"P1 replaced by  "<<P1<<" with good reduction"<<endl;
	  cout<<"index is now "<<m<<endl;
#endif
	  Plist[0]=P0;
	  Plist[j]=P1;
	}
      else // lcm(n0,n1)>n0: we gain something
	{
	  long a,b,g=gcd(n0,n1);
	  // Now find u (coprime to g) s.t. (n1/g)P1 == u* (n0/g)P0
	  Point Q0=(n0/g)*P0; Point Q=Q0; // (P0.getcurve());
	  Point Q1=(n1/g)*P1;
	  long u=1; //0;
#ifdef DEBUG_EGR
	  cout<<"Looking for u"<<endl;
#endif
	  while(gcd(u,CGOrder)>1 || (!InSameComponent(Q,Q1,p))) {u++; Q=Q+Q0;}
#ifdef DEBUG_EGR
	  cout<<" u="<<u<<endl;
#endif
	  g=bezout(u*n0,n1,a,b);
#ifdef DEBUG_EGR
	  cout<<" a="<<a<<", b="<<b<<", g="<<g<<endl;
#endif
	  Point newP0 = b*P0+a*P1;
	  P1=(-u*n0/g)*P0+(n1/g)*P1;
	  P0=newP0;
#ifdef DEBUG_EGR
	  cout<<"new index = lcm="<<((n0*n1)/g)<<endl;
#endif
	  m= ((n0*n1)/g);
	  Plist[0]=P0; // this now has index m
	  Plist[j]=P1; // this has good reduction
	}
    }
  // At this point all points after Plist[0] have good reduction, and
  // Plist[0] has index m
  Plist[0]=m*Plist[0];
#ifdef DEBUG_EGR
  cout<<" gr1prime returns index "<<m<<endl;
  //  bigfloat reg1 = regulator(Plist);
  //  cout<<" new regulator = "<<reg1<<endl;
  //  cout<<" ratio = "<<reg1/reg0<<endl;
#endif
  return m;
}

// replaces the (independent) points with a new set which spans the
// subgroup of the original with good reduction at all p in plist,
// returning the overall index
int grprimes(vector<Point>& Plist, const CurveRed& E, const vector<bigint>& plist)
{
#ifdef DEBUG_EGR
  cout<<"in grprimes with plist="<<plist<<endl;
  //  bigfloat reg0 = regulator(Plist);
  //  cout<<"regulator = "<<reg0<<endl;
#endif
  int n=Plist.size();
  if(n==0) return 1;
  int m=1;
  for(unsigned int j=0; j<plist.size(); j++) m*=gr1prime(Plist,E,plist[j]);
#ifdef DEBUG_EGR
  cout<<" grprimes returns index "<<m<<endl;
  //  bigfloat reg1 = regulator(Plist);
  //  cout<<" new regulator = "<<reg1<<endl;
  //  cout<<" ratio = "<<reg1/reg0<<endl;
#endif
  return m;
}

// replaces the (independent) points with a new set which spans the
// subgroup of the original with good reduction at all p,
// returning the overall index
int egr_subgroup(vector<Point>& Plist, int real_too)
{
  int n=Plist.size();
  if(n==0) return 1;
  Point P0=Plist[0];
  CurveRed E = CurveRed(P0.getcurve());
  vector<bigint> plist = getbad_primes(E);
  if(real_too && (getconncomp(E)==2)) plist.push_back(BIGINT(0));
#ifdef DEBUG_EGR
  cout<<"Using primes "<<plist<<endl;
#endif
  return grprimes(Plist,E,plist);
}

// Given a list of points P1,...,Pn and a prime p, this returns a
// vector [c1,c2,...,cn] where the image of Pi in the component group
// is ci mod m, where m is the exponent of the component group at p.
// 
// Each ci is a vector of length 1 or 2 (the latter for when the
// component group is C2xC2), not just an integer.
//
// If p=0 then m=1 or 2 (m=2 iff there are two real components and at
// least one P_i is not in the connected component)
//

vector<vector<int> >  MapPointsToComponentGroup(const vector<Point>& Plist, 
						const bigint& p)
{
  int i,j,k,kk,n=Plist.size();
  vector<vector<int> > images;
  images.resize(n);
  if (n==0) return images;
  
  Point P0=Plist[0];
  CurveRed E = CurveRed(P0.getcurve());
  vector<int> G=ComponentGroup(E,p);
  cout<<"Component group = "<<G;
  int m = G.size();
  int cyclic = (m==1);
  int orderG = G[0]; if(!cyclic) orderG*=G[1];
  for(i=0; i<n; i++)  
    {
      images[i].resize(G.size());
      for(j=0; j<m; j++) images[i][j]=0;
    }
  if (orderG==1) return images;

  if (cyclic) 
    {
      cout<< "cyclic case (order "<<orderG<<")"<<endl;

// Now G is cyclic and nontrivial

      int orderH=1;        // order of current image
      Point GPointGen(E);  // point generating current image
      Point P=Plist[0];
      cout<< "\nprocessing point #1: "<<P<<": ";
      j=OrderInComponentGroup(P,p,orderG);
      cout<< "image of "<<P<<" has order "<<j<<" in component group"<<endl;
      if (orderG%j)
	{
	  cout<< "Error:  order of "<<P<<" in the component group"
	      << "does not divide the order "<<orderG<<" of the group!"<<endl;
	  return images;
	}

      images[0][0]=(orderG / j)% orderG;
      orderH=j;
      GPointGen=P;  // point generating current image
      cout<< "After processing first point, image has size "<<orderH;
      cout<< "generated by image of "<<P<<endl;
 
      for(k=1;k<n; k++)
	{
	  Point Pk=Plist[k];
	  cout<< "processing point#"<<k<<": "<<Pk<<endl;
	  j=OrderInComponentGroup(Pk,p,orderG);
	  cout<< "image has order "<<j<<" in component group"<<endl;
	  if (orderG%j)
	    {
	      cout<< "Error:  order of "<<P
		  <<" in the component group does not divide the order "
		  <<orderG<<" of the group!"<<endl;
	      return images;
	    }
	  
	  long a,b;
	  long g=bezout(orderH,j,a,b);
	  cout<<"g,a,b = "<<g<<","<<a<<","<<b<<endl;
	  long new_orderH=(j*orderH) / g;
	  long new_index=orderG / new_orderH;
	  GPointGen=a*Pk+b*GPointGen;
	  orderH=new_orderH;

	  cout<< "After processing this point:"<<endl;
	  cout<< "Order of image is now "<<new_orderH<<endl; 
	  cout<< "Index of image is now "<<new_index<<endl; 
	  cout<< "Image is generated by image of "<<GPointGen<<endl;

	  for (kk=0;kk<k; kk++)
	    images[kk][0]*=(new_orderH / orderH); 
	  images[k][0]=(new_orderH / j)%orderG;
	  orderH=new_orderH;
	}
    }
  else
    {
      cout<< "non-cyclic case"<<endl;

      vector<Point> GPointGens;

      for (k=0; k<n; k++)
	{
	  Point Pk=Plist[k];
	  cout<< "processing point#"<<k<<": "<<Pk<<endl;
	  j=OrderInComponentGroup(Pk,p,2);
	  cout<< "image has order "<<j<<" in component group"<<endl;
	  if (2%j)
	    {
	      cout<< "Error:  order of "<<Pk
		  <<" in the component group is not 1 or 2!"<<endl;
	      return images;
	    }

	  if (j==1) continue; // good reduction, nothing to do

	  switch(GPointGens.size())
	    {
	    case 0:  // no generators yet, we have the first generator
	      {
		GPointGens.push_back(Pk);
		images[k][0]=1;
		images[k][1]=0;
		break;
	      } // end of case 0
	      
	    case 1: // one generator so far: do we have a second generator?
	      {
		if (InSameComponent(Pk,GPointGens[0],p)) // no
		  {
		    images[k][0]=1;
		    images[k][1]=0;
		  }
		else                                  // yes
		  {
		    GPointGens.push_back(Pk);
		    images[k][0]=0;
		    images[k][1]=1;
		  }
		break;
	      } // end of case 1
	    
	    case 2: // we already have 2 generators
	      {
		if (InSameComponent(Pk,GPointGens[0],p))
		  {
		    images[k][0]=1;
		    images[k][1]=0;
		  }
		else
		  if (InSameComponent(Pk,GPointGens[1],p))
		    {
		      images[k][0]=0;
		      images[k][1]=1;
		    }
		  else
		    if (InSameComponent(Pk,GPointGens[0]+GPointGens[1],p))
		      { 
			images[k][0]=1;
			images[k][1]=1;
		      }
	      } // end of case 2
	    }   // end of switch 
	}     // loop on points
    } // else ... noncyclic case
  return images; 
}

