"""
Free algebras 

AUTHOR: David Kohel, 2005-09
"""

#*****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful, 
#    but WITHOUT ANY WARRANTY; without even the implied warranty 
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# 
#  See the GNU General Public License for more details; the full text 
#  is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.rings.ring import Ring
from sage.ext._integer import Integer
from sage.monoids.free_monoid import FreeMonoid
from sage.monoids.free_monoid_element import FreeMonoidElement
from sage.algebras.algebra import Algebra
from sage.algebras.free_algebra_element import FreeAlgebraElement

class FreeAlgebra(Algebra):
    def __init__(self, R, n, names = None):
        """
        Returns the free algebra on $n$ generators.
        
        EXAMPLES:
            sage: F = FreeAlgebra(QQ,ZZ(3),names=("x","y","z"))
            sage: mul([ F.gen(i) for i in range(3) ])
            x*y*z
            sage: mul([ F.gen(i%3) for i in range(12) ])
            x*y*z*x*y*z*x*y*z*x*y*z
            sage: (x,y,z) = F.gens()
            sage: (2 + x*z + x**2)**2 + (x - y)**2
            4 + 3*x^2 - x*y + 2*x*z - y*x + y^2 + x^4 + x^3*z + x*z*x^2 + x*z*x*z
        """
        if not isinstance(R, Ring):
            raise TypeError, "Argument R must be a ring."
        self.__base_ring = R
        self.__monoid = FreeMonoid(n, names=names)
        self.__ngens = n
        self.assign_names(names)

    def __repr__(self):
        return "Free algebra on %s generators %s over %s"%(
            self.__ngens, self.gens(), self.__base_ring)

    def __call__(self, x, canonical=False):
        if isinstance(x, FreeAlgebraElement) and x.parent() == self: 
            return x
        F = self.__monoid
        R = self.__base_ring 
        if isinstance(x, (int, long, Integer)):
            if x == 0:
                return FreeAlgebraElement(self,{})
            else:
                return FreeAlgebraElement(self,{F(1):R(x)})
        elif isinstance(x, Ring) and x.parent() is R:
            if x == 0:
                return FreeAlgebraElement(self,{})
            else:
                return FreeAlgebraElement(self,{F(1):x})
        elif isinstance(x, FreeMonoidElement) and x.parent() is F:
            return FreeAlgebraElement(self,{x:R(1)})
        else:
            raise TypeError, "Argument x (= %s) if of invalid type."%x
                                
    def __contains__(self, x):
        return isinstance(x, FreeAlgebraElement) and x.parent() == self
            
    def gen(self,i):
        """
        The i-th generator of the algebra.
        """
        n = self.__ngens
        if i < 0 or not i < n: 
            raise IndexError, "Argument i (= %s) must be between 0 and %s."%(i, n-1)
        R = self.__base_ring
        F = self.__monoid
        return FreeAlgebraElement(self,{F.gen(i):R(1)})

    def ngens(self):
        """
        The number of generators of the algebra.
        """
        return self.__ngens

    def assign_names(self,names):
        """
        Assign the printing names for the generators; this will have the unfortunate
        effect of overwriting the names for the covering algebra; this also does not
        overwrite the return value of names() for the Algebra.
        """
        #self._Algebra.assign_names(names) # where is this located?
        self.monoid().assign_names(names)


    def base_ring(self):
        return self.__base_ring

    def monoid(self):
        """
        The free monoid of generators of the algebra.
        """
        return self.__monoid
    
# doctests
if __name__ ==  '__main__':
    import doctest, sys
    from sage.all import QQ, ZZ, mul
    doctest.testmod(sys.modules[__name__])
    
