"""
Categories

AUTHORS:  David Kohel and William Stein

Every SAGE object lies in a category.  Categories in SAGE are modeled on
the mathematical idea of category, and are distinct from Python
classes, which are a programming construct.

In most cases, typing \code{x.category()} returns the category to
which $x$ belongs.  If $C$ is a category and $x$ is any object, $C(x)$
tries to make an object in $C$ from $x$.

EXAMPLES:
    We create a couple of categories.

    sage: Sets()
    Category of sets
    sage: GSets()
    Category of G-sets
    sage: Semigroups()
    Category of semigroups
    sage: VectorSpaces(FiniteField(11))
    Category of vector spaces over Finite field of size 11
    sage: Ideals(IntegerRing())
    Category of ring ideals in Integer Ring

The default category for elements $x$ of an objects $O$ is the
category of all objects of $O$.  For example,

    sage: V = VectorSpace(RationalField(), 3)
    sage: x = V.gen(1)
    sage: x.category()
    Category of elements of Vector space of dimension 3 over Rational Field

"""

#*****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu> and
#                     William Stein <wstein@math.ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful, 
#    but WITHOUT ANY WARRANTY; without even the implied warranty 
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# 
#  See the GNU General Public License for more details; the full text 
#  is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

__doc_exclude = ['ref', 'k', 'uniq', 'uniq1', 'latex', \
                 'Category', 'Category_idea', 'Category_in_ambient', \
                 'Category_module', 'Category_over_base', \
                 'Category_over_base_ring']

from weakref import ref

import sage.rings.all
import sage.algebras.all
from sage.misc.latex import latex


class Category:
    """
    The base class for all 
    """
    def __init__(self, s=None):
        if s is None:  # figure out from the type name!
            t = str(type(self))
            t = t[t.rfind('.')+1:]
            s = t[:t.rfind("'")]
            self.__label = s
            i = -1
            while i < len(s)-1:
                for i in range(len(s)):
                    if s[i].isupper():
                        s = s[:i] + " " + s[i].lower() + s[i+1:]
                        break
            s = s.lstrip()
        elif isinstance(s, str):
            self.__label = s
        else:
            raise TypeError, "Argument string must be a string."
        self.__category = s

    def __call__(self, x):
        if x in self:
            return x
        raise NotImplementedError

    def __repr__(self):
        return "Category of %s"%self.__category

    def _latex_(self):
        return "\\mbox{\\bf %s}"%self.__label

    def __hash__(self):
        return hash(self.__category)

    def short_name(self):
        return ''.join([x.capitalize() for x in self.__category.split()])

    def __contains__(self, x):
        try:
            c = x.category()
        except AttributeError:
            return False
        return c.is_subcategory(self)

    def is_abelian(self):
        return False

    def is_subcategory(self, c):
        """
        Returns True if self is naturally embedded as a subcategory of c.
        
        EXAMPLES:
            sage: Rings  = Rings()
            sage: AbGrps = AbelianGroups()
            sage: Rings.is_subcategory(AbGrps)
            True
            sage: AbGrps.is_subcategory(Rings)
            False

        The \code{is_subcategory} function takes into account the base.
            sage: M3 = VectorSpaces(FiniteField(3))
            sage: M9 = VectorSpaces(FiniteField(9))
            sage: M3.is_subcategory(M9)
            False
            
        """
        if not isinstance(c, Category):
            raise TypeError, "Argument c (= %s, type = %s) must be a category"%(c, type(c))
        if self == c: return True
        if category_hierarchy.has_key(self.__class__):
            S = category_hierarchy[self.__class__]
            if not c.__class__ in S:
                return False
            return c._parameters().issubset(self._parameters())
        return False

    def _is_subclass(self, c,):
        if isinstance(c, Category):
            return self.is_subcategory(c)
        if self.__class__ == c:
            return True
        return c in category_hierarchy[self.__class__]

    def __eq__(self, c):
        if self is c:
            return True
        return False

    def _parameters(self):
        return set([])

    def category(self):
        return Objects()

def is_Category(x):
    """
    Returns True if x is a category.
    """
    return isinstance(x, Category)


#############################################################
# ...
#############################################################
cache = {}
class uniq(object):
    def __new__(cls):
        if cache.has_key(cls):
            return cache[cls]
        O = object.__new__(cls)
        cache[cls] = O
        return O


cache1 = {}
class uniq1(object):
    def __new__(cls, arg1):
        if cache.has_key((cls, arg1)):
            X = cache[(cls, arg1)]()
            if X: return X
        O = object.__new__(cls)
        cache[(cls,arg1)] = ref(O)
        return O


####################################################################
#   Different types of categories
####################################################################
    
#############################################################
# Category of objects
#############################################################
class Category_uniq(Category, uniq):
    pass

#############################################################
# Category of elements of some object
#############################################################
class Elements(Category, uniq1):
    """
    The category of all elements of a given object.

    EXAMPLES:
        sage: a = IntegerRing()(5)
        sage: C = a.category(); C
        Category of elements of Integer Ring
        sage: a in C
        True
        sage: 2/3 in C
        False
    """
    def __init__(self, object):
        Category.__init__(self)
        self.__object = object

    def object(self):
        return self.__object
    
    def __repr__(self):
        return "Category of elements of %s"%self.object()

    def _latex_(self):
        return "\\mbox{\\bf{}Elt}_{%s}"%latex(self.__object)


#############################################################
# Category of objects over some base object
#############################################################
class Category_over_base(Category, uniq1):
    def __init__(self, base, name=None):
        Category.__init__(self, name)
        self.__base = base

    def base(self):
        """
        Return the base over which elements of this category are
        defined.
        """
        return self.__base

    def __repr__(self):
        return Category.__repr__(self) + " over %s"%self.__base

    def _latex_(self):
        return "\\mbox{\\bf %s}_{%s}"%(self.__label, latex(self.__base))
    

#############################################################
# Category of objects over some base ring
#############################################################
class Category_over_base_ring(Category_over_base):
    def __init__(self, base, name=None):
        Category_over_base.__init__(self, base, name)
        if not sage.rings.all.is_Ring(base):
            raise TypeError, "base must be a ring"
        self.__base = base

    def base_ring(self):
        """
        Return the base ring over which elements of this category are
        defined.
        """
        return self.base()

    def _parameters(self):
        return set([self.__base])


class AbelianCategory:
    def is_abelian(self):
        return True


#############################################################
# Category of objects in some ambient object
#############################################################
class Category_in_ambient(Category, uniq1):
    def __init__(self, ambient, name=None):
        Category.__init__(self, name)
        self.__ambient = ambient

    def ambient(self):
        """
        Return the ambient object in which objects of this category are
        embedded.
        """
        return self.__ambient

    def __repr__(self):
        return Category.__repr__(self) + " in %s"%self.__ambient

    def _parameters(self):
        return set([self.__ambient])


class Category_module(AbelianCategory, Category_over_base_ring):
    def __init__(self,  base, name=None):
        Category_over_base_ring.__init__(self, base, name)


class Category_ideal(Category_in_ambient):
    def __init__(self, ring, name=None):
        Category_in_ambient.__init__(self, ring, name)
        
    def ring(self):
        return self.ambient()

    def __contains__(self, x):
        """
        EXAMPLES:
            sage: C = Ideals(IntegerRing())
            sage: IntegerRing().zero_ideal() in C
            True
        """
        if sage.rings.all.is_Ideal(x) and x.ring() == self.ring():
            return True
        return False

    def __call__(self, v):
        if v in self:
            return v
        return self.ring().ideal(v)




####################################################################
# Actual categories (below)
####################################################################
        

#############################################################
# Generic category (default when requesting category of
# an object using misc.functional.category
#############################################################
class Objects(Category_uniq):
    """
    The category of all SAGE objects.

    EXAMPLES:
        sage: Objects()
        Category of objects
    """
    def __contains__(self, x):
        return True




#############################################################
# Sets
#############################################################

class Sets(Category_uniq):
    """
    The category of sets.

    EXAMPLES:
        sage: Sets()
        Category of sets
    """
    def __call__(self, X):
        import sage.sets.all
        return sage.sets.all.Set(X)



#############################################################
# GSets
#     $G$-Sets play an important rule in permutation groups.
#############################################################
class GSets(Category_uniq):
    """
    The category of $G$-sets, for a group $G$.

    EXAMPLES:
        sage: GSets()
        Category of G-sets
    """
    def __init__(self):
        Category.__init__(self, "G-sets")
        


#############################################################
# Semigroup
#############################################################

class Semigroups(Category_uniq):
    """
    The category of all semigroups.

    EXAMPLES:
        sage: Semigroups()
        Category of semigroups
    """



#############################################################
# Monoid
#############################################################
class Monoids(Category_uniq):
    """
    The category of monoids.

    EXAMPLES:
        sage: Monoids()
        Category of monoids
    """
    

#############################################################
# Group
#############################################################
class Groups(Category_uniq):
    """
    The category of groups.

    EXAMPLES:
        sage: Groups()
        Category of groups
    """
    


#############################################################
# AbelianSemigroup
#############################################################
class AbelianSemigroups(Category_uniq):
    """
    The category of all abelian semigroups.

    
    EXAMPLES:
        sage: AbelianSemigroups()
        Category of abelian semigroups
    """


#############################################################
# AbelianMonoid
#############################################################
class AbelianMonoids(Category_uniq):
    """
    The category of all monoids.

    EXAMPLES:
        sage: AbelianMonoids()
        Category of abelian monoids
    """
        


#############################################################
# AbelianGroup
#############################################################
class AbelianGroups(AbelianCategory, Category_uniq):
    """
    The category of all abelian groups.

    EXAMPLES:
        sage: AbelianGroups()
        Category of abelian groups
    """
        

#############################################################
# Ring
#############################################################
class Rings(Category_uniq):
    """
    The category of all rings.

    EXAMPLES:
        sage: Rings()
        Category of rings
    """



#############################################################
# CommutativeRing
#############################################################
class CommutativeRings(Category_uniq):
    """
    The category of commutative rings.
    
    EXAMPLES:
        sage: CommutativeRings()
        Category of commutative rings
    """
        


#############################################################
# Field
#############################################################
class Fields(Category_uniq):
    """
    The category of fields.

    EXAMPLES:
        sage: K = Fields()
        sage: K
        Category of fields
        sage: K(IntegerRing())
        Rational Field
        sage: K(PolynomialRing(GF(3)))
        Fraction field of Univariate Polynomial Ring in x over Finite field of size 3
        sage: K(RealField())
        Real Field with 53 bits of precision
    """
    def __contains__(self, x):
        return sage.rings.all.is_Field(x)

    def __call__(self, x):
        if x in self:
            return x
        try:
            return x.fraction_field()
        except AttributeError:
            raise TypeError, "unable to associate a field to %s"%x
            


#############################################################
# FiniteField
#############################################################
class FiniteFields(Category_uniq):
    """
    The category of all finite fields.

    EXAMPLES:
        sage: K = FiniteFields()
        sage: K
        Category of finite fields
        sage: FiniteField(17) in K
        True
        sage: RationalField() in K
        False
        sage: K(RationalField())
        Traceback (most recent call last):
        ...
        TypeError: unable to canonically associate a finite field to Rational Field
    """
    def __contains__(self, x):
        return sage.rings.all.is_Field(x) and x.is_finite()

    def __call__(self, x):
        if x in self:
            return x
        raise TypeError, "unable to canonically associate a finite field to %s"%x
        # TODO: local dvr ring?



#############################################################
# NumberField
#############################################################
class NumberFields(Category_uniq):
    r"""
    The category of number fields.
    
    EXAMPLES:
    We create the category of number fields.
        sage: C = NumberFields()
        sage: C
        Category of number fields

    Notice that the rational numbers $\Q$ are not considered as
    an object in this category.
        sage: RationalField() in C
        False

    However, we can define a degree 1 extension of $\Q$, which is in
    this category.
        sage: x = PolynomialRing(RationalField()).gen()
        sage: K = NumberField(x - 1); K
        Number Field in x with defining polynomial x - 1
        sage: K in C
        True

    We can also coerce $\Q$ into the category of number fields.
        sage: C(RationalField())
        Number Field in x with defining polynomial x - 1

    Number fields all lie in this category, irregardless of the name
    of the variable.
        sage: K = NumberField(x**2 - 1, name='a')
        sage: K in C
        True
    """
    def __contains__(self, x):
        return sage.rings.all.is_NumberField(x)

    def __call__(self, x):
        if x in self:
            return x
        try:
            return x.number_field()
        except AttributeError:
            raise TypeError, "unable to canonically associate a number field to %s"%x
        
        

#############################################################
# Algebra
#############################################################
class Algebras(Category_over_base_ring):
    """
    The category of algebras over a fixed base ring.

    
    """
    def __init__(self, R):
        Category_over_base_ring.__init__(self, None, base=R)

    def __contains__(self, x):
        return sage.algebras.all.is_Algebra(x) and x.base_field() == self.base_field()

    def base_field(self):
        """
        Return the base field over which the algebras of this category
        are all defined.
        """
        return self.base_ring()


#############################################################
# CommutativeAlgebra
#############################################################
class CommutativeAlgebras(Category_over_base_ring):
    """
    The category of commutative algebras over a given base ring.
    EXAMPLES:
        sage: M = CommutativeAlgebras(GF(19))
        sage: M
        Category of commutative algebras over Finite field of size 19
    """
    

#############################################################
# MonoidAlgebra
#############################################################
class MonoidAlgebras(Category_over_base_ring):
    """
    The category of all monoid algebras over a given base ring.
    EXAMPLES:
        sage: MonoidAlgebras(GF(2))
        Category of monoid algebras over Finite field of size 2
    """


#############################################################
# GroupAlgebra
#############################################################
class GroupAlgebras(Category_over_base_ring):
    """
    EXAMPLES:
        sage: GroupAlgebras(IntegerRing())
        Category of group algebras over Integer Ring
    """
        


#############################################################
# MatrixAlgebra
#############################################################
class MatrixAlgebras(Category_over_base_ring):
    """
    The category of matrix algebras over a field.
    
    EXAMPLES:
        sage: MatrixAlgebras(RationalField())
        Category of matrix algebras over Rational Field
    """



#############################################################
# RingModule
#############################################################
class RingModules(Category_module):
    """
    The category of all modules over a base ring.

    EXAMPLES:
        sage: Modules(RationalField())
        Category of ring modules over Rational Field

        sage: Modules(Integers(9))
        Category of ring modules over Ring of integers modulo 9
    """

Modules = RingModules



#############################################################
# FreeModule
#############################################################
class FreeModules(Category_module):
    """
    The category of free modules over a base ring.
    
    EXAMPLES:
        sage: FreeModules(IntegerRing())
        Category of free modules over Integer Ring
    """
    def __init__(self, R):
        if not sage.rings.all.is_Ring(R):
            raise TypeError, "R (=%s) must be a ring"%R
        Category_module.__init__(self, R)
        
    def __call__(self, x):
        try:
            M = x.free_module()
            if M.base_ring() != self.base_ring():
                M = M.change_ring(self.base_ring())
        except TypeError, AttributeError:
            raise TypeError, "unable to coerce x (=%s) into %s"%(x,self)
        return M

    def is_abelian(self):
        return self.base_ring().is_field()


#############################################################
# VectorSpace
#############################################################
class VectorSpaces(Category_module):
    """
    The category of vector spaces over a specified field,
    with an emedding in an ambient vector space.
    
    EXAMPLES:
        sage: VectorSpaces(RationalField())
        Category of vector spaces over Rational Field
    """
    def __init__(self, K):
        if not sage.rings.all.is_Field(K):
            raise TypeError, "K (=%s) must be a field"%K
        Category_module.__init__(self, K)
        
    def __call__(self, x):
        try:
            V = x.vector_space()
            if V.base_field() != self.base_field():
                V = V.change_ring(self.base_field())
        except TypeError, AttributeError:
            raise TypeError, "unable to coerce x (=%s) into %s"%(x,self)
        return V

    def base_field(self):
        return self.base_ring()

    
#############################################################
# HeckeModules
#############################################################
class HeckeModules(Category_module):
    r"""
    The category of Hecke modules.

    A Hecke module is a module $M$ over the \emph{anemic} Hecke
    algebra, i.e., the Hecke algebra generated by Hecke operators
    $T_n$ with $n$ coprime to the level of $M$.  (Every Hecke module
    defines a level function, which is a positive integer.)  The
    reason we require that $M$ only be a module over the anemic Hecke
    algebra is that many natural maps, e.g., degeneracy maps,
    Atkin-Lehner operators, etc., are $\T$-module homomorphisms; but
    they are homomorphisms over the anemic Hecke algebra.
    
    EXAMPLES:
    We create the category of Hecke modules over $\Q$.
        sage: C = HeckeModules(RationalField()); C
        Category of Hecke modules over Rational Field
        
    Note that the base ring can be an arbitrary commutative ring.
        sage: HeckeModules(IntegerRing())
        Category of Hecke modules over Integer Ring
        sage: HeckeModules(FiniteField(5))
        Category of Hecke modules over Finite field of size 5

    The base ring doesn't have to be a principal ideal domain.
        sage: HeckeModules(PolynomialRing(IntegerRing()))
        Category of Hecke modules over Univariate Polynomial Ring in x over Integer Ring
    """
    def __init__(self, R):
        if not (sage.rings.all.is_Ring(R) and R.is_commutative()):
            raise TypeError, "R (=%s) must be a commutative ring"%R
        Category_module.__init__(self, R, "Hecke modules")
        


#############################################################
# RingIdeal
#############################################################
class RingIdeals(Category_ideal):
    """
    The category of all ideals in a fixed ring.
    
    EXAMPLES:
        sage: Ideals(Integers(200))
        Category of ring ideals in Ring of integers modulo 200
        sage: C = Ideals(IntegerRing()); C
        Category of ring ideals in Integer Ring
        sage: I = C([8,12,18])
        sage: I
        Principal ideal of Integer Ring generated by 2
    """
    def __init__(self, R):
        if not sage.rings.all.is_Ring(R):
            raise TypeError, "R (=%s) must be a ring"%R
        Category_ideal.__init__(self, R)

    
Ideals = RingIdeals
        

#############################################################
# CommutativeRingIdeal
#############################################################
class CommutativeRingIdeals(Category_ideal):
    """
    The category of ideals in a fixed commutative ring.
    
    EXAMPLES:

        sage: C = CommutativeRingIdeals(IntegerRing())
        sage: C
        Category of commutative ring ideals in Integer Ring
    """
    def __init__(self, R):
        if not sage.rings.all.is_CommutativeRing(R):
            raise TypeError, "R (=%s) must be a commutative ring"%R
        Category_ideal.__init__(self, R)
        




#############################################################
# AlgebraModule
#############################################################
class AlgebraModules(Category_module):
    """
    The category of modules over a fixed algebra $A$.
    """
    def __init__(self, A):
        if not sage.algebras.all.is_Algebra(A):
            raise TypeError, "A (=%s) must be an algebra"%A
        Category_module.__init__(self, A)

    def algebra(self):
        return self.base_ring()
    

#############################################################
# AlgebraIdeal
#############################################################
class AlgebraIdeals(Category_ideal):
    """
    The category of ideals in a fixed algebra $A$.
    """
    def __init__(self, A):
        if not sage.algebras.all.is_Algebra(A):
            raise TypeError, "A (=%s) must be an algebra"%A
        Category_ideal.__init__(self, A)

    def algebra(self):
        return self.ambient()

    

#############################################################
# CommutativeAlgebraIdeal
#############################################################
class CommutativeAlgebraIdeals(Category_ideal):
    """
    The category of ideals in a fixed commutative algebra $A$.
    """
    def __init__(self, A):
        if not sage.algebras.all.is_CommutativeAlgebra(A):
            raise TypeError, "A (=%s) must be an algebra"%A
        Category_in_ambient.__init__(self, A)

    def algebra(self):
        return self.ambient()
    



###################################################################
#
# Natural inclusions between categories.
#
###################################################################

category_hierarchy = {\
    Sets                   : [],\
    GSets                  : [Sets],\
    Semigroups             : [Sets],\
    Monoids                : [Semigroups, Sets],\
    Groups                 : [Monoids, Semigroups, Sets],\
    AbelianSemigroups      : [Semigroups, Sets],\
    AbelianMonoids         : [Monoids, Sets],\
    AbelianGroups          : [Groups, AbelianMonoids, AbelianSemigroups, Sets],\
    Rings                  : [AbelianGroups, Sets],\
    CommutativeRings       : [Rings, Sets],\
    Fields                 : [CommutativeRings, Rings, Sets],\
    FiniteFields           : [Fields, CommutativeRings, Rings, Sets],\
    NumberFields           : [CommutativeRings, Rings, Sets],\
    Algebras               : [Rings, RingModules, Sets],\
    CommutativeAlgebras    : [Algebras, CommutativeRings, Rings, Sets],\
    MonoidAlgebras         : [Algebras, Sets],\
    GroupAlgebras          : [MonoidAlgebras, Algebras, Sets],\
    MatrixAlgebras         : [Algebras, Sets],\
    RingModules            : [AbelianGroups, Sets],\
    FreeModules            : [RingModules, AbelianGroups, Sets],\
    VectorSpaces           : [FreeModules, RingModules, AbelianGroups, Sets],\
    HeckeModules           : [FreeModules, RingModules, AbelianGroups, Sets],\
    AlgebraModules         : [RingModules, AbelianGroups, Sets],\
    RingIdeals             : [RingModules, AbelianGroups, Sets],\
    CommutativeRingIdeals  : [RingIdeals, AbelianGroups, Sets],\
    AlgebraIdeals          : [AlgebraModules, AbelianGroups, Sets],\
    CommutativeAlgebraIdeals:[CommutativeRingIdeals, RingIdeals, AbelianGroups, Sets], \
    Elements               : [Sets]}

for k in category_hierarchy:
    category_hierarchy[k].append(Objects)

if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.all import *
    from sage.modular.all import *
    from sage.modules.all import VectorSpace
    doctest.testmod(sys.modules[__name__])
