"""
Hom spaces between objects of a category

AUTHORS: David Kohel and William Stein
"""

#*****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu>, William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful, 
#    but WITHOUT ANY WARRANTY; without even the implied warranty 
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# 
#  See the GNU General Public License for more details; the full text 
#  is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import weakref
import categories

_cache = {}
def Hom(X, Y, cat=None):
    """
    Create the space of homomorphisms from X to Y in the category cat.

    INPUT:
        X -- anything
        Y -- anything
        cat -- (optional) category in which to coerce X and Y

    OUTPUT:
        a homspace in cat
    """
    global _cache
    key = (X,Y,cat)
    if _cache.has_key(key):
        H = _cache[key]()
        if H: return H

    cat_X = X.category()
    cat_Y = Y.category()
    if cat is None:
        if cat_X.is_subcategory(cat_Y):
            cat = cat_X
        if cat_Y.is_subcategory(cat_X):
            if not (cat is None) and not (cat_X is cat_Y):
                raise ValueError, "No unambiguous category found for Hom from %s to %s."%(X,Y)
            cat = cat_Y
        else:
            raise TypeError, "No suitable category found for Hom from %s to %s."%(X,Y)
    elif isinstance(cat, categories.Category):
        if not isinstance(cat, categories.Category):
            raise TypeError, "Argument cat (= %s) must be a category."%cat
        if not cat_X.is_subcategory(cat) \
               or not cat_Y.is_subcategory(cat):
            raise TypeError, \
                  "Argument cat (= %s) is incompatible with %s and %s."%(cat, X, Y)
    else:
        raise TypeError, "Argument cat (= %s) must be a category."%cat
    

    X = cat(X)
    Y = cat(Y)
    
    # construct H
    if cat._is_subclass(categories.HeckeModules):

        from sage.modular.hecke.homspaces import HeckeModuleHomspace
        H = HeckeModuleHomspace(X, Y)
        
    elif cat._is_subclass(categories.FreeModules):

        from sage.modules.free_module_homspace import FreeModuleHomspace
        H = FreeModuleHomspace(X, Y, cat)

    else:  # default
    
        H = Homspace(X, Y, cat)
    ##_cache[key] = weakref.ref(H)
    _cache[(X, Y, cat)] = weakref.ref(H)
    
    return H

def End(X, cat=None):
    """
    Create the space of endomorphisms of X in the category cat.

    INPUT:
        X -- anything
        cat -- (optional) category in which to coerce X 

    OUTPUT:
        a space of endomorphisms in cat
    """
    return Hom(X,X, cat)

class Homspace:
    """
    The class for collections of morphisms in a category.
    """
    def __init__(self, X, Y, cat):
        if not isinstance(cat, categories.Category):
            raise TypeError, "cat (=%s) must be a category"%cat
        self.__domain = X
        self.__codomain = Y
        self.__category = cat

    def __repr__(self):
        return "All morphisms from %s to %s in %s"%(
            self.__domain, self.__codomain, self.__category)

    def __call__(self, x, y=None):
        """
        Construct a morphism in this homspace from x if possible.
        """
        if x in self:
            return x
        raise TypeError, "Unable to coerce x (=%s) to a morphism in %s"%(x,self)

    def __cmp__(self, other):
        if self.__domain == other.__domain and self.__codomain == other.__codomain \
               and self.__category == other.__category:
            return 0
        return cmp(self.__domain, other.__domain)

    def __contains__(self, x):
        try:
            return x.parent() == self
        except AttributeError:
            pass
        return False

    def category(self):
        return self.__category

    def domain(self):
        return self.__domain

    def codomain(self):
        return self.__codomain

    def is_endomorphism_space(self):
        """
        Return True if the domain and codomain of self.
        """
        return self.__domain == self.__codomain

    def reversed(self):
        """
        Return the corresponding homspace, but with the domain and codomain
        reversed.
        """
        return Homspace(self.__codomain, self.__domain, self.__category)


def is_Homspace(x):
    """
    Return True if x is a space of homomorphisms in a category.
    """
    return isinstance(x, Homspace)

def is_Endspace(x):
    """
    Return True if x is a space of endomorphisms in a category.
    """
    return isinstance(x, Homspace) and x.is_endomorphism_space()
