"""
Classes that defines morphisms
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import homspaces

import sage.rings.coerce
import sage.rings.arith as arith

import sage.modules.module_element


class Morphism(sage.modules.module_element.ModuleElement):
    def __init__(self, parent, f, f_inverse=None):
        """
        The first argument f is a function or class with a call method
        (of one argument) that takes objects in domain and returns
        objects in codomain.  The Map object, when called with
        argument x, tries to coerce it into the domain, if possible,
        and raises an error if it is not.
        """
        if not isinstance(parent, homspaces.Homspace):
            raise TypeError, "parent (=%s) must be a Homspace"%parent
        self.__f = f
        self.__f_inverse = f_inverse
        self.__parent = parent

    def __cmp__(self, other):
        if not isinstance(other, Morphism) or self.parent() != other.parent():
            return sage.rings.coerce.cmp(self, other)
        if self is other:
            return 0
        return -1
        
    def __repr__(self):
        return "Morphism: %s --> %s"%(self.domain(),self.codomain())
        
    def domain(self):
        return self.__parent.domain()
        
    def codomain(self):
        return self.__parent.codomain()

    def parent(self):
        return self.__parent

    def category(self):
        return self.__parent.category()

    def is_endomorphism(self):
        return self.__parent.is_endomorphism_space()

    def __invert__(self):  # notation in python is (~f) for the inverse of f.
        assert self.__f_inverse != None, "Inverse of %s not known."%self
        return parent.reversed()(self.__f_inverse, self.__f)
                        
    def __call__(self, x):
        try:
            y = self.__domain(x)
        except TypeError:
            raise TypeError, "%s must be coercible into %s"%(x,self.domain())
        return self.__f(y)

    def _add_function(self, other): # used by __add__ and overloaded in derived classes
        raise ValueError, "addition of morphisms in %s is not supported"%self.category

    def __add__(self, other):
        if not isinstance(other,Morphism) or (other.__parent != self.__parent):
            return sage.rings.coerce.bin_op(self, other, operator.add)
        return self.__parent(self._add_function(other))        

    def _sub_function(self, other): # used by __sub__ and overloaded in derived classes
        raise ValueError, "addition of morphisms in %s is not supported"%self.category        

    def __sub__(self, other):
        if not isinstance(other,Morphism) or (other.__parent != self.__parent):
            return sage.rings.coerce.bin_op(self, other, operator.sub)
        return self.__parent(self._sub_function(other))

    def _mul_function(self, other): # used by __mul__ and overloaded in derived classes
        def f(x):
            return self(other(x))
        return f

    def __mul__(self, other):
        r"""
        The multiplication * operator is operator composition.  Thus
        if $s$ and $t$ are Hecke operators, then $(s*t)(x) = s(t(x))$.

        Note that if the morphisms are defined using matrices, then
        this is the reverse of matrix multiplication, since matrices
        act from the right.

        INPUT:
            self -- Morphism
            other -- Morphism

        OUTPUT:
            The morphism $x \mapsto self(other(x))$.
        """
        if not isinstance(other, Morphism):
            raise TypeError, "other must be a morphism"
        if other.codomain() != self.domain():
            raise ValueError, "self domain must equal other codomain"
        H = homspaces.Hom(other.domain(), self.codomain(), self.__parent.category())
        f = self._mul_function(other)
        return H(f)

    def __pow__(self, n):
        if not self.is_endomorphism():
            raise TypeError, "self must be an endomorphism."
        return arith.generic_power(self, n)

        

def is_Morphism(x):
    return isinstance(x, Morphism)

