"""
Abstract base class for ring elements.
"""

##################################################################
# Generic ring element, so all this functionality must be defined
# by any ring element.  Derived class must call __init__
##################################################################

import operator

import sage.structure.element_py

import sage.ext.coerce
from sage.ext.coerce import gcd, lcm, xgcd

# This classes uses element.pxd.  To add data members, you
# must change that file. 

cdef class Element:
    """
    Generic element base class, so all this functionality must
    be defined by any ring element.
    """

    def __init__(self, parent):
        self.__parent = parent

    def __reduce__(self):
        return sage.rings.element_py.init, (self.parent(),)

    def __repr__(self):
        return "Generic element of a ring"

    def __hash__(self):
        return hash(str(self))

    def base_ring(self):
        return self.parent().base_ring()

    def category(self):
        from sage.categories.categories import Elements
        return Elements(self.parent())

    def parent(self, x=None):
        if x==None:
            return self.__parent
        else:
            return self.__parent(x)

    def __xor__(self, right):
        raise RuntimeError, "Use ** for exponentiation, not '^', which means xor\n"+\
              "in Python, and has the wrong precedence."

    def _coeff_repr(self, no_space=True):
        if self._is_atomic():
            s = str(self)
        else:
            s = "(%s)"%self
        if no_space:
            return s.replace(' ','')
        return s
    
    def _is_atomic(self):
        if self.parent().is_atomic_repr():
            return True
        s = str(self)
        return bool(s.find("+") == -1 and s.find("-") == -1 and s.find(" ") == -1)

    def _rich_to_bool(self, int op, int n):
        if op == 0:
            return bool(n < 0)
        elif op == 1:
            return bool(n <= 0)
        elif op == 2:
            return bool(n == 0)
        elif op == 3:
            return bool(n != 0)
        elif op == 4:
            return bool(n > 0)
        elif op == 5:
            return bool(n >= 0)
        assert False, "bug in _rich_to_bool"

    #def __richcmp__(self, right, int op):
        # Warning -- this is *not* inherited by derived Pyrex types,
        # so you must copy it to them.  It works fine for Python
        # derived classes (even of derived Pyrex types).  Weird.
    #    print "hi", self, right, op
    #    cdef int n
    #    if not isinstance(right, Element) or right.parent() != self.parent():
    #        try:
    #            n = sage.ext.coerce.cmp(self, right)
    #        except TypeError:
    #            n = -1
    #    else:
    #        n = self.__cmp__(right)
    #    return self._rich_to_bool(op, n)

    #def __cmp__(self, other):
    #    return self._cmp(other)


cdef class ModuleElement(Element):
    """
    Generic element of a module.
    """
    ##################################################    
    def is_zero(self):
        return self == self.__parent(0)

    def is_nonzero(self):
        return not self.is_zero()

    def __add__(self, right):
        if not isinstance(self, Element) or \
               not isinstance(right, Element) or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.add)
        return self._add(right)

    def _add(self, right):
        raise NotImplementedError

    def __sub__(self, right):
        if not isinstance(self, Element) or \
               not isinstance(right, Element) or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.sub)
        return self._sub(right)

    def _sub(self, right):
        return self + (-right)

    def __neg__(self):
        return sage.ext.coerce.bin_op(-1, self, operator.mul)

    def __pos__(self):
        return self

    ##################################################
    def order(self):
        """
        Return the additive order of self.
        """
        return self.additive_order()   

    def additive_order(self):
        """
        Return the additive order of self.
        """
        raise NotImplementedError


cdef class MonoidElement(Element):
    """
    Generic element of a monoid.
    """
    def order(self):
        """
        Return the multiplicative order of self.
        """
        return self.multiplicative_order()

    def multiplicative_order(self):
        """
        Return the multiplicative order of self.
        """
        raise NotImplementedError
    
    def __mul__(self, right):
        if not isinstance(self, Element) or not isinstance(right, Element) \
               or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.mul)
        return self._mul(right)

    def _mul(self, right):
        raise NotImplementedError
    

cdef class AdditiveGroupElement(ModuleElement):
    """
    Generic element of an additive group.
    """
    def order(self):
        """
        Return additive order of element
        """
        return self.additive_order()

    def _mul(self, x):
        raise ArithmeticError, "multiplication not defined in an additive group"
    

cdef class MultiplicativeGroupElement(MonoidElement):
    """
    Generic element of a multiplicative group.
    """
    def order(self):
        """
        Return the multiplicative order of self.
        """
        return self.multiplicative_order()

    def _add(self, x):
        raise ArithmeticError, "addition not defined in a multiplicative group"


cdef class RingElement(Element):
    ##################################################
    def is_zero(self):
        return self == self.__parent(0)

    def is_nonzero(self):
        return not self.is_zero()

    def __add__(self, right):
        if not isinstance(self, Element) or \
               not isinstance(right, Element) or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.add)
        return self._add(right)

    def _add(self, right):
        raise NotImplementedError

    def __sub__(self, right):
        if not isinstance(self, Element) or \
               not isinstance(right, Element) or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.sub)
        return self._sub(right)

    def _sub(self, right):
        return self + (-right)

    def __mul__(self, right):
        if not isinstance(self, Element) or not isinstance(right, Element) \
               or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.mul)
        return self._mul(right)
    
    def _mul(self, right):
        raise NotImplementedError

    def __div__(self, right):
        if not isinstance(self, Element) or not isinstance(right, Element) \
               or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, operator.div)
        return self._div(right)

    def _div(self, right):
        return self.parent().fraction_field()(self, right)
    
    def __neg__(self):
        return sage.ext.coerce.bin_op(-1, self, operator.mul)

    def __pos__(self):
        return self

    def __invert__(self):
        return 1/self

    ##################################################

    def order(self):
        """
        Return the additive order of self.
        """
        return self.additive_order()

    def additive_order(self):
        """
        Return the additive order of self.
        """
        raise NotImplementedError

    def multiplicative_order(self):
        r"""
        Return the multiplicative order of self, if self is a unit, or raise
        \code{ArithmeticError} otherwise.
        """
        if not self.is_unit():
            raise ArithmeticError, "self (=%s) must be a unit to have a multiplicative order."
        raise NotImplementedError

    def is_unit(self):
        if self == 1 or self == -1:
            return True
        raise NotImplementedError
    
    def __pow__(self, n, dummy):
        cdef int i
        import _integer
        if isinstance(n, float):
            raise TypeError, "raising %s to the power of the float %s not defined"%(self, n)

        n = int(n)

        a = self
        power = self.parent()(1)
        if n < 0:
            n = -n
            a = ~self
        elif n == 0:
            return power
        i = 0
        apow2 = a
        while (n>>i) > 0:
            if (n>>i) & 1:
                power = power * apow2
            if n == 0: break   # to not waste time doing an extra multiplication/increment
            apow2 = apow2 * apow2
            i = i+1
        return power




cdef class CommutativeRingElement(RingElement):
    pass

cdef class IntegralDomainElement(CommutativeRingElement):
    pass

cdef class DedekindDomainElement(IntegralDomainElement):
        
    pass
        
cdef class PrincipalIdealDomainElement(DedekindDomainElement):
    def lcm(self, right):
        """
        Returns the least common multiple of self and right. 
        """
        if not isinstance(self, Element) or not isinstance(right, Element) \
               or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, lcm)
        return self._lcm(right)

    def gcd(self, right):
        """
        Returns the gcd of self and right, or 0 if both are 0.
        """
        if not isinstance(self, Element) or not isinstance(right, Element) \
               or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, gcd)
        return self._gcd(right)

    def xgcd(self, right):
        r"""
        Return the extended gcd of self and other, i.e., elements $r, s, t$ such that
        $$
           r = s \cdot self + t \cdot other.
        $$
        """
        if not isinstance(self, Element) or not isinstance(right, Element) \
               or right.parent() != self.parent():
            return sage.ext.coerce.bin_op(self, right, xgcd)
        return self._xgcd(right)

cdef class EuclideanDomainElement(PrincipalIdealDomainElement):

    def degree(self):
        raise NotImplementedError

    def _gcd(self, other):
        """
        Return the greatest common divisor of self and other.

        Algorithm 3.2.1 in Cohen, GTM 138.
        """
        A = self
        B = other
        while B.is_nonzero():
            Q, R = A.quo_rem(B)
            A = B
            B = R
        return A

    def leading_coefficient(self):
        raise NotImplementedError

    def quo_rem(self, other):
        raise NotImplementedError

    def __floordiv__(self,right):
        """
        Quotient of division of self by other.  This is denoted //.
        """
        Q, _ = self.quo_rem(right)
        return Q
        
    def __mod__(self, other):
        """
        Remainder of division of self by other.
        EXAMPLES:
            sage: x = PolynomialRing(IntegerRing()).gen()
            sage: x % (x+1)
            -1
            sage: (x**3 + x - 1) % (x**2 - 1)
            2*x - 1
        """
        _, R = self.quo_rem(other)
        return R
        
    

cdef class FieldElement(RingElement):

    def is_unit(self):
        return not self.is_zero()
    
    def _gcd(self, FieldElement other):
        """
        Return the greatest common divisor of self and other.
        """
        if self.is_zero() and other.is_zero():
            return self
        else:
            return self.parent()(1)

    def _lcm(self, FieldElement other):
        """
        Return the least common multiple of self and other.
        """
        if self.is_zero() and other.is_zero():
            return self
        else:
            return self.parent()(1)

    def _xgcd(self, FieldElement other):
        R = self.parent()
        if not self.is_zero():
            return R(1), ~self, R(0)
        elif not other.is_zero():
            return R(1), R(0), ~self
        else: # both are 0
            return self, self, self
        

    def quo_rem(self, right):
        if not isinstance(right, FieldElement) or right.parent() != self:
            right = self.parent()(right)
        return self/right, 0

    
cdef class AlgebraElement(RingElement):
    pass

cdef class CommutativeAlgebraElement(AlgebraElement):
    pass    

cdef class InfinityElement(RingElement):
    pass
