"""
Real numbers
"""


#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

cimport _element
import _element

include "interrupt.pxi"  # ctrl-c interrupt block support

import operator

import sage.rings.real_field0
import sage.misc.misc as misc

import sage.rings.coerce

# cdefs.pxi is in sage/ext.  It does *not* export all the
# functionality of GMP.  When you need to use a GMP function
# not listed in gmp.pxi, you have to add it to gmp.pxi.
# Use the GMP manual to find the function declaration,
# then possibly modify it slightly and put it in gmp.pxi (see
# the other examples there).

include "cdefs.pxi"  

cdef class RealNumber0(_element.Element)

cdef public void set_from_mpf(RealNumber0 self, mpf_t value):
    mpf_set(self.value, value)
    
cdef object RealNumber0_mul(RealNumber0 a, RealNumber0 b):
    cdef RealNumber0 x
    x = RealNumber0()

    _sig_on
    mpf_mul(x.value, a.value, b.value)
    _sig_off    
    
    return x

cdef object RealNumber0_div(RealNumber0 a, RealNumber0 b):
    cdef RealNumber0 x
    x = RealNumber0()

    _sig_on
    mpf_div(x.value, a.value, b.value)
    _sig_off
    
    return x

cdef RealNumber0_add(RealNumber0 self, RealNumber0 other):
    cdef RealNumber0 x
    x = RealNumber0()
    _sig_on
    mpf_add(x.value, self.value, other.value)
    _sig_off
    return x

cdef RealNumber0_sub(RealNumber0 self, RealNumber0 other):
    cdef RealNumber0 x
    x = RealNumber0()

    _sig_on
    mpf_sub(x.value, self.value, other.value)
    _sig_off
    
    return x
    
cdef class RealNumber0(_element.Element):
    cdef mpf_t value

    def __new__(self, x=None):
        mpf_init(self.value)

    def __init__(self, x=None):
        cdef int i
        if not (x is None):
            self.__set_value(x)

    def __reduce__(self):
        return sage.rings.rational_field.Q, (str(self),)

    def __set_value(self, x):
        if isinstance(x, RealNumber0):
            set_from_RealNumber0(self, x)
        elif isinstance(x, int):
            i = x
            mpf_set_si(self.value, i, 1)
        elif isinstance(x, long):
            s = "%x"%x
            mpf_set_str(self.value, s, 16)
        elif isinstance(x, _integer.Integer):
            set_from_Integer(self, x)

        elif isinstance(x, str):
            _sig_on
            mpf_set_str(self.value, x, 0)
            mpf_canonicalize(self.value)
            _sig_off
            
        elif isinstance(x, tuple) and len(x) == 2:

            n, d = x
            if int(d)==0:
                raise ZeroDivisionError, "division by 0."
            s = "%s/%s"%(n,d)

            _sig_on
            mpf_set_str(self.value, s, 0)
            mpf_canonicalize(self.value)
            _sig_off

        elif isinstance(x, list) and len(x) == 1:
            
            self.__set_value(x[0])
            
        elif isinstance(x, sage.libs.pari.gen):
            # TODO: figure out how to convert to pari integer in base 16
            s = str(x)
            _sig_on
            if mpf_set_str(self.value, s, 0):
                _sig_off
                raise TypeError, "Unable to coerce %s (%s) to RealNumber0"%(x,type(x))
            _sig_off
            
        else:
            raise TypeError, "Unable to coerce %s (%s) to RealNumber0"%(x,type(x))

    cdef void set_from_mpf(RealNumber0 self, mpf_t value):
        mpf_set(self.value, value)

    cdef cmp(RealNumber0 self, RealNumber0 x):
        cdef int i
        i = mpf_cmp(self.value, x.value)
        if i < 0:
            return -1
        elif i == 0:
            return 0
        else:
            return 1

    def __cmp__(self, x):
        return self.cmp(x)

    def __richcmp__(RealNumber0 self, x, int op):
        cdef int n
        if not isinstance(x, RealNumber0):
            try:
                x = RealNumber0(x)
            except TypeError:
                n = sage.rings.coerce.cmp(self, x)
            else:
                n = self.cmp(x)
        else:
            n = self.cmp(x)
        if op == 0:
            return bool(n < 0)
        elif op == 1:
            return bool(n <= 0)
        elif op == 2:
            return bool(n == 0)
        elif op == 3:
            return bool(n != 0)
        elif op == 4:
            return bool(n > 0)
        elif op == 5:
            return bool(n >= 0)
            
    def copy(self):
        cdef RealNumber0 z
        z = RealNumber0()
        mpf_set(z.value, self.value)
        return z

    def  __dealloc__(self):
        mpf_clear(self.value)
        
    def __repr__(self):
        return self.str()
        
    def str(self, int base=10):
        if base < 2 or base > 36:
            raise ValueError, "base (=%s) must be between 2 and 36"%base
        cdef size_t n
        cdef char *s
        _sig_on
        mpf_get_str(s, base, self.value)
        _sig_off
        return PyString_FromString(s)

    def __float__(self):
        return mpf_get_d(self.value)
        
    def __hash__(self):
        return hash(mpf_get_d(self.value))
    
    def set_si(self, signed long int n):
        mpf_set_si(self.value, n, 1)

    def set_str(self, s, base=10):
        valid = mpf_set_str(self.value, s, base)
        if valid != 0:
            raise ValueError, "invalid literal:" + s

    def __add(RealNumber0 self, RealNumber0 other):
        cdef RealNumber0 x
        x = RealNumber0()
        _sig_on
        mpf_add(x.value, self.value, other.value)
        _sig_off
        return x
        
    def __add__(x, y):
        if isinstance(x, RealNumber0) and isinstance(y, RealNumber0):
            return x.__add(y)
        return sage.rings.coerce.bin_op(x, y, operator.add)

    def __sub(RealNumber0 self, RealNumber0 other):
        cdef RealNumber0 x
        x = RealNumber0()
        _sig_on
        mpf_sub(x.value, self.value, other.value)
        _sig_off
        return x
        
    def __sub__(x, y):
        if isinstance(x, RealNumber0) and isinstance(y, RealNumber0):
            return x.__sub(y)
        return sage.rings.coerce.bin_op(x, y, operator.sub)
        
    def __mul(RealNumber0 self, RealNumber0 other):
        cdef RealNumber0 x
        x = RealNumber0()
        _sig_on
        mpf_mul(x.value, self.value, other.value)
        _sig_off
        return x
    
    def __mul__(x, y):
        if isinstance(x, RealNumber0) and isinstance(y, RealNumber0):
            return x.__mul(y)
        return sage.rings.coerce.bin_op(x, y, operator.mul)
    
    def __div(RealNumber0 self, RealNumber0 other):
        if not other:
            raise ZeroDivisionError, "RealNumber0 division by zero"
        cdef RealNumber0 x
        x = RealNumber0()
        _sig_on
        mpf_div(x.value, self.value, other.value)
        _sig_off
        return x
        
    def __div__(x, y):
        if isinstance(x, RealNumber0) and isinstance(y, RealNumber0):
            return x.__div(y)
        return sage.rings.coerce.bin_op(x, y, operator.div)
    
    def __invert__(self):
        raise NotImplementedError
    
    def __pow__(RealNumber0 self, n, dummy):
        raise NotImplementedError
    
    def __pos__(self):
        return self

    def __neg__(self):
        raise NotImplementedError        

    def __nonzero__(self):
        raise NotImplementedError

    def __abs__(self):
        raise NotImplementedError

    def __mod__(RealNumber0 self, other):
        raise NotImplementedError

    def parent(self):
        return self.__parent

    def __int__(self):
        raise NotImplementedError

    def __long__(self):
        raise NotImplementedError

