"""
Expect interface to other computer algebra systems (base class).
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import os
import pexpect

from sage.misc.misc import SAGE_HOME
from sage.structure.element import RingElement

class Expect:
    def __init__(self, name, prompt, command=None, maxread=10000, 
                 script_subdirectory="", restart_on_ctrlc=False,
                 verbose_start=False):

        if command == None:
            command = name
        self.__maxread = maxread
        self.__script_subdirectory = script_subdirectory
        self.__name = name
        self.__command = command
        self.__prompt = prompt
        self.__restart_on_ctrlc = restart_on_ctrlc
        self.__verbose_start = verbose_start
        self.__path = '%s/data/scripts/%s/%s'%(SAGE_HOME,self.__name, self.__script_subdirectory)
        self.__initialized = False
        self.__seq = -1

    def __repr__(self):
        return "A running instance of %s"%self.__name

    def _temp_file(self, x):
        T = self.__path + "/tmp/"
        if not os.path.exists(T):
            os.makedirs(T)
        return T + str(x)


    def path(self):
        return self.__path

    def expect(self):
        return self.__expect

    def interact(self):
        self.__expect.interact()

    def _change_prompt(self, prompt):
        self.__prompt = prompt

    def _start(self):
	if os.uname()[0][:6] == "CYGWIN":  # extra verbose on Windows, since is very slow
            print "Starting %s (slow on Windows only; takes about 15 seconds)"%self.__name
        current_path = os.path.abspath('.')
        dir = self.__path
        if not os.path.exists(dir):
            os.makedirs(dir)
        os.chdir(dir)
        if self.__verbose_start:
            print "Starting %s"%self.__command.split()[0]
        try:
            self.__expect = pexpect.spawn(self.__command)
        except pexpect.ExceptionPexpect:
            raise RuntimeError, "Unable to start %s because the command '%s' failed."%(self.__name, self.__command)
        os.chdir(current_path)
        self.__expect.timeout = 9999999  # don't make this bigger, or it breaks on OS X
        self.__expect.setmaxread(self.__maxread)
        self.__expect.expect(self.__prompt)

    def __del__(self):
        # The following two lines program around a bug in pexpect.
        def dummy(): pass
        try:
            self.__expect.close = dummy
        except AttributeError:
            pass

    def eval_line(self, line):
        if line.find('\n') != -1:
            raise ValueError, "line must not contain any newlines"
        try:
            try:
                E = self.__expect
            except AttributeError:
                self._start()
                E = self.__expect
            E.sendline(line)                
            if len(line)>0:
                self.__expect.expect(self.__prompt)
                out = self.__expect.before
            else:
                out = '\n\r'
        except KeyboardInterrupt:
            if self.__restart_on_ctrlc:
                self.__expect.close()
                self._start()
                raise KeyboardInterrupt, "Restarting %s"%self.__name.upper()
            else:
                self.__expect.sendline(chr(3))  # send ctrl-c
                self.__expect.expect(self.__prompt)
                self.__expect.expect(self.__prompt)
                raise KeyboardInterrupt, "Ctrl-c pressed"

        i = out.find("\n")
        j = out.rfind("\r")
        return out[i+1:j].replace('\r\n','\n')

    def __call__(self, code):
        return '\n'.join([self.eval_line(L) for L in str(code).split('\n')])

    def _coerce_(self, x):
        if not isinstance(x, ExpectObject):
            raise TypeError


    ############################################################
    #         Functions for working with variables.
    #  The first three must be overloaded by derived classes,
    #  and the definition depends a lot on the class.  But
    #  the functionality one gets from this is very nice. 
    ############################################################
    
    def set(self, var, value):
        """
        Set the variable var to the given value.
        """
        raise NotImplementedError

    def get(self, var):
        """
        Get the value of the variable var.
        """
        raise NotImplementedError

    def clear(self, var):
        """
        Clear the variable named var.
        """
        raise NotImplementedError
    
    def _next_var_name(self):
        self.__seq += 1
        return "v%s"%self.__seq

    def _create(self, value):
        name = self._next_var_name()
        self.set(name, value)
        return name

    def new(self, x):
        r"""
        Create an expect object X obtained by evaluating the string x using self.

        The object X returned can be used like any SAGE object, and
        wraps an object in self.  The standard arithmetic operators
        work.  Morever if foo is a function then
                      X.foo(y,z,...)
        calls foo(X, y, z, ...) and returns the corresponding object.

        EXAMPLES:
        We illustrate objects that wrap GP objects (gp is the PARI interpreter):
        
            sage: M = gp.new('[1,2;3,4]')
            sage: M
            [1, 2; 3, 4]
            sage: M * M
            [7, 10; 15, 22]
            sage: M + M
            [2, 4; 6, 8]
            sage: M.matdet()
            -2
            sage: E = gp.new('ellinit([1,2,3,4,5])')
            sage: E.ellglobalred()
            [10351, [1, -1, 0, -1], 1]
            sage: E.ellan(20)
            [1, 1, 0, -1, -3, 0, -1, -3, -3, -3, -1, 0, 1, -1, 0, -1, 5, -3, 4, 3]
        """
        if isinstance(x, ExpectObject) and x.parent() == self:
            return x
        return ExpectObject(self, x)

    def function_call(self, function, args):
        if not isinstance(args, list):
            args = [args]
        for i in range(len(args)):
            if not isinstance(args[i], ExpectObject):
                args[i] = self.new(args[i])
        return self.new("%s(%s)"%(function, ",".join([s.name() for s in args])))



class FunctionObject:
    def __init__(self, obj, name):
        self._obj = obj
        self._name = name

    def __repr__(self):
        return "member function %s"%self._name

    def __call__(self, *args):
        return self._obj.parent().function_call(self._name, [self._obj.name()] + list(args))


class ExpectObject(RingElement):
    def __init__(self, parent, value):
        self._name = parent._create(value)
        RingElement.__init__(self, parent)

    def __del__(self):
        try:
            self.parent().clear(self._name)
        except RuntimeError:
            pass

    def __repr__(self):
        return self.parent().get(self._name)

    def __getattr__(self, attrname):
        return FunctionObject(self, attrname)

    def __getitem__(self, n):
        return self.parent().new('%s[%s]'%(self._name, n))

    def name(self):
        return self._name

    def _add(self, right):
        return self.parent().new('%s + %s'%(self._name, right._name))
        
    def _sub(self, right):
        return self.parent().new('%s - %s'%(self._name, right._name))

    def _mul(self, right):
        return self.parent().new('%s * %s'%(self._name, right._name))

    def _div(self, right):
        return self.parent().new('%s / %s'%(self._name, right._name))

    def __pow__(self, n):
        return self.parent().new('%s ^ %s'%(self._name,n))



if __name__ ==  '__main__':
    import doctest, sys
    from gp import gp
    doctest.testmod(sys.modules[__name__])
