"""
Interface to the GAP interpreter

SAGE provides an interface to the GAP system.  This system provides
extensive group theory, combinatorics, etc.
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from expect import Expect
from sage.misc.misc import SAGE_HOME
import os
DB_HOME = "%s/data/"%SAGE_HOME
VERSION='4.4.6'
WORKSPACE = "%s/data/scripts/gap/empty_workspace"%SAGE_HOME

class GAP(Expect):
    def __init__(self, maxread=10000, script_subdirectory="user",
                 use_workspace_cache = True):

        self.__use_workspace_cache = use_workspace_cache
        if use_workspace_cache and os.path.exists(WORKSPACE):
            cmd = "gap -T -n -b -L %s"%WORKSPACE
            self.__make_workspace = False            
        else:
            cmd = "gap -T -n -b"
            self.__make_workspace = True

        Expect.__init__(self,
                        name = 'gap',
                        prompt = 'gap> ',
                        command = cmd,
                        maxread = maxread,
                        script_subdirectory = script_subdirectory,
                        restart_on_ctrlc = True,
                        verbose_start = False)


    def _start(self):
        Expect._start(self)
        if self.__use_workspace_cache and self.__make_workspace:
            self('SaveWorkspace("%s");'%WORKSPACE)

    def __call__(self, x, newlines=True):
        r"""
        Send the code in the string s to the GAP interpreter and return
        the output as a string.

        INPUT:
            s -- string containing GAP code.
            newlines -- bool (default: True); if False, remove all
                      backslash-newlines inserted by the GAP output formatter.
        """
        x = x.rstrip()
        if len(x) == 0 or x[len(x) - 1] != ';':
            x += ';'
        s = Expect.__call__(self, x)
        if newlines:
            return s
        else:
            return s.replace("\\\n","")

    def set(self, var, value):
        """
        Set the variable var to the given value.
        """
        self('%s:=%s;;'%(var,value))

    def get(self, var):
        """
        Get the value of the variable var.
        """
        return self('%s;'%var)
        
    def clear(self, var):
        """
        Clear the variable named var.
        """
        self('Unbind(%s)'%var)


def gap_reset_workspace():
    """
    Call this when you include any new GAP code into SAGE.
    
    The first time you start GAP from SAGE, it saves the
    startup state of GAP in the file

        data/scripts/gap/empty_workspace
        
    This is useful, since then subsequent startup of GAP
    is at least 10 times as fast.  Unfortunately, if you
    install any new code for GAP, it won't be noticed,
    unless you call this function (and restart SAGE).
    """
    g = GAP(use_workspace_cache=False)
    g('SaveWorkspace("%s");'%WORKSPACE)
        



###########

gap = GAP()


import os
def gap_console(use_workspace_cache=True):
    if use_workspace_cache and os.path.exists(WORKSPACE):
        cmd = "gap -L %s"%WORKSPACE
    else:
        cmd = "gap"
    os.system(cmd)
