r"""
Interface to Singular

This interface is extremely flexible, since it's exactly like typing
into the Singular interpreter, and anything that works there should
work here. 

\begin{notice} The \code{pexpect} module must be installed and the
\code{Singular} command must be in your path.
Both pexpect and \code{Singular} are built and installed
when you install SAGE.
\end{notice}

AUTHORS: David Joyner and William Stein

EXAMPLES:
First we illustrate multivariate polynomial factorization:

    sage: R1 = singular.ring(0, '(x,y)', 'dp')
    sage: R1
    //   characteristic : 0
    //   number of vars : 2
    //        block   1 : ordering dp
    //                  : names    x y 
    //        block   2 : ordering C
    sage: f = singular.new('9x16 - 18x13y2 - 9x12y3 + 9x10y4 - 18x11y2 + 36x8y4 + 18x7y5 - 18x5y6 + 9x6y4 - 18x3y6 - 9x2y7 + 9y8')
    sage: f
    9x16-18x13y2-9x12y3+9x10y4-18x11y2+36x8y4+18x7y5-18x5y6+9x6y4-18x3y6-9x2y7+9y8
    sage: f.parent()
    A running instance of singular

    sage: F = f.factorize(); F
    [1]:
       _[1]=9
       _[2]=x6-2x3y2-x2y3+y4
       _[3]=-x5+y2
    [2]:
       1,1,2

    sage: F[1]
    9,
    x6-2x3y2-x2y3+y4,
    -x5+y2
    sage: F[1][2]
    x6-2x3y2-x2y3+y4

We can convert f and each exponent back to SAGE objects as well.

    sage: x, y = MPolynomialRing(RationalField(), 2).gens()
    sage: g = eval(f.sage_polystring()); g
    9*x_1^8 - 9*x_0^2*x_1^7 - 18*x_0^3*x_1^6 - 18*x_0^5*x_1^6 + 9*x_0^6*x_1^4 + 18*x_0^7*x_1^5 + 36*x_0^8*x_1^4 + 9*x_0^10*x_1^4 - 18*x_0^11*x_1^2 - 9*x_0^12*x_1^3 - 18*x_0^13*x_1^2 + 9*x_0^16
    sage: eval(F[1][2].sage_polystring())
    x_1^4 - x_0^2*x_1^3 - 2*x_0^3*x_1^2 + x_0^6

This example illustrates polynomial GCD's:
    sage: R2 = singular.ring(0, '(x,y,z)', 'lp')
    sage: a = singular.new('3x2*(x+y)')
    sage: b = singular.new('9x*(y2-x2)')
    sage: g = a.gcd(b)
    sage: g
    x2+xy


This example illustrates computation of a Groebner basis:

    sage: R3 = singular.ring(0, '(a,b,c,d)', 'lp')
    sage: I = singular.ideal(['a+b+c+d', 'ab+ad+bc+cd', 'abc+abd+acd+bcd', 'abcd-1'])
    sage: I2 = I.groebner()
    sage: I2
    c2d6-c2d2-d4+1,
    c3d2+c2d3-c-d,
    bd4-b+d5-d,
    bc-bd5+c2d4+cd-d6-d2,
    b2+2bd+d2,
    a+b+c+d


The following example is the same as the one in the Singular - Gap
interface documentation:

    sage: R  = singular.ring(0, '(x0,x1,x2)', 'lp')
    sage: I1 = singular.ideal(['x0*x1*x2 -x0^2*x2', 'x0^2*x1*x2-x0*x1^2*x2-x0*x1*x2^2', 'x0*x1-x0*x2-x1*x2'])
    sage: I2 = I1.groebner()
    sage: I2
    x1^2*x2^2,
    x0*x2^3-x1^2*x2^2+x1*x2^3,
    x0*x1-x0*x2-x1*x2,
    x0^2*x2-x0*x1*x2

This example illustrates moving a polynomial from one ring to another.
It also illustrates calling a method of an object with an argument.

    sage: R = singular.ring(0, '(x,y,z)', 'dp')
    sage: f = singular.new('x3+y3+(x-y)*x2y2+z2')
    sage: f
    x3y2-x2y3+x3+y3+z2
    sage: R1 = singular.ring(0, '(x,y,z)', 'ds')
    sage: f = R.fetch(f)
    sage: f
    z2+x3+y3+x3y2-x2y3

We can calculate the Milnor number of $f$:
    sage: _=singular('LIB "sing.lib";')     # assign to _ to suppress printing
    sage: f.milnor()
    4

The Jacobian applied twice yields the Hessian matrix of $f$,
with which we can compute.
    sage: H = f.jacob().jacob()
    sage: H
    6x+6xy2-2y3,6x2y-6xy2,  0,
    6x2y-6xy2,  6y+2x3-6x2y,0,
    0,          0,          2 
    sage: H.det()
    72xy+24x4-72x3y+72xy3-24y4-48x4y2+64x3y3-48x2y4

The 1x1 and 2x2 minors:
    sage: H.minor(1)
    2,
    6y+2x3-6x2y,
    6x2y-6xy2,
    6x2y-6xy2,
    6x+6xy2-2y3
    sage: H.minor(2)
    12y+4x3-12x2y,
    12x2y-12xy2,
    12x2y-12xy2,
    12x+12xy2-4y3,
    -36xy-12x4+36x3y-36xy3+12y4+24x4y2-32x3y3+24x2y4

    sage: _=singular('option(redSB)')
    sage: H.minor(1).groebner()
    1
"""

#*****************************************************************************
#       Copyright (C) 2005 David Joyner and William Stein
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from expect import Expect, ExpectObject, FunctionObject

from sage.structure.element import RingElement

class Singular(Expect):
    def __init__(self, maxread=10000, script_subdirectory="user"):
        Expect.__init__(self,
                        name = 'singular',
                        prompt = '> ',
                        command = "Singular",
                        maxread = maxread,
                        script_subdirectory = script_subdirectory,
                        restart_on_ctrlc = True,
                        verbose_start = False)
    def __call__(self, x):
        """
        Send the code x to the Singular interpreter and return the output
        as a string.
        """
        x = x.rstrip()
        if len(x) == 0 or x[len(x) - 1] != ';':
            x += ';'
        s = Expect.__call__(self, x)
        if s.find("error occurred in STDIN") != -1:
            raise RuntimeError, "Singular reported the following error:\n" + s
        return s

    def set(self, type, name, value):
        """
        Set the variable with given name to the given value.
        """
        self('%s %s=%s;'%(type, name, value))

    def get(self, var):
        """
        Get string representation of variable named var.
        """
        return self('print(%s);'%var)
        
    def clear(self, var):
        """
        Clear the variable named var.
        """
        self('kill %s;'%var)

    def _create(self, value, type='def'):
        name = self._next_var_name()
        self.set(type, name, value)
        return name


    def new(self, x, type='def'):
        """
        Create an expect object X with given type determined by the string x.
        This returns var, where var is built using the Singular statement
                  type var = ... x ...

        The object X returned can be used like any SAGE object, and
        wraps an object in self.  The standard arithmetic operators
        work.  Morever if foo is a function then
                      X.foo(y,z,...)
        calls foo(X, y, z, ...) and returns the corresponding object.

        EXAMPLES:
        """
        if isinstance(x, SingularObject):
            return x
        return SingularObject(self, type, x)


    ###################################################################
    # Singular libraries
    ###################################################################
    def LIB(self, lib):
        """
        Load the Singular library named lib.
        """
        self('LIB "%s"'%lib)
    lib = LIB

    ###################################################################
    # constructors
    ###################################################################
    def ideal(self, gens):
        """
        Return the ideal generated by gens.
        
        INPUT:
            gens -- list of Singular objects (or objects that can be
                    made into Singular objects via evaluation)
        OUTPUT:
            the Singular ideal generated by the given list of gens
        """
        if not isinstance(gens, list):
            raise TypeError, "gens must be a list"

        gens2 = []
        for g in gens:
            if not isinstance(g, SingularObject):
                gens2.append(self.new(g))
            else:
                gens2.append(g)
        return self.new(",".join([g.name() for g in gens2]), 'ideal')
    
    def list(self, x):
        return self.new(x, 'list')

    def matrix(self, nrows, ncols, entries):
        """
        EXAMPLES:
            sage: singular.lib("matrix.lib")
            sage: R = singular.ring(0, '(x,y,z)', 'dp')
            sage: A = singular.matrix(3,2,'1,2,3,4,5,6')
            sage: A
            1,2,
            3,4,
            5,6 
            sage: A.gauss_col()
            2,-1,
            1,0, 
            0,1  
        """
        name = self._next_var_name()
        self('matrix %s[%s][%s] = %s'%(name, nrows, ncols, entries))
        return SingularObject(self, None, None, name)

    def ring(self, base, vars, order='lp'):
        r"""
        Create a Singular ring.

        INPUT:
            base -- determines the base ring; e.g., an integer to
                    to give the characteristic, etc. (see examples below)
            vars -- a string that defines the variable names
            order -- string -- the monomial order (default: 'lp')

        OUTPUT:
            a Singular ring

        EXAMPLES:
        We first declare $\Q[x,y,z]$ with degree reverse lexicographic ordering.
        
            sage: R = singular.ring(0, '(x,y,z)', 'dp')
            sage: R
            //   characteristic : 0
            //   number of vars : 3
            //        block   1 : ordering dp
            //                  : names    x y z 
            //        block   2 : ordering C

            sage: R1 = singular.ring(32003, '(x,y,z)', 'dp')
            sage: R2 = singular.ring(32003, '(a,b,c,d)', 'lp')

        This is a ring in variables named x(1) through x(10) over the finite
        field of order $7$:
            sage: R3 = singular.ring(7, '(x(1..10))', 'ds')

        This is a polynomial ring over the transcendtal extension $\Q(a)$ of $\Q$:
            sage: R4 = singular.ring('(0,a)', '(mu,nu)', 'lp')

        This is a ring over the field of single-precision floats:
            sage: R5 = singular.ring('real', '(a,b)', 'lp')

        This is over 50-digit floats:
            sage: R6 = singular.ring('(real,50)', '(a,b)', 'lp')
            sage: R7 = singular.ring('(complex,50,i)', '(a,b)', 'lp')

        To use a ring that you've defined, use the setring() method on
        the ring.  This sets the ring to be the ``current ring''.  For
        example,
            sage: R = singular.ring(7, '(a,b)', 'ds')
            sage: S = singular.ring('real', '(a,b)', 'lp')
            sage: singular.new('10*a')
            1.000e+01a
            sage: R.setring()
            sage: singular.new('10*a')
            3a
        """
        return self.new('%s,%s,%s'%(base,vars,order), 'ring')

    def string(self, x):
        return self.new(x, 'string')


class SingularObject(ExpectObject):
    def __init__(self, parent, type, value, _name=None):
        if _name is None:
            self._name = parent._create( value, type)
        else:
            self._name = _name
        RingElement.__init__(self, parent)

    def sage_polystring(self):
        s = str(self).replace('^','**')
        i = 0; t = ''
        while i < len(s):
            if s[i].isdigit() and i > 0 and s[i-1].isalpha():
                t += "**"
            elif s[i].isalpha() and i >0 and (s[i-1].isdigit() or s[i-1].isalpha()):
                t += "*"
            t += s[i]
            i += 1
        return t

    def setring(self):
        self.parent()('setring %s'%(self._name))


        

        
##################################

singular = Singular()

import os
def singular_console():
    os.system('Singular')



if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.all import MPolynomialRing, RationalField
    doctest.testmod(sys.modules[__name__])


