"""
NTL ZZ_p integers module p
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

include 'interrupt.pxi'
include 'misc.pxi'
include 'decl.pxi'

cimport ntl_ZZ
import ntl_ZZ

cdef class ntl_ZZ_p:
    r"""
    The \class{ZZ_p} class is used to represent integers modulo $p$.
    The modulus $p$ may be any positive integer, not necessarily prime.
    
    Objects of the class \class{ZZ_p} are represented as a \code{ZZ} in the
    range $0, \ldots, p-1$.

    An executing program maintains a "current modulus", which is set to p
    with ntl_ZZ_p.init(p).  The current modulus should be initialized before
    any ZZ_p objects are created.

    The modulus may be changed, and a mechanism is provided for saving and
    restoring a modulus (see classes ZZ_pBak and ZZ_pContext below).
    """
    # See ntl_ZZ_p.pxd for definition of data members

    def __dealloc__(self):
        del_ZZ_p(self.x)

    def __repr__(self):
        _sig_on
        return string(ZZ_p_to_str(self.x))

    def __cmp__(ntl_ZZ_p self, ntl_ZZ_p other):
        cdef int t
        _sig_on
        t = ZZ_p_eq(self.x, other.x)
        _sig_off
        if t:
            return 0
        return 1

    def __invert__(ntl_ZZ_p self):
        _sig_on
        return make_ZZ_p(ZZ_p_inv(self.x))
        

    def __mul__(ntl_ZZ_p self, other):
        cdef ntl_ZZ_p y
        if not isinstance(other, ntl_ZZ_p):
            other = ntl_ZZ_p(other)
        y = other
        _sig_on
        return make_ZZ_p(ZZ_p_mul(self.x, y.x))
            
    def __sub__(ntl_ZZ_p self, other):
        cdef ntl_ZZ_p y
        if not isinstance(other, ntl_ZZ_p):
            other = ntl_ZZ_p(other)
        y = other
        _sig_on
        return make_ZZ_p(ZZ_p_sub(self.x, y.x))

    def __add__(ntl_ZZ_p self, other):
        cdef ntl_ZZ_p y
        if not isinstance(other, ntl_ZZ_p):
            other = ntl_ZZ_p(other)
        y = other
        _sig_on
        return make_ZZ_p(ZZ_p_add(self.x, y.x))

    def __neg__(ntl_ZZ_p self):
        _sig_on
        return make_ZZ_p(ZZ_p_neg(self.x))
            
    def __pow__(ntl_ZZ_p self, long e, ignored):
        _sig_on
        return make_ZZ_p(ZZ_p_pow(self.x, e))

    cdef set(self, void *y):  # only used internally for initialization; assumes self.x not set yet!
        self.x = <ZZ_p*> y
        
        
cdef public make_ZZ_p(ZZ_p* x):
    cdef ntl_ZZ_p y
    _sig_off
    y = ntl_ZZ_p()
    y.x = x
    return y

def make_new_ZZ_p(x='0'):
    s = str(x)
    cdef ntl_ZZ_p n
    n = ntl_ZZ_p()
    _sig_on
    n.x = str_to_ZZ_p(s)
    _sig_off
    return n

def set_modulus(ntl_ZZ.ntl_ZZ p):
    ntl_ZZ_set_modulus(<ZZ*>p.x)


def random():
    """
    Return a random number modulo p.
    """
    _sig_on
    return make_ZZ_p(ZZ_p_random())
