"""
NTL ZZ_pX polynomials over the integers
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

include 'interrupt.pxi'
include 'decl.pxi'
include 'misc.pxi'




cdef class ntl_ZZ_pX:
    r"""
    The class \class{ZZ_pX} implements polynomial arithmetic modulo $p$.

    Polynomial arithmetic is implemented using the FFT, combined with
    the Chinese Remainder Theorem.  A more detailed description of the
    techniques used here can be found in [Shoup, J. Symbolic
    Comp. 20:363-397, 1995].

    Small degree polynomials are multiplied either with classical 
    or Karatsuba algorithms.
    """
    # See ntl_ZZ_pX.pxd for definition of data members
    def __init__(self):
        """
        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20)) 
            sage: f = ntl.ZZ_pX([1,2,5,-9])
            sage: f
            [1 2 5 11]
            sage: g = ntl.ZZ_pX([0,0,0]); g
            []
            sage: g[10]=5
            sage: g
            [0 0 0 0 0 0 0 0 0 0 5]
            sage: g[10]
            5
        """
        return 

    def __dealloc__(self):
        if self.x:
            ZZ_pX_dealloc(self.x)
                

    def __repr__(self):
        return str(ZZ_pX_repr(self.x))

    def copy(self):
        return make_ZZ_pX(ZZ_pX_copy(self.x))

    cdef set(self, void* x):  # only used internally for initialization; assumes self.x not set yet!
        self.x = <ZZ_pX*>x

    def __setitem__(self, long i, a):
        if i < 0:
            raise IndexError, "index (i=%s) must be >= 0"%i
        a = str(int(a))
        ZZ_pX_setitem(self.x, i, a)

    def __getitem__(self, unsigned int i):
        cdef char* t
        t = ZZ_pX_getitem(self.x,i)
        return int(string(t))

    def __add__(ntl_ZZ_pX self, ntl_ZZ_pX other):
        """
        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))         
            sage: ntl.ZZ_pX(range(5)) + ntl.ZZ_pX(range(6))
            [0 2 4 6 8 5]
        """
        return make_ZZ_pX(ZZ_pX_add(self.x, other.x))

    def __sub__(ntl_ZZ_pX self, ntl_ZZ_pX other):
        """
        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))         
            sage: ntl.ZZ_pX(range(5)) - ntl.ZZ_pX(range(6))
            [0 0 0 0 0 15]
        """
        return make_ZZ_pX(ZZ_pX_sub(self.x, other.x))

    def __mul__(ntl_ZZ_pX self, ntl_ZZ_pX other):
        """
        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))         
            sage: ntl.ZZ_pX(range(5)) * ntl.ZZ_pX(range(6))
            [0 0 1 4 10 0 10 14 11]
        """
        _sig_on
        return make_ZZ_pX(ZZ_pX_mul(self.x, other.x))

    def __div__(ntl_ZZ_pX self, ntl_ZZ_pX other):
        """
        Compute quotient self / other, if the quotient is a polynomial.
        Otherwise an Exception is raised.
        
        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))         
            sage: f = ntl.ZZ_pX([1,2,3]) * ntl.ZZ_pX([4,5])**2
            sage: g = ntl.ZZ_pX([4,5])
            sage: f/g
            [4 13 5 15]
            sage: ntl.ZZ_pX([1,2,3]) * ntl.ZZ_pX([4,5])
            [4 13 5 15]

            sage: f = ntl.ZZ_pX(range(10)); g = ntl.ZZ_pX([-1,0,1])
            sage: f/g
            Traceback (most recent call last):
            ...
            ArithmeticError: self (=[0 1 2 3 4 5 6 7 8 9]) is not divisible by other (=[16 0 1])
        """
        _sig_on
        cdef int divisible
        cdef ZZ_pX* q
        q = ZZ_pX_div(self.x, other.x, &divisible)
        if not divisible:
            raise ArithmeticError, "self (=%s) is not divisible by other (=%s)"%(self, other)
        return make_ZZ_pX(q)

    def __mod__(ntl_ZZ_pX self, ntl_ZZ_pX other):
        """
        Given polynomials a, b in ZZ[X], there exist polynomials q, r
        in QQ[X] such that a = b*q + r, deg(r) < deg(b).  This
        function returns q if q lies in ZZ[X], and otherwise raises an
        Exception.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))         
            sage: f = ntl.ZZ_pX([2,4,6]); g = ntl.ZZ_pX([2])
            sage: f % g   # 0
            []

            sage: f = ntl.ZZ_pX(range(10)); g = ntl.ZZ_pX([-1,0,1])
            sage: f % g
            [3 8]
        """
        _sig_on
        return make_ZZ_pX(ZZ_pX_mod(self.x, other.x))

    def quo_rem(self, ntl_ZZ_pX other):
        """
        Returns the unique integral q and r such that self = q*other +
        r, if they exist.  Otherwise raises an Exception.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))         
            sage: f = ntl.ZZ_pX(range(10)); g = ntl.ZZ_pX([-1,0,1])
            sage: q, r = f.quo_rem(g)
            sage: q, r
            ([3 7 1 4 14 16 8 9], [3 8])
            sage: q*g + r == f
            True
        """
        cdef ZZ_pX *r, *q
        _sig_on
        ZZ_pX_quo_rem(self.x, other.x, &r, &q)
        return (make_ZZ_pX(q), make_ZZ_pX(r))

    def square(self):
        """
        Return f*f.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))
            sage: f = ntl.ZZ_pX([-1,0,1])
            sage: f*f
            [1 0 15 0 1]
        """
        _sig_on
        return make_ZZ_pX(ZZ_pX_square(self.x))

    def __pow__(ntl_ZZ_pX self, long n, ignored):
        """
        Return the n-th nonnegative power of self.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))
            sage: g = ntl.ZZ_pX([-1,0,1])
            sage: g**10
            [1 0 10 0 5 0 0 0 10 0 8 0 10 0 0 0 5 0 10 0 1]
        """
        if n < 0:
            raise NotImplementedError
        import sage.rings.arith
        return sage.rings.arith.generic_power(self, n, make_new_ZZ_pX([1]))

    def __cmp__(ntl_ZZ_pX self, ntl_ZZ_pX other):
        """
        Decide whether or not self and other are equal.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([1,2,3])
            sage: g = ntl.ZZ_pX([1,2,3,0])
            sage: f == g
            True
            sage: g = ntl.ZZ_pX([0,1,2,3])
            sage: f == g
            False
        """
        if ZZ_pX_equal(self.x, other.x):
            return 0
        return -1

    def is_zero(self):
        """
        Return True exactly if this polynomial is 0.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([0,0,0,20])
            sage: f.is_zero()
            True
            sage: f = ntl.ZZ_pX([0,0,1])
            sage: f
            [0 0 1]
            sage: f.is_zero()
            False
        """
        return bool(ZZ_pX_is_zero(self.x))

    def is_one(self):
        """
        Return True exactly if this polynomial is 1.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([1,1])
            sage: f.is_one()
            False
            sage: f = ntl.ZZ_pX([1])
            sage: f.is_one()
            True
        """
        return bool(ZZ_pX_is_one(self.x))

    def is_monic(self):
        """
        Return True exactly if this polynomial is monic.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([2,0,0,1])
            sage: f.is_monic()
            True
            sage: g = f.reverse()
            sage: g.is_monic()
            False
            sage: g
            [1 0 0 2]
        """
        return bool(ZZ_pX_is_monic(self.x))

    def __neg__(self):
        """
        Return the negative of self.
        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))
            sage: f = ntl.ZZ_pX([2,0,0,1])
            sage: -f
            [18 0 0 19]
        """
        return make_ZZ_pX(ZZ_pX_neg(self.x))

    def left_shift(self, long n):
        """
        Return the polynomial obtained by shifting all coefficients of
        this polynomial to the left n positions.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([2,0,0,1])
            sage: f
            [2 0 0 1]
            sage: f.left_shift(2)
            [0 0 2 0 0 1]
            sage: f.left_shift(5)
            [0 0 0 0 0 2 0 0 1]

        A negative left shift is a right shift.
            sage: f.left_shift(-2)
            [0 1]
        """
        return make_ZZ_pX(ZZ_pX_left_shift(self.x, n))
        
    def right_shift(self, long n):
        """
        Return the polynomial obtained by shifting all coefficients of
        this polynomial to the right n positions.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([2,0,0,1])
            sage: f
            [2 0 0 1]
            sage: f.right_shift(2)
            [0 1]
            sage: f.right_shift(5)
            []
            sage: f.right_shift(-2)
            [0 0 2 0 0 1]
        """
        return make_ZZ_pX(ZZ_pX_right_shift(self.x, n))

    def gcd(self, ntl_ZZ_pX other):
        """
        Return the gcd d = gcd(a, b), where by convention the leading coefficient
        of d is >= 0.  We uses a multi-modular algorithm.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))        
            sage: f = ntl.ZZ_pX([1,2,3]) * ntl.ZZ_pX([4,5])**2
            sage: g = ntl.ZZ_pX([1,1,1])**3 * ntl.ZZ_pX([1,2,3])
            sage: f.gcd(g)
            [6 12 1]
            sage: g.gcd(f)
            [6 12 1]
        """
        _sig_on
        return make_ZZ_pX(ZZ_pX_gcd(self.x, other.x))

    def xgcd(self, ntl_ZZ_pX other, plain=True):
        """
        Returns r,s,t such that r = s*self + t*other.
            
        Here r is the resultant of a and b; if r != 0, then this
        function computes s and t such that: a*s + b*t = r; otherwise
        s and t are both 0. 

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))                
            sage: f = ntl.ZZ_pX([1,2,3]) * ntl.ZZ_pX([4,5])**2
            sage: g = ntl.ZZ_pX([1,1,1])**3 * ntl.ZZ_pX([1,2,3])
            sage: f.xgcd(g)   # nothing since they are not coprime
            ([6 12 1], [15 13 6 8 7 9], [4 13])
    
        In this example the input quadratic polynomials have a common root modulo 13.
            sage: f = ntl.ZZ_pX([5,0,1])
            sage: g = ntl.ZZ_pX([18,0,1])
            sage: f.xgcd(g)
            ([1], [13], [4])
            """
        cdef ZZ_pX *s, *t, *r
        _sig_on
        if plain:
            ZZ_pX_plain_xgcd(&r, &s, &t, self.x, other.x)
        else:
            ZZ_pX_xgcd(&r, &s, &t, self.x, other.x)            
        return (make_ZZ_pX(r), make_ZZ_pX(s), make_ZZ_pX(t))

    def degree(self):
        """
        Return the degree of this polynomial.  The degree of the 0
        polynomial is -1.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))                
            sage: f = ntl.ZZ_pX([5,0,1])
            sage: f.degree()
            2
            sage: f = ntl.ZZ_pX(range(100))
            sage: f.degree()
            99
            sage: f = ntl.ZZ_pX()
            sage: f.degree()
            -1
            sage: f = ntl.ZZ_pX([1])
            sage: f.degree()
            0
        """
        return ZZ_pX_degree(self.x)

    def leading_coefficient(self):
        """
        Return the leading coefficient of this polynomial.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))                
            sage: f = ntl.ZZ_pX([3,6,9])
            sage: f.leading_coefficient()
            9
            sage: f = ntl.ZZ_pX()
            sage: f.leading_coefficient()
            0
        """
        return make_ZZ_p(ZZ_pX_leading_coefficient(self.x))

    def constant_term(self):
        """
        Return the constant coefficient of this polynomial.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([3,6,9])
            sage: f.constant_term()
            3
            sage: f = ntl.ZZ_pX()
            sage: f.constant_term()
            0
        """
        cdef char* t
        t = ZZ_pX_constant_term(self.x)
        return int(string(t))

    def set_x(self):
        """
        Set this polynomial to the monomial "x".

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))                
            sage: f = ntl.ZZ_pX()
            sage: f.set_x()
            sage: f
            [0 1]
            sage: g = ntl.ZZ_pX([0,1])
            sage: f == g
            True

        Though f and g are equal, they are not the same objects in memory:
            sage: f is g
            False
            
        """
        ZZ_pX_set_x(self.x)

    def is_x(self):
        """
        True if this is the polynomial "x".

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX()
            sage: f.set_x()
            sage: f.is_x()
            True
            sage: f = ntl.ZZ_pX([0,1])
            sage: f.is_x()
            True
            sage: f = ntl.ZZ_pX([1])
            sage: f.is_x()
            False
        """
        return bool(ZZ_pX_is_x(self.x))

    def derivative(self):
        """
        Return the derivative of this polynomial.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([1,7,0,13])
            sage: f.derivative()
            [7 0 19]
        """
        return make_ZZ_pX(ZZ_pX_derivative(self.x))

    def factor(self, verbose=False):
        cdef ZZ_pX** v
        cdef long* e
        cdef long i, n
        _sig_on
        ZZ_pX_factor(&v, &e, &n, self.x, verbose)
        _sig_off
        F = []
        for i from 0 <= i < n:
            F.append((make_ZZ_pX(v[i]), e[i]))
        free(v)
        free(e)
        return F

    def linear_roots(self):
        """
        Assumes that input is monic, and has deg(f) distinct roots.
        Returns the list of roots.
        """
        cdef ZZ_p** v
        cdef long i, n
        _sig_on
        ZZ_pX_linear_roots(&v, &n, self.x)
        _sig_off
        F = []
        for i from 0 <= i < n:
            F.append(make_ZZ_p(v[i]))
        free(v)
        return F

    def reverse(self, hi=None):
        """
        Return the polynomial obtained by reversing the coefficients
        of this polynomial.  If hi is set then this function behaves
        as if this polynomial has degree hi.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))                
            sage: f = ntl.ZZ_pX([1,2,3,4,5])
            sage: f.reverse()
            [5 4 3 2 1]
            sage: f.reverse(hi=10)
            [0 0 0 0 0 0 5 4 3 2 1]
            sage: f.reverse(hi=2)
            [3 2 1]
            sage: f.reverse(hi=-2)
            []
        """
        if not (hi is None):
            return make_ZZ_pX(ZZ_pX_reverse_hi(self.x, int(hi)))
        else:
            return make_ZZ_pX(ZZ_pX_reverse(self.x))

    def truncate(self, long m):
        """
        Return the truncation of this polynomial obtained by
        removing all terms of degree >= m.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))                
            sage: f = ntl.ZZ_pX([1,2,3,4,5])
            sage: f.truncate(3)
            [1 2 3]
            sage: f.truncate(8)
            [1 2 3 4 5]
            sage: f.truncate(1)
            [1]
            sage: f.truncate(0)
            []
            sage: f.truncate(-1)
            []
            sage: f.truncate(-5)
            []
        """
        if m <= 0:
            return make_new_ZZ_pX()
        return make_ZZ_pX(ZZ_pX_truncate(self.x, m))

    def multiply_and_truncate(self, ntl_ZZ_pX other, long m):
        """
        Return self*other but with terms of degree >= m removed.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([1,2,3,4,5])
            sage: g = ntl.ZZ_pX([10])
            sage: f.multiply_and_truncate(g, 2)
            [10]
            sage: g.multiply_and_truncate(f, 2)
            [10]
        """
        if m <= 0:
            return make_new_ZZ_pX()
        return make_ZZ_pX(ZZ_pX_multiply_and_truncate(self.x, other.x, m))

    def square_and_truncate(self, long m):
        """
        Return self*self but with terms of degree >= m removed.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([1,2,3,4,5])
            sage: f.square_and_truncate(4)
            [1 4 10]
            sage: (f*f).truncate(4)
            [1 4 10]
        """
        if m < 0:
            return make_new_ZZ_pX()
        return make_ZZ_pX(ZZ_pX_square_and_truncate(self.x, m))

    def invert_and_truncate(self, long m):
        """
        Compute and return the inverse of self modulo $x^m$.
        The constant term of self must be 1 or -1.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: f = ntl.ZZ_pX([1,2,3,4,5,6,7])
            sage: f.invert_and_truncate(20)
            [1 18 1 0 0 0 0 8 17 2 13 0 0 0 4 0 17 10 9]
            sage: g = f.invert_and_truncate(20)
            sage: g * f
            [1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 8 4 4 3]
        """
        if m < 0:
            raise ArithmeticError, "m (=%s) must be positive"%m
        n = self.constant_term()
        if n != 1 and n != -1:
            raise ArithmeticError, \
                  "The constant term of self must be 1 or -1."
        _sig_on        
        return make_ZZ_pX(ZZ_pX_invert_and_truncate(self.x, m))
        
    def multiply_mod(self, ntl_ZZ_pX other, ntl_ZZ_pX modulus):
        """
        Return self*other % modulus.  The modulus must be monic with
        deg(modulus) > 0, and self and other must have smaller degree.
        
        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))        
            sage: modulus = ntl.ZZ_pX([1,2,0,1])    # must be monic
            sage: g = ntl.ZZ_pX([-1,0,1])
            sage: h = ntl.ZZ_pX([3,7,13])
            sage: h.multiply_mod(g, modulus)
            [10 6 4]
        """
        _sig_on
        return make_ZZ_pX(ZZ_pX_multiply_mod(self.x, other.x, modulus.x))

    def trace_mod(self, ntl_ZZ_pX modulus):
        """
        Return the trace of this polynomial modulus the modulus.
        The modulus must be monic, and of positive degree degree bigger
        than the the degree of self.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))                
            sage: f = ntl.ZZ_pX([1,2,0,3])
            sage: mod = ntl.ZZ_pX([5,3,-1,1,1])
            sage: f.trace_mod(mod)
            3
        """
        _sig_on
        return make_ZZ_p(ZZ_pX_trace_mod(self.x, modulus.x))

    def trace_list(self):
        """
        Return the list of traces of the powers $x^i$ of the
        monomial x modulo this polynomial for i = 0, ..., deg(f)-1.
        This polynomial must be monic.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(20))                
            sage: f = ntl.ZZ_pX([1,2,0,3,0,1])
            sage: f.trace_list()
            [5, 0, 14, 0, 10]

        The input polynomial must be monic or a ValueError is raised:
            sage: ntl.set_modulus(ntl.ZZ(20))                
            sage: f = ntl.ZZ_pX([1,2,0,3,0,2])
            sage: f.trace_list()
            Traceback (most recent call last):
            ...
            ValueError: polynomial must be monic.
        """
        if not self.is_monic():
            raise ValueError, "polynomial must be monic."
        _sig_on
        cdef char* t
        t = ZZ_pX_trace_list(self.x)
        return eval(string(t).replace(' ', ','))

    def resultant(self, ntl_ZZ_pX other):
        """
        Return the resultant of self and other.
        
        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))                
            sage: f = ntl.ZZ_pX([17,0,1,1])
            sage: g = ntl.ZZ_pX([34,-17,18,2])
            sage: f.resultant(g)
            0
        """
        _sig_on
        return make_ZZ_p(ZZ_pX_resultant(self.x, other.x))

    def norm_mod(self, ntl_ZZ_pX modulus):
        """
        Return the norm of this polynomial modulo the modulus.  The
        modulus must be monic, and of positive degree strictly greater
        than the degree of self. 


        EXAMPLE:
            sage: ntl.set_modulus(ntl.ZZ(17))                
            sage: f = ntl.ZZ_pX([1,2,0,3])
            sage: mod = ntl.ZZ_pX([-5,2,0,0,1])
            sage: f.norm_mod(mod)
            11

        The norm is the constant term of the characteristic polynomial.
            sage: f.charpoly_mod(mod)
            [11 1 8 14 1]
        """
        _sig_on
        return make_ZZ_p(ZZ_pX_norm_mod(self.x, modulus.x))

    def discriminant(self):
        r"""
        Return the discriminant of a=self, which is by definition
        $$
                (-1)^{m(m-1)/2} {\mbox{\tt resultant}}(a, a')/lc(a),
        $$      
        where m = deg(a), and lc(a) is the leading coefficient of a.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))                
            sage: f = ntl.ZZ_pX([1,2,0,3])
            sage: f.discriminant()
            1
        """
        cdef long m
        
        c = ~self.leading_coefficient()
        m = self.degree()
        if (m*(m-1)/2) % 2:
            c = -c
        return c*self.resultant(self.derivative())

    def charpoly_mod(self, ntl_ZZ_pX modulus):
        """
        Return the characteristic polynomial of this polynomial modulo
        the modulus.  The modulus must be monic of degree bigger than
        self.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))                
            sage: f = ntl.ZZ_pX([1,2,0,3])
            sage: mod = ntl.ZZ_pX([-5,2,0,0,1])
            sage: f.charpoly_mod(mod)
            [11 1 8 14 1]
        """
        _sig_on
        return make_ZZ_pX(ZZ_pX_charpoly_mod(self.x, modulus.x))

    def minpoly_mod(self, ntl_ZZ_pX modulus):
        """
        Return the minimal polynomial of this polynomial modulo the
        modulus.  The modulus must be monic of degree bigger than
        self.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))
            sage: f = ntl.ZZ_pX([0,0,1])
            sage: g = f*f
            sage: f.charpoly_mod(g)
            [0 0 0 0 1]

        However, since $f^2 = 0$ moduluo $g$, its minimal polynomial
        is of degree $2$.
            sage: f.minpoly_mod(g)
            [0 0 1]
        """
        _sig_on
        return make_ZZ_pX(ZZ_pX_minpoly_mod(self.x, modulus.x))

    def clear(self):
        """
        Reset this polynomial to 0.  Changes this polynomial in place.

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))        
            sage: f = ntl.ZZ_pX([1,2,3])
            sage: f
            [1 2 3]
            sage: f.clear()
            sage: f
            []
        """
        ZZ_pX_clear(self.x)

    def preallocate_space(self, long n):
        """
        Pre-allocate spaces for n coefficients.  The polynomial that f
        represents is unchanged.  This is useful if you know you'll be
        setting coefficients up to n, so memory isn't re-allocated as
        the polynomial grows.  (You might save a millisecond with this
        function.)

        EXAMPLES:
            sage: ntl.set_modulus(ntl.ZZ(17))        
            sage: f = ntl.ZZ_pX([1,2,3])
            sage: f.preallocate_space(20)
            sage: f
            [1 2 3]
            sage: f[10]=5  # no new memory is allocated
            sage: f
            [1 2 3 0 0 0 0 0 0 0 5]
        """
        _sig_on
        ZZ_pX_preallocate_space(self.x, n)
        _sig_off


## TODO: NTL's ZZ_pX has minpolys of linear recurrence sequences!!!

    
cdef make_ZZ_pX(ZZ_pX* x):
    cdef ntl_ZZ_pX y
    _sig_off
    y = ntl_ZZ_pX()
    y.x = x
    return y

cimport ntl_ZZ_p
import ntl_ZZ_p
cdef make_ZZ_p(ZZ_p* x):
    cdef ntl_ZZ_p.ntl_ZZ_p y
    _sig_off
    y = ntl_ZZ_p.ntl_ZZ_p()
    y.set(<void*> x)
    return y

def make_new_ZZ_pX(v=[]):
    s = str(v).replace(',',' ').replace('L','')
    cdef ntl_ZZ_pX z
    z = ntl_ZZ_pX()
    _sig_on
    z.x = str_to_ZZ_pX(s)
    _sig_off
    return z
