r"""
Matrices

Elements of matrix spaces are of class \code{Matrix}.  They can be
either sparse or dense, and can be defined over any base ring.

EXAMPLES:

We create the $2\times 3$ matrix
$$
   \left(\begin{matrix} 1&2&3\\4&5&6 \end{matrix}\right)
$$
as an element of a matrix space over $\Q$:

    sage: M = MatrixSpace(RationalField(),2,3)
    sage: A = M([1,2,3, 4,5,6])
    sage: A
    [1 2 3]
    [4 5 6]
    sage: A.parent()
    Full MatrixSpace of 2 by 3 dense matrices over Rational Field

We next change the top-right entry of $A$.  Note that matrix indexing
is $0$-based in SAGE, so the top right entry is $(0,2)$, which should
be thought of as ``row number $0$, column number 2''.

    sage: A[0,2] = 389
    sage: A
    [  1   2 389]
    [  4   5   6]

Also notice how matrices print.  All columns have the same width and
entries in a given column are right justified.   Next we compute the
reduced row echelon form of $A$.

    sage: A.echelon_form()
    [      1       0 -1933/3]
    [      0       1  1550/3]


\subsection{Design} Matrices are either immutable or not.  When
initially created, matrices are typically mutable, so one can change
their entries.  However, nothing about mutable matrices is cached.
Once a matrix $A$ is made immutable using \code{A.set_immutable()} the
entries of $A$ cannot be changed.  However, properies of $A$ such as
its rank, characteristic polynomial, etc., are all cached so
computations involving $A$ may be more efficient.  Once $A$ is made
immutable it cannot be changed back.  However, one can obtain a
mutable copy of $A$ using \code{A.copy()}.

The echelon form method always returns immutable matrices with known
rank.
"""


#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import sage.rings.arith 
import sage.misc.misc as misc
import sage.misc.latex as latex
import sage.matrix.dense_matrix_pyx as dense_matrix_pyx
import sage.matrix.sparse_matrix
import sage.modules.free_module_element
import sage.modules.free_module
import sage.matrix.matrix_space
import matrix_space
import sage.libs.pari
import sage.modules.module_element  as module_element
import berlekamp_massey
import sage.rings.polynomial_ring as polynomial_ring
import sage.rings.integer_ring as integer_ring
import sage.rings.rational_field as rational_field
import sage.rings.number_field as number_field
import sage.rings.coerce as coerce

cputime = misc.cputime

def is_Matrix(x):
    return isinstance(x, Matrix)

class Matrix(module_element.ModuleElement):
    r"""
    The \class{Matrix} class is the base class for all matrix
    classes.  To create a \class{Matrix}, first create a
    \class{MatrixSpace}, then coerce a list of elements into the
    \class{MatrixSpace}.  See the documentation of
    \class{MatrixSpace} for more details.
    """
    def __init__(self, parent):
        module_element.ModuleElement.__init__(self, parent)
        self.__nrows = parent.nrows()
        self.__ncols = parent.ncols()
        self.__is_immutable = False

    def set_immutable(self):
        self.__is_immutable = True

    def is_immutable(self):
        return self.__is_immutable
        
    def __repr__(self):
        nr = self.nrows()
        nc = self.ncols()
        if nr == 0 or nc == 0:
            return "[]"
        # compute column widths
        S = [str(x) for x in self.list()]
        width = max([len(x) for x in S])
        rows = []
        m = 0
        
        # compute rows
        for r in xrange(nr):
            s = ""
            for c in xrange(nc):
                if c == nc-1:
                    sep=""
                else:
                    sep=" "
                entry = S[r*nc+c]
                if c == 0:
                    m = max(m, len(entry))
                entry = " "*(width-len(entry)) + entry
                s += entry + sep
            rows.append(s)
            
        # Remove leading spaces
##        n = width-m
##        if n > 0:
##            for r in xrange(nr):
##                rows[r] = rows[r][n:]
        
        # Put brackets around in a single string
        s = "\n".join(["[%s]"%row for row in rows])
        return s
        
    def _latex_sparse(self, variable="x"):
        """
        Return a latex string that represents this matrix as a sparse
        matrix.  The rows are printed as sums sum a_i x_i, where
        x is the variable.
        """
        nr = self.nrows()
        nc = self.ncols()
        s = "\\left(\\begin{align*}\n"
        for i in xrange(nr):
            v = [(j, self[i,j]) for j in xrange(nc) if self[i,j] != 0]
            for j in xrange(len(v)):
                s += "%s*%s_{%s}"%(v[j][1], variable, v[j][0])
                if j == 0:
                    s += "& + "
                elif j < len(v) - 1:
                    s += " + "
                else:
                    s += "\\\\\n"
        s += "\n\\end{align*}"
        return s
        
    def _latex_(self):
        nr = self.nrows()
        nc = self.ncols()
        if nr == 0 or nc == 0:
            return "()"
        # compute column widths
        S = [str(x) for x in self.list()]
        width = max([len(x) for x in S])
        S = self.list()
        rows = []
        m = 0
        
        # compute rows
        for r in xrange(nr):
            s = ""
            for c in xrange(nc):
                if c == nc-1:
                    sep=""
                else:
                    sep="&"
                entry = latex.latex(S[r*nc+c])
                if c == 0:
                    m = max(m, len(entry))
                s += entry + sep
            rows.append(s)
            
        # Put brackets around in a single string
        s = "\\\\\n".join(["%s"%row for row in rows])
        return "\\left(\\begin{array}{%s}\n"%('r'*nc) + s + "\n\\end{array}\\right)"

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash(self.__str__())

    def _require_mutable(self):
        if self.__is_immutable:
            raise TypeError, "matrix is immutable; please change a copy instead."
    
    ###################################################
    ## PARI
    ###################################################
    def _pari_(self):
        return sage.libs.pari.pari.matrix(self.nrows(), self.ncols(), self.list())


    ###################################################
    ## Access:
    ##   These *must* be overridden by each subclass
    ###################################################
    def __getitem__(self, ij):
        """
        INPUT:
            A[i, j] -- the i,j of A, and
            A[i]    -- the i-th row of A.
        """
        raise NotImplementedError
    
    def get(self, ij):
        """
        Entry access, but possibly without bounds checking (for
        efficiency).
        INPUT:
            A[i, j] -- the i,j of A, and
            A[i]    -- the i-th row of A.
        """
        return self[ij]
    
    def __setitem__(self, ij, x):
        """
        INPUT:
            A[i, j] = value -- set the (i,j) entry of A
            A[i] = value    -- set the ith row of A
        """
        raise NotImplementedError
        
    def set(self, ij, x):
        self[ij] = x
        
    ###################################################
    ## Convenience
    ###################################################
    def matrix_space(self, nrows=None, ncols=None, sparse=None):
        if nrows == None: nrows = self.nrows()
        if ncols == None: ncols = self.ncols()
        if sparse is None:
            sparse = sparse=self.is_sparse()
        return self.parent().matrix_space(nrows, ncols, sparse=sparse)
        
    def new_matrix(self, nrows=None, ncols=None, entries=0, 
                coerce_entries=True, copy=True):
        """
        Create a matrix in the parent of this space with the given
        number of rows, columns, etc.

        WARNING: This function called with no arguments returns the 0
        matrix by default, not the matrix self.
        """
        return self.matrix_space(nrows, ncols).matrix(
                entries, coerce_entries, copy)
        
    ###################################################
    ## Properties
    ###################################################
    def ncols(self):
        return self.__ncols
    
    def nrows(self):
        return self.__nrows
    

    ###################################################
    ## Functions
    ###################################################
    def add_multiple_of_row(self, i, j, s):
        """
        Replace row i by s times row j.
        """
        self._require_mutable()
        s = self.base_ring()(s)
        for c in xrange(self.ncols()):
            self[i,c] += s*self[j,c]
            
    def add_multiple_of_column(self, i, j, s):
        """
        Replace column i by s times column j.
        """
        self._require_mutable()
        s = self.base_ring()(s)        
        for r in xrange(self.nrows()):
            self[r,i] += s*self[r,j]
    
    def augment(self, other):
        """
        Return the augmented matrix of the form [self | other].
        """
        if not isinstance(other, Matrix):
            raise TypeError, "other must be a Matrix"
        if self.base_ring() != other.base_ring():
            raise TypeError, "base rings must be the same"
        if self.nrows() != other.nrows():
            raise TypeError, "number of rows must be the same"
        Z = self.new_matrix(ncols = self.ncols() + other.ncols())
        for r in xrange(self.nrows()):
            for c in xrange(self.ncols()):
                Z[r,c] = self[r,c]
        nc = self.ncols()
        for r in xrange(other.nrows()):
            for c in xrange(other.ncols()):
                Z[r,c + nc] = other[r,c]
        return Z
        
    def block_sum(self, other):
        """
        Return the block matrix that has self and other on the diagonal:
        [self |    0  ]
        [  0  | other ]
        """
        if not isinstance(other, Matrix):
            raise TypeError, "other must be a Matrix"
        top = self.augment(self.new_matrix(ncols=other.ncols()))
        bottom = other.new_matrix(ncols=self.ncols()).augment(other)
        return top.stack(bottom)
        
    def change_ring(self, ring):
        """
        Return the matrix obtained by coercing the entries of this
        matrix into the given ring.
        """
        if ring == self.base_ring():
            return self
        M = sage.matrix.matrix_space.MatrixSpace(ring, self.__nrows, self.__ncols)
        return M(self.list())
        
    def columns(self):
        """
        Returns the list of columns of self, as vectors.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(), 2,3) (xrange(6))
            sage: A
            [0 1 2]
            [3 4 5]
            sage: A.columns()
            [(0, 3), (1, 4), (2, 5)]
        """
        return self.transpose().rows()
        
    def commutator(self, other):
        """
        Return the commutator self*other - other*self.
        """
        return self*other - other*self
      
    def copy(self):
        """
        Return a copy of this matrix.  Changing the entries of the
        copy will not change the entries of this matrix. 
        """
        return self.new_matrix(entries=self._entries(), coerce_entries=False)
        
    def dense_matrix(self):
        """
        If this matrix is sparse, return a dense matrix with the same
        entries.  If this matrix is dense, return this matrix (not a
        copy).

        NOTE: The definition of"dense" and "sparse" in SAGE have
        nothing to do with the number of nonzero entries.  Sparse
        and dense are properties of the underlying representation
        of the matrix.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),2, sparse=True)([1,2,0,1])
            sage: A.is_sparse()
            True
            sage: B = A.dense_matrix()
            sage: B.is_sparse()
            False
            sage: A*B
            [1 4]
            [0 1]
            sage: A.parent()
            Full MatrixSpace of 2 by 2 sparse matrices over Rational Field
            sage: B.parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: (A*B).parent()
            Full MatrixSpace of 2 by 2 sparse matrices over Rational Field
            sage: (B*A).parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
        """
        if self.is_dense():
            return self
        M = self.matrix_space(self.nrows(), self.ncols(), sparse=False)
        return M(self.list())
        
    def sparse_matrix(self):
        """
        If this matrix is dense, return a sparse matrix with
        the same entries.  If this matrix is sparse, return this
        matrix (not a copy). 

        NOTE: The definition of "dense" and "sparse" in SAGE have
        nothing to do with the number of nonzero entries.  Sparse
        and dense are properties of the underlying representation
        of the matrix.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),2, sparse=False)([1,2,0,1])
            sage: A.is_sparse()
            False
            sage: B = A.sparse_matrix()
            sage: B.is_sparse()
            True
            sage: A
            [1 2]
            [0 1]
            sage: B
            [1 2]
            [0 1]
            sage: A*B
            [1 4]
            [0 1]
            sage: A.parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: B.parent()
            Full MatrixSpace of 2 by 2 sparse matrices over Rational Field
            sage: (A*B).parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
            sage: (B*A).parent()
            Full MatrixSpace of 2 by 2 sparse matrices over Rational Field
        """
        if self.is_sparse():
            return self
        M = self.matrix_space(self.nrows(), self.ncols(), sparse=True)
        return M(self._dict())

    def dense_row(self, n):
        if n < 0 or n >= self.nrows():
            raise IndexError, "n must be between 0 and the number of rows"
        V = sage.modules.free_module.VectorSpace(\
                 self.base_ring(), self.ncols(), sparse=False)
        return V([self[n, i] for i in xrange(self.ncols())])
        
    def _dict(self):
        try:
            return self.__dict
        except AttributeError:
            v = {}
            for ij in self.nonzero_positions():
                v[ij] = self.get(ij)
            if self.is_immutable():
                self.__dict = v
            return v

    def _entries(self):
        """
        This must be defined in the derived class.  It is the
        underlying representation of elements, which may be a list or
        dict or something else.
        """
        raise NotImplementedError
        
    def is_dense(self):
        return self.parent().is_dense()
        
    def is_sparse(self):
        return self.parent().is_sparse()
        
    def is_square(self):
        return self.nrows() == self.ncols()
        
    def list(self):
        return [self[i,j] for i in xrange(self.nrows()) for j in xrange(self.ncols())]

    def nonpivots(self):
        """
        Return the list of i such that the i-th column of self 
        is NOT a pivot column of the reduced row echelon form 
        of self.
        
        OUTPUT:
            list -- sorted list of integers
        """
        X = set(self.pivots())
        return [j for j in xrange(self.ncols()) if not (j in X)]
        
    def nonzero_positions(self):
        """
        Returns the set of pairs (i,j) such that self[i,j] != 0.
        """
        z = self.base_ring()(0)
        return set([(i,j) for i in xrange(self.nrows()) \
                          for j in xrange(self.ncols()) \
                          if self[i,j] != z])
        
    def nonzero_positions_in_column(self, i):
        """
        Return the integers j such that self[j,i] is nonzero, i.e.,
        such that the j-th position of the i-th column is nonzero.
        """
        z = self.base_ring()(0)
        return set([j for j in xrange(self.nrows()) if self[j,i] != z])
        
    def nonzero_positions_in_row(self, i):
        """
        Return the integers j such that self[i,j] is nonzero, i.e.,
        such that the j-th position of the i-th row is nonzero.
        """
        z = self.base_ring()(0)        
        return set([j for j in xrange(self.ncols()) if self[i,j] != z])
        
    def rescale_row(self, i, s):
        """
        Replace i-th row of self by s times i-th row of self.
        """
        if s == 1: return
        self._require_mutable()
        for j in xrange(self.ncols()):
            self[i,j] *= s
        
    def row(self, i):
        V = sage.modules.free_module.FreeModule(self.base_ring(), self.ncols())
        return V([self[i,j] for j in xrange(self.ncols())])
        
    def rows(self):
        return [self.row(i) for i in xrange(self.nrows())]

    def linear_combination_of_rows(self, v):
        R = self.rows()
        return sum([v[i]*R[i] for i in xrange(len(v))], 0)
    
    def sparse_columns(self):
        try:
            return self.__sparse_columns
        except AttributeError:
            C = [[] for _ in xrange(len(self.ncols()))]
            for i, j in self.nonzero_positions():
                x = self.get(i,j)
                C[j].append((i,x))
            if self.is_immutable():
                self.__sparse_columns = C
            return C
            
    def sparse_rows(self):
        try:
            return self.__sparse_rows
        except AttributeError:
            R = [[] for _ in xrange(len(self.ncols()))]
            for i, j in self.nonzero_positions():
                x = self.get(i,j)
                R[i].append((j,x))
            if self.is_immutable():
                self.__sparse_rows = R
            return R
        
    def stack(self, other):
        """
        Return the augmented matrix self on top of other:
           [ self  ]
           [ other ]

        EXAMPLES:
            sage: M = Matrix(RationalField(), 2, 3, range(6))
            sage: N = Matrix(RationalField(), 1, 3, [10,11,12])
            sage: M.stack(N)
            [ 0  1  2]
            [ 3  4  5]
            [10 11 12]
        """
        if not isinstance(other, Matrix):
            raise TypeError, "other must be a matrix"
        if self.base_ring() != other.base_ring():
            raise TypeError, "base rings must be the same"
        if self.ncols() != other.ncols():
            raise TypeError, "number of columns must be the same"
        Z = self.new_matrix(nrows = self.nrows() + other.nrows())
        for r in xrange(self.nrows()):
            for c in xrange(self.ncols()):
                Z[r,c] = self[r,c]
        nr = self.nrows()
        for r in xrange(other.nrows()):
            for c in xrange(other.ncols()):
                Z[r+nr,c] = other[r,c]
        return Z

    def submatrix_from_columns(self, columns):
        """
        Return the submatrix of self of columns col[i] for i in
        the list of columns.
        """
        if not isinstance(columns, list):
            raise TypeError, "columns (=%s) must be a list of integers"%columns
        A = self.new_matrix(ncols = len(columns))
        k = 0
        for i in columns:
            i = int(i)
            #if not isinstance(i, int):
            #    raise TypeError, "each element of columns must be an int"
            if i < 0 or i >= self.ncols():
                raise IndexError, "column %s out of range"%i
            for r in xrange(self.nrows()):
                A[r,k] = self[r,i]
            k += 1
        return A
        
    def submatrix_from_rows(self, rows):
        """
        Return the submatrix of self of rows row[i] for i in
        the list of rows.
        """
        if not isinstance(rows, list):
            raise TypeError, "rows must be a list of integers"
        A = self.new_matrix(nrows = len(rows))
        k = 0
        for i in rows:
            #if not isinstance(i, int):
            #    raise TypeError, "each element of rows must be an int"
            i = int(i)
            if i < 0 or i >= self.nrows():
                raise IndexError, "row %s out of range"%i
            for c in xrange(self.ncols()):
                A[k,c] = self[i,c]
            k += 1
        return A
        
    def swap_columns(self, c1, c2):
        """
        Swap columns c1 and c2 of self.
        """
        self._require_mutable()
        nc = self.ncols()
        if c1 < 0 or c1 >= nc:
            raise IndexError, "c1 must satisfy 0<=c1<%s"%nc
        if c2 < 0 or c2 >= nc:
            raise IndexError, "c2 must satisfy 0<=c2<%s"%nc
        if c1 == c2: return
        for r in xrange(self.nrows()):
            a, b = self.get((r,c2)), self.get((r,c1))
            self.set((r,c1), a)
            self.set((r,c2), b)
        
    def swap_rows(self, r1, r2):
        """
        Swap rows r1 and r2 of self.
        """
        self._require_mutable()
        nr = self.nrows()
        if r1 < 0 or r1 >= nr:
            raise IndexError, "r1 must satisfy 0<=r1<%s"%nr
        if r2 < 0 or r2 >= nr:
            raise IndexError, "r2 must satisfy 0<=r2<%s"%nr
        if r1 == r2: return
        for c in xrange(self.ncols()):
            a, b = self.get((r2,c)), self.get((r1,c))
            self.set((r1,c), a)
            self.set((r2,c), b)
        
    def trace(self):
        """
        Return the trace of self, which is the sum of the
        diagonal entries of self.
        INPUT:
            self -- a square matrix
        OUTPUT:
            element of the base ring of self
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "matrix must be square"
        R = self.base_ring()
        return misc.add([self[i,i] for i in xrange(self.nrows())],R(0))
        
    def vector_matrix_multiply(self, v):
        """
        Returns the vector times matrix product.

        INPUT:
             v -- a free module element.

        OUTPUT:
            The the vector times matrix product v*A.

        EXAMPLES:
            sage: MS = MatrixSpace(RationalField(), 2,2)
            sage: B = MS.matrix([1,2,1,2])
            sage: V = VectorSpace(RationalField(), 2)
            sage: v = V([1,2])
            sage: B.vector_matrix_multiply(v)     # computes v*B
            (3, 6)
            sage: Bt = B.transpose()
            sage: Bt.vector_matrix_multiply(v)    # computes B*v
            (5, 5)
        """
        M = sage.modules.free_module.FreeModule(self.base_ring(), self.ncols())
        if not isinstance(v, sage.modules.free_module_element.FreeModuleElement):
            v = M(v)
        if self.base_ring() != v.base_ring():
            raise ArithmeticError, "base rings must be equal"
        if self.nrows() != v.degree():
            raise ArithmeticError, "number of rows of matrix must equal degree of vector"
        s = M(0)
        for i in xrange(self.nrows()):
            if v[i] != 0:
                s += v[i]*self.row(i)
        return s

    def iterates(self, v, n):
        """
        Let $A$ be this matrix and $v$ be a free module element.
        Return a vector whose rows are the entries of the following
        vectors:
        $$
           v, v A, v A^2, \ldots, v A^{n-1}.
        $$

        INPUT:
            v -- free module element
            n -- nonnegative integer

        EXAMPLES:
            sage: A = MatrixSpace(IntegerRing(), 2)([1,1,3,5]); A
            [1 1]
            [3 5]
            sage: v = FreeModule(IntegerRing(), 2)([1,0])
            sage: A.iterates(v,0)
            []
            sage: A.iterates(v,5)
            [  1   0]
            [  1   1]
            [  4   6]
            [ 22  34]
            [124 192]
        """
        n = int(n)
        if n >= 2 and self.nrows() != self.ncols():
            raise ArithmeticError, "matrix must be square if n >= 2."
        if n == 0:
            return self.matrix_space(n, self.ncols())(0)
        M = sage.modules.free_module.FreeModule(self.base_ring(), self.ncols())
        if not isinstance(v, sage.modules.free_module_element.FreeModuleElement):
            v = M(v)
        X = [v]
        for _ in range(n-1):
            X.append(X[len(X)-1]*self)
        MS = self.matrix_space(n, self.ncols())
        return MS(X)

        
        
    ###################################################
    ## Arithmetic
    ###################################################
    def __abs__(self):
        return self.determinant()

    def __is_compatible(self, other):
        return isinstance(other, Matrix) and self.parent() == other.parent()
    
    def _add(self, right):
        A = self.new_matrix()
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                A[i,j] = self[i,j] + right[i,j]
        return A
            
    def __cmp__(self, right):
        if not isinstance(right, Matrix) or right.parent() != self.parent():
            return coerce.cmp(self, right)
        return cmp(self._entries(), right._entries())

    def _div(self, right):
        return self*(~right)
        
    def __mod__(self, p):
        raise NotImplementedError
    
    def _scalar_multiply(self, x):
        if not x in self.base_ring():
            x = self.base_ring()(x)
        M = self.new_matrix()
        nc = self.ncols()
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                M[i,j] = x * self[i,j]
        return M
    
    def _right_scalar_multiply(self, x):
        if not x in self.base_ring():
            x = self.base_ring()(x)
        M = self.new_matrix()
        nc = self.ncols()
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                M[i,j] = self[i,j] * x
        return M

    def __mul__(self, right):
        if not isinstance(right, Matrix):
            return self._right_scalar_multiply(right)
        if self.base_ring() != right.base_ring():
            raise ArithmeticError, "base rings must be equal"
        if self.ncols() != right.nrows():
            raise ArithmeticError, "number of columns of self (=%s) must equal number of rows of right (=%s)."%(
                self.ncols(), right.nrows())
        nr = self.nrows()
        snc = self.ncols()
        nc = right.ncols()
        M = self.new_matrix(nr, nc)
        if snc == 0: return M
        zero = self.base_ring()(0)
        for i in xrange(nr):
            for j in xrange(nc):
                M[i,j] = misc.add([self[i,k]*right[k,j] for k in range(snc)],zero)
        return M
        
    def __neg__(self):
        return self.__rmul__(self.base_ring()(-1))
        
    def __pos__(self):
        return self
        
    def __pow__(self, n):
        """
        EXAMPLES:
            sage: MS = MatrixSpace(RationalField(), 3, 3)
            sage: A = MS([0, 0, 1, 1, 0, '-2/11', 0, 1, '-3/11'])
            sage: A * A**(-1) == 1
            True
            sage: A**4
            [      -3/11     -13/121   1436/1331]
            [    127/121   -337/1331 -4445/14641]
            [    -13/121   1436/1331 -8015/14641]
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "self must be square"
        #if not isinstance(n, int):
        #    raise TypeError, "n must be an int"
        n = int(n)
        if n == 0:
            return self.parent()(1)
        if n < 0:
            return (~self)**(-n)
        ans = self.parent()(1)
        apow = self
        while n != 0:
            if n%2 != 0:
                ans = ans * apow
            n /= 2
            if n == 0:  # to not waste time doing an extra multiplication/increment
                break
            apow = apow * apow
        return ans
        
    def __rmul__(self, left):        
        A = self.new_matrix()
        for i, j in self.nonzero_positions():
            A[i,j] = left*self[i,j]
        return A
        
    def _sub(self, right):
        A = self.new_matrix()
        for i in xrange(self.nrows()):
            for j in xrange(self.ncols()):
                A[i,j] = self[i,j] - right[i,j]
        return A
 
    

#############################################
## Generic matrices over an integral domain
############################################# 

class Matrix_domain(Matrix):
    def __init__(self, parent):
        Matrix.__init__(self, parent)

    def charpoly(self):
        r"""
        Return the characteristic polynomial of self, as a polynomial
        over the fraction field of the base ring.

        EXAMPLES:
        First a matrix over $\Z$:
            sage: A = MatrixSpace(IntegerRing(),2)( [[1,2], [3,4]] )
            sage: f = A.charpoly()
            sage: f
            x^2 - 5*x - 2
            sage: f.parent()
            Univariate Polynomial Ring in x over Rational Field

        We compute the characteristic polynomial of a matrix over
        the polynomial ring $\Z[a]$:
            sage: R = PolynomialRing(IntegerRing(),'a'); a = R.gen()
            sage: M = MatrixSpace(R,2)([[a,1], [a,a+1]])
            sage: M
            [    a     1]
            [    a a + 1]
            sage: f = M.charpoly()
            sage: f
            x^2 + (-2*a - 1)*x + a^2
            sage: f.parent()
            Univariate Polynomial Ring in x over Fraction field of Univariate Polynomial Ring in a over Integer Ring
            sage: M.trace()
            2*a + 1
            sage: M.determinant()
            a^2

        We compute the characteristic polynomial of a matrix over the
        multi-variate polynomial ring $\Z[x,y]$:
            sage: R = MPolynomialRing(IntegerRing(),2); x,y = R.gens()
            sage: A = MatrixSpace(R,2)([x,y,x**2,y**2])
            sage: f = A.charpoly()
            sage: f
            x^2 + (-1*x_1^2 - x_0)*x + x_0*x_1^2 - x_0^2*x_1
            
        It's a little difficult to distinguish the variables.  To fix this,
        we rename the indeterminate $Z$:
            sage: f.parent().assign_names("Z")
            sage: f
            Z^2 + (-1*x_1^2 - x_0)*Z + x_0*x_1^2 - x_0^2*x_1
        """
        f = self.matrix_over_field().charpoly()
        return f

    def determinant(self):
        """
        Return the determinant of this matrix, which must be square.

        EXAMPLES:
        We create a matrix over $\Z[x,y]$ and compute its determinant.
            sage: R = MPolynomialRing(IntegerRing(),2); x,y = R.gens()
            sage: A = MatrixSpace(R,2)([x,y,x**2,y**2])
            sage: A.determinant()
            x_0*x_1^2 - x_0^2*x_1
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "Matrix must be square, but is %sx%s"%(
                self.nrows(), self.ncols())
        # Use stupid slow but completely general method.  
        d = (-1)**self.nrows() * self.charpoly()[0]
        return self.base_ring()(d)

    # shortcut
    det = determinant  

    def is_invertible(self):
        r"""
        Return True if this matrix is invertible.

        EXAMPLES:
        The following matrix is invertible over $\Q$ but not over $\Z$.
            sage: A = MatrixSpace(IntegerRing(), 2)(range(4))
            sage: A.is_invertible()
            False
            sage: A.matrix_over_field().is_invertible()
            True

        The inverse function is a constructor for matrices over the
        fraction field, so it can work even if A is not invertible.
            sage: ~A   # inverse of A
            [-3/2  1/2]
            [   1    0]

        The next matrix is invertible over $\Z$.
            sage: A = MatrixSpace(IntegerRing(),2)([1,10,0,-1])
            sage: A.is_invertible()
            True
            sage: ~A                # compute the inverse
            [ 1 10]
            [ 0 -1]

        The following nontrivial matrix is invertible over $\Z[x]$.
            sage: R = PolynomialRing(IntegerRing())
            sage: x = R.gen()
            sage: A = MatrixSpace(R,2)([1,x,0,-1])
            sage: A.is_invertible()
            True
            sage: ~A
            [ 1  x]
            [ 0 -1]
        """
        return self.is_square() and self.determinant().is_unit()

    def __invert__(self):
        r"""
        Return this inverse of this matrix, as a matrix over the fraction field.

        Raises a \code{ZeroDivisionError} if the matrix has zero
        determinant, and raises an \code{ArithmeticError}, if the
        inverse doesn't exist because the matrix is nonsquare.

        EXAMPLES:
            sage: A = MatrixSpace(IntegerRing(), 2)([1,1,3,5])
            sage: ~A
            [ 5/2 -1/2]
            [-3/2  1/2]

        Even if the inverse lies in the base field, the result is still a matrix
        over the fraction field.
            sage: I = MatrixSpace(IntegerRing(),2)( 1 )  # identity matrix
            sage: ~I
            [1 0]
            [0 1]
            sage: (~I).parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
        """
        return ~self.matrix_over_field()
            

    def matrix_over_field(self):
        """
        Return this matrix, but with entries viewed as elements
        of the fraction field of the base ring.

        EXAMPLES:
            sage: A = MatrixSpace(IntegerRing(),2)([1,2,3,4])
            sage: B = A.matrix_over_field()
            sage: B
            [1 2]
            [3 4]
            sage: B.parent()
            Full MatrixSpace of 2 by 2 dense matrices over Rational Field
        """
        return self.change_ring(self.base_ring().fraction_field())

    
#############################################
## Generic matrices over a PID
############################################# 
class Matrix_pid(Matrix_domain):
    def __init__(self, parent):
        Matrix_domain.__init__(self, parent)

    def column_module(self):
        """
        Return the free module over the base ring spanned by the
        columns of this matrix.
        """
        return self.transpose().row_space()
        
    def decomposition(self, is_diagonalizable=False, dual=False):
        """
        Returns the decomposition of the free module on which this
        matrix acts from the right, along with whether this matrix
        acts irreducibly on each factor.  The factors are guaranteed
        to be sorted in the same way as the corresponding factors of
        the characteristic polynomial.
        
        Let A be the matrix acting from the on the vector space V of
        column vectors.  Assume that A is square.  This function
        computes maximal subspaces W_1, ..., W_n corresponding to
        Galois conjugacy classes of eigenvalues of A.  More precisely,
        let f(X) be the characteristic polynomial of A.  This function
        computes the subspace $W_i = ker(g_(A)^n)$, where g_i(X) is an
        irreducible factor of f(X) and g_i(X) exactly divides f(X).
        If the optional parameter is_diagonalizable is True, then we
        let W_i = ker(g(A)), since then we know that ker(g(A)) =
        $ker(g(A)^n)$.

        If dual is True, also returns the corresponding decomposition
        of V under the action of the transpose of A.  The factors are
        guarenteed to correspond.
        
        OUTPUT:
            list -- list of pairs (V,t), where V is a vector spaces
                    and t is a bool, and t is True exactly when the
                    charpoly of self on V is irreducible.
                    
            (optional) list -- list of pairs (W,t), where W is a vector
                    space and t is a bool, and t is True exactly
                    when the charpoly of the transpose of self on W
                    is irreducible.
        """
        if not self.is_square():
            raise ArithmeticError, "self must be a square matrix"
        
        f = self.charpoly()
        E = []
        
        # Idea: For optimization, could compute powers of self
        #       up to max degree of any factor.  Then get g(self)
        #       by taking a linear combination.   ??????

        if dual:
            Edual = []
        F = f.factor()
        if len(F) == 1:
            V = sage.modules.free_module.FreeModule(
                              self.base_ring(), self.nrows()) 
            m = F[0][1]
            if dual:
                return [(V,m==1)], [(V,m==1)]
            else:
                return [(V,m==1)]
        F.sort()
        for g, m in f.factor():
            if is_diagonalizable:
                B = g(self)
            else:
                B = g(self) ** m
            E.append((B.kernel(), m==1))
            if dual:
                Edual.append((B.transpose().kernel(), m==1))
        if dual:
            return E, Edual
        return E

    def echelon_form(self, include_zero_rows=True):
        """
        Return the echelon form of this matrix over the integers.

        This is a matrix over the base ring (a PID) which is, \emph{by
        definition}, what is also called the Hermite normal form.
        """
        raise NotImplementedError


    def image(self):
        """
        Return the image of the homomorphism on rows defined by this matrix.
        """
        return self.row_module()
    
    def row_module(self):
        """
        Return the free module over the base ring spanned by the rows
        of self.

        EXAMPLES:
            sage: A = MatrixSpace(IntegerRing(), 2)([1,2,3,4])
            sage: A.row_module()
            Free module of degree 2 and rank 2 over Integer Ring
            Echelon basis matrix:
            [1 0]
            [0 2]
        """
        M = sage.modules.free_module.FreeModule(self.base_ring(), self.ncols())
        return M.submodule(self.rows())
        
    def kernel_on(self, V, poly=None, check=False):
        """
        Return the kernel of self restricted to the invariant subspace V.
        The result is a vector subspace of V, which is also a subspace
        of the ambient space.

        INPUT:
            V -- vector subspace
            check -- (optional) default: False
            poly -- (optional) default: None; if not None, compute instead
                    the kernel of poly(self) on V. 
            
        OUTPUT:
            a subspace

        WARNING: This function does \emph{not} check that V is in fact
        invariant under self, unless check is True (not the default).
        """
        A = self.restrict(V, check=check)
        if not poly is None:
            A = poly(A)
        W = A.kernel()
        if V.is_ambient():
            return W
        else:
            A = V.basis_matrix()
            B = W.basis_matrix()
            C = B*A
            return C.row_module()


#############################################
## Generic matrices over the integers
#############################################

def _parimatrix_to_strlist(A):
    s = str(A)
    s = s.replace('Mat(','').replace(')','')
    s = s.replace(';',',').replace(' ','')
    s = s.replace(",", "','")
    s = s.replace("[", "['")
    s = s.replace("]", "']")
    return eval(s)

def _parimatrix_to_reversed_strlist(A):
    s = str(A)
    if s.find('Mat') != -1:
        return _parimatrix_to_strlist(A)
    s = s.replace('[','').replace(']','').replace(' ','')
    v = s.split(';')
    v.reverse()
    s = "['" + (','.join(v)) + "']"
    s = s.replace(",", "','")
    return eval(s)


class Matrix_integer(Matrix_pid):
    def __init__(self, parent):
        Matrix_pid.__init__(self, parent)

    def echelon_form(self, include_zero_rows=True):
        r"""
        Return the echelon form of this matrix over the integers.

        EXAMPLES:
            sage: A = MatrixSpace(IntegerRing(),2)([1,2,3,4])
            sage: A.echelon_form()
            [1 0]
            [0 2]

            sage: A = MatrixSpace(IntegerRing(),5)(range(25))
            sage: A.echelon_form()
            [  5   0  -5 -10 -15]
            [  0   1   2   3   4]
            [  0   0   0   0   0]
            [  0   0   0   0   0]
            [  0   0   0   0   0]
        """
        if self.nrows() == 0 or self.ncols() == 0:
            self.__rank = 0
            return self
        # The following complicated sequence of column reversals
        # and transposes is needed since PARI's Hermite Normal Form
        # does column operations instead of row operations.
        n = self.ncols()
        r = [n-i for i in range(n)]
        v = self._pari_()
        v = v.vecextract(r) # this reverses the order of columns
        v = v.mattranspose()
        w = v.mathnf(1)
        def convert_parimatrix(z):
            n = z.ncols(); r = [n-i for i in range(n)]        
            z = z.vecextract(r)  
            z = z.mattranspose()
            n = z.ncols(); r = [n-i for i in range(n)]        
            z = z.vecextract(r)
            return _parimatrix_to_strlist(z)
            
        H = convert_parimatrix(w[0])

        # We can do a 'fast' change of the above into a list of ints,
        # since we know all entries are ints:
        (nr,nc) = (self.nrows(), self.ncols())
        num_missing_rows = (nr*nc - len(H)) / nc
        rank = nr - num_missing_rows        
        if include_zero_rows:
            H += ['0']*(num_missing_rows*nc)
            H = self.new_matrix(nrows=nr, ncols=nc, entries=H, coerce_entries=True)
        else:
            H = self.new_matrix(nrows=rank, ncols=nc, entries=H, coerce_entries=True)
        H.__rank = rank
        H.set_immutable()
        return H

    def rank(self):
        """
        Return the rank of self, which is the rank of the space
        spanned by the rows of self.
        """
        try:
            return self.__rank
        except AttributeError:
            r = self.change_ring(rational_field.RationalField()).rank()
            if self.is_immutable():
                self.__rank = r
            return r


    def elementary_divisors(self):
        """
        Return the elementary divisors of self, in order.

        The elementary divisors are the invariants of the finite
        abelian group that is the cokernel of this matrix.  They are
        ordered in reverse by divisibility.

        INPUT:
            matrix
        OUTPUT:
            list of int's
        
        EXAMPLES:
            sage: A = MatrixSpace(IntegerRing(), 3)(range(9))
            sage: A.elementary_divisors()
            [0, 3, 1]
            
        SEE ALSO: smith_form
        """
        try:
            return self.__elementary_divisors
        except AttributeError:
            if self.nrows() == 0 or self.ncols() == 0:
                return []
            d = self._pari_().matsnf(0).python()
            if self.is_immutable():
                self.__elementary_divisors = d
            return d

    def smith_form(self, transformation=False):
        """
        Returns matrices S, U, and V such that S = U*self*V, and S
        is in Smith normal form.  Thus S is diagonal with diagonal
        entries the ordered elementary divisors of S.

        The elementary divisors are the invariants of the finite
        abelian group that is the cokernel of this matrix.  They are
        ordered in reverse by divisibility.

        EXAMPLES:
            sage: A = MatrixSpace(IntegerRing(), 3)(range(9))
            sage: D, U, V = A.smith_form()
            sage: D
            [0 0 0]
            [0 3 0]
            [0 0 1]
            sage: U
            [-1  2 -1]
            [ 0 -1  1]
            [ 0  1  0]
            sage: V
            [ 1  4 -1]
            [-2 -3  1]
            [ 1  0  0]
            sage: U*A*V
            [0 0 0]
            [0 3 0]
            [0 0 1]
            
        SEE ALSO: elementary_divisors
        """
        v = self._pari_().matsnf(1).python()
        M = self.parent()
        return M(v[2]), M(v[0]), M(v[1]), 

    def kernel(self, LLL=False):
        r"""
        Return the kernel of this matrix, as a module over the integers.

        INPUT:
           LLL -- bool (default: False); if True the basis is an LLL
                  reduced basis; otherwise, it is an echelon basis.

        EXAMPLES:
            sage: M = MatrixSpace(IntegerRing(),4,2)(range(8))
            sage: M.kernel()
            Free module of degree 4 and rank 2 over Integer Ring
            Echelon basis matrix:
            [ 1  0 -3  2]
            [ 0  1 -2  1]
        """

        Z = self.base_ring()
        
        if self.nrows() == 0:    # from a 0 space
            M = sage.modules.free_module.FreeModule(Z, self.nrows())
            return M.zero_subspace()

        elif self.ncols() == 0:  # to a 0 space
            return sage.modules.free_module.FreeModule(Z, self.nrows())
            
        A = self._pari_().mattranspose()
        B = A.matkerint()


        n = self.nrows()
        Z = integer_ring.IntegerRing()
        M = sage.modules.free_module.FreeModule(Z, n)

        if B.ncols() == 0:
            return M.zero_submodule()

        # Now B is a basis for the LLL-reduced integer kernel as a
        # PARI object.  The basis vectors or B[0], ..., B[n-1],
        # where n is the dimension of the kernel.
        B = [M([Z(x) for x in b]) for b in B]
        if LLL:
            return M.submodule_with_basis(B)
        else:
            return M.submodule(B)



#############################################
## Generic matrices over a field
############################################# 
class Matrix_field(Matrix_pid):
    def __init__(self, parent):
        Matrix_pid.__init__(self, parent)

    def __invert__(self):
        """
        Return this inverse of this matrix.

        Raises a ZeroDivisionError if the matrix has zero determinant, and
        raises an ArithmeticError, if the inverse doesn't exist
        because the matrix is nonsquare.

        EXAMPLES:
        
        """
        if not self.is_square():
            raise ArithmeticError, "self must be a square matrix"
        A = self.augment(self.parent().identity_matrix())
        B = A.echelon_form()
        if B[self.nrows()-1,self.ncols()-1] != 1:
            raise ZeroDivisionError, "self is not invertible"
        return B.submatrix_from_columns(range(self.ncols(), 2*self.ncols()))

    def charpoly(self):
        """
        Return the characteristic polynomial of this matrix.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),3)(range(9))
            sage: A.charpoly()
            x^3 - 12*x^2 - 18*x
            sage: A.trace()
            12
            sage: A.determinant()
            0
        """
        try:
            return self.__charpoly
        except AttributeError:
            pass
        if self.nrows() != self.ncols():
            raise ArithmeticError, "charpoly of non-square matrix not defined."
        
        R = polynomial_ring.PolynomialRing(self.base_ring())
        zero = R(0)
        if self.nrows() == 0:
            self.__charpoly = zero
            return self.__charpoly
        time = misc.verbose(t=0)
        H = self.hessenberg_form()
        n = self.nrows()
        c = [zero for i in range(n+1)]
        c[0] = R(1)
        X = R.gen()
        for m in range(1,n+1):
            c[m] = (X - R(H[m-1,m-1]))*c[m-1]
            t = 1
            for i in range(1,m):
                t = t*H[m-i, m-i-1]
                c[m] = c[m] - R(t*H[m-i-1,m-1])*c[m-i-1]
        misc.verbose('computed characteristic polynomial of %sx%s matrix'%
                     (self.nrows(), self.ncols()), time)
        f = c[n]
        if self.is_immutable():
            self.__charpoly = f
        return f

    def column_space(self):
        """
        Return the vector space over the base ring spanned by the
        columns of this matrix.
        """
        return self.column_module()
        
    def decomposition_of_subspace(self, M, is_diagonalizable=False):
        """
        Suppose the right action of self on M leaves M
        invariant. Return the decomposition of M as a list of pairs
        (W, is_irred) where is_irred is True if the charpoly of self
        acting on the factor W is irreducible.
        """
        if not self.is_square():
            raise ArithmeticError, "matrix must be square"
        if M.base_ring() != self.base_ring():
            raise ArithmeticError, "base rings are incompatible"
        if M.degree() != self.ncols():
            raise ArithmeticError, \
               "M must be a subspace of an %s-dimensional space"%self.ncols()

        time = misc.verbose(t=0)

        # 1. Restrict
        B = self.restrict(M)
        time0 = misc.verbose("restrict -- ", time)

        # 2. Decompose restriction
        D = B.decomposition(is_diagonalizable=is_diagonalizable, dual=False)
        
        assert sum([A.dimension() for A,_ in D]) == M.dimension(), "bug in decomposition; " + \
               "the sum of the dimensions of the factors must equal the dimension of the space acted on"

        # 3. Lift decomposition to subspaces of ambient vector space.
        # Each basis vector for an element of D defines a linear combination
        # of the basis of W, and these linear combinations define the
        # corresponding subspaces of the ambient space M.
        
        misc.verbose("decomposition -- ", time0)        
        C = M.basis_matrix()
        Z = M.ambient_vector_space()
        
        D = [(Z.subspace([x*C for x in W.basis()]), is_irred) for W, is_irred in D]
            
        misc.verbose(t=time)
        return D

    def denominator(self):
        r"""
        Return the least common multiple of the denominators of the
        elements of self.

        If there is no denominator function for the base field, or no
        LCM function for the denominators, raise a TypeError.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),2)(['1/2', '1/3', '1/5', '1/7'])
            sage: A.denominator()
            210

        Denominators are note defined for real numbers:
            sage: A = MatrixSpace(RealField(),2)([1,2,3,4])
            sage: A.denominator()
            Traceback (most recent call last):
            ...
            TypeError: denominator not defined for elements of the base ring

        We can even compute the denominator of matrix over the fraction field
        of $\Z[x]$.
            sage: K = FractionField(PolynomialRing(IntegerRing()))
            sage: x = K.gen()
            sage: A = MatrixSpace(K,2)([1/x, 2/(x+1), 1, 5/(x**3)])
            sage: A.denominator()
            x^4 + x^3
        """
        if self.nrows() == 0 or self.ncols() == 0:
            return 1
        R = self.base_ring()
        x = self.list()
        try:
            d = x[0].denominator()
        except AttributeError:
            raise TypeError, "denominator not defined for elements of the base ring"
        try:
            for y in x:
                d = d.lcm(y.denominator())
        except AttributeError:
            raise TypeError, "lcm function not defined for elements of the base ring"
        return d

    def echelon_form(self, include_zero_rows=True):
        """
        Returns the reduced row echelon form of self.

        INPUT:
            matrix -- an element A of a MatrixSpace

        OUTPUT:
            matrix -- The reduced row echelon form of A.
            Note that self is *not* changed by this command.

        EXAMPLES:
           sage: MS = MatrixSpace(RationalField(),2,3)
           sage: C = MS.matrix([1,2,3,4,5,6])
           sage: C.rank()
           2
           sage: C.nullity()
           1
           sage: C.echelon_form()
           [ 1  0 -1]
           [ 0  1  2]
        """
        try:
            return self.__echelon_form
        except AttributeError:
            pass
        t = misc.verbose("Generic echelon...")
        pivot_positions = []
        start_row = 0
        A = self.copy()
        nrows = A.nrows()
        ncols = A.ncols()
        cleared_a_column = False
        for c in range(ncols):
            misc.verbose("column %s of %s"%(c, ncols),t, level=2)
            for r in range(start_row, nrows):
                if A.get((r,c)) != 0:
                    pivot_positions.append(c)
                    # Divide row r through by 1/A[r,c], so leading coefficient
                    # is 1.
                    A.rescale_row(r, ~A.get((r,c)))
                    # Swap
                    A.swap_rows(r,start_row)
                    # Clear column
                    cleared_a_column = True
                    for i in range(nrows):
                        if i != start_row:
                            x = A.get((i,c))
                            if x != 0:
                                # Add -x times start row to i
                                A.add_multiple_of_row(i, start_row, -x)
                if cleared_a_column:
                    start_row += 1
                    cleared_a_column = False
                    break
        if not include_zero_rows:
            A = A.submatrix_from_rows(range(len(pivot_positions)))
        misc.verbose("Finished generic echelon.",t)
        A.__pivots = pivot_positions
        A.__rank = len(pivot_positions)
        A.set_immutable()
        return A
        
    def fcp(self):
        """
        Return the factorization of the characteristic polynomial of
        self.
        """
        return self.charpoly().factor()
        
    def hessenberg_form(self):
        """
        Return the Hessenberg form of self.

        The hessenberg form of a matrix $A$ is a matrix that is
        similar to $A$, so has the same characteristic polynomial as
        $A$, and is upper triangular except possible for entries right
        below the diagonal.

        ALGORITHM: See Henri Cohen's first book.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),3)([2, 1, 1, -2, 2, 2, -1, -1, -1])
            sage: A.hessenberg_form()
            [  2 3/2   1]
            [ -2   3   2]
            [  0  -3  -2]

            sage: A = MatrixSpace(RationalField(),4)([2, 1, 1, -2, 2, 2, -1, -1, -1,1,2,3,4,5,6,7])
            sage: A.hessenberg_form()
            [    2  -7/2 -19/5    -2]
            [    2   1/2 -17/5    -1]
            [    0  25/4  15/2   5/2]
            [    0     0  58/5     3]
        """
        if not self.is_square():
            raise ArithmeticError, "self must be square"
        n = self.nrows()
        tm = misc.verbose("Computing Hessenberg Normal Form of %sx%s matrix"%(n,n))
        h = self.copy()
        for m in range(1,n-1):
            # Search for a non-zero entry in column m-1
            i = False
            for r in range(m+1,n):
                if h[r,m-1] != 0:
                    i = r
                    break
            if i: 
                # Found a nonzero entry in column m-1 that is strictly below row m
                # Now set i to be the first nonzero position >= m in column m-1 
                if h[m,m-1] != 0:
                    i = m
                t = h[i,m-1]
                if i>m:
                    h.swap_rows(i,m)
                    # We must do the corresponding column swap to 
                    # maintain the characteristic polynomial (which is
                    # an invariant of Hessenberg form)
                    h.swap_columns(i,m)  
                # Now the nonzero entry in position (m,m-1) is t.
                # Use t to clear the entries in column m-1 below m.
                for j in range(m+1,n):
                    if h[j,m-1] != 0:
                        u = h[j,m-1]/t
                        h.add_multiple_of_row(j, m, -u)
                        # To maintain charpoly, do the corresponding column operation,
                        # which doesn't mess up the matrix, since it only changes
                        # column m, and we're only worried about column m-1 right now.
                        # Add u*column_j to column_m.
                        h.add_multiple_of_column(m, j, u)
        misc.verbose("Finished Hessenberg Normal Form of %sx%s matrix"%(n,n),tm)
        return h
        
    def kernel(self):
        r"""
        Return the kernel of this matrix, as a vector space.

        \algorithm{Elementary row operations don't change the kernel,
        since they are just right multiplication by an invertible
        matrix, so we instead compute kernel of the column echelon
        form.  More precisely, there is a basis vector of the kernel
        that corresponds to each non-pivot row.  That vector has a 1
        at the non-pivot row, 0's at all other non-pivot rows, and for
        each pivot row, the negative of the entry at the non-pivot row
        in the column with that pivot element.}
        
        \note{Since we view matrices as acting on the right, but have
        functions for reduced \emph{row} echelon forms, we instead
        compute the reduced row echelon form of the transpose of this
        matrix, which is the reduced column echelon form.}

        EXAMPLES:

        A kernel of dimension one over $\Q$:x
            sage: A = MatrixSpace(RationalField(),3)(range(9))
            sage: A.kernel()
            Vector space of degree 3 and dimension 1 over Rational Field
            Basis matrix:
            [ 1 -2  1]

        A trivial kernel:
            sage: A = MatrixSpace(RationalField(),2)([1,2,3,4])
            sage: A.kernel()
            Vector space of degree 2 and dimension 0 over Rational Field
            Basis matrix:
            []

        Kernel of a zero matrix:
            sage: A = MatrixSpace(RationalField(),2)(0)
            sage: A.kernel()
            Vector space of degree 2 and dimension 2 over Rational Field
            Basis matrix:
            [1 0]
            [0 1]

        Kernel of a non-square matrix:
            sage: A = MatrixSpace(RationalField(),3,2)(range(6))
            sage: A.kernel()
            Vector space of degree 3 and dimension 1 over Rational Field
            Basis matrix:
            [ 1 -2  1]
            
        The 2-dimensional kernel of a matrix over a cyclotomic field:
            sage: K = CyclotomicField(12); a=K.gen()
            sage: M = MatrixSpace(K,4,2)([1,-1, 0,-2, 0,-a**2-1, 0,a**2-1])
            sage: M
            [             1             -1]
            [             0             -2]
            [             0 -zeta_12^2 - 1]
            [             0  zeta_12^2 - 1]
            sage: M.kernel()
            Vector space of degree 4 and dimension 2 over Cyclotomic Field of order 12 and degree 4
            Basis matrix:
            [               0                1                0     -2*zeta_12^2]
            [               0                0                1 -2*zeta_12^2 + 1]

        A nontrivial kernel over a complicated base field.
            sage: K = FractionField(MPolynomialRing(RationalField(),2))
            sage: M = MatrixSpace(K, 2)([[K.gen(1),K.gen(0)], [K.gen(1), K.gen(0)]])
            sage: M
            [x_1 x_0]
            [x_1 x_0]
            sage: M.kernel()
            Vector space of degree 2 and dimension 1 over Fraction field of Polynomial ring in x_0, x_1 over Rational Field
            Basis matrix:
            [(-1*x_1^2)/(-1*x_1^2)          x_1/(-1*x_1)]
        """

        R = self.base_ring()

        if self.nrows() == 0:    # from a 0 space
            V = sage.modules.free_module.VectorSpace(R, self.nrows())
            return V.zero_subspace()

        elif self.ncols() == 0:  # to a 0 space
            return sage.modules.free_module.VectorSpace(R, self.nrows())
            
        if isinstance(R, number_field.NumberField_generic):
            A = self._pari_().mattranspose()
            B = A.matker()
            n = self.nrows()
            V = sage.modules.free_module.VectorSpace(R, n)
            basis = [V([R(x) for x in b]) for b in B]
            return V.subspace(basis)
            
        E = self.transpose().echelon_form()
        pivots = E.pivots()
        pivots_set = set(pivots)
        basis = []
        VS = sage.modules.free_module.VectorSpace
        V = VS(R, self.nrows())
        ONE = R(1)
        for i in xrange(self.nrows()):
            if not (i in pivots_set):
                v = V(0)
                v[i] = ONE
                for r in range(len(pivots)):
                    v[pivots[r]] = -E[r,i]
                basis.append(v)
        return V.subspace(basis)

    def integer_kernel(self):
        """
        Return the integer kernel of this matrix.

        Assume that the base field of this matrix has a numerator and
        denominator functions for its elements, e.g., it is the
        rational numbers or a fraction field.  This function computes
        a basis over the integers for the kernel of self.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),4)(range(16))
            sage: A.integer_kernel()
            Free module of degree 4 and rank 2 over Integer Ring
            Echelon basis matrix:
            [ 1  0 -3  2]
            [ 0  1 -2  1]

        The integer kernel even makes sense for matrices with
        fractional entries:
            sage: A = MatrixSpace(RationalField(),2)(['1/2',0,  0, 0])
            sage: A.integer_kernel()
            Free module of degree 2 and rank 1 over Integer Ring
            Echelon basis matrix:
            [0 1]
        """
        d = self.denominator()
        A = self*d
        R = d.parent()
        M = sage.matrix.matrix_space.MatrixSpace(R, self.nrows(), self.ncols())(A)
        return M.kernel()


    def is_invertible(self):
        """
        Return True if this matrix is invertible.
        """
        return self.is_square() and self.rank() == self.nrows()
        
    def maxspin(self, v):
        """
        Computes the largest integer n such that the list of vectors
        $S=[v, A(v), ..., A^n(v)]$ are linearly independent, and returns
        that list.

        INPUT:
            self -- Matrix
            v -- Vector
        OUTPUT:
            list -- list of Vectors

        ALGORITHM:
            The current implementation just adds vectors to a vector
            space until the dimension doesn't grow.  This could be
            optimized by directly using matrices and doing an
            efficient Echelon form.  Also, when the base is Q, maybe
            we could simultaneously keep track of what is going on in
            the reduction modulo p, which might make things much
            faster.
        """
        if v == 0: return []
        VS = sage.modules.free_module.VectorSpace
        V = VS([v])
        w = v
        S = [v]
        while True:
            w = w*self
            W = V + VS([w])
            if W.dimension() == V.dimension():
                return S
            V = W
            S.append(w)

    def numeric_array(self, typecode=None):
        """
        Return the Numeric array associated to this field, if possible, and Numeric
        is installed.
        """
        import Numeric
        if typecode is None:
            typecode = Numeric.Float64
        A = Numeric.array(self.list(), typecode=typecode)
        return Numeric.resize(A,(self.nrows(), self.ncols()))
        
    def nullity(self):
        # Use that rank + nullity = number of columns
        return self.ncols() - self.rank()
        
    def pivots(self):
        """
        Return the i such that the i-th column of self is a
        pivot column of the reduced row echelon form of self.
        
        OUTPUT:
            list -- sorted list of integers
        """
        try:
            return self.__pivots
        except AttributeError:
            P = self.echelon_form().pivots()
            if self.is_immutable():
                self.__pivots = P
            return P
        
    def rank(self):
        try:
            return self.__rank
        except AttributeError:
            rank = self.echelon_form().rank()
            if self.is_immutable():
                self.__rank = rank
            return rank

    def _set_rank(self, r):
        self.__rank = r

    def _set_pivots(self, pivots):
        self.__pivots = pivots
        
    def restrict(self, V, check=False):
        """
        Returns the matrix that defines the action of self on the
        invariant subspace V.  If V is the ambient vector space,
        returns self (not a copy of self).

        INPUT:
            V -- vector subspace
            check -- (optional) default: False; if True but V is not
                     invariant, raise an ArithmeticError exception.
        OUTPUT:
            a matrix
            
        WARNING:
        This function returns an nxn matrix, where V has dimension n.
        It does \emph{not} check that V is in fact invariant under
        self, unless check is True (not the default).

        EXAMPLES:
            sage: V = VectorSpace(RationalField(),3)
            sage: M = MatrixSpace(RationalField(),3)
            sage: A = M([1,2,0, 3,4,0, 0,0,0])
            sage: W = V.subspace([[1,0,0], [0,1,0]])
            sage: A.restrict(W)
            [1 2]
            [3 4]
            sage: A.restrict(W, check=True)
            [1 2]
            [3 4]

        We illustrate the warning about invariance not being checked
        by default, by giving a non-invariant subspace.  With the default
        check=False this function returns the 'restriction' matrix, which
        is meaningless as check=True reveals.
            sage: W2 = V.subspace([[1,0,0], [0,1,1]])
            sage: A.restrict(W2)
            [1 2]
            [3 4]
            sage: A.restrict(W2, check=True)
            Traceback (most recent call last):
            ...
            ArithmeticError: subspace is not invariant under matrix
        """
        if not isinstance(V, sage.modules.free_module.FreeModule_generic_field):
            raise TypeError, "V must be a Vector Space"
        if V.base_ring() != self.base_ring():
            raise TypeError, "base rings must be the same"
        if V.degree() != self.nrows():
            raise IndexError, "degree of V must equal number of rows of self"
        if V.dimension() == 0:
            return self.new_matrix(nrows=0, ncols=0)
        if V.is_full():
            return self

        if not check:
            B = V.basis_matrix()
            P = B.pivots()
            return B*self.submatrix_from_columns(P)
        else:
            n = V.dimension()
            try:
                C = [V.coordinate_vector(b*self) for b in V.basis()]
            except ArithmeticError:
                raise ArithmeticError, "subspace is not invariant under matrix"
            return self.new_matrix(n, n, C)

    def restrict_domain(self, V):
        """
        Compute the matrix relative to the basis for V on the domain
        obtained by restricting self to V, but not changing the
        codomain of the matrix.  This is the matrix whose rows are the
        images of the basis for V.

        INPUT:
            V -- vector space (subspace of ambient space on which self acts)

        SEE ALSO: restrict()

        EXAMPLES:
            sage: V = VectorSpace(RationalField(),3)
            sage: M = MatrixSpace(RationalField(),3)
            sage: A = M([1,2,0, 3,4,0, 0,0,0])
            sage: W = V.subspace([[1,0,0], [1,2,3]])
            sage: A.restrict_domain(W)
            [1 2 0]
            [3 4 0]
            sage: W2 = V.subspace_with_basis([[1,0,0], [1,2,3]])
            sage: A.restrict_domain(W2)
            [ 1  2  0]
            [ 7 10  0]
        """
        return self.new_matrix(V.dimension(), self.ncols(), [b*self for b in V.basis()])

        
    def row_space(self):
        """
        Return the vector space over the base field spanned by the
        rows of self.  

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),2,2)([1,2,3,4])
            sage: A.row_space()
            Vector space of degree 2 and dimension 2 over Rational Field
            Basis matrix:
            [1 0]
            [0 1]
            sage: A.row_span(IntegerRing())
            Free module of degree 2 and rank 2 over Integer Ring
            Echelon basis matrix:
            [1 0]
            [0 2]
        """
        return self.row_module()

    def row_span(self, R=None):
        r"""
        Return the R-module spanned by the rows of self.

        INPUT:
            R -- (optional) principal ideal ring, such that the
                 entries of self coerce into R.  The default
                 is the base ring.
        OUTPUT:
            a free R-module.

        EXAMPLES:
        
        We define a $2\times 3$ matrix over $\Q$, then consider its row span
        over both $\Q$ and $\Z$.
        
            sage: A = MatrixSpace(RationalField(),2,3)([1,2,3, '1/3', 4, 2])
            sage: A
            [  1   2   3]
            [1/3   4   2]
            sage: M1 = A.row_span(); M1
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [   1    0 12/5]
            [   0    1 3/10]
            sage: M2 = A.row_span(IntegerRing()); M2
            Free module of degree 3 and rank 2 over Integer Ring
            Echelon basis matrix:
            [1/3   4   2]
            [  0  10   3]

        Note that the determinants of the inner product matrices are
        different, though their discriminants differ by a square:

            sage: M1.inner_product_matrix()
            [ 169/25   18/25]
            [  18/25 109/100]
            sage: d1 = M1.discriminant(); d1
            137/20
            sage: d2 = M2.discriminant(); d2
            685/9
            sage: d2/d1
            100/9
        """
        if R is None:
            return self.row_module()
        M = sage.modules.free_module.FreeModule(R, self.ncols())
        return M.submodule(self.rows())

    def wiedemann(self, i, t=0):
        """
        Application of Wiedemann's algorithm to the i-th standard
        basis vector.

        If the optimal parameter t is nonzero, use only the first t
        linear recurrence relations.
        """
        i = int(i); t=int(t)
        if self.nrows() != self.ncols():
            raise ArithmeticError, "matrix must be square."
        n = self.nrows()
        v = sage.modules.free_module.VectorSpace(self.base_ring(), n).gen(i)
        tm = misc.verbose('computing iterates...')
        cols = self.iterates(v, 2*n).columns()
        tm = misc.verbose('computed iterates', tm)
        f = None
        # Compute the minimal polynomial of the linear recurrence
        # sequence corresponding to the 0-th entries of the iterates,
        # then the 1-th entries, etc.
        if t == 0:
            R = range(n)
        else:
            R = [t]
        for i in R:
            tm = misc.verbose('applying berlekamp-massey')
            g = berlekamp_massey.berlekamp_massey(cols[i].list())
            misc.verbose('berlekamp-massey done', tm)
            if f is None:
                f = g
            else:
                f = f.lcm(g)
            if f.degree() == n:
                break
        return f
        

#############################################
## Generic *DENSE* matrices over any field
############################################# 

def _convert_dense_entries_to_list(entries):
    # Create matrix from a list of vectors
    e = []
    for v in entries:
        e += v.list()
    copy = False
    return e

class Matrix_generic_dense(Matrix):
    """
    The \\class{Matrix_generic_dense} class derives from
    \\class{Matrix}, and defines functionality for dense matrices over
    any base ring.  Matrices are represented by a list of elements in
    the base ring, and element access operations are implemented in
    this class.
    """
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix.__init__(self, parent)
        self.__nrows = parent.nrows()
        self.__ncols = parent.ncols()
        
        
        R = self.base_ring()
        if not isinstance(entries, list):
            x = R(entries)
            zero = R(0)
            entries = [zero for _ in xrange(self.nrows()*self.ncols())]
            if x != zero:
                if self.nrows() != self.ncols():
                    raise TypeError, "scalar matrix must be square"
                
                nc = self.ncols()
                for i in xrange(self.nrows()):
                    entries[i*nc + i] = x
            
        if len(entries) > 0 and hasattr(entries[0], "is_vector"):
            entries = _convert_dense_entries_to_list(entries)
            
        if len(entries) != self.nrows() * self.ncols():
            if len(entries) != self.nrows() * self.ncols():
                raise ArithmeticError, "entries must be a list of length %s"%\
                       (self.nrows()*self.ncols())
        if coerce_entries:
            try:
                entries = [R(x) for x in entries]
            except TypeError:
                raise TypeError, "Unable to coerce elements of entries (=%s) to %s"%(entries, R)
        elif copy:
            # Make a copy
            entries = list(entries) 
            
        self.__entries = entries

        
    def __getitem__(self, ij):
        """
        INPUT:
            A[i, j] -- the i,j of A, and
            A[i]    -- the i-th row of A.
        """
        if isinstance(ij, int):
            return self.row(ij)
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, "index must be a 2-tuple (i,j)"
        i,j = ij
        if i < 0 or i >= self.__nrows:
            raise IndexError, "row index (i=%s) is out of range"%i
        if j < 0 or j >= self.__ncols:
            raise IndexError, "colums index (j=%s) is out of range"%j
        return self.__entries[int(i*self.ncols() + j)]
    
    def __setitem__(self, ij, value):
        """
        INPUT:
            A[i, j] = value -- set the (i,j) entry of A
            A[i] = value    -- set the ith row of A
        """
        self._require_mutable()
        if isinstance(ij, int):
            # Set the whole row.
            if ij < 0 or ij >= self.__nrows:
                raise IndexError, "row index (=%s) is out of range"%ij                
            for j in self.ncols():
                self[ij,j] = value[j]
            return
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, "index must be a 2-tuple (i,j)"
        i,j=ij
        if i < 0 or i >= self.__nrows:
            raise IndexError, "row index (i=%s) is out of range"%i
        if j < 0 or j >= self.__ncols:
            raise IndexError, "colums index (j=%s) is out of range"
        self.__entries[int(i*self.ncols() + j)] = self.base_ring()(value)

    def _entries(self):
        return self.__entries
        
    def list(self):
        return list(self.__entries)

    def antitranspose(self):
        f = []
        e = self.list()
        (nc, nr) = (self.ncols(), self.nrows())
        for j in reversed(xrange(nc)):
            for i in reversed(xrange(nr)):
                f.append(e[i*nc + j])
        return self.new_matrix(nrows = nc, ncols = nr,
                               entries = f, copy=False, coerce_entries=False)

    def transpose(self):
        """
        Returns the transpose of self, without changing self.
        
        EXAMPLES:
        We create a matrix, compute its transpose, and note that the
        original matrix is not changed.
            sage: M = MatrixSpace(RationalField(), 2)
            sage: A = M([1,2,3,4])
            sage: B = A.transpose()
            sage: print B
            [1 3]
            [2 4]
            sage: print A
            [1 2]
            [3 4]
        """
        f = []
        e = self.list()
        (nc, nr) = (self.ncols(), self.nrows())
        for j in xrange(nc):        
            for i in xrange(nr):
                f.append(e[i*nc + j])
        return self.new_matrix(nrows = nc, ncols = nr,
                               entries = f, copy=False, coerce_entries=False)
    
        
#############################################
## Generic sparse matrices over any field
#############################################  
def _sparse_dot_product(v, w):
    """
    v and w are dictionaries with integer keys.
    """
    x = set(v.keys()).intersection(set(w.keys()))
    return sum([v[k]*w[k] for k in x])
    
def _convert_sparse_entries_to_dict(entries):
    e = {}
    for i in xrange(len(entries)):
        for j, x in (entries[i].dict()).iteritems():
            e[(i,j)] = x
    return e
    
class Matrix_generic_sparse(Matrix):
    """
    The \\class{Matrix_generic_sparse} class derives from
    \\class{Matrix}, and defines functionality for dense matrices over
    any base ring.  A generic sparse matrix is represented using a
    dictionary with keys pairs $(i,j)$ and values in the base ring.
    """
    ##WARNING: We do not check that the i,j pairs satisfy
    ##    0 <= i < nrows,  0 <= j < ncols.
    def __init__(self, parent, 
                 entries=0, 
                 coerce_entries=True, 
                 copy=True):
        Matrix.__init__(self, parent)
        R = self.base_ring()
        self.__nrows = parent.nrows()
        self.__ncols = parent.ncols()
        
        if not isinstance(entries, (list, dict)):
            x = R(entries)
            zero = R(0)
            entries = {}
            if x != zero:
                if self.nrows() != self.ncols():
                    raise TypeError, "scalar matrix must be square"
                for i in xrange(self.nrows()):
                    entries[(i,i)] = x
            
        if isinstance(entries, list) and len(entries) > 0 and \
                hasattr(entries[0], "is_vector"):
            entries = _convert_sparse_entries_to_dict(entries)

        if isinstance(entries, list):
            if len(entries) != self.nrows() * self.ncols():
                raise TypeError, "entries has the wrong length"
            x = entries
            entries = {}
            k = 0
            for i in xrange(self.nrows()):
                for j in xrange(self.ncols()):
                    if x[k] != 0:
                        entries[(i,j)] = x[k]
                    k += 1
            copy = False
            
        if not isinstance(entries, dict):
            raise TypeError, "entries must be a dict"
        if coerce_entries:
            try:
                for k, x in entries.iteritems():
                    entries[k] = R(x)
            except TypeError:
                raise TypeError, "Unable to coerce entries to %s"%R
        elif copy:
            # Make a copy
            entries = dict(entries)
        self.__entries = entries
        self.__zero = R(0)
        
    def __getitem__(self, ij):
        """
        INPUT:
            A[i, j] -- the i,j of A, and
            A[i]    -- the i-th row of A.
        """
        if isinstance(ij, int):
            return self.row(ij)
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, "index must be a 2-tuple (i,j)"
        # the ij might not be int's, which would cause a problem:
        ij = (int(ij[0]), int(ij[1]))   
        (i,j) = ij[0], ij[1]
        if i < 0 or i >= self.__nrows:
            raise IndexError, "row index (i=%s) is out of range"%i
        if j < 0 or j >= self.__ncols:
            raise IndexError, "colums index (j=%s) is out of range"%j
        if self.__entries.has_key(ij):
            return self.__entries[ij]
        return self.__zero
        
    def _dict(self):
        return self._entries()
                
    def get(self, ij):
        """
        Like __getitem__ but with no type or bounds checking.
        For (i,j) access, returns 0 if access is out of bounds.
        """
        if isinstance(ij, int):
            return self.row(ij)
        if self.__entries.has_key(ij):
            return self.__entries[ij]
        return self.__zero
        
    def set(self, ij, x):
        """
        Like __setitem__ but with no type or bounds checking.  Only
        works for single entries, not whole rows.
        """
        self._require_mutable()
        if x == 0:
            if self.__entries.has_key(ij):
                del self.__entries[ij]
            return
        self.__entries[ij] = x
    
    def __setitem__(self, ij, value):
        """
        INPUT:
            A[i, j] = value -- set the (i,j) entry of A
            A[i] = value    -- set the ith row of A
        """
        self._require_mutable()
        if isinstance(ij, int):
            # Set the whole row.
            if ij < 0 or ij >= self.__nrows:
                raise IndexError, "row index (=%s) is out of range"%ij
            try:
                for j in value.nonzero_positions():
                    self[ij,j] = value[j]
            except AttributeError:
                for j in self.ncols():
                    self[ij,j] = value[j]
            return
        
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise TypeError, "index must be a 2-tuple (i,j)"
        # the ij might not be int's, which would cause a problem, so we coerce
        ij = (int(ij[0]), int(ij[1]))   
        i,j=ij
        if i < 0 or i >= self.__nrows:
            raise IndexError, "row index (i=%s) is out of range"%i
        if j < 0 or j >= self.__ncols:
            raise IndexError, "colums index (j=%s) is out of range"
        x = self.base_ring()(value)
        if x == 0:
            if self.__entries.has_key(ij):
                del self.__entries[ij]
            return
        self.__entries[ij] = x
        
    def __mul__(self, right):
        if not isinstance(right, Matrix):
            return self._right_scalar_multiply(right)        
        if not isinstance(right, Matrix_generic_sparse):
            right = self.matrix_space(right.nrows(), right.ncols())(right)
        if self.base_ring() != right.base_ring():
            raise ArithmeticError, "base rings must be equal"
        if self.ncols() != right.nrows():
            raise ArithmeticError, "number of columns of self (=%s) must equal number of rows of right (=%s)."%(
                self.ncols(), right.nrows())
        nr = self.nrows()
        snc = self.ncols()
        nc = right.ncols()
        zero = self.base_ring()(0)
        E = {}
        rows = self.sparse_rows()
        cols = right.sparse_columns()
        for i in range(nr):
            row = rows[i]
            if len(row) == 0: continue
            for j in range(nc):
                x = _sparse_dot_product(row, cols[j])
                if x != 0: E[(i,j)] = x
        return self.new_matrix(nr, nc, entries = E, coerce_entries=False, copy=False)

    def hessenberg_form(self):
        """
        Return the Hessenberg form of this matrix.
        """
        # Current implementation is way too slow on sparse matrix.
        # On the dense it is reasonable.
        return self.dense_matrix().hessenberg_form()
        
    def sparse_columns(self):
        try:
            return self.__sparse_columns
        except AttributeError:
            C = [{} for _ in xrange(self.ncols())]
            for ij, x in self.__entries.iteritems():
                C[ij[1]][ij[0]] = x
            if self.is_immutable():
                self.__sparse_columns = C
            return C
            
    def sparse_rows(self):
        try:
            return self.__sparse_rows
        except AttributeError:
            R = [{} for _ in xrange(self.nrows())]
            for ij, x in self.__entries.iteritems():
                R[ij[0]][ij[1]] = x
            if self.is_immutable():
                self.__sparse_rows = R
            return R
    
    def __rmul__(self, left):
        left = self.base_ring()(left)
        X = {}
        for ij, x in self._entries().iteritems():
            X[ij] = left*x
        return self.new_matrix(entries=X, copy=False, coerce_entries=False)
        
    def denominator(self):
        R = self.base_ring()
        x = self._entries()
        if len(x) == 0:
            return 1
        Z = x.iteritems()
        d = Z.next()[1].denominator()
        for _, y in Z:
            d = d.lcm(y.denominator())
        return d
        
    def _entries(self):
        return self.__entries
        
    def nonzero_positions(self):
        """
        Returns the set of pairs (i,j) such that self[i,j] != 0.
        """
        return set(self.__entries.keys())
    
    def submatrix_from_columns(self, columns):
        """
        Return the submatrix of self of columns col[i] for i in
        the list of columns.
        """
        if not isinstance(columns, list):
            raise TypeError, "columns must be a list of integers"
            
        # Let A be this matrix and let B denote this matrix with
        # the columns deleted.
        # ALGORITHM:
        # 1. Make a table that encodes the function
        #          f : Z --> Z,
        #    f(j) = column of B where the j-th column of A goes
        # 2. Build new list of entries and return resulting matrix.
        C = set(columns)
        X = []
        j = 0
        for i in xrange(self.ncols()):
            if i in C:
                X.append(j)
                j += 1
            else:
                X.append(-1)    # column to be deleted.
        entries = {}
        E = self._entries()
        for ij in E.iterkeys():
            if ij[1] in C:
                i,j=ij
                entries[(i,X[j])] = E[ij]
                
        return self.new_matrix(ncols = len(columns), entries = entries, 
                    copy=False, coerce_entries=False)

    def submatrix_from_rows(self, rows):
        """
        Return the submatrix of self of rows row[i] for i in
        the list of rows.
        """
        if not isinstance(rows, list):
            raise TypeError, "rows must be a list of integers"
        R = set(rows)
        if not R.issubset(set(xrange(self.nrows()))):
            raise ArithmeticError, "Invalid rows."
        X = []
        i = 0
        for j in xrange(self.nrows()):
            if j in R:
                X.append(i)
                i += 1
            else:
                X.append(-1)    # row to be deleted.
        entries2 = []
        E = self._entries()
        for ij in E.iterkeys():
            if ij[0] in R:
                i,j=ij
                entries[(X[i],j)] = E[ij]
                
        return self.new_matrix(
                    nrows = len(rows),
                    entries = entries, 
                    copy=False, coerce_entries=False)
        
    def swap_rows(self, r1, r2):
        """
        Swap rows r1 and r2 of self.
        """
        self._require_mutable()
        nr = self.nrows()
        if r1 < 0 or r1 >= nr:
            raise IndexError, "r1 must satisfy 0<=r1<%s"%nr
        if r2 < 0 or r2 >= nr:
            raise IndexError, "r2 must satisfy 0<=r2<%s"%nr
        if r1 == r2: return
        for c in xrange(self.ncols()):
            self[r1,c], self[r2,c]  =  self[r2,c], self[r1,c]
        
    def transpose(self):
        """
        Returns the transpose of self, without changing self.
        
        EXAMPLES:
        We create a matrix, compute its transpose, and note that the
        original matrix is not changed.
            sage: M = MatrixSpace(RationalField(), 2, sparse=True)
            sage: A = M([1,2,3,4])
            sage: B = A.transpose()
            sage: print B
            [1 3]
            [2 4]
            sage: print A
            [1 2]
            [3 4]
        """
        X = {}
        for ij, x in self.__entries.iteritems():
            X[(ij[1],ij[0])] = x
        return self.new_matrix(nrows = self.ncols(), ncols = self.nrows(),
                           entries = X, copy=False, coerce_entries=False)
        

############################################################
# Generic matrices over special types of commutative rings
############################################################

class Matrix_generic_dense_domain(Matrix_domain, Matrix_generic_dense):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix_generic_dense.__init__(self, parent, entries, coerce_entries, copy)

class Matrix_generic_sparse_domain(Matrix_domain, Matrix_generic_sparse):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix_generic_sparse.__init__(self, parent, entries, coerce_entries, copy)

    
class Matrix_generic_dense_pid(Matrix_pid, Matrix_generic_dense):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix_generic_dense.__init__(self, parent, entries, coerce_entries, copy)

class Matrix_generic_sparse_pid(Matrix_pid, Matrix_generic_sparse):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix_generic_sparse.__init__(self, parent, entries, coerce_entries, copy)


class Matrix_generic_dense_field(Matrix_field, Matrix_generic_dense):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix_generic_dense.__init__(self, parent, entries, coerce_entries, copy)


class Matrix_generic_sparse_field(Matrix_field, Matrix_generic_sparse):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix_generic_sparse.__init__(self, parent, entries, coerce_entries, copy)


#############################################
## Dense matrices over the integers
##
##       Stored using PARI.
#############################################        
class Matrix_dense_integer(Matrix_integer, Matrix_generic_dense):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix_generic_dense.__init__(self, parent, entries, coerce_entries, copy)


#############################################
## Sparse matrices over the integers
#############################################        
class Matrix_sparse_integer(Matrix_integer, Matrix_generic_sparse):
    def __init__(self, parent, entries=0, 
                       coerce_entries=True, 
                       copy=True):
        Matrix_generic_sparse.__init__(self, parent, entries, coerce_entries, copy)


#############################################
## Dense matrices over the rational numbers
#############################################        
class Matrix_dense_rational(Matrix_generic_dense_field):
    """
    The \\class{Matrix_dense_rational} class derives from
    \\class{Matrix}, and defines functionality for dense matrices over
    the field $\\Q$ of rational numbers.
    """
    def __init__(self, 
                    parent, 
                    entries=0, 
                    coerce_entries=True, 
                    copy=True):
        Matrix.__init__(self, parent)
                
        if isinstance(entries, dense_matrix_pyx.Matrix_rational):
            if copy:
                entries = entries.copy()
            self.__matrix = entries
            return
            
        if not isinstance(entries, list):
            entries = self.base_ring()(entries)
            
        if isinstance(entries, list) and len(entries) > 0 and \
               hasattr(entries[0],"is_vector"):
            entries = _convert_dense_entries_to_list(entries)
        
        if isinstance(entries, list):
            if len(entries) != parent.nrows() * parent.ncols():
                raise TypeError, "entries has wrong length"

        self.__matrix = dense_matrix_pyx.Matrix_rational(
                    parent.nrows(), parent.ncols(), entries)
            
    def __getitem__(self, ij):
        """
        Use \\code{A[i,j]} to obtain the the $(i,j)$th entry of $A$, and
        \\code{A[i]} to obtain the $i$-th row of $A$.
        """
        if isinstance(ij, int):
            return self.row(ij)
        return self.__matrix[ij]
        
    def __setitem__(self, ij, x):
        r"""
        Use \code{A[i,j]=x} to set the $(i,j)$th entry of $A$ to $x$,
        and \code{A[i]=v} to set the $i$th row of $A$ to the entries
        of $v$.
        """
        self._require_mutable()
        if isinstance(ij, int):
            # Set the whole row.
            if ij < 0 or ij >= self.__nrows:
                raise IndexError, "row index (=%s) is out of range"%ij                
            for j in self.ncols():
                self.__matrix[ij,j] = value[j]
            return
        self.__matrix[ij] = x
               
    def __mul__(self, right):
        if not isinstance(right, Matrix_dense_rational):
            if isinstance(right, Matrix):
                right = self.matrix_space(right.nrows(), right.ncols())(right)
            else:
                return self._right_scalar_multiply(right)
        if self.ncols() != right.nrows():
            raise ArithmeticError, "number of columns of self (=%s) must equal number of rows of right (=%s)."%(
                self.ncols(), right.nrows())
        M = self.matrix_space(self.nrows(), right.ncols())
        A = self.__matrix * right.__matrix
        return Matrix_dense_rational(M, A, copy=False)

    def iterates(self, v, n):
        """
        Let A be this matrix.   Return a matrix with rows
        $$
           v, v*A, v*A^2, ..., v*A^(n-1).
        $$
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "matrix must be square."
        I =  self.__matrix.iterates(list(v), n)
        M = self.matrix_space(n, self.ncols())
        return Matrix_dense_rational(M, I, copy=False)
        
        
    def hessenberg_form(self):
        """
        Return the Hessenberg form of this matrix.
        """
        time = misc.verbose(t=0)
        A = self.copy()
        A.__matrix.hessenberg_form()
        misc.verbose("finished computing Hessenberg form",time)
        return A

    def charpoly(self, bound=None):
        """
        Return the characteristic polynomial of this matrix, computed
        using the standard multimodular Hessenberg algorithm.

        The multimodular algorithm works by first computing a bound B,
        then computing the characteristic polynomial (using Hessenberg
        form) modulo enough primes so that their product is bigger
        than B.  One then uses the Chinese Remainder Theorem to
        recover the characteristic polynomial.  If the optional bound
        is specified, that bound is used for B instead of a potentially
        much worse general bound.
        """
        if self.nrows() != self.ncols():
            raise ArithmeticError, "the matrix must be square."

        try:
            return self.__charpoly
        except AttributeError:
            pass

        if self.nrows() == 0:
            R = polynomial_ring.PolynomialRing(self.base_ring())
            f = R(0)
            if self.is_immutable():
                self.__charpoly = f
            return f

        d = self.denominator()
        if d != 1:
            A = self*d
        else:
            A = self
        if bound is None:
            # The following code (inspired by NTL's mat_poly_ZZ.c) computes
            # a bound that is vastly better than the Hadamard bound.  Victor
            # Shoup's comment, 'This bound is computed via interpolation
            # through complex roots of unity' means.  Reference: Mathieu and
            # Ford (1990, Section 6), 'On p-adic computation of the rational
            # form of a matrix.'.
            time = misc.verbose("computing bound")
            bound = 1
            zero = integer_ring.IntegerRing()(0)
            for i in range(A.nrows()):
                t1 = zero
                for j in range(A.ncols()):
                    t1 += (A[i,j]*A[i,j]).numerator()
                t1 += abs(A[i,i].numerator()) + 1
                if t1 > 1:
                    t1 = t1.isqrt() + 1
                bound *= t1
            bound += 1
            misc.verbose("bound = %s"%bound, time)
        else:
            bound = int(bound)

        p = 46337   # todo -- don't hardcode, so on 64-bit machine will be twice as fast!
        
        prod = 1
        X = []
        primes = []
        while prod < bound:
            time = misc.verbose("using p = %s"%p)
            B = A.__matrix.matrix_modint_nodenom(p)
            primes.append(p)
            fmodp = B.charpoly()   # vector of entries of the characteristic polynomial
            misc.verbose("got charpoly mod %s"%p, time)
            X.append(fmodp)
            p = sage.rings.arith.previous_prime(p)
            prod *= p
            
        time = misc.verbose("final CRT")
        w = sage.rings.arith.CRT_vectors(X, primes)
        misc.verbose("done with CRT",time )
        modulus = misc.prod(primes)
        m = modulus // 2
        for i in range(len(w)):
            if w[i] >= m:
                w[i] = w[i] - modulus
        g = polynomial_ring.PolynomialRing(rational_field.RationalField())(w)
        f = ((1/d)**self.nrows()) * g.rescale(d)
        if self.is_immutable():
            self.__charpoly = f
        return f

    def list(self):
        return list(self.__matrix.list())

    def _entries(self):
        return self.__matrix

    def minpoly(self):
        return self._pari_().minpoly()
        
    def transpose2(self):
        T = self.__matrix.transpose()
        return Matrix_dense_rational(self.matrix_space(self.ncols(), self.nrows()), T, copy=False)
        
    def echelon_form(self, height_guess=None, include_zero_rows=True):
        """
        Return the echelon form of this matrix over the rational
        numbers, computed using a multi-modular algorithm.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),3)(range(9))
            sage: A.echelon_form()
            [ 1  0 -1]
            [ 0  1  2]
            [ 0  0  0]
        """
        try:
            return self.__echelon_form
        except AttributeError:
            pass
        A = self.__matrix.echelon_modular(height_guess=height_guess)
        pivots = A.pivots()
        r = len(pivots)
        if not include_zero_rows:
            A = A.matrix_from_rows(range(len(pivots)))
            nr = r
        else:
            nr = self.nrows()
        E = Matrix_dense_rational(self.matrix_space(nrows=nr), A)
        E._set_pivots(pivots)
        E._set_rank(len(pivots))
        E.set_immutable()
        if self.is_immutable():
            self.__echelon_form = E
        return E
        
        
#############################################
## Sparse matrices over the rational numbers
#############################################
class Matrix_sparse_rational(Matrix_generic_sparse_field):
    """
    The \\class{Matrix_sparse_rational} class derives from
    \\class{Matrix}, and defines functionality for sparse matrices
    over the field $\\Q$ of rational numbers.
    """
    def __init__(self, 
                    parent, 
                    entries=0, 
                    coerce_entries=True, 
                    copy = True):
        Matrix_generic_sparse.__init__(self, parent, entries, 
                        coerce_entries, copy)
            

    def _sparse_list(self):
        try:
            return self.__sparse_list
        except AttributeError:
            L = [(ij[0], ij[1], x) for 
                ij, x in self._entries().iteritems()]
            L.sort()
            if self.is_immutable():
                self.__sparse_list = L
            return L
            
    def _sparse_matrix_rational(self):
        try:
            return self.__sparse_matrix_rational
        except AttributeError:
            S = sage.matrix.sparse_matrix.SparseMatrix(
                self.base_ring(), self.nrows(), self.ncols(), self._sparse_list(), 
                coerce=False, sort=False, copy=False)
            if self.is_immutable():
                self.__sparse_matrix_rational = S
            return S

    def echelon_form(self, height_guess=None, include_zero_rows=True):
        """
        Return the echelon form of this sparse matrix over the
        rational numbers, computed using a sparse multi-modular
        algorithm.

        The height guess is a guess for a bound on the height of the
        entries of the echelon form.  If you know for some reason that
        the entries of the echelon form are bounded, giving a good
        height bound can speed up this function.  At the end of the
        computation the result is checked for correctness and the
        height bound increased if necessary, so giving too small of a
        guess will not lead to incorrect results (though it may slow
        down the algorithm).

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),3, sparse=True)(range(9))
            sage: A.echelon_form()
            [ 1  0 -1]
            [ 0  1  2]
            [ 0  0  0]
            sage: A = 9999999*MatrixSpace(RationalField(),3, sparse=True)(range(9))
            sage: A
            [       0  9999999 19999998]
            [29999997 39999996 49999995]
            [59999994 69999993 79999992]
            sage: A.echelon_form(height_guess=1)
            [ 1  0 -1]
            [ 0  1  2]
            [ 0  0  0]
        """
        try:
            return self.__echelon_form
        except AttributeError:
            pass
        if self.nrows() == 0:
            E = self.copy()
            E._set_pivots ([])
            E._set_rank(0)
        else:
            t0 = misc.verbose()
            A = self._sparse_matrix_rational()
            t = misc.verbose("start")
            B = A.echelon_form(height_guess=height_guess)
            t1 = misc.verbose("done", t)
            pivots = B.pivots()
            X = B.entries()
            Y = {}
            for i,j,x in X:
                Y[(i,j)] = x
            if include_zero_rows:
                nr = self.nrows()
            else:
                nr = len(pivots)
            E = self.new_matrix(nrows=nr, coerce_entries=False, entries = Y, copy=False)
            E._set_pivots(pivots)
            E._set_rank(len(pivots))
            misc.verbose("overhead", t0+(t1-t))
        if self.is_immutable():
            self.__echelon_form = E
        E.set_immutable()  
        return E

    def hessenberg_form(self):
        """
        Return the Hessenberg form of this sparse matrix.

        EXAMPLES:
            sage: A = MatrixSpace(RationalField(),3, sparse=True)(range(9))
            sage: A.hessenberg_form()
            [ 0  5  2]
            [ 3 14  5]
            [ 0 -5 -2]
        """
        return self.dense_matrix().hessenberg_form()

    def charpoly(self):
        """
        Return the characteristic polynomial of this matrix. 
        """
        try:
            return self.__charpoly
        except AttributeError:
            f = self.dense_matrix().charpoly()
            if self.is_immutable():
                self.__charpoly = f
        return f
        

        
#############################################
import sage.rings.polynomial_ring
def berlekamp(v):
    """
    Find minimal polynomial of a linear recurrence sequence.
    """
    if len(v) % 2 != 0 or len(v) == 0:
        raise TypeError, "len of v must be even and positive"
    n = len(v)/2
    S = sage.rings.polynomial_ring.PolynomialRing(v[0].parent())
    x = S._0
    R0 = x**(2*n)
    R1 = S(v)
    V0 = S(0)
    V1 = S(1)
    while R1.degree() >= n:
        Q, R = R0/R1   
        print "\nQ = ", Q
        print "R = ", R
        assert R0 == Q*R1 + R
        V0, V1, R0, R1 = V1, V0-Q*V1, R1, R
        print "V0 = ", V0
        print "V1 = ", V1
        print "R0 = ", R0
        print "R1 = ", R1
    d = max(V1.degree(), 1+R1.degree())
    print "d = ", d
    P = V1.reverse()
    print "p rev = ", P
    P = x**(d-P.degree())*P
    c = P.leading_coefficient()
    print "c = ", c
    return (1/c)*P
    
        
if __name__ ==  '__main__':
    import doctest, sys
    from sage.matrix.matrix_space import MatrixSpace
    from sage.matrix.all import Matrix
    from sage.modules.free_module import *
    from sage.rings.all import RationalField, IntegerRing, CyclotomicField, \
         PolynomialRing, FractionField, MPolynomialRing, RealField
    doctest.testmod(sys.modules[__name__])
 
