"""
Generic spaces of modular symbols.

All the spaces of modular symbols derive from this class.  This class
is an abstract base class.
"""

#*****************************************************************************
#       SAGE: System for Algebra and Geometry Experimentation
#
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import math
import random
import weakref

import sage.modules.free_module as free_module
import sage.matrix.matrix_space as matrix_space
import sage.modules.free_module_morphism as free_module_morphism
import sage.misc.misc as misc
import sage.modular.dims as dims
import sage.modular.hecke.all as hecke
import sage.modular.modsym.element
import sage.rings.all as rings
import sage.structure.gens as gens
import sage.rings.arith as arith


class ModularSymbolsSpace(hecke.HeckeModule_free_module):
    def __init__(self, group, weight, character, sign, base_ring):
        self.__group = group
        self.__character = character
        self.__sign = sign
        hecke.HeckeModule_free_module.__init__(self, base_ring, group.level(), weight)
        
    def __cmp__(self, other):
        """
        Compare self and other.
        """
        if not isinstance(other, ModularSymbolsSpace):
            return -1
        if self.__group > other.__group:
            return -1
        elif self.__group < other.__group:
            return 1
        elif self.weight() < other.weight():
            return -1
        elif self.weight() > other.weight():
            return 1
        elif self.__character < other.__character:
            return -1
        elif self.__character > other.__character:
            return 1
        elif self.__sign < other.__sign:
            return -1
        elif self.__sign > other.__sign:
            return 1
        elif self.base_ring() < other.base_ring():
            return -1
        elif self.base_ring() > other.base_ring():
            return 1
        if self.is_ambient() == other.is_ambient():
            return 0
        if self.ambient_hecke_module() < other.ambient_hecke_module():
            return -1
        elif self.ambient_hecke_module() > other.ambient_hecke_module():
            return 1
        
        # ambients same and at most one of the two spaces is ambient
        if self.rank() < other.rank():
            return -1
        elif self.rank() > other.rank():
            return 1
            
        # neither is ambient; but they're in the same ambient space
        if self.free_module() == other.free_module():
            return 0
        return -1

    def character(self):
        return self.__character

    def cuspidal_submodule(self):
        raise NotImplementedError

    def cuspidal_subspace(self):
        """
        Synonym for cuspidal_submodule.
        """
        return self.cuspidal_submodule()

    def new_subspace(self, p=None):
        """
        Synonym for new_submodule.
        """
        return self.new_submodule(p)

    def eisenstein_subspace(self):
        """
        Synonym for eisenstein_submodule.
        """
        return self.eisenstein_submodule()
    
    def dual_star_involution_matrix(self):
        """
        Return the matrix of the dual star involution, which is
        induced by complex conjugation on the linear dual of modular
        symbols.
        """
        raise NotImplementedError

    def group(self):
        """
        Returns the group of this modular symbols space.

        INPUT:
           ModularSymbols self -- an arbitrary space of modular symbols
           
        OUTPUT:
           CongruenceSubgroup -- the congruence subgroup that this is a space
                              of modular symbols for.

        ALGORITHM:
           The group is recorded when this space is created.

        EXAMPLES:
            sage: m = ModularSymbols(20)
            sage: m.group()
            Gamma0(20)
        """
        return self.__group

    def is_cuspidal(self):
        raise NotImplementedError

    def is_simple(self):
        try:
            return self._is_simple
        except AttributeError:
            D = self.factorization()
            if len(D) <= 0 or len(D) == 1 and D[0][1] == 1:
                self._is_simple = True
            else:
                self._is_simple = False
            return self._is_simple
        
    def multiplicity(self, S):
        """
        Return the multiplicity of the simple modular symbols space S
        in self.  S must be a simple anemic Hecke module.

        ASSUMPTION: self is an anemic Hecke module with the same weight and
        group as S, and S is simple.
        """
        if self.level() % S.level() != 0 or S.weight() != self.weight():
            return 0
        if not S.is_simple():
            raise ArithmeticError, "S must be simple"
        if not self.is_anemic_hecke_module():
            raise ArithmeticError, "S must be an anemic Hecke module"
        A = self.ambient_hecke_module()
        B = A.subspace_generated_by_images(S)
        C = self.intersection(B)
        d = C.rank()
        n = S.rank()
        assert d % n == 0, "the dimension of intersection must be a multiple of dimension of simple space.  bug!"
        return d//n
        
    def ngens(self):
        """
        The number of generators of self.
        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols.
        OUTPUT:
           int -- the number of generators, which is the same as the
                  dimension of self.
        ALGORITHM:
           Call the dimension function.
        EXAMPLES:
        sage: m = ModularSymbols(33)
        sage: m.ngens()
        9
        sage: m.rank()
        9
        sage: ModularSymbols(100, weight=2, sign=1).ngens()
        18
        """
        return self.rank()

    def q_expansion(self, prec):
        """
        Returns the q-expansion to precision prec of a newform
        associated to self, where self must be new, cuspidal, and
        simple.
        """
        try:
            f = self.__q_expansion
        except AttributeError:
            if not self.is_cuspidal():
                raise ArithmeticError, "self must be cuspidal."
            if not self.is_simple():
                raise ArithmeticError, "self must be simple."
            a2 = self.eigenvalue(2)
            R = rings.PowerSeriesRing(a2.parent(), "q")
            q = R.gen(0)
            f = q + a2*q**2 + rings.O(q**3)
        if f.prec() < prec:
            R = f.parent()
            ext = [self.eigenvalue(n) for n in range(f.prec(),prec)]
            f = R(list(f) + ext)
            self.__qeigenform = f.add_bigoh(prec)
            return self.__qeigenform
        else:
            return f.O(prec)

    def q_expansion_basis(self, prec):
        """
        Returns a basis of q-expansions (as power series) to precision
        prec of the space of modular forms associated to self.  The
        q-expansions are defined over the same base ring as prec.
        """
        try:
            return self.__q_expansion_basis
        except AttributeError:
            pass

        if not self.is_cuspidal():
            raise ArithmeticError, "space must be cuspidal"
        
        if self.is_simple():
            f = self.q_expansion(prec)
            R = PowerSeriesRing(self.base_ring(), 'q')
            B = [R([f[i][j] for i in xrange(prec)]) for j in range(self.rank())]
            self.__q_expansion_basis = B
            return B

        # Here use a slow general algorithm that involves computing
        # functionals etc. like in Merel's 1585 paper. (?)
        raise NotImplementedError

    def factorization(self):
        """
        Returns a list of pairs $(S,e)$ where $S$ is simple spaces of
        modular symbols and self is isomorphic to the direct sum of
        the $S^e$ as a module over the \emph{anemic} Hecke algebra
        adjoin the star involution.

        ASSUMPTION: self is a module over the anemic Hecke algebra.
        """
        try:
            return self._factorization
        except AttributeError:
            raise NotImplementedError

    def hecke_module_of_level(self, level):
        """
        See the documentation for self.modular_symbols_of_level(level)
        """
        return self.modular_symbols_of_level(level)
            
    def sign(self):
        """
        Returns the sign of self.

        For efficiency reasons, it is often useful to compute in the
        (largest) quotient of modular symbols where the * involution
        acts as +1, or where it acts as -1.
        

        INPUT:
           ModularSymbols self -- arbitrary space of modular symbols.
           
        OUTPUT:
           int -- the sign of self, either -1, 0, or 1.
                  -1 -- largest quotient where * acts as -1,
                  +1 -- largest quotient where * acts as +1, 
                   0 -- full space of modular symbols (no quotient).

        ALGORITHM:
           Call the dimension function.

        EXAMPLES:
        sage: m = ModularSymbols(33)
        sage: m.rank()
        9
        sage: m.sign()
        0
        sage: m = ModularSymbols(33, sign=0)
        sage: m.sign()
        0
        sage: m.rank()
        9
        sage: m = ModularSymbols(33, sign=-1)
        sage: m.sign()
        -1
        sage: m.rank()
        3
        """
        return self.__sign

    def simple_factors(self):
        """
        Returns a list modular symbols spaces $S$ where $S$ is simple
        spaces of modular symbols (for the anemic Hecke algebra) and
        self is isomorphic to the direct sum of the $S$ with some
        multiplicities, as a module over the \emph{anemic} Hecke
        algebra.  For the multiplicities use factorization() instead.

        ASSUMPTION: self is a module over the anemic Hecke algebra.
        """
        return [S for S,_ in self.factorization()]

    def star_eigenvalues(self):
        """
        Returns the eigenvalues of the star involution acting on self.

        EXAMPLES:
            sage: M = ModularSymbols(11)
            sage: D = M.decomposition()
            sage: M.star_eigenvalues()
            [1, -1]
            sage: D[0].star_eigenvalues()
            [1]
            sage: D[1].star_eigenvalues()
            [1, -1]
            sage: D[1].plus_subspace().star_eigenvalues()
            [1]
            sage: D[1].minus_subspace().star_eigenvalues()
            [-1]
        """
        try:
            return self.__star_eigenvalues
        except AttributeError:
            pass
        if self.sign() != 0:
            return self.sign()
        M = self.star_involution().matrix()
        R = self.base_ring()
        if M == 1:
            self.__star_eigenvalues = [R(1)]
        elif M == -1:
            self.__star_eigenvalues = [R(-1)]
        else:
            self.__star_eigenvalues = [R(1), R(-1)]
        return self.__star_eigenvalues
    
    def star_decomposition(self):
        """
        """
        S = self.star_involution()
        return S.decomposition()

    def sturm_bound(self):
        r"""
        Returns the Sturm bound for this space of modular symbols.

        Returns a positive integer~$n$ such that the Hecke operators
        $T_1,\ldots, T_n$ acting on \emph{cusp forms} generate the
        Hecke algebra as a $\Z$-module when the character is trivial
        or quadratic.  Otherwise, $T_1,\ldots,T_n$ generate the Hecke
        algebra at least as a $\Z[\eps]$-module, where $\Z[\eps]$ is
        the ring generated by the values of the Dirichlet character
        $\eps$.  Alternatively, this is a bound such that if two cusp
        forms associated to this space of modular symbols are
        congruent modulo $(\lambda, q^n)$, then they are congruent
        modulo $\lambda$.

        REFERENCES:
        See the Agashe-Stein appendix to Lario and Schoof, \emph{Some
        computations with Hecke rings and deformation rings},
        Experimental Math., 11 (2002), no. 2, 303-311.  This result
        originated in the paper Sturm, \emph{On the congruence of
        modular forms}, Springer LNM 1240, 275--280, 1987.

        REMARK:

        Kevin Buzzard pointed out to me (William Stein) in Fall 2002
        that the above bound is fine for $\Gamma_1(N)$ with character,
        as one sees by taking a power of $f$.  More precisely, if $f
        \con 0 \pmod{p}$ for first $s$ coefficients, then $f^r \con 0
        \pmod{p}$ for first $sr$ coefficents.  Since the weight of
        $f^r$ is $r\cdot k(f)$, it follows that if $s \geq b$, where
        $b$ is the Sturm bound for $\Gamma_0(N)$ at weight $k(f)$, then
        $f^r$ has valuation large enough to be forced to be $0$ at
        $r*k(f)$ by Sturm bound (which is valid if we choose $r$
        correctly).  Thus $f \con 0 \pmod{p}$.  Conclusion: For
        $\Gamma_1(N)$ with fixed character, the Sturm bound is
        \emph{exactly} the same as for $\Gamma_0(N)$.

        A key point is that we are finding $\Z[\eps]$ generators for
        the Hecke algebra here, not $\Z$-generators.  So if one wants
        generators for the Hecke algebra over $\Z$, this bound must
        be suitably modified.
        """
        # For Gamma_0(N), n = \frac{k}{12}[\SL_2(\Z):\Gamma_0(N)]
        try:
            return self.__sturm_bound
        except:
            self.__sturm_bound = \
                  int(math.ceil((self.weight() + dims.idxG0(self.level()))/12.0))
        return self.__sturm_bound

    def plus_subspace(self, compute_dual=True):
        """
        Return the subspace of self on which the star involution acts as +1.

        INPUT:
            compute_dual -- bool (default: True) also compute dual subspace.
                            This are useful for many algorithms.
                            
        OUTPUT:
            subspace of modular symbols
        """
        return self.sign_subspace(+1, compute_dual)
    
    def minus_subspace(self, compute_dual=True):
        """
        Return the subspace of self on which the star involution acts as -1.

        INPUT:
            compute_dual -- bool (default: True) also compute dual subspace.
                            This are useful for many algorithms.
        OUTPUT:
            subspace of modular symbols

        """
        return self.sign_subspace(-1, compute_dual)

    def sign_subspace(self, sign, compute_dual=True):
        """
        Return the subspace of self that is fixed under the star involution.

        INPUT:
            sign -- int (either -1, 0 or +1)
            compute_dual -- bool (default: True) also compute dual subspace.
                            This are useful for many algorithms.
        OUTPUT:
            subspace of modular symbols
        """
        sign = int(sign)
        if not sign in [-1, 0, 1]:
            raise ValueError, "sign must be -1, 0 or 1"
        if self.sign() == sign:  # an easy case
            return self
        if self.sign() == -sign:  # another easy case
            return self.zero_subspace()
        if sign == 0:
            # if sign is zero then self.sign() isn't 0 because
            # of the above checks.  
            raise ArithmeticError, "There is no sign 0 subspace of a space of modular symbols with nonzero sign."
        try:
            return self.__plus_subspace[(sign, compute_dual)]
        except AttributeError:
            self.__plus_subspace = {}
        except KeyError:
            pass
        P = self._compute_sign_subspace(sign, compute_dual)
        P.__star_eigenvalue = sign
        self.__plus_subspace[(sign,compute_dual)] = P
        return P
            
    def star_involution(self):
        """
        Return the star involution on self, which is induced by complex
        conjugation on modular symbols.
        """
        raise NotImplementedError

    def hecke_bound(self):
        r"""
        Returns the Sturm bound for this space of modular symbols.
        
        Returns a positive integer~$n$ such that the Hecke operators
        $T_1,\ldots, T_n$ acting on \emph{cusp forms} generate the
        Hecke algebra as a $\Z$-module when the character is trivial
        or quadratic.  Otherwise, $T_1,\ldots,T_n$ generate the Hecke
        algebra at least as a $\Z[\eps]$-module, where $\Z[\eps]$ is
        the ring generated by the values of the Dirichlet character
        $\eps$.  Alternatively, this is a bound such that if two cusp
        forms associated to this space of modular symbols are
        congruent modulo $(\lambda, q^n)$, then they are congruent
        modulo $\lambda$.
        
        REFERENCES:
        See the Agashe-Stein appendix to Lario and Schoof's \emph{Some
        computations with Hecke rings and deformation rings},
        Experimental Math., 11 (2002), no. 2, 303-311.  This result
        originated in the paper Sturm, \emph{On the congruence of
        modular forms}, Springer LNM 1240, 275--280, 1987.
        
        REMARK:

        Kevin Buzzard pointed out to me (William Stein) in Fall 2002
        that the above bound is fine for $\Gamma_1(N)$ with character,
        as one sees by taking a power of $f$.  More precisely, if $f
        \con 0 \pmod{p}$ for first $s$ coefficients, then $f^r \con 0
        \pmod{p}$ for first $sr$ coefficents.  Since the weight of
        $f^r$ is $r\cdot k(f)$, it follows that if $s \geq b$, where
        $b$ is the Sturm bound for $\Gamma_0(N)$ at weight $k(f)$, then
        $f^r$ has valuation large enough to be forced to be $0$ at
        $r*k(f)$ by Sturm bound (which is valid if we choose $r$
        correctly).  Thus $f \con 0 \pmod{p}$.  Conclusion: For
        $\Gamma_1(N)$ with fixed character, the Sturm bound is
        \emph{exactly} the same as for $\Gamma_0(N)$.

        A key point is that we are finding $\Z[\eps]$ generators for
        the Hecke algebra here, not $\Z$-generators.  So if one wants
        generators for the Hecke algebra over $\Z$, this bound must
        be suitably modified. 
        """
        # For Gamma_0(N), n = \frac{k}{12}[\SL_2(\Z):\Gamma_0(N)]
        try:
            return self.__sturm_bound
        except AttributeError:
            self.__sturm_bound = \
                  int(math.ceil((self.weight() + dims.idxG0(self.level()))/12.0))
        return self.__sturm_bound

        


if __name__ ==  '__main__':
    import doctest, sys
    from modsym import *
    doctest.testmod(sys.modules[__name__])
