"""
Free modules

SAGE supports computation with free modules over an arbitrary
commutative ring.  Nontrivial functionality is available over $\Z$ and
fields.  All free modules over an integral domain are equipped with an
embedding in an ambient vector space and an inner product, which you
can specify.

The following example illustrates the creation of both a
vector spaces and a free module over the integers and a submodule of
it.

EXAMPLES:
    sage: V = VectorSpace(RationalField(),3)
    sage: W = V.subspace([V([1,2,7]), V([1,1,0])])
    sage: W
    Vector space of degree 3 and dimension 2 over Rational Field
    Basis matrix:
    [ 1  0 -7]
    [ 0  1  7]
    sage: C = VectorSpaces(FiniteField(7))
    sage: C
    Category of vector spaces over Finite field of size 7
    sage: C(W)
    Vector space of degree 3 and dimension 2 over Finite field of size 7
    Basis matrix:
    [1 0 0]
    [0 1 0]

    
    sage: M = FreeModule(IntegerRing(), 3)
    sage: C = VectorSpaces(FiniteField(7))
    sage: C(M)
    Vector space of dimension 3 over Finite field of size 7
    sage: W = M.subspace([M([1,2,7]), M([8,8,0])])
    sage: C(W)
    Vector space of degree 3 and dimension 2 over Finite field of size 7
    Basis matrix:
    [1 0 0]
    [0 1 0]
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

# python imports
import random
import weakref


# SAGE imports
import sage.modules.free_module_element
import module

import sage.matrix.matrix_space
import sage.matrix.matrix

import sage.misc.latex as latex

import sage.rings.commutative_ring as commutative_ring
import sage.rings.principal_ideal_domain as principal_ideal_domain
import sage.rings.field as field
import sage.rings.integral_domain as integral_domain
import sage.rings.ring as ring
import sage.rings.integer as integer
import sage.structure.gens as gens


###############################################################################
#
# Constructor functions
#
###############################################################################
_cache = {}
def FreeModule(base_ring, rank, sparse=False, inner_product_matrix=None):
    """
    Create the free module over the given commutative ring of the given rank.

    INPUT:
        base_ring -- a commutative ring
        rank -- a nonnegative integer
        sparse -- bool; (default False)
        inner_product_matrix -- the inner product matrix (default None)

    OUTPUT:
        a free module

    EXAMPLES:

    First we illustrate creating free modules over various base
    fields.  Notice how the base field affects the free module that is
    created.  For example, free modules over a field are vector spaces,
    and free modules over a principal ideal domain are special in that
    more functionality is available for them.
    
        sage: FreeModule(Integers(8),10)
        Ambient free module of rank 10 over Ring of integers modulo 8
        sage: FreeModule(RationalField(),10)
        Vector space of dimension 10 over Rational Field
        sage: FreeModule(IntegerRing(),10)
        Ambient free module of rank 10 over the principal ideal domain Integer Ring
        sage: FreeModule(FiniteField(5),10)
        Vector space of dimension 10 over Finite field of size 5
        sage: FreeModule(Integers(7),10)
        Vector space of dimension 10 over Ring of integers modulo 7
        sage: FreeModule(PolynomialRing(RationalField()),5)
        Ambient free module of rank 5 over the principal ideal domain Univariate Polynomial Ring in x over Rational Field
        sage: FreeModule(PolynomialRing(IntegerRing()),5)
        Ambient free module of rank 5 over the integral domain Univariate Polynomial Ring in x over Integer Ring

    Of course we can make rank 0 free modules:
    
        sage: FreeModule(RealField(100),0)
        Vector space of dimension 0 over Real Field with 100 bits of precision

    Next we create a free module with sparse representation of
    elements.  These work exactly as with dense representations, but
    they may use less memory and arithmetic may be faster.  

        sage: M = FreeModule(IntegerRing(),200,sparse=True)
        sage: M.is_sparse()
        True
        sage: type(M.gen(0))
        <class 'sage.modules.free_module_element.FreeModuleElement_generic_sparse'>

    The default is dense.
        sage: M = FreeModule(IntegerRing(),200)
        sage: type(M.gen(0))
        <class 'sage.modules.free_module_element.FreeModuleElement_generic_dense'>


    Note that matrices associated in some way to sparse free modules
    are sparse by default, and the default representation is dense:
        sage: M = FreeModule(Integers(8), 2)
        sage: A = M.basis_matrix()
        sage: A.is_sparse()
        False
        sage: Ms = FreeModule(Integers(8), 2, sparse=True)
        sage: M == Ms   # as mathematical objects they are equal
        True
        sage: Ms.basis_matrix().is_sparse()
        True
        

    We can also specify an inner product matrix, which is used when
    computing inner products of elements.
        sage: Z = IntegerRing()
        sage: A = MatrixSpace(Z,2)([[1,0],[0,-1]])
        sage: M = FreeModule(IntegerRing(),2,inner_product_matrix=A)
        sage: v, w = M.gens()
        sage: v.inner_product(w)
        0
        sage: v.inner_product(v)
        1
        sage: w.inner_product(w)
        -1
        sage: (v+2*w).inner_product(w)
        -2
        
    """
    global _cache
    rank = int(rank)
    key = (base_ring, rank, sparse, inner_product_matrix)
    if _cache.has_key(key):
        M = _cache[key]()
        if not (M is None):
            return M

    if not isinstance(base_ring, commutative_ring.CommutativeRing):
        raise TypeError, "base_ring must be a commutative ring"

    if isinstance(base_ring, field.Field):
        M = FreeModule_ambient_field(base_ring, rank,
                                        sparse=sparse, inner_product_matrix=inner_product_matrix)

    elif isinstance(base_ring, principal_ideal_domain.PrincipalIdealDomain):
        M = FreeModule_ambient_pid(base_ring, rank,
                                   sparse=sparse, inner_product_matrix=inner_product_matrix)
    elif isinstance(base_ring, integral_domain.IntegralDomain):
        M = FreeModule_ambient_domain(base_ring, rank,
                                         sparse=sparse, inner_product_matrix=inner_product_matrix)
    else:
        M = FreeModule_ambient(base_ring, rank,
                                  sparse=sparse, inner_product_matrix=inner_product_matrix)

    _cache[key] = weakref.ref(M)
    return M
    

def VectorSpace(K, dimension, sparse=False,  inner_product_matrix=None):
    """
    EXAMPLES:

    The base can be complicated:
    
        sage: V = VectorSpace(FractionField(PolynomialRing(IntegerRing())),3)
        sage: V
        Vector space of dimension 3 over Fraction field of Univariate Polynomial Ring in x over Integer Ring
        sage: V.basis()
        [(1, 0, 0), (0, 1, 0), (0, 0, 1)]

    The base must be a field or a TypeError is raised.
    sage: VectorSpace(IntegerRing(),5)
    Traceback (most recent call last):
    ...
    TypeError: K must be a field
    """
    
    if not isinstance(K, field.Field):
        raise TypeError, "K must be a field"
    return FreeModule(K, dimension, sparse, inner_product_matrix)


###############################################################################
#
# Base class for all free modules
#
###############################################################################

def is_FreeModule(x):
    return isinstance(x, FreeModule_generic)

class FreeModule_generic(module.Module):
    """
    Base class for all free modules.
    """
    def __init__(self, base_ring, rank, degree, sparse=False, inner_product_matrix=None):
        """
        Create the free module of given rank over the given base_ring.

        INPUT:
            base_ring -- a commutative ring
            rank -- a non-negative integer
        """
        if not isinstance(base_ring, commutative_ring.CommutativeRing):
            raise TypeError, "base_ring must be a commutative ring"
        rank = int(rank)
        if rank < 0:
            raise ValueError, "rank must be nonnegative"
        degree = int(degree)
        if degree < 0:
            raise ValueError, "degree must be nonnegative"
        
        self.__base_ring = base_ring
        self.__rank = rank
        self.__degree = degree
        self.__is_sparse = sparse
        if inner_product_matrix != None:
            self.__inner_product_matrix = inner_product_matrix
        else:
            self.__inner_product_is_dot_product = True

        if self.is_sparse():
            self.__element_class = sage.modules.free_module_element.FreeModuleElement_generic_sparse
            return
        self.__element_class = sage.modules.free_module_element.FreeModuleElement_generic_dense

    def __call__(self, x, coerce_entries=True, copy=True, check_element=True):
        if isinstance(x, (int, long, integer.Integer)) and x==0:
            return self.zero_vector()
        elif isinstance(x, sage.modules.free_module_element.FreeModuleElement):
            if x.parent() == self:
                if copy:
                    return x.copy()
                else:
                    return x
            x = x.list()
        w = self.__element_class(self, x, coerce_entries, copy)
        if check_element:
            self.coordinates(w)
        return w

    def __cmp__(self, right):
        raise NotImplementedError


    def __contains__(self, v):
        r"""
        EXAMPLES:

        We create the module $\Z^3$, and the submodule generated by
        one vector $(1,1,0)$, and check whether certain elements are
        in the submodule.
        
            sage: R = FreeModule(IntegerRing(), 3)
            sage: V = R.submodule([R.gen(0) + R.gen(1)])
            sage: R.gen(0) + R.gen(1) in V
            True
            sage: R.gen(0) + 2*R.gen(1) in V
            False
            
            sage: Q = RationalField()
            sage: w = Q('1/2')*(R.gen(0) + R.gen(1))
            sage: w
            (1/2, 1/2, 0)
            sage: w.parent()
            Vector space of dimension 3 over Rational Field
            sage: w in V
            False
            sage: V.coordinates(w)
            [1/2]
            
        """
        if not isinstance(v, sage.modules.free_module_element.FreeModuleElement):
            return False
        if v.parent() == self:
            return True
        try:
            c = self.coordinates(v)
        except ArithmeticError:
            return False
        # Finally, check that each coordinate lies in the base ring.
        R = self.base_ring()
        if not self.base_ring().is_field():
            for a in c:
                try:
                    b = R(a)
                except TypeError:
                    return False
        return True

    def __len__(self):
        return len(self.base_ring())**self.rank()

    def ambient_module(self):
        """
        Return the ambient module associated to this module.
        """
        return FreeModule(self.base_ring(), self.degree())
    
    def base_ring(self):
        """
        Return the base ring of this module.
        """
        return self.__base_ring

    def basis(self):
        """
        Return the basis of this module.
        """
        raise NotImplementedError

    def basis_matrix(self):
        """
        Return the matrix whose rows are the basis for this free module.
        """
        try:
            return self.__basis_matrix
        except AttributeError:
            MAT = sage.matrix.matrix_space.MatrixSpace(self.base_ring(),
                            len(self.basis()), self.degree(),
                            sparse = self.is_sparse())
            self.__basis_matrix = MAT(self.basis())
            return self.__basis_matrix

    def category(self):
        """
        Return the category to which this free module belongs.
        """
        import sage.categories.all        
        return sage.categories.all.FreeModules(self.base_ring())

    def matrix(self):
        """
        Return the basis matrix of this module, which is the matrix
        whose rows are a basis for this module.

        EXAMPLES:
            sage: M = FreeModule(IntegerRing(), 2)
            sage: M.matrix()
            [1 0]
            [0 1]
            sage: M.submodule([M.gen(0) + M.gen(1), M.gen(0) - 2*M.gen(1)]).matrix()
            [1 1]
            [0 3]
        """
        return self.basis_matrix()

    def coordinates(self, v):
        """
        Write v in terms of the basis for self.

        Returns a list c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.

        EXAMPLES:
            sage: M = FreeModule(IntegerRing(), 2); M0,M1=M.gens()
            sage: W = M.submodule([M0 + M1, M0 - 2*M1])
            sage: W.coordinates(2*M0-M1)
            [2, -1]
        """
        return self.coordinate_vector(v).list()

    def coordinate_vector(self, v):
        """
        Return the a vector whose cofficients give v as a linear combination
        of the basis for self.

        EXAMPLES:
            sage: M = FreeModule(IntegerRing(), 2); M0,M1=M.gens()
            sage: W = M.submodule([M0 + M1, M0 - 2*M1])
            sage: W.coordinate_vector(2*M0 - M1)
            (2, -1)
        """
        raise NotImplementedError

    def degree(self):
        """
        Return the degree of this free module.  This is the dimension of the
        ambient vector space in which it is embedded.

        EXAMPLES:
            sage: M = FreeModule(IntegerRing(), 10)
            sage: W = M.submodule([M.gen(0), 2*M.gen(3) - M.gen(0), M.gen(0) + M.gen(3)])
            sage: W.degree()
            10
            sage: W.rank()
            2
        """
        return self.__degree
        
    def dimension(self):
        """
        Return the dimension of this free module.

        EXAMPLES:
            sage: M = FreeModule(FiniteField(19), 100)
            sage: W = M.submodule([M.gen(50)])
            sage: W.dimension()
            1
        """
        return self.rank()

    def discriminant(self):
        """
        Return the discriminant of this free module.
        """
        return self.inner_product_matrix().determinant()

    def free_module(self):
        return self

    def gen(self, i=0):
        if i < 0 or i >= self.rank():
            raise ValueError, "Generator %s not defined."%i
        return self.basis()[int(i)]

    def inner_product_matrix(self):
        try:
            return self.__inner_product_matrix
        except AttributeError:
            # Not set, so induced by the ambient space or the identity matrix
            # if this space is ambient.
            if isinstance(self, FreeModule_ambient):
                M = sage.matrix.matrix_space.MatrixSpace(
                     self.base_ring(), self.rank(), sparse=True)(1)
            else:
                # TODO: probably much better way to do this using matrix multiplication!
                A = self.ambient_vector_space()
                I = [A(v).inner_product(A(w)) for v in self.basis() for w in self.basis()]
                M = sage.matrix.matrix_space.MatrixSpace( \
                      self.base_field(), self.rank(), sparse=False)(I)
                
            self.__inner_product_matrix = M 
        return self.__inner_product_matrix

    def inner_product_is_dot_product(self):
        try:
            return self.__inner_product_is_dot_product
        except AttributeError:
            self.__inner_product_is_dot_product  = (self.inner_product_matrix() == 1)
            return self.__inner_product_is_dot_product

    def set_inner_product_matrix(self, A):
        """
        Set the inner product matrix of self to A.
        """
        if not isinstance(A, sage.matrix.matrix.Matrix):
            raise TypeError, "A must be a matrix."
        if A.nrows() != self.rank():
            raise ArithmeticError, "A must have %s rows but it has %s rows"%(self.rank(),
                                                                             A.nrows())
        if A.ncols() != self.rank():
            raise ArithmeticError, "A must have %s columns but it has %s rows"%(self.rank(),
                                                                             A.ncols())
        self.__inner_product_matrix = A
        del self.__inner_product_is_dot_product
    
    def is_dense(self):
        return not self.__is_sparse
        
    def is_full(self):
        """
        True if the rank equals the degree.
        """
        return self.rank() == self.degree()

    def is_finite(self):
        return self.base_ring().is_finite() or self.rank() == 0
        
    def is_sparse(self):
        return self.__is_sparse

    def ngens(self):
        try:
            return self.__ngens
        except AttributeError:
            self.__ngens = len(self.basis())
        return self.__ngens

    def nonembedded_free_module(self):
        return FreeModule(self.base_ring(), self.rank())

    def random_element(self, bound=0, prob=1.0):
        """
        Returns a random element of self.
        """
        R = self.base_ring()
        v = self(0)
        prob = float(prob)
        for i in xrange(self.degree()):
            if random.random() <= prob:
                v[i] += R.random_element(bound)
        return v

    def rank(self):
        """
        Return the rank of this free module.
        """
        return self.__rank

    def zero_vector(self):
        return self.__element_class(self, 0)


class FreeModule_generic_pid(FreeModule_generic):
    """
    Base class for all free modules over a PID
    """
    def __init__(self, base_ring, rank, degree, sparse=False, inner_product_matrix=None):
        if not isinstance(base_ring, principal_ideal_domain.PrincipalIdealDomain):
            raise TypeError, "base_ring must be a principal ideal domain"
        FreeModule_generic.__init__(self, base_ring, rank, degree, sparse, inner_product_matrix)

    def __mul__(self, other):
        """
        Return the product of this module by the number other, which
        is the module spanned by other times each basis vector.

        EXAMPLES:
            sage: M = FreeModule(IntegerRing(), 3)
            sage: 2*M
            Free module of degree 3 and rank 3 over Integer Ring
            Echelon basis matrix:
            [2 0 0]
            [0 2 0]
            [0 0 2]

            sage: a = RationalField()('1/3')
            sage: a*M
            Free module of degree 3 and rank 3 over Integer Ring
            Echelon basis matrix:
            [1/3   0   0]
            [  0 1/3   0]
            [  0   0 1/3]
        """
        if other == 0:
            return self.zero_submodule()
        if other == 1 or other == -1:
            return self
        return self.ambient_module().submodule([v*other for v in self.basis()])

    def is_vector(self):   # for purposes of multiplication by a scalar
        return True

    def __rmul__(self, left):
        return self*left
        
    def __add__(self, other):
        r"""
        Return the sum of self and other, where both self and other
        must be submodules of the ambient vector space.

        EXAMPLES:
        We add two vector spaces:
        
            sage: V  = VectorSpace(RationalField(), 3)
            sage: W  = V.subspace([V([1,1,0])])
            sage: W2 = V.subspace([V([1,-1,0])])
            sage: W + W2
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [1 0 0]
            [0 1 0]
            
        We add two free $\Z$-modules.
            sage: M = FreeModule(IntegerRing(), 3)
            sage: W = M.submodule([M([1,0,2])])
            sage: W2 = M.submodule([M([2,0,-4])])
            sage: W + W2
            Free module of degree 3 and rank 2 over Integer Ring
            Echelon basis matrix:
            [1 0 2]
            [0 0 8]

        We can also add free $\Z$-modules embedded non-integrally
        into an ambient space.

            sage: V = VectorSpace(RationalField(), 3)
            sage: W = M.submodule([Q('1/2')*V.gen(0) - Q('1/3')*V.gen(1)])
            sage: W2 = M.submodule([Q('1/3')*V.gen(0) + V.gen(1)])
            sage: W + W2
            Free module of degree 3 and rank 2 over Integer Ring
            Echelon basis matrix:
            [ 1/6  7/3    0]
            [   0 11/3    0]
        """
        if not isinstance(other, FreeModule_generic):
            raise TypeError, "other (=%s) must be a free module"%other
        if not (self.ambient_vector_space() is other.ambient_vector_space()):
            raise TypeError, "ambient vector spaces must be equal"
        return self.ambient_module().submodule(self.basis() + other.basis())
        
    
    def base_field(self):
        return self.base_ring().fraction_field()

    def basis_matrix(self):
        """
        Return the matrix whose rows are the basis for this free module.
        """
        try:
            return self.__basis_matrix
        except AttributeError:
            MAT = sage.matrix.matrix_space.MatrixSpace(self.base_field(),
                            len(self.basis()), self.degree(),
                            sparse = self.is_sparse())
            self.__basis_matrix = MAT(self.basis())
            return self.__basis_matrix

    def intersection(self, other):
        raise NotImplementedError

    def is_submodule(self, other):
        """
        True if this module is a submodule of other.
        """
        if not isinstance(other, FreeModule_generic):
            return False
        if self.ambient_vector_space() != other.ambient_vector_space():
            return False
        if other.is_ambient_vector_space():
            return True
        if other.rank() < self.rank():
            return False
        for b in self.basis():
            if not (b in other):
                return False
        return True

    def zero_submodule(self):
        return FreeModule_submodule_pid(self.ambient_vector_space(),[], check=False)

    def submodule(self, gens, check=True):
        r"""
        Create the R-submodule of the ambient vector space with given
        generators, where R is the base ring of self.

        INPUT:
            gens -- a list of vector in self
            check -- whether or not to verify that each gen is in
                     the ambient vector space
            
        OUTPUT:
            FreeModule -- the submodule spanned by the vectors in the list gens.
            The basis for the subspace is always put in reduced row echelon form.

        EXAMPLES:
        We create a submodule of $\Z^3$:
            sage: M = FreeModule(IntegerRing(), 3)
            sage: B = M.basis()
            sage: W = M.submodule([B[0]+B[1], 2*B[1]-B[2]])
            sage: W
            Free module of degree 3 and rank 2 over Integer Ring
            Echelon basis matrix:
            [ 1  1  0]
            [ 0  2 -1]
        """
# need general HNF for this:
#
#        Next we create a submodule of a free module over the principal ideal
#        domain $\Q[x]$:
#            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
#            sage: M = FreeModule(R, 3)
#            sage: B = M.basis()
#            sage: W = M.submodule([x*B[0], 2*B[1]- x*B[2]])
#            sage: W
#            Free module of degree 3 and rank 2 over Integer Ring
#            Echelon basis matrix:
#            [1 1 0]
#            [2 0 1]
#        """
        return FreeModule_submodule_pid(self, gens, check=check)
        
    def submodule_with_basis(self, basis, check=True):
        """
        Create the R-submodule of the ambient vector space with given
        basis, where R is the base ring of self.

        INPUT:
            basis -- a list of linearly independent vectors
            check -- whether or not to verify that each gen is in
                     the ambient vector space

        OUTPUT:
            FreeModule -- the R-submodule with given basis

        EXAMPLES:
        First we create a submodule of $\Z^3$:
            sage: M = FreeModule(IntegerRing(), 3)
            sage: B = M.basis()
            sage: W = M.submodule_with_basis([B[0]+B[1], 2*B[1]-B[2]])
            sage: W
            Free module of degree 3 and rank 2 over Integer Ring
            User basis matrix:
            [ 1  1  0]
            [ 0  2 -1]

        """
# need general HNF for this:
#
#        Next we create a submodule of a free module over the principal ideal
#        domain $\Q[x]$:
#            sage: R = PolynomialRing(RationalField(), 'x'); x = R.gen()
#            sage: M = FreeModule(R, 3)
#            sage: B = M.basis()
#            sage: W = M.submodule_with_basis([x*B[0], 2*B[1]- x*B[2]])
#            sage: W
#            ???
#        """
        return FreeModule_submodule_with_basis_pid(
                                    self, basis, check=check)

    def subspace(self, gens, check=True):
        """
        Create the vector subspace of the ambient vector space with
        given generators.

        INPUT:
            gens -- a list of vector in self
            check -- whether or not to verify that each gen is in
                     the ambient vector space
            
        OUTPUT:
            a vector subspace

        EXAMPLES:
            sage: V = VectorSpace(RationalField(), 3)
            sage: B = V.basis()
            sage: W = V.subspace([B[0]+B[1], 2*B[1]-B[2]])
            sage: W
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [   1    0  1/2]
            [   0    1 -1/2]

            sage: x = PolynomialRing(RationalField()).gen()
            sage: K = NumberField(x**2 + 1, 'a'); a = K.gen()
            sage: V = VectorSpace(K, 3)
            sage: V.subspace([2*V.gen(0) + 3*V.gen(2)])
            Vector space of degree 3 and dimension 1 over Number Field in a with defining polynomial x^2 + 1
            Basis matrix:
            [  1   0 3/2]
        """
        if not isinstance(gens, list):
            raise TypeError, "gens must be a list"
        return FreeModule_submodule_field(self.ambient_vector_space(),
                                        gens, check=check)

    def subspace_with_basis(self, basis, check=True):
        """
        Create the vector subspace of the ambient vector space with
        given basis.

        INPUT:
            basis -- a list of linearly independent vectors
            check -- whether or not to verify that each gen is in
                     the ambient vector space

        OUTPUT:
            a vector subspace with user-specified basis
            
        EXAMPLES:
            sage: V = VectorSpace(RationalField(), 3)
            sage: B = V.basis()
            sage: W = V.subspace_with_basis([B[0]+B[1], 2*B[1]-B[2]])
            sage: W
            Vector space of degree 3 and dimension 2 over Rational Field
            User basis matrix:
            [ 1  1  0]
            [ 0  2 -1]
        """
        return FreeModule_submodule_with_basis_field(
                    self.ambient_vector_space(), basis, check=check)        



class FreeModule_generic_field(FreeModule_generic_pid):
    """
    Base class for all free modules over fields (i.e., vector spaces).
    """
    def __init__(self, base_field, dimension, degree, sparse=False, inner_product_matrix=None):
        if not isinstance(base_field, field.Field):
            raise TypeError, "base_field (=%s) must be a field"%base_field
        FreeModule_generic_pid.__init__(self, base_ring, dimension, degree, sparse, inner_product_matrix)

    def __mul__(self, other):
        """
        Return the product of this module by the number other, which is the module
        spanned by other times each basis vector.
        """
        if other == 0:
            return self.zero_submodule()
        return self

    def __rmul__(self, left):
        return self*left

    def __add__(self, other):
        if not isinstance(other, FreeModule_generic_field):
            raise TypeError, "other must be a Vector Space"
        V = self.ambient_vector_space() 
        if V != other.ambient_vector_space():
            raise ArithmeticError, "self and other must have the same ambient space"
        return V.subspace(self.basis() + other.basis())

    def category(self):
        """
        Return the category to which this vector space belongs.
        """
        import sage.categories.all
        return sage.categories.all.VectorSpaces(self.base_field())

    def intersection(self, other):
        """
        Return the intersection of self and other, which must be
        R-submodules of a common ambient vector space.

        EXAMPLES:
            sage: V  = VectorSpace(RationalField(),3)
            sage: W1 = V.subspace([V.gen(0), V.gen(0) + V.gen(1)])
            sage: W2 = V.subspace([V.gen(1), V.gen(2)])
            sage: W1.intersection(W2)
            Vector space of degree 3 and dimension 1 over Rational Field
            Basis matrix:
            [0 1 0]
            sage: W2.intersection(W1)
            Vector space of degree 3 and dimension 1 over Rational Field
            Basis matrix:
            [0 1 0]
            sage: V.intersection(W1)
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [1 0 0]
            [0 1 0]
            sage: W1.intersection(V)
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [1 0 0]
            [0 1 0]
            sage: Z = V.subspace([])
            sage: W1.intersection(Z)
            Vector space of degree 3 and dimension 0 over Rational Field
            Basis matrix:
            []
        """
        if self.ambient_vector_space() != other.ambient_vector_space():
            raise ArithmeticError, "self and other must have the same ambient space."

        # dispense with the three easy cases
        if self.is_ambient_vector_space():
            return other
        elif other.is_ambient_vector_space():
            return self
        elif self.dimension() == 0 or other.dimension() == 0:
            return self.zero_submodule()

        # standard algorithm for computing intersection of general subspaces
        if self.dimension() <= other.dimension():
            V1 = self; V2 = other
        else:
            V1 = other; V2 = self
        A1 = V1.basis_matrix()
        A2 = V2.basis_matrix()
        S  = A1.stack(A2)
        K  = S.kernel()
        n = V1.dimension()
        B = [A1.linear_combination_of_rows(v.list()[:n]) for v in K.basis()]
        return self.ambient_vector_space().subspace(B, check=False)

    def is_subspace(self, other):
        """
        True if this vector space is a subspace of other.

        EXAMPLES:
            sage: V = VectorSpace(RationalField(),3)
            sage: W = V.subspace([V.gen(0), V.gen(0) + V.gen(1)])
            sage: W2 = V.subspace([V.gen(1)])
            sage: W.is_subspace(V)
            True
            sage: W2.is_subspace(V)
            True
            sage: W.is_subspace(W2)
            False
            sage: W2.is_subspace(W)
            True
        """
        return self.is_submodule(other)

    def submodule(self, gens, check=True):
        return self.subspace(gens, check)

    def submodule_with_basis(self, gens, check=True):
        return self.subspace_with_basis(gens, check)

    def vector_space(self):
        return self

    def zero_submodule(self):
        return self.zero_subspace()
        
    def zero_subspace(self):
        return FreeModule_submodule_field(self.ambient_vector_space(),[], check=False)
        

###############################################################################
#
# Generic ambient free modules, i.e., of the form R^n for some commutative ring R.
#
###############################################################################

class FreeModule_ambient(FreeModule_generic):
    """
    Ambient free module over a commutative ring.
    """
    def __init__(self, base_ring, rank, sparse=False, inner_product_matrix=None):
        """
        Create the free module of given rank over the given base_ring.

        INPUT:
            base_ring -- a commutative ring
            rank -- a non-negative integer
        """
        FreeModule_generic.__init__(self, base_ring, rank,
                                    rank, sparse, inner_product_matrix)

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, FreeModule_generic):
            return 1
        c = cmp(self.rank(), other.rank())
        if c: return c
        c = cmp(self.base_ring(), other.base_ring())
        if c: return c
        if isinstance(other, FreeModule_ambient):
            return 0
        else:  # now other is not ambient
            return -cmp(other, self)
        
    def __repr__(self):
        return "Ambient free module of rank %s over %s"%(self.rank(), self.base_ring())

    def _latex_(self):
        t = "%s"%latex.latex(self.base_ring())
        if t.find(" ") != -1:
            t = "(%s)"%t
        return "%s^{\\oplus %s}"%(t, self.rank())

    def is_ambient(self):
        return True

    def echelon_coordinate_vector(self, v):
        """
        Write v in terms of the basis for self.  This is trivial
        since self is an ambient space.
        """
        if not isinstance(v, sage.modules.free_module_element.FreeModuleElement):
            raise TypeError, "v must be a Vector"
        if v.parent().ambient_vector_space() != self.ambient_vector_space():
            raise ArithmeticError, "v is not in self"
        return list(v.list())

    def basis(self):
        try:
            return self.__basis
        except  AttributeError:
            ZERO = self(0)
            self.__basis = []
            for n in range(self.rank()):
                v = ZERO.copy()
                v[n] = 1
                self.__basis.append(v)
            return self.__basis

    def change_ring(self, R):
        """
        Return the ambient free module over R of the same rank as self.
        """
        if self.base_ring() == R:
            return self
        return FreeModule(R, self.rank())

    def linear_combination_of_basis(self, v):
        """
        Return the linear combination of the basis for self
        obtained from the elements of the list v.
        """
        return self(v)

    def coordinate_vector(self, v):
        """
        Write v in terms of the user basis for self.

        Returns a vector c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        return self.ambient_vector_space()(v)
    
    def echelon_coordinate_vector(self, v):
        return self.coordinate_vector(v)

    
    
###############################################################################
#
# Ambient free modules over an integral domain.  
#
###############################################################################

class FreeModule_ambient_domain(FreeModule_ambient):
    """
    Ambient free module over an integral domain.
    """
    def __init__(self, base_ring, rank, sparse=False, inner_product_matrix=None):
        FreeModule_ambient.__init__(self, base_ring, rank, sparse, inner_product_matrix)

    def __repr__(self):
        return "Ambient free module of rank %s over the integral domain %s"%(
            self.rank(), self.base_ring())

    def base_field(self):
        return self.base_ring().fraction_field()

    def ambient_vector_space(self):
        """
        Returns the ambient vector space, which is this free module tensored
        with its fraction field.
        """
        try:
            return self.__ambient_vector_space
        except AttributeError:
            if self.inner_product_is_dot_product():
                mat = None
            else:
                mat = self.inner_product_matrix()
            self.__ambient_vector_space = FreeModule(self.base_field(), self.rank(),
                                             sparse = self.is_sparse(),
                                             inner_product_matrix = mat)
            return self.__ambient_vector_space
        
    def is_ambient_vector_space(self):
        """
        Returns True if this is the ambient vector space.
        """
        return True

    def vector_space(self):
        return self.change_ring(self.base_ring().fraction_field())

###############################################################################
#
# Ambient free modules over a principal ideal domain.
#
###############################################################################

class FreeModule_ambient_pid(FreeModule_generic_pid, FreeModule_ambient_domain):
    """
    Ambient free module over a principal ideal domain.
    """
    def __init__(self, base_ring, rank, sparse=False, inner_product_matrix=None):
        """
        Create the ambient free module of given rank over the given
        principal ideal domain

        INPUT:
            base_ring -- a principal ideal domain
            rank -- a non-negative integer
        """
        FreeModule_ambient_domain.__init__(self, base_ring, rank, sparse, inner_product_matrix)
    
    def __repr__(self):
        return "Ambient free module of rank %s over the principal ideal domain %s"%(
            self.rank(), self.base_ring())

    def is_ambient_vector_space(self):
        """
        Returns True if this is the ambient vector space.
        """
        return False


###############################################################################
#
# Ambient free modules over a field (i.e., a vector space).
#
###############################################################################

class FreeModule_ambient_field(FreeModule_generic_field, FreeModule_ambient_pid):
    def __init__(self, base_field, dimension, sparse=False, inner_product_matrix=None):
        FreeModule_ambient_pid.__init__(self, base_field, dimension, sparse, inner_product_matrix)

    def __repr__(self):
        return "Vector space of dimension %s over %s"%(self.dimension(), self.base_ring())

    def ambient_vector_space(self):
        """
        Returns the ambient vector space.
        """
        return self

    def base_field(self):
        return self.base_ring()


###############################################################################
#
# R-Submodule of K^n where K is the fraction field of a principal ideal domain $R$.
#
###############################################################################

class FreeModule_submodule_with_basis_pid(FreeModule_generic_pid):
    """
    An $R$-submodule of $K^n$ with distinguished basis, where $K$ is
    the fraction field of a principal ideal domain $R$.
    """
    def __init__(self, ambient, basis, check=True, echelonize=False, inner_product_matrix=None):
        if not isinstance(ambient, FreeModule_ambient_pid):
            raise TypeError, "ambient (=%s) must be ambient."%ambient
        if not isinstance(basis, list):
            raise TypeError, "basis (=%s) must be a list"%basis
        d = self._denominator(basis)
        if echelonize:
            MAT = sage.matrix.matrix_space.MatrixSpace(ambient.base_ring(),
                            len(basis), ambient.degree(),
                            sparse = ambient.is_sparse())
            if d > 1:
                basis = [x*d for x in basis]
            A = MAT(basis)
            E = A.echelon_form(include_zero_rows=False)
            if d != 1:
                E = E.matrix_over_field()*(~d)   # divide out denominator
            basis = E.rows()
            
        FreeModule_generic.__init__(self, ambient.base_ring(), len(basis), ambient.degree(),
                                    ambient.is_sparse(), inner_product_matrix=inner_product_matrix)
        V = ambient.ambient_vector_space()
        if check:
            for i in range(len(basis)):
                x = basis[i]
                if not (x in V):
                    try:
                        basis[i] = V(x)
                    except TypeError:
                        raise TypeError, "each element of basis must be in the ambient vector space"
        self.__ambient_vector_space = V
        self.__basis = basis
        if not echelonize:
            E = self.basis_matrix().echelon_form(include_zero_rows=False)
        self.__echelonized_basis = E.rows()
        if check and len(basis) > 0 and self.__echelonized_basis[len(basis)-1] == 0:
            raise ArithmeticError, "basis vectors must be linearly independent."

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, FreeModule_generic):
            return -1
        c = cmp(self.ambient_vector_space(), other.ambient_vector_space())
        if c: return c
        c = cmp(self.rank(), other.rank())
        if c: return c
        c = cmp(self.base_ring(), other.base_ring())
        if c: return c
        return cmp(self.basis(), other.basis())
            
    def __repr__(self):
        s = "Free module of degree %s and rank %s over %s\n"%(
            self.degree(), self.rank(), self.base_ring()) + \
            "User basis matrix:\n%s"%self.basis_matrix()
        return s

    def _latex_(self):
        #return "\\mbox{\\rm Span}_{%s}(%s)"%(latex.latex(self.base_ring()), latex.latex(self.basis()))
        return "\\text{Free module over %s spanned by the rows of:}%s."%(
            latex.latex(self.base_ring()),latex.latex(self.basis_matrix()))

    def _denominator(self, B):  # for internal use
        if len(B) == 0:
            return 1
        d = B[0].denominator()
        for x in B[1:]:
            d = d.lcm(x.denominator())
        return d

    def echelon_coordinates(self, v):
        """
        Write $v$ in terms of the echelon basis for self.

        Returns a list $c$ such that if $B$ is the basis for self, then
        $$
                \sum c_i B_i = v.
        $$        
        If $v$ is not in self, raises an \code{ArithmeticError} exception.
        """
        if not isinstance(v, sage.modules.free_module_element.FreeModuleElement):
            raise TypeError, "v must be a free module element"
        if v.degree() != self.degree():
            raise ArithmeticError, "v (=%s) is not in self"%v
        nz = v.nonzero_positions()
        zero = self.base_ring()(0)
        if len(nz) == 0:
            return [zero]*self.dimension()
        c = []
        v = v.parent().ambient_vector_space()(v)
        for b in self.echelonized_basis():
            nz = b.nonzero_positions()
            j = nz[0]
            if v[j].is_zero():
                c.append(zero)
                continue
            q = v[j] / b[j]
            c.append(q)
            # NOTE: It is *very* important that the product be written
            # b*v[j] instead of v[j]*b below.  Otherwise the
            # arithmetic goes through some coercision tests, which
            # test inclusion, which in some cases call the coordinates
            # function and when get a huge (but not infinite) wasteful
            # recursion.
            v -= b*q
        if not v.is_zero():
            raise ArithmeticError, "v (=%s) is not in self"%v
        return c

    def user_to_echelon_matrix(self):
        try:
            return self.__user_to_echelon_matrix
        except AttributeError:
            rows = sum([self.echelon_coordinates(b) for b in self.basis()], [])
            M = sage.matrix.matrix_space.MatrixSpace(self.base_ring().fraction_field(),
                                         self.dimension(),
                                         sparse = self.is_sparse())
            self.__user_to_echelon_matrix = M(rows)
        return self.__user_to_echelon_matrix
        

    def vector_space(self):
        return self.change_ring(self.base_ring().fraction_field())

    def echelon_to_user_matrix(self):
        try:
            return self.__echelon_to_user_matrix
        except AttributeError:
            self.__echelon_to_user_matrix = ~self.user_to_echelon_matrix()
            return self.__echelon_to_user_matrix

    def ambient_vector_space(self):
        return self.__ambient_vector_space

    def basis(self):
        return self.__basis

    def change_ring(self, R):
        """
        Return the free module over R obtained by coercing each
        element of self into a vector over the fraction field of R,
        then taking the resulting R-module.

        INPUT:
            R -- a principal ideal domain

        EXAMPLES:
            sage: Q = RationalField()
            sage: V = VectorSpace(Q, 3)
            sage: W = V.subspace([V.gen(0) + Q('1/2')*V.gen(1)])
            sage: W.change_ring(GF(7))
            Vector space of degree 3 and dimension 1 over Finite field of size 7
            Basis matrix:
            [1 4 0]
        """
        if self.base_ring() == R:
            return self
        K = R.fraction_field()
        V = VectorSpace(K, self.degree())
        B = [V(b) for b in self.basis()]
        M = FreeModule(R, self.degree())
        return M.submodule(B)

    def coordinate_vector(self, v):
        """
        Write v in terms of the user basis for self.

        Returns a vector c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        # First find the coordinates of v wrt echelon basis.
        w = self.echelon_coordinate_vector(v)
        # Next use transformation matrix from echelon basis to
        # user basis.
        T = self.echelon_to_user_matrix()
        return T.linear_combination_of_rows(w)

    def echelonized_basis(self):
        return self.__echelonized_basis
    
    def echelon_coordinate_vector(self, v):
        """
        Write v in terms of the user basis for self.

        Returns a vector c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        return FreeModule(self.base_ring().fraction_field(), self.rank())(self.echelon_coordinates(v))

    def is_ambient_vector_space(self):
        return False

    def linear_combination_of_basis(self, v):
        """
        Return the linear combination of the basis for self
        obtained from the coordinates of v.
        """
        return self.basis_matrix().linear_combination_of_rows(v)


class FreeModule_submodule_pid(FreeModule_submodule_with_basis_pid):
    """
    An $R$-submodule of $K^n$ where $K$ is the fraction field of a
    principal ideal domain $R$.
    """
    def __init__(self, ambient, gens, check=True, inner_product_matrix=None):
        FreeModule_submodule_with_basis_pid.__init__(self, ambient,
                                                     gens, echelonize=True,
                                                     inner_product_matrix=inner_product_matrix)

    def __repr__(self):
        s = "Free module of degree %s and rank %s over %s\n"%(
            self.degree(), self.rank(), self.base_ring()) + \
            "Echelon basis matrix:\n%s"%self.basis_matrix()
        return s

    def coordinate_vector(self, v):
        """
        Write v in terms of the user basis for self.

        Returns a vector c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        return self.echelon_coordinate_vector(v)
    

class FreeModule_submodule_with_basis_field(FreeModule_generic_field, FreeModule_submodule_with_basis_pid):
    """
    A vector subspace with distinguished basis.
    """
    def __init__(self, ambient, basis, check=True, echelonize=False,
                 inner_product_matrix=None):
        FreeModule_submodule_with_basis_pid.__init__(self, ambient, basis,
                                                     check,
                                                     inner_product_matrix=inner_product_matrix,
                                                     echelonize=echelonize)

    def __repr__(self):
        return "Vector space of degree %s and dimension %s over %s\n"%(
                self.degree(), self.dimension(), self.base_field()) + \
                "User basis matrix:\n%s"%self.basis_matrix()

    def _denominator(self, B):  # for internal use
        return 1

    def is_ambient(self):
        return False

    def is_ambient_vector_space(self):
        """
        Returns True if this is the ambient vector space.
        """
        return self.dimension() == self.degree()



class FreeModule_submodule_field(FreeModule_submodule_with_basis_field):
    """
    A vector subspace.
    """
    def __init__(self, ambient, gens, check=True, inner_product_matrix=None):
        FreeModule_submodule_with_basis_field.__init__(self, ambient, gens, check,
                                                       echelonize=True,
                    inner_product_matrix=inner_product_matrix)

    def __repr__(self):
        return "Vector space of degree %s and dimension %s over %s\n"%(
                self.degree(), self.dimension(), self.base_field()) + \
                "Basis matrix:\n%s"%self.basis_matrix()

    def echelon_coordinates(self, v):
        """
        Write v in terms of the user basis for self.

        Returns a vector c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        if not isinstance(v, sage.modules.free_module_element.FreeModuleElement):
            raise TypeError, "v must be a free module element"
        if v.degree() != self.degree():
            raise ArithmeticError, "v (=%s) is not in self"%v
        nz = v.nonzero_positions()
        zero = self.base_field()(0)
        if len(nz) == 0:
            return [zero]*self.dimension()
        c = []
        for b in self.echelonized_basis():
            j = b.nonzero_positions()[0]
            c.append(v[j])
            # NOTE: It is *very* important that the product be written b*v[j] instead
            # of v[j]*b below.  Otherwise the arithmetic goes through some coercision
            # tests, which test inclusion, which in some cases call the coordinates
            # function and when get a huge (but not infinite) wasteful recursion.
            v -= b*v[j]    
        if v != 0:
            raise ArithmeticError, "v (=%s) is not in self"%v
        return c

    def coordinate_vector(self, v):
        """
        Write v in terms of the user basis for self.

        Returns a vector c such that if B is the basis for self, then 
                sum c[i] B[i] = v
        If v is not in self, raises an ArithmeticError exception.
        """
        return self.echelon_coordinate_vector(v)

    

if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.all import *
    from sage.matrix.all import *
    from sage.categories.all import *
    doctest.testmod(sys.modules[__name__])
