"""
A complex number
"""

import math
import operator

from coerce import bin_op
import complex_field
import sage.rings.ring_element as ring_element
import real_field
import sage.rings.arith
import sage.misc.misc
import sage.libs.pari
#from sage.databases.odlyzko import zeta_zeroes

class ComplexNumber(ring_element.RingElement):
    """
    A complex number.
    """
    def __init__(self, real, imag=None, prec=53):
        if imag == None:
            if isinstance(real, list) or isinstance(real,tuple):
                real, imag = real
            elif isinstance(real, complex):
                real, imag = real.real, real.imag
            elif isinstance(real, sage.libs.pari.gen):
                x = real
                real = str(x.real())
                imag = str(x.imag())
            else:
                imag = 0
        ring_element.RingElement.__init__(self, complex_field.CC)
        try:
            R = real_field.RealField(prec)
            self.__re = R(real)
            self.__im = R(imag)
        except TypeError:
            raise TypeError, "unable to coerce (%s,%s) to a ComplexNumber"%(real,imag)
        self.__repr = None
        
    def __repr__(self):
        if self.__repr == None:
            s=""
            if self.__re != 0:
                s=str(self.__re)
            if self.__im != 0:
                y = self.__im
                if s!="":
                    if y < 0:
                        s=s+" - "
                        y=-y
                    else:
                        s=s+" + "
                s=s+"%s*I"%y
            if len(s) == 0: s = "0"
            self.__repr = s
        return self.__repr
    
    def __add__(self, right):
        if not isinstance(right, ComplexNumber):
            return bin_op(self, right, operator.add)
        return ComplexNumber(self.__re + right.__re, self.__im + right.__im)
    
    def __sub__(self, right):
        if not isinstance(right, ComplexNumber):
            return bin_op(self, right, operator.sub)
        return ComplexNumber(self.__re - right.__re, self.__im - right.__im)
    
    def __mul__(self, right):
        if not isinstance(right, ComplexNumber):
            return bin_op(self, right, operator.mul)
        return ComplexNumber(self.__re * right.__re - self.__im * right.__im,
                             self.__re * right.__im + self.__im * right.__re)
    
    def norm(self):
        return self.__re*self.__re + self.__im*self.__im
    
    def __div__(self, right):
        if not isinstance(right, ComplexNumber):
            return bin_op(self, right, operator.div)
        right_nm = right.norm()
        a = right.__re/right_nm
        b = right.__im/right_nm
        re =  a * self.__re + b * self.__im
        im = -b * self.__re + a * self.__im
        return ComplexNumber(re, im)
    
    def __rdiv__(self, left):
        return ComplexNumber(left)/self
    
    def __pow__(self, right):
        # TODO: this is silly; should probably implement
        # complex numbers using PARI's implementation.
        z = self._pari_()
        right = sage.libs.pari.pari(right)
        w = z**right
        return w.python()
        
        #return sage.rings.arith.generic_power(self, right)
    
    def __neg__(self):
        return ComplexNumber(-self.__re, -self.__im)
    
    def __pos__(self):
        return self
    
    def __abs__(self):
        R = self.__re.parent()
        x = R(self.__re**2 + self.__im**2)
        return x.sqrt()
    
    def __invert__(self):
        a = abs(self)
        return ComplexNumber(self.__re/a, -self.__im/a)
    
    def __int__(self):
        if self.__im == 0:
            return int(self.__re)
        raise TypeError
    
    def __long__(self):
        if self.__im == 0:
            return long(self.__re)
        raise TypeError
    
    def __float__(self):
        if self.__im == 0:
            return float(self.__re)
        raise TypeError

    def __complex__(self):
        return complex(float(self.__re), float(self.__im))
    
    def __cmp__(self, other):
        if not isinstance(other, ComplexNumber):
            other = self.parent()(other)
        return sage.misc.misc.generic_cmp((self.__re, self.__im) , (other.__re, other.__im))

    #def argument(self):
    #    s = self._argument()
    #    if s < 0:
    #        s += 2*math.pi
    #    return s
    
    def argument(self):
        """
        The argument (angle) of the complex number, normalized
        so that -pi < arg <= pi.

        WARNING: We use the floating point atan function, so there is precision loss.
        """
        #print "computing arg(%s):"%self
        if self == 0:
            return 0
        if self.__re == 0:
            if self.__im > 0:
                return math.pi/2
            else:
                return -math.pi/2
        if self.__im == 0:
            if self.__re > 0:
                return 0
            else:
                return math.pi
            
        # compute principal argument: 0 <= a < pi/2
        a = math.atan(abs(float(self.__im)/float(self.__re)))
        if self.__im >= 0:
            if self.__re >= 0:
                return a
            else:  # if self.__re < 0:
                return math.pi - a
        else: # if self.__im <= 0:
            if self.__re >= 0:
                return -a
            else:  # if self.__re < 0
                return a - math.pi

    def arg(self):
        """
        Same as argument.
        """
        return self.argument()

    def conjugate(self):
        return ComplexNumber(self.__re, -self.__im)

    def exp(self, prec=None):
        """
        Compute exp(z) using Fourier series.

        If prec is given, compute until the nth summand
        is less than $10^{-prec}$.
        """
        return self._pari_().exp().python()
##         if prec==None:
##             prec = self.real().prec() + 4
##         zn = ComplexNumber(1,0)
##         n = 0
##         n_factorial = 1
##         s = zn
##         t = zn
##         pr = 10.0**(-prec)
##         while abs(t) > pr:
##             n += 1
##             zn = zn*self
##             n_factorial *= n
##             t = zn/n_factorial
##             print "t = ", t
##             s += t
##         return s

    def log(self):
        """
        Complex logarithm of z with branch chosen as follows:
        Write z = rho*exp(i*theta) with -pi <= theta < pi.  Then
               log(z) = log(rho) + i*theta.

        WARNING: Currently the real log is computed using floats, so there is
        potential precision loss.  
        """
        #print "log(%s) = ?"%self
        theta = self.argument()
        #print "arg = ", theta
        rho = abs(self)
        #print "abs = ", rho
        #assert -math.pi <= theta and theta < math.pi
        return ComplexNumber(rho.log(), theta)

    def order(self):
        if self.__re == 1 and self.__im == 0:
            return 1
        if self.__re == -1 and self.__im == 0:
            return 2
        raise NotImplementedError, "general order over complex field not implemented."

    def sqrt(self):
        raise NotImplementedError

    def prec(self):
        return self.__re.prec()

    def real(self):
        return self.__re

    def imag(self):
        return self.__im

    def _pari_(self):
        return sage.libs.pari.pari(str(self))
