"""
Element of finite field.

EXAMPLES:
    sage: K = FiniteField(2)
    sage: V = VectorSpace(K,3)
    sage: w = V([0,1,2])
    sage: K(1)*w
    (0, 1, 0)
"""


import operator

import sage.structure.element as element
import finite_field
import arith
import integer_ring
from integer import Integer
import rational
import polynomial_ring
from sage.libs.pari import pari, gen
from sage.rings.coerce import bin_op
import field_element

class FiniteFieldElement(field_element.FieldElement):
    """
    An element of a finite field.

    Create elements by first defining the finite field F, then use
    the notation F(n), for n an integer. or let a = F.gen() and
    write the element in terms of a.
    """
    def __init__(self, parent, value):
        """
        Create element of a finite field.

        EXAMPLES:
            sage: k = GF(9)
            sage: a = k(11); a
            2
            sage: a.parent()
            Finite field in x of size 3^2
        """
        field_element.FieldElement.__init__(self, parent)
        self.__parent = parent
        if isinstance(value, str):
            raise TypeError, "value must not be str"
        try:
            if isinstance(value, gen):
                if value.type()[-3:] == "MOD" :
                    self.__value = value
                else:
                    self.__value = value.Mod(parent._pari_modulus())*parent._pari_one()
                return
            elif isinstance(value, FiniteFieldElement):
                if parent != value.parent():
                    raise TypeError, "No coercion of element %s into %s defined."%(value, parent)
                else:
                    self.__value = value.__value
                    return
            self.__value = pari(value).Mod(parent._pari_modulus())*parent._pari_one()
        except AttributeError, TypeError:
            raise TypeError, 'Unable to coerce %s into %s.'%(value, parent)

    def polynomial(self):
        """
        Elements of a finite field are represented as a polynomial
        modulo a modulus.  This functions returns the representating
        polynomial as an element of the polynomial ring over the
        prime finite field, with the same variable as the finite field.
        
        EXAMPLES:
        The default variable is a:
            sage: k = GF(3**2)
            sage: k.gen().polynomial()
            x
            
        The variable can be any string.
            sage: k = FiniteField(3**4, "alpha")
            sage: a = k.gen()
            sage: a.polynomial()
            alpha
            sage: (a**2 + 1).polynomial()
            alpha^2 + 1
            sage: (a**2 + 1).polynomial().parent()
            Univariate Polynomial Ring in alpha over Finite field of size 3
        """
        return self.parent().polynomial_ring()(\
            list(reversed(self.__value.lift().list())))

    def is_square(self):
        """
        Returns True if and only if this element is a perfect square.

        EXAMPLES:
            sage: k = GF(3**2)
            sage: a = k.gen()
            sage: a.is_square()
            False
            sage: (a**2).is_square()
            True
            sage: k = GF(2**2)
            sage: a = k.gen()
            sage: (a**2).is_square()
            True
            sage: k = GF(17**5); a = k.gen()
            sage: (a**2).is_square()
            True
            sage: a.is_square()
            False
        """
        K = self.parent()
        if K.characteristic() == 2:
            return True
        n = K.order() - 1
        a = self**(n / 2)
        return a == 1
        
    def rational_reconstruction(self):
        """
        If the parent field is a prime field, uses rational reconstruction to
        try to find a lift of this element to the rational numbers.

        EXAMPLES:
            sage: k = GF(97)
            sage: a = k(RationalField()('2/3'))
            sage: a
            33
            sage: a.rational_reconstruction()
            2/3
        """
        if self.parent().degree() != 1:
            raise ArithmeticError, "The finite field must be prime, but has degree %s"%self.parent().degree()
        t = arith.rational_reconstruction(int(self), self.parent().characteristic())
        if t == None or t[1] == 0:
            raise ZeroDivisionError, "Unable to compute rational reconstruction."
        return rational.Rational((t[0],t[1]))

    def multiplicative_order(self):
        """
        Returns the order of self.
        """
        # todo: this is currently computed using a very *very* stupid
        # algorithm!  change to use factorization of order of group.
        try:
            return self.__multiplicative_order
        except AttributeError:
            x = self.copy()
            one = self.__parent(1)
            self.__multiplicative_order = 1
            while x != one:
                self.__multiplicative_order += 1
                x *= self
                if self.__multiplicative_order == 100:
                    print "WARNING: Using IDIOTIC multiplicative_order algorithm in FiniteFieldElement.multiplicative_order!"
            return self.__multiplicative_order
    
    def copy(self):
        """
        Return a copy of this element.

        EXAMPLES:
            sage: k = GF(3**3)
            sage: a = k(5)
            sage: a
            2
            sage: a.copy()
            2
            sage: b = a.copy()
            sage: a == b
            True
            sage: a is b
            False
            sage: a is a
            True
        """
        return FiniteFieldElement(self.__parent, self.__value)

    def _pari_(self):
        """
        Return PARI object corresponding to this finite field element.

        EXAMPLES:
            sage: k = GF(3**3)
            sage: a = k.gen()
            sage: b = a**2 + 2*a + 1
            sage: b._pari_()
            Mod(Mod(1, 3)*x^2 + Mod(2, 3)*x + Mod(1, 3), Mod(1, 3)*x^3 + Mod(2, 3)*x + Mod(1, 3))

        Looking at the PARI representation of a finite field element, it's no wonder people
        find PARI difficult to work with directly.  Compare our representation:
        
            sage: b
            x^2 + 2*x + 1
            sage: b.parent()
            Finite field in x of size 3^3
        """
        return self.__value

    def charpoly(self):
        """
        Returns the characteristic polynomial of this element.

        EXAMPLES:
            sage: k = GF(3**3)
            sage: a = k.gen()
            sage: a.charpoly()
            x^3 + 2*x + 1
            sage: k.modulus()
            x^3 + 2*x + 1
            sage: b = a**2 + 1
            sage: b.charpoly()
            x^3 + x^2 + 2*x + 1
        """
        R = polynomial_ring.PolynomialRing(self.parent().prime_subfield())
        return R(list(reversed(self.__value.charpoly().lift().list())))

    def trace(self):
        """
        Returns the trace of this element.

        EXAMPLES:
            sage: k = GF(3**3); a = k.gen()
            sage: b = a**2 + 2
            sage: b.charpoly()
            x^3 + x^2 + 2
            sage: b.trace()
            2
            sage: b.norm()
            2
        """
        return self.parent().prime_subfield()(self.__value.trace().lift())

    def norm(self):
        """
        Returns the norm of this element, which is the constant term
        of the characteristic polynomial, i.e., the determinant of left
        multiplication. 

        EXAMPLES:
            sage: k = GF(3**3); a = k.gen()
            sage: b = a**2 + 2
            sage: b.charpoly()
            x^3 + x^2 + 2
            sage: b.trace()
            2
            sage: b.norm()
            2
        """
        return self.charpoly()[0]

    def order(self):
        """
        Return the additive order of this finite field element.
        """
        if self.is_zero():
            return Integer(1)
        return self.parent().characteristic()
        
    def __repr__(self):
        return "%s"%(self.__value.lift().lift())
    
    def __compat(self, other):
        if self.parent() != other.parent():
            raise TypeError, "Parents of finite field elements %s and %s must be equal."%(self, other)
        
    def __add__(self, right):
        if not isinstance(right, FiniteFieldElement):
            return bin_op(self, right, operator.add)
        self.__compat(right)
        return FiniteFieldElement(self.__parent, self.__value + right.__value)
    
    def __sub__(self, right):
        if not isinstance(right, FiniteFieldElement):
            return bin_op(self, right, operator.sub)
        self.__compat(right)        
        return FiniteFieldElement(self.__parent, self.__value - right.__value)
    
    def __mul__(self, right):
        if not isinstance(right, FiniteFieldElement):
            return bin_op(self, right, operator.mul)
        self.__compat(right)                
        return FiniteFieldElement(self.__parent, self.__value * right.__value)

    def __div__(self, right):
        if not isinstance(right, FiniteFieldElement):
            return bin_op(self, right, operator.div)
        self.__compat(right)
        if right.__value == 0:
            raise ZeroDivisionError
        return FiniteFieldElement(self.__parent, self.__value / right.__value)

    def __int__(self):
        try:
            return int(self.__value.lift().lift())
        except ValueError:
            raise TypeError, "Cannot coerce %s to int"%self

    def _integer_(self):
        return self.lift()

    def __long__(self):
        try:
            return long(self.__value.lift().lift())
        except ValueError:
            raise TypeError, "Cannot coerce %s to int"%self

    def __float__(self):
        try:
            return float(self.__value.lift().lift())
        except ValueError:
            raise TypeError, "Cannot coerce %s to float"%self

    def __rdiv__(self, left):
        return self.parent()(left)/self

    # commented out because PARI (used for .__value) prints
    # out crazy warnings when the exponent is LARGE -- this
    # is even a problem in gp!!!
    # (Commenting out causes this to use a generic algorithm)
    #def __pow__(self, right):
    #    right = int(right)
    #    return FiniteFieldElement(self.__parent, self.__value**right)

    def __neg__(self):
        return FiniteFieldElement(self.__parent, -self.__value)
    
    def __pos__(self):
        return self
    
    def __abs__(self):
        raise ArithmeticError, "Absolute value of element (=%s) of %s not defined."%(self, self.parent())
    
    def __invert__(self):
        """
        EXAMPLES:
            sage: k = GF(9); a = k.gen()
            sage: ~a
            x + 2
            sage: (a+1)*a
            2*x + 1
        """

        if self.__value == 0:
            raise ZeroDivisionError, "Cannot invert 0"
        return FiniteFieldElement(self.__parent, ~self.__value)
    
    def lift(self):
        """
        If this element lies in a prime finite field, return a lift of this
        element to an integer.

        EXAMPLES:
            sage: k = GF(next_prime(10**10))
            sage: a = k(17)/k(19)
            sage: b = a.lift(); b
            7894736858
            sage: b.parent()
            Integer Ring
        """
        return integer_ring.IntegerRing()(self.__value.lift().lift())
    
    
    def __cmp__(self, other):
        """

        Compare an element of a finite field with other.  If other is
        not an element of a finite field, an attempt is made to coerce
        it so it is one.

        EXAMPLES:
            sage: k = GF(3**3); a = k.gen()
            sage: a == 1
            False
            sage: a**0 == 1
            True
            sage: a == a
            True
            sage: a < a**2
            True
            sage: a > a**2
            False
        """
        if self is other: return 0
        if not isinstance(other, FiniteFieldElement):
            try:
                other = self.parent()(other)
            except TypeError:
                return -1
        #if self.parent() != other.parent():
        #    return -1
        if self.__value == other.__value:
            return 0
        return -1
    
    
if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.all import FiniteField, GF, RationalField
    from sage.rings.arith import next_prime
    from sage.modules.all import *
    doctest.testmod(sys.modules[__name__])
