"""
Ideals

SAGE provides extremely limited functionality for computing with
ideals.  One can create an ideal in any commutative ring $R$ by giving
a list of generators, using the notation \code{R.ideal([a,b,...])}.
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import sage.misc.latex as latex
import sage.rings.ring
import sage.rings.principal_ideal_domain
import commutative_ring

def Ideal(R, gens=[], coerce=True):
    """
    Create the ideal in ring with given generators.

    INPUT:
        R -- a ring
        gens -- list of elements
        coerce -- bool (default: True); whether gens need to be coerced into ring.

    Alternatively, one can also call this function with the syntax
         Ideal(gens)
    where gens is a nonempty list of generators or a single generator.
        
    OUTPUT:
        The ideal of ring generated by gens.
    """
    if not commutative_ring.is_CommutativeRing(R):
        raise TypeError, "R must be a commutative ring"
    if isinstance(R, Ideal_generic):
        return Ideal(R.ring(), R.gens())
    
    if isinstance(R, list) and len(R) > 0:
        return Ideal(R[0].parent(), R)        
        try:
            return Ideal(R[0].parent(), R)
        except AttributeError:
            pass

    if not isinstance(R, sage.rings.ring.Ring):
        try:
            S = R.parent()
        except AttributeError:
            raise TypeError, "ring (=%s) must be a ring, list, or element."%R
        if isinstance(S, sage.rings.ring.Ring):
            return Ideal(S, [R])
        else:
            raise TypeError, "ring (=%s) must be a ring, list, or element."%R

    if not isinstance(gens, list):
        gens = [R(gens)]
        coerce = False
    elif len(gens) == 0:
        gens = [R(0)]
        coerce = False

    if coerce:
        gens = [R(g) for g in gens]

    gens = list(set(gens))

    if isinstance(R, sage.rings.principal_ideal_domain.PrincipalIdealDomain):
        # Use GCD algorithm to obtain a principal ideal
        g = gens[0]
        for h in gens[1:]:
            g = R.gcd(g, h)
        return Ideal_pid(R, g)
        
    if len(gens) == 1:
        return Ideal_principal(R, gens[0])

    
    return Ideal_generic(R, gens)

def is_Ideal(x):
    return isinstance(x, Ideal_generic)

    
class Ideal_generic:
    """
    An ideal.
    """
    def __init__(self, ring, gens):
        self.__ring = ring
        self.__gens = gens
        
    def __repr__(self):
        return "Ideal of %s with generators %s"%(self.ring(), self.gens())

    def __cmp__(self, other):
        if not isinstance(other, Ideal_generic):
            return -1
        other = self.ring().ideal(other)
        if set(self.gens()) == set(other.gens()):
            return 0
        raise NotImplementedError

    def __contains__(self, x):
        raise NotImplementedError

    def is_zero(self):
        return self.gens() == [self.ring()(0)]

    def _latex_(self):
        return '\\left(%s\\right)%s'%(", ".join([latex.latex(g) for g in \
                                                 self.gens()]),
                                      latex.latex(self.ring()))

    def ring(self):
        """
        Return the ring in which this ideal is contained.
        """
        return self.__ring

    def gens(self):
        return self.__gens

    def is_maximal(self):
        raise NotImplementedError

    def is_prime(self):
        raise NotImplementedError

    def is_principal(self):
        raise NotImplementedError

    def category(self):
        """
        Return the category of this ideal.

        
        """
        import sage.categories.all 
        return sage.categories.all.Ideals(self.__ring)

    def __add__(self, other):
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)
        return Ideal(self.ring(), self.gens() + other.gens())
    
    def __radd__(self, other):
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)
        return Ideal(self.ring(), self.gens() + other.gens())

    def __mul__(self, other):
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)
        return Ideal(self.ring(), [x*y for x in self.gens() for y in other.gens()])

    def __rmul__(self, other):
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)
        return Ideal(self.ring(), [y*x for x in self.gens() for y in other.gens()])
    
    
class Ideal_principal(Ideal_generic):
    """
    A principal ideal.
    """
    def __init__(self, ring, gen):
        Ideal_generic.__init__(self, ring, [gen])

    def __repr__(self):
        return "Principal ideal of %s generated by %s"%(self.ring(), self.gen())

    def is_principal(self):
        return True

    def gen(self):
        return self.gens()[0]
    
    def __contains__(self, x):
        if self.gen().is_zero():
            return x.is_zero()
        return self.gen().divides(x)
    
    def __cmp__(self, other):
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)

        if not other.is_principal():
            return -1
        
        if self.is_zero():
            if not other.is_zero():
                return -1
            return 0

        # is other.gen() / self.gen() a unit in the base ring?
        g0 = other.gen()
        g1 = self.gen()
        if g0.divides(g1) and g1.divides(g0):
            return 0
        return 1

    def divides(self, other):
        """
        Returns True if self divides other.
        """
        if isinstance(other, Ideal_principal):
            return self.gen().divides(other.gen())
        raise NotImplementedError
            
class Ideal_pid(Ideal_principal):
    """
    An ideal of a principal ideal domain.
    """
    def __init__(self, ring, gen):
        Ideal_principal.__init__(self, ring, gen)

    def __add__(self, other):
        if not isinstance(other, Ideal_generic):
            other = self.ring().ideal(other)
        return self.ring().ideal(self.gcd(other))
        

class Ideal_fractional(Ideal_generic):
    def __init__(self, ring, gen):
        Ideal_generic.__init__(self, ring, [gen])
    def __repr__(self):
        return "Fractional ideal of %s with generators %s"%(self.ring(), self.gens())
    
