"""
Element of integers modulo n.
"""

__doc_exclude = ["bin_op"]

import operator

import integer_mod_ring
import arith
import rational
from sage.libs.pari import pari
import integer_ring
import integer
import commutative_ring_element

import sage.rings.coerce as coerce

def Mod(n, m):
    """
    Return the equivalence class of n modulo m
    as an element of $\Z/m\Z$.

    EXAMPLES:
        sage: x = Mod(12345678, 32098203845329048)
        sage: x
        12345678
        sage: x**100
        1017322209155072
    """
    
    return IntegerMod(integer_mod_ring.IntegerModRing(m),n)
mod=Mod

    
class IntegerMod(commutative_ring_element.CommutativeRingElement):
    def __init__(self, parent, value, construct=False):
        commutative_ring_element.CommutativeRingElement.__init__(self, parent)
        if construct:
            self.__value = value
            return
        if isinstance(value, rational.Rational):
            value = value % parent.modulus()
        self.__value = pari('Mod(%s,%s)'%(value, parent.modulus()))


    def __cmp__(self, right):
        if not isinstance(self, IntegerMod) or not isinstance(right, IntegerMod) \
               or right.parent() != self.parent():
            return coerce.cmp(self, right)
        return self._cmp(right)

    def modulus(self):
        return self.parent().modulus()

    def is_square(self):
        return bool(self.__value.issquare())

    def is_unit(self):
        return self.lift().gcd(self.modulus()) == 1

    def sqrt(self):
        #if not self.is_square():
        #    raise ValueError, "%s must be a square, but it isn't."%self
        return self.parent()(self.__value.sqrt(), construct=True)

    def _pari_(self):
        return self.__value
        
    def rational_reconstruction(self):
        """
        EXAMPLES:
            sage: R = IntegerModRing(97)
            sage: a = R(2) / R(3)
            sage: a
            33
            sage: a.rational_reconstruction()
            2/3
        """
        t = arith.rational_reconstruction(int(self), self.modulus())
        if t == None or t[1] == 0:
            raise ZeroDivisionError, "Unable to compute rational reconstruction."
        return rational.Rational((t[0],t[1]))

    def crt(self, other):
        """
        Use the Chinese Remainder Theorem to find an element of the
        integers modulo the product of the moduli that reduces to self
        and to other.  The modulus of other must be coprime to the
        modulus of self.
        """
        if not isinstance(other, IntegerMod):
            raise TypeError, "other must be an integer mod"
        return self.parent()(self.__value.chinese(other.__value), construct=True)
    
    def order(self):
        """
        Returns the additive order of self.
        """
        n = self.parent().modulus()
        return integer.Integer(n//int(self.__value.lift().gcd(n)))
    
    def multiplicative_order(self):
        """
        Returns the additive order of self.
        """
        if not self.is_unit():
            raise ArithmeticError, "self (=%s) must be a unit"%self
        return integer.Integer(self.__value.order())  # pari's "order" is by default multiplicative
    
    def copy(self):
        return self.parent()(self.__value, construct=True)
    
    def __repr__(self):
        return str(self.__value.lift())
    
    def _add(self, right):
        return self.parent()(self.__value + right.__value, construct=True)
    
    def _sub(self, right):
        return self.parent()(self.__value - right.__value, construct=True)

    def _mul(self, right):
        return self.parent()(self.__value * right.__value, construct=True)

    def _div(self, right):
        return self.parent()(self.__value / right.__value, construct=True)

    def __int__(self):
        return int(self.__value.lift())

    def _integer_(self):
        return integer.Integer(self.__value.lift())

    def _rational_(self):
        return rational.Rational(self.__value.lift())

    def __long__(self):
        return long(self.__value.lift())

    def __mod__(self, right):
        right = int(right)
        if self.modulus() % right != 0:
            raise ZeroDivisionError, "Error - reduction mod %s not defined."%right
        return integer_mod_ring.IntegerModRing(right)(self)
    
    # commented out because PARI (used for .__value) prints
    # out crazy warnings when the exponent is LARGE -- this
    # is even a problem in gp!!!
    #def __pow__(self, right):
    #    right = int(right)
    #    return self.parent()(self.__value**right, construct=True)

    def __neg__(self):
        return self.parent()(-self.__value, construct=True)
    
    def __invert__(self):
        return self.parent()(~self.__value, construct=True)

    def lift(self):
        return integer.Integer(self.__value.lift())
    
    def __float__(self):
        return float(self.__value.lift())
    
    def __cmp__(self, other):
        if not isinstance(other, IntegerMod) or self.parent() != other.parent():
            return coerce.cmp(self, other)
        return cmp(self.__value, other.__value)

if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.all import IntegerModRing
    doctest.testmod(sys.modules[__name__])
