"""
Ring of integers modulo n.

EXAMPLES:
    sage: R = Integers(97)
    sage: a = R(5)
    sage: a**100000000000000000000000000000000000000000000000000000000000000
    61
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

__doc_exclude = ["Zmod"]

import random
import weakref

import arith
import commutative_ring
import field
import integer_mod
import integer
import rational

_objsIntegerModRing = {}
def IntegerModRing(modulus):
    global _objsIntegerModRing
    if _objsIntegerModRing.has_key(modulus):
        x = _objsIntegerModRing[modulus]()
        if x != None: return x
    if arith.is_prime(modulus):
        R = IntegerModRing_field(modulus)
    else:
        R = IntegerModRing_generic(modulus)
    _objsIntegerModRing[modulus] = weakref.ref(R)
    return R

def is_IntegerModRing(x):
    """
    Return True if x is an integer modulo ring.
    """
    return isinstance(x, IntegerModRing_generic)

class IntegerModRing_generic(commutative_ring.CommutativeRing):
    """
    The ring of integers modulo N, e.g., when N is prime
    this is a prime finite field.
    """
    def __init__(self, modulus):
        """
        Create with the command
              IntegerModRing(modulus)

        INPUT:
            modulus -- an integer > 1

        OUTPUT:
            IntegerModRing -- the ring of integers modulo N.

        EXAMPLES:
        
        First we compute with integers modulo $17$.
            sage: FF = IntegerModRing(17)
            sage: FF
            Ring of integers modulo 17
            sage: FF.is_field()
            True
            sage: FF.characteristic()
            17
            sage: FF.modulus()
            17
            sage: gens = FF.unit_gens()
            sage: a = gens[0]
            sage: a
            3
            sage: a.is_square()
            False
            sage: def pow(i): return a**i
            sage: [pow(i) for i in range(16)]
            [1, 3, 9, 10, 13, 5, 15, 11, 16, 14, 8, 7, 4, 12, 2, 6]

        Next we compute with the integers modulo $16$.
            sage: Z16 = IntegerModRing(16)
            sage: Z16.is_field()
            False
            sage: Z16.modulus()
            16
            sage: Z16.characteristic() 
            16
            sage: gens = Z16.unit_gens()
            sage: gens
            [15, 5]
            sage: a = gens[0]
            sage: b = gens[1]
            sage: def powa(i): return a**i
            sage: def powb(i): return b**i
            sage: gp_exp = FF.unit_group_exponent()
            sage: gp_exp
            16
            sage: [powa(i) for i in range(15)]
            [1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1, 15, 1]
            sage: [powb(i) for i in range(15)]
            [1, 5, 9, 13, 1, 5, 9, 13, 1, 5, 9, 13, 1, 5, 9]
            sage: a.multiplicative_order()
            2
            sage: b.multiplicative_order()
            4
        """
        modulus = integer.Integer(modulus)
        if modulus == 0:
            raise ZeroDivisionError, "modulus must be nonzero"
        self.__modulus = modulus
        self.__unit_group_exponent = None
        self.__factored_modulus = None
        
    def is_finite(self):
        return True

    def is_field(self):
        return False
        
    def factored_modulus(self):
        if self.__factored_modulus != None:
            return self.__factored_modulus
        self.__factored_modulus = arith.factor(self.__modulus, int_=True)
        return self.__factored_modulus
    
    def characteristic(self):
        return self.__modulus
    
    def __repr__(self):
        return "Ring of integers modulo %s"%self.__modulus
    
    def modulus(self):
        return self.__modulus
    
    def __call__(self, *args, **kwds):
        return integer_mod.IntegerMod(self, *args, **kwds)

    def _coerce_(self, x):
        if isinstance(x, (int, long, integer.Integer, rational.Rational)):
            return integer_mod.IntegerMod(self, x)
        if isinstance(x, integer_mod.IntegerMod) and x.parent().modulus() % self.modules() == 0:
            return integer_mod.IntegerMod(self, x)
        raise TypeError
    
    def __cmp__(self, other):
        if not isinstance(other, IntegerModRing_generic):
            return -1
        if self.__modulus < other.__modulus:
            return -1
        elif self.__modulus > other.__modulus:
            return 1
        return 0
    
    def __getitem__(self, n):
        if n < 0 or n >= self.modulus():
            raise IndexError
        return self(n)

    # The following __unit_gens functions are here since I just factored
    # them out from the unit_gens function.  They are only called by
    # the unit_gens function.
    def __unit_gens_primecase(self, p):
        if p==2:
            return integer_mod.Mod(1,p)
        P = arith.prime_divisors(p-1)
        ord = p-1
        one = integer_mod.Mod(1,p)
        x = 2
        while x < p:
            generator = True
            z = integer_mod.Mod(x,p)
            for q in P:
                if z**(ord/q) == one:
                    generator = False
                    break
            if generator:
                return z
            x += 1
        #end for
        assert False, "didn't find primitive root for p=%s"%p

    def __unit_gens_primepowercase(self, p, r):
        r"""
        Find smallest generator for $(\Z/p^r\Z)^*$.
        """
        if r==1:
            return [self.__unit_gens_primecase(p)]
        elif p==2:
            if r==1:
                return []
            elif r==2:
                return [integer_mod.Mod(-1,2**r)]
            elif r>=3:
                pr=2**r
                a = integer_mod.Mod(5, pr)
                return [integer_mod.Mod(-1,pr), a]
            assert False, "p=2, r=%s should be >=1"%r
        else:  # odd prime
            pr = p**r
            R = IntegerModRing(pr)
            x = R(self.__unit_gens_primecase(p).lift())
            n = p**(r-2)*(p-1)
            one = integer_mod.Mod(1,pr)
            for b in range(0,p):
                z = x+R(b*p)
                if z**n != one:
                    a = integer_mod.Mod(z,pr)
                    return [a]
            assert False, "p=%s, r=%s, couldn't find generator"%(p,r)
        
    def unit_gens(self):
        r"""
        Returns generators for the unit group $(\Z/N\Z)^*$.
        
        We compute the list of generators using a deterministic
        algorithm, so the generators list will always be the same.
        Each generator corresponds to a prime divisor of $N$ (or
        possibly two prime divisors for p=2).

        INPUT: (none)
        OUTPUT:
            list -- a list of elements of self
        """
        try:
            return self.__unit_gens
        except AttributeError:
            self.__unit_gens = []
        n = self.__modulus
        if n == 1:
            self.__unit_gens = [self(1)]
            return self.__unit_gens
        for p,r in self.factored_modulus():
            m = n/(p**r)
            for g in self.__unit_gens_primepowercase(p, r):
                x = g.crt(integer_mod.Mod(1,m))
                self.__unit_gens.append(x)
        return self.__unit_gens
    
    def unit_group_exponent(self):
        if self.__unit_group_exponent != None:
            return self.__unit_group_exponent
        a = []
        for p, r in self.factored_modulus():
            if p != 2:
                a.append((p-1)*(p**(r-1)))   # phi(p**r)
            else:  # p=2
                if r==2:
                    a.append(2)
                elif r>2:
                    a.append(2**(r-2))
            #endif
        #endfor
        self.__unit_group_exponent = int(arith.LCM(a))
        return self.__unit_group_exponent
        
    def random_element(self, bound=None):
        """
        Return a random element of this ring.

        If bound is not None, return the coercion of an integer in the
        interval [-bound, bound] into this ring.
        """
        if not (bound is None):
            return commutative_ring.CommutativeRing.random_element(self, bound)
        a = random.randrange(0,self.__modulus-1)
        return self(a)

class IntegerModRing_field(field.Field, IntegerModRing_generic):
    def __init__(self, modulus):
        IntegerModRing_generic.__init__(self, modulus)


Zmod = IntegerModRing
Integers = IntegerModRing

def GF(p):
    if not arith.is_prime(p):
        raise NotImplementedError, "Only prime fields currently implemented."
    return IntegerModRing(p)

def crt(v):
    """CRT(v): v is a lift of elements of rings.IntegerMod(n), for
               various coprime moduli n.
    """
    if len(v) == 0:
        return IntegerModRing(1)(1)
    x = v[0]
    for i in range(1,len(v)):
        x = x.crt(v[i])
    return x

    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
