"""
Multivariate polynomial element

AUTHORS: David Joyner and William Stein
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import operator

import arith

from sage.structure.element import RingElement
from coerce import bin_op, cmp as coerce_cmp

import sage.misc.misc as misc
import integer

import polylist
import polydict

class MPolynomial(RingElement):
    def __init__(self, parent, x):
        RingElement.__init__(self, parent)
        self.__element = x

    def __repr__(self):
        return "%s"%self.__element

    def __call__(self, *x):
        """
        Evaluate this multi-variate polynomial at $x$, where $x$ is
        either the tuple of values to substitute in, or one can use
        functional notation $f(a_0,a_1,a_2, \ldots)$ to evaluate $f$
        with the ith variable replaced by $a_i$.

        EXAMPLES:
            sage: x, y = MPolynomialRing(RationalField(),2).gens()
            sage: f = x**2 + y**2
            sage: f(1,2)
            5
            sage: f((1,2))
            5

            sage: x = MPolynomialRing(RationalField(),3).gens()
            sage: f = x[0] + x[1] - 2*x[1]*x[2]
            sage: f
            x_1 - 2*x_1*x_2 + x_0
            sage: f(1,2,0)
            3
            sage: f(1,2,5)
            -17

        AUTHOR: David Kohel, 2005-09-27
        """
        if len(x) == 1 and isinstance(x[0], tuple):
            x = x[0]
        n = self.parent().ngens()
        if len(x) != n:
            raise TypeError, "x must be a tuple of length %s"%n
        if n == 0:
            return self
        try:
            K = x[0].parent()
        except AttributeError:
            K = self.parent().base_ring()
        y = K(0) 
        for (m,c) in self.element().dict().iteritems():  
            y += K(c)*misc.mul([ x[i]**m[i] for i in range(n) ])      
        return y 

    def __cmp__(self, right):
        if not isinstance(right, MPolynomial) or right.parent() != self.parent():
            return coerce_cmp(self, right)
        return self.__element.__cmp__(right.__element)

    def _add(self, right):
        return self.parent()(self.__element + right.__element)

    def _sub(self, right):
        return self.parent()(self.__element - right.__element)

    def _mul(self, right):
        return self.parent()(self.__element * right.__element)

    def _div(self, right):
        return self.parent().fraction_field()(self.__element, right.__element)

    def __pow__(self, n):
        if not isinstance(n, (int, long, integer.Integer)):
            raise TypeError, "The exponent (n=%s) must be an integer."%n
        return self.parent()(self.__element**n)

    def __rpow__(self, n):
        if not isinstance(n, (int, long, integer.Integer)):
            raise TypeError, "The exponent (n=%s) must be an integer."%n
        return self.parent()(self.__element**n)

    def element(self):
        return self.__element


class MPolynomial_polylist(MPolynomial):
    def __init__(self, parent, x):
        if not isinstance(x, polylist.PolyList):
            x = polylist.PolyList(x, 
                                  parent.names(),
                                  parent._monomial_order_function(),
                                  parent.base_ring()(0))
        MPolynomial.__init__(self, parent, x)

    def __repr__(self):
        return self.element().poly_repr()





class MPolynomial_polydict(MPolynomial):
    def __init__(self, parent, x):
        if not isinstance(x, polydict.PolyDict):
            x = polydict.PolyDict(x, parent.base_ring()(0), remove_zero=True)
        MPolynomial.__init__(self, parent, x)

    def __neg__(self):
        return self*(-1)

    def __repr__(self):
        return self.element().poly_repr(self.parent().names(),
                                        atomic_coefficients=self.parent().base_ring().is_atomic_repr())

    def _latex_(self):
        return self.element().latex(self.parent().latex_names(),
                                    atomic_coefficients=self.parent().base_ring().is_atomic_repr())



##############################################################
# For checking that documentation strings are correct.
##############################################################
if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.all import RationalField, MPolynomialRing
    doctest.testmod(sys.modules[__name__])
