"""
Multivariate Polynomials

AUTHORS: David Joyner and William Stein
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import weakref

import commutative_ring
import integral_domain

import fraction_field
import fraction_field_element

import multi_polynomial
import polylist, polydict

import sage.misc.latex as latex


def MPolynomialRing(base_ring, n=1, names=None, repr='dict'):
    """
    Create a Multivariate polynomial ring over a commutative base ring.

    INPUT:
        base_ring -- CommutativeRing

        n -- int, number of variables  (default: 1)
        
        names -- list or str; list of n variable names or a string;
               if a string, names the variables var_0, var_1, etc.
               (default: None)
        repr -- str; choice of underlying representation
                of polynomials.
               (default: 'dict')
               The choices are
                   'dict' -- Use a Python dictionary-based
                             representation  (much faster)
                   'list' -- Use a Python list-based representation

                   
    EXAMPLES:
        sage: R = MPolynomialRing(RationalField(), 3)
        sage: R
        Polynomial ring in x_0, x_1, x_2 over Rational Field
        sage: x_0,x_1,x2 = R.gens()
        sage: x_0.element()
        PolyDict with representation {(1, 0, 0): 1}
        sage: x_0 + x_1 + x2
        x_2 + x_1 + x_0
        sage: (x_0 + x_1 + x2)**2
        x_2^2 + 2*x_1*x_2 + x_1^2 + 2*x_0*x_2 + 2*x_0*x_1 + x_0^2

    Next we specify all the variables and do some additional examples of arithmetic.
    
        sage: R = MPolynomialRing(RationalField(), 3, repr='list', names=['a','b','c'])
        sage: R
        Polynomial ring in a, b, c over Rational Field
        sage: a,b,c = R.gens()
        sage: a.element()
        PolyList with distributive representation [[1, [1, 0, 0]]]
        sage: (a+b)**2
        2*a*b + a^2 + b^2
        sage: (a + 2*b + 3*c**2)**3
        27*a*c^4 + 8*b^3 + 27*c^6 + 9*a^2*c^2 + 36*a*b*c^2 + a^3 + 54*b*c^4 + 6*a^2*b + 12*a*b^2 + 36*b^2*c^2


    We can construct multi-variate polynomials rings over completely
    arbitrary SAGE rings.  In this example, we construct a polynomial
    ring S in 3 variables over a polynomial ring in 2 variables over
    GF(9).  Then we construct a polynomial ring in 20 variables over S!
    
        sage: R = MPolynomialRing(GF(9),2, names=['n1','n2']); n1,n2=R.gens()
        sage: n1**2 + 2*n2
        2*n2 + n1^2
        sage: S = MPolynomialRing(R,3, names='a'); a0,a1,a2=S.gens()
        sage: S
        Polynomial ring in a_0, a_1, a_2 over Polynomial ring in n1, n2 over Finite field in x of size 3^2
        sage: x = (n1+n2)*a0 + 2*a1**2
        sage: x
        2*a_1^2 + (n2 + n1)*a_0
        sage: x**3
        2*a_1^6 + (n2^3 + n1^3)*a_0^3
        sage: T = MPolynomialRing(S, 20)
        sage: T
        Polynomial ring in x_0, x_1, x_2, x_3, x_4, x_5, x_6, x_7, x_8, x_9, x_10, x_11, x_12, x_13, x_14, x_15, x_16, x_17, x_18, x_19 over Polynomial ring in a_0, a_1, a_2 over Polynomial ring in n1, n2 over Finite field in x of size 3^2
    """
    if not isinstance(base_ring, commutative_ring.CommutativeRing):
        raise TypeError, "Base ring (=%s) must be a commutative ring."%base_ring
    if isinstance(base_ring, integral_domain.IntegralDomain):
        if repr == 'list':
            return MPolynomialRing_polylist_domain(base_ring, n, names)
        elif repr == 'dict':
            return MPolynomialRing_polydict_domain(base_ring, n, names)
    else:
        if repr == 'list':
            return MPolynomialRing_polylist(base_ring, n, names)
        elif repr == 'dict':
            return MPolynomialRing_polydict(base_ring, n, names)



def is_MPolynomialRing(x):
    return isinstance(x, MPolynomialRing_generic)

class MPolynomialRing_generic(commutative_ring.CommutativeRing):
    def __init__(self, base_ring, n=1, names=None):
        if not isinstance(base_ring, commutative_ring.CommutativeRing):
            raise TypeError, "Base ring (=%s) must be a commutative ring."%base_ring
        self.__base_ring = base_ring
        n = int(n)
        if n < 0:
            raise ValueError, "Multivariate Polynomial rings must " + \
                  "have more than 0 variables."
        self.__ngens = n
        self.assign_names(names)

    def __repr__(self):
        return "Polynomial ring in %s over %s"%(", ".join(self.names()), self.base_ring())

    def _latex_(self):
        return "%s%s"%(latex.latex(self.base_ring()), 
                         latex.latex(self.names()))

    def base_ring(self):
        return self.__base_ring

    def characteristic(self):
        """
        Return the characteristic of this polynomial ring. 
        EXAMPLES:
            sage: R = PolynomialRing(RationalField(),3)
            sage: R.characteristic()
            0
            sage: R = PolynomialRing(GF(7),20)
            sage: R.characteristic()
            7
        """
        return self.__base_ring.characteristic()

    def gen(self, n=0):
        if n < 0 or n >= self.__ngens:
            raise ValueError, "Generator %s not defined."%n
        return self._gens[int(n)]

    def krull_dimension(self):
        return self.base_ring().krull_dimension() + self.ngens()

    def ngens(self):
        return self.__ngens

    def _monomial_order_function(self):
        raise NotImplementedError

class MPolynomialRing_polylist(MPolynomialRing_generic):

    def __init__(self, base_ring, n=1, names=None):
        MPolynomialRing_generic.__init__(self, base_ring, n, names)
        self.__monomial_order_function = polylist.lex_monord
        # Construct the generators
        v = [0 for _ in xrange(n)]
        one = base_ring(1)
        self._gens = []
        for i in xrange(n):
            v[i] = 1
            self._gens.append(multi_polynomial.MPolynomial_polylist(self, [[one, list(v)]]))
            v[i] = 0
        self.__zero_list = v
            
    def _monomial_order_function(self):
        return self.__monomial_order_function

    def __call__(self, x):
        """
        Coerce x into this multivariate polynomial ring.
        """
        if isinstance(x, multi_polynomial.MPolynomial) and x.parent() == self:
            return x
        elif isinstance(x, polylist.PolyList):
            return multi_polynomial.MPolynomial_polylist(self, x)
        c = self.base_ring()(x)
        return multi_polynomial.MPolynomial_polylist(self, [[c, list(self.__zero_list)]])

class MPolynomialRing_polylist_domain(integral_domain.IntegralDomain, MPolynomialRing_polylist):
    def __init__(self, base_ring, n=1, names=None):
        MPolynomialRing_polylist.__init__(self, base_ring, n, names)
        
                
class MPolynomialRing_polydict(MPolynomialRing_generic):

    def __init__(self, base_ring, n=1, names=None):
        MPolynomialRing_generic.__init__(self, base_ring, n, names)
        # Construct the generators
        v = [0 for _ in xrange(n)]
        one = base_ring(1);
        self._gens = []
        for i in xrange(n):
            v[i] = 1  # int's!
            self._gens.append(multi_polynomial.MPolynomial_polydict(self, {tuple(v):one}))
            v[i] = 0
        self.__zero_tuple = tuple(v)
            
    def _monomial_order_function(self):
        return self.__monomial_order_function

    def __call__(self, x):
        """
        Coerce x into this multivariate polynomial ring.
        """
        if isinstance(x, multi_polynomial.MPolynomial) and x.parent() == self:
            return x
        elif isinstance(x, polydict.PolyDict):
            return multi_polynomial.MPolynomial_polydict(self, x)
        elif isinstance(x, fraction_field_element.FractionFieldElement) and x.parent().ring() == self:
            if x.denominator() == 1:
                return x.numerator()
            else:
                raise TypeError, "unable to coerce in %s since the denominator is not 1"%x
        c = self.base_ring()(x)
        return multi_polynomial.MPolynomial_polydict(self, {self.__zero_tuple:c})

class MPolynomialRing_polydict_domain(integral_domain.IntegralDomain, MPolynomialRing_polydict):
    def __init__(self, base_ring, n=1, names=None):
        MPolynomialRing_polydict.__init__(self, base_ring, n, names)
        
    
    

if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.rational_field import RationalField
    from sage.rings.all import GF
    PolynomialRing = MPolynomialRing
    doctest.testmod(sys.modules[__name__])
