"""
Univariate Polynomial
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import sage.rings.rational_field
import sage.rings.integer_ring
import sage.rings.rational
import integer
import finite_field
import padic_field
import sage.rings.polynomial_ring
from sage.rings.coerce import bin_op, cmp as coerce_cmp
import arith
import sage.rings.ring_element as ring_element
import sage.rings.euclidean_domain_element as euclidean_domain_element
import sage.rings.integral_domain_element as integral_domain_element
import integer_ring
import integer_mod_ring
import polynomial_pyx
import rational_field
import complex_field
import fraction_field_element
from infinity import infinity
import sage.misc.misc as misc
from sage.libs.pari import pari, gen
from sage.libs.ntl import ZZ as ntl_ZZ, ZZX, ZZX_class, ZZ_p, ZZ_pX, ZZ_pX_class, set_modulus
import sage.misc.latex as latex
import sage.structure.factorization as factorization

from coerce import bin_op

QQ = rational_field.RationalField()        

ZZ = integer_ring.IntegerRing()


class Polynomial(ring_element.RingElement):
    """
    Polynomial base class.
    """
    def __init__(self, parent, is_gen = False): 
        """
        The following examples illustrate creation of elements of
        polynomial rings, and some basic arithmetic.

        First we make a polynomial over the integers and do some arithmetic:
            sage: x = PolynomialRing(IntegerRing()).gen()
            sage: f = x**5 + 2*x**2 + (-1); f
            x^5 + 2*x^2 - 1
            sage: f**2
            x^10 + 4*x^7 - 2*x^5 + 4*x^4 - 4*x^2 + 1

        Next we do arithmetic in a sparse polynomial ring over the integers:
            sage: R = PolynomialRing(IntegerRing(), "x"); x = R.gen(); R
            Univariate Polynomial Ring in x over Integer Ring
            sage: S = PolynomialRing(R, "Z"); Z = S.gen(); S
            Univariate Polynomial Ring in Z over Univariate Polynomial Ring in x over Integer Ring
            sage: f = Z**3 + (x**2-2*x+1)*Z - 3; f
            Z^3 + (x^2 - 2*x + 1)*Z + -3
            sage: f*f
            Z^6 + (2*x^2 - 4*x + 2)*Z^4 + (-6)*Z^3 + (x^4 - 4*x^3 + 6*x^2 - 4*x + 1)*Z^2 + (-6*x^2 + 12*x - 6)*Z + 9
            sage: f**3 == f*f*f
            True

        To have the element print as 'y', give 'y' as the
        second argument to the PolynomialRing constructor.
            sage: y = PolynomialRing(IntegerRing(), 'y').gen()
            sage: y**3 - 2*y
            y^3 - 2*y
        """
        ring_element.RingElement.__init__(self, parent)
        self._is_gen = is_gen

    def _add(self, right):
        if self.degree() >= right.degree():
            x = list(self.list())
            y = right.list()
        else:
            x = list(right.list())
            y = self.list()
            
        for i in xrange(len(y)):
            x[i] += y[i]
        
        return self.polynomial(x)
        
    def __call__(self, a):
        """
        Evaluate polynomial at x=a using Horner's rule

        INPUT:
            a -- ring element a; need not be in the coefficient
                 ring of the polynomial.
        OUTPUT:
            ring element, in the parent of a, if a is a ring element.

        EXAMPLES:
            sage: x = PolynomialRing(QQ, 'x').gen()
            sage: f = x - 5
            sage: f(3)
            -2
            sage: f = (x-1)**5
            sage: f(2)
            1
            
        AUTHOR: David Joyner, 2005-04-10.
        """
        try:
            P = a.parent()
            ans = P(self[0])
        except AttributeError:
            P = self.parent().base_ring()
            ans = self[0]
        d = self.degree()
        result = P(self[d])
        i = d-1
        while i >= 0:
            result = result * a + P(self[i])
            i -= 1
        return result

    def __cmp__(self, other):
        if not isinstance(other, Polynomial) or other.parent() != self.parent():
            return coerce_cmp(self, other)
        if self.degree() < other.degree():
            return -1
        elif self.degree() > other.degree():
            return 1
        if self.list() < other.list():
            return -1
        elif self.list() > other.list():
            return 1
        else:
            return 0
            
    def __getitem__(self, n):
        raise NotImplementedError
        
    def __hash__(self):
        return hash((self.degree(), self[0]))
        
    def __float__(self):
         if self.degree() > 0:
             raise TypeError, "cannot coerce nonconstant polynomial to float"
         return float(self[0])

    def __int__(self):
        if self.degree() > 0:
            raise TypeError, "cannot coerce nonconstant polynomial to int"
        return int(self[0])

    def _integer_(self):
        if self.degree() > 0:
            raise TypeError, "cannot coerce nonconstant polynomial %s to int"%self
        return integer.Integer(self[0])

    def __invert__(self):
        return self.parent()(1)/self

    def inverse_of_unit(self):
        if self.degree() > 0:
            raise ValueError, "self is not a unit."
        return self.parent()(~(self[0]))

    def __long__(self):
        if self.degree() > 0:
            raise TypeError, "cannot coerce nonconstant polynomial to long"
        return long(self[0])

    def _mul(self, right):
        """
        EXAMPLES:
            sage: x = PolynomialRing(IntegerRing()).gen()
            sage: (x - 4)*(x**2 - 8*x + 16)
            x^3 - 12*x^2 + 48*x - 64
        """
        if right == 0 or self == 0:
            return self.polynomial(0)
        return self._mul_karatsuba(right)
        
    def __pow__(self, right):
        right = int(right)
        if self.degree() <= 0:
            return self.parent()(self[0]**right)
        if right < 0:
            return (~self)**(-right)
        if self._is_gen:   # special case x**n should be faster!
            z = self.parent()(0)
            z[right] = 1
            return z
        return arith.generic_power(self, right)
    
    def __repr__(self, name=None):
        s = " "
        m = self.degree() + 1
        r = reversed(xrange(m))
        if name is None:
            name = self.parent().name()
        atomic_repr = self.parent().base_ring().is_atomic_repr()
        coeffs = self.list()
        for n in reversed(xrange(m)):
            x = coeffs[n]
            if x != 0:
                if n != m-1:
                    s += " + "
                x = str(x)
                if not atomic_repr and n > 0 and (x.find("+") != -1 or x.find("-") != -1):
                    x = "(%s)"%x
                if n > 1:
                    var = "*%s^%s"%(name,n)
                elif n==1:
                    var = "*%s"%name
                else:
                    var = ""
                s += "%s%s"%(x,var)
        if atomic_repr:
            s = s.replace(" + -", " - ")
        s = s.replace(" 1*"," ")
        s = s.replace(" -1*", " -")
        if s==" ":
            return "0"
        return s[1:]

    def _latex_(self, name=None):
        s = " "
        m = self.degree() + 1
        r = reversed(xrange(m))
        if name is None:
            name = self.parent().name()
        atomic_repr = self.parent().base_ring().is_atomic_repr()
        coeffs = self.list()
        for n in reversed(xrange(m)):
            x = coeffs[n]
            if x != 0:
                if n != m-1:
                    s += " + "
                x = latex.latex(x)
                if not atomic_repr and n > 0 and (x.find("+") != -1 or x.find("-") != -1):
                    x = "\\left(%s\\right)"%x
                if n > 1:
                    var = "|%s^{%s}"%(name,n)
                elif n==1:
                    var = "|%s"%name
                else:
                    var = ""
                s += "%s%s"%(x,var)
        if atomic_repr:
            s = s.replace(" + -", " - ")
        s = s.replace(" 1|"," ")
        s = s.replace(" -1|", " -")
        s = s.replace("|","")
        if s==" ":
            return "0"
        return s[1:]
        
       
    def __setitem__(self, n, x):
        raise NotImplentedError


    def __floordiv__(self,right):
        """
        Quotient of division of self by other.  This is denoted //.
        """
        Q, _ = self.quo_rem(right)
        return Q
        
    def __mod__(self, other):
        """
        Remainder of division of self by other.
        EXAMPLES:
            sage: x = PolynomialRing(IntegerRing()).gen()
            sage: x % (x+1)
            -1
            sage: (x**3 + x - 1) % (x**2 - 1)
            2*x - 1
        """
        _, R = self.quo_rem(other)
        return R

    def _is_atomic(self):
        return self.degree() == self.valuation()
        
    def _mul_generic(self, right):
        d1 = self.degree()
        d2 = right.degree()
        d = d1 + d2
        w = [sum([self[i]*right[k-i] for i in range(0,min(d1,k)+1) if \
                  i <= d1 and k-i <= d2 and self[i]!=0 and right[k-i]!=0]) \
                for k in range(d+1)]
        return Polynomial(self.parent(), w)

    def _mul_karatsuba(self, right):
        r"""
        Returns the product of two polynomials using the Karatsuba
        divide and conquer multiplication algorithm.  This is only
        used over a generic base ring.  (Special libraries like NTL
        are used, e.g., for the integers and rationals, which are much
        faster.)

        INPUT:
           self: Polynomial
           right: Polynomial (over same base ring as self)

        OUTPUT: Polynomial
           The product self*right.

        ALGORITHM:
           The basic idea is to use that
           $$
               (aX + b) (cX + d) = acX^2 + ((a+b)(c+d)-ac-bd)X + bd
           $$  
           where ac=a*c and bd=b*d, which requires three
           multiplications instead of the naive four.  (In my examples,
           strangely just doing the above with four multiplications
           does tend to speed things up noticeably.)
           Given f and g of arbitrary degree bigger than one, let e
           be min(deg(f),deg(g))/2.  Write
           $$
                  f = a X^e + b   \text{ and }   g = c X^e + d
           $$       
           and use the identity
           $$
                 (aX^e + b) (cX^e + d) = ac X^{2e} +((a+b)(c+d) - ac - bd)X^e + bd
           $$     
           to recursively compute $fg$.

        TIMINGS:
        On a Pentium M 1.8Ghz laptop:
           f=R.random(1000,bound=100)
           g=R.random(1000,bound=100)
           time h=f._mul_karatsuba(g)
           Time: 0.42 seconds
           The naive multiplication algorithm takes 14.58 seconds.
           In contrast, MAGMA does this sort of product almost
           instantly, and can easily deal with degree 5000.  Basically
           MAGMA is 100 times faster at polynomial multiplication.

           Over Z using NTL, multiplying two polynomials constructed
           using R.random(10000,bound=100) takes 0.10 seconds.  Using
           MAGMA V2.11-10 the same takes 0.14 seconds.

           Over Q using PARI, multiplying two polynomials constructed
           using R.random(10000,bound=100) takes 1.23 seconds.  Not
           good!  TODO: use NTL polynomials over Z with a denominator
           instead of PARI.

        NOTES:
         * Karatsuba multiplication of polynomials is also implemented in PARI in
                src/basemath/polarit3.c
         * The MAGMA documentation appears to give no information about how
           polynomial multiplication is implemented.
        """
        
        def sum(v,w):
            if len(v)>=len(w):
                x = list(v)
                y = w
            else:
                x = list(w)
                y = v
            for i in range(len(y)):
                x[i] = x[i] + y[i]
            return x
        def dif(v,w):
            if len(v)>=len(w):
                x = list(v)
                y = w
            else:
                x = list(w)
                y = v
            for i in range(len(y)):
                x[i] -= y[i]
            return x
        def do_karatsuba(left, right):
            if len(left) == 0 or len(right) == 0:
                return []
            if len(left) == 1:
                return [left[0]*a for a in right]
            if len(right) == 1:
                return [right[0]*a for a in left]
            if len(left) == 2 and len(right) == 2:
                b = left[0]
                a = left[1]
                d = right[0]
                c = right[1]
                ac = a*c
                bd = b*d
                return [bd,(a+b)*(c+d)-ac-bd,ac]
            e = min(len(left), len(right))/2
            assert e>=1, "bug in karatsuba"
            a, b = left[e:], left[:e]
            c, d = right[e:], right[:e]
            ac = do_karatsuba(a,c)
            bd = do_karatsuba(b,d)
            zeros = [0 for _ in range(e)]
            t2 = zeros + zeros + ac
            t1 = zeros + dif(do_karatsuba(sum(a,b),sum(c,d)),sum(ac,bd))
            t0 = bd
            return sum(t0,sum(t1,t2))
        return self.parent()(do_karatsuba(self.list(), right.list()))
        
    def base_ring(self):
        """
        Return the base ring of the parent of self.
        
        EXAMPLES:
            sage: x = PolynomialRing(ZZ).gen()
            sage: x.base_ring()
            Integer Ring
            sage: (2*x+3).base_ring()
            Integer Ring
        """
        return self.parent().base_ring()
        
    def copy(self):
        """
        Return a copy of self.
        
        EXAMPLES:
        We create the polynomial $f=x+3$, then set $g=f$, and change
        the coefficient of $x$ in $g$, which also changes the coefficient
        of $x$ in $f$.  If we instead copy $f$, then changing the
        coefficient of $x$ of $g$ does not change $f$.

            sage: x = PolynomialRing(IntegerRing()).gen()
            sage: f = x+3
            sage: g = f
            sage: g[1]=3
            sage: f
            3*x + 3
            sage: g = f.copy()
            sage: g[1]=5
            sage: f
            3*x + 3
            sage: g
            5*x + 3
        """
        return self.polynomial(self)

    def degree(self):
        """
        Return the degree of this polynomial.  The zero polynomial
        has degree -1.
        
        EXAMPLES:
            sage: x = PolynomialRing(ZZ).gen()
            sage: f = x**93 + 2*x + 1
            sage: f.degree()
            93
            sage: x = PolynomialRing(QQ, sparse=True).gen()
            sage: f = x**100000 
            sage: f.degree()
            100000
        """
        raise NotImplementedError

    def denominator(self):
        """
        Return the least common multiple of the denominators of
        the entries of self, when this makes sense, i.e., when the
        coefficients have a denominator function.  
        
        WARNING: This is not the denominator of the rational function 
        defined by self, which would always be 1 since self is a polynomial.
        
        EXAMPLES:
        First we compute the denominator of a polynomial with integer
        coefficients, which is of course 1.
            sage: x = PolynomialRing(IntegerRing()).gen()
            sage: f = x**3 + 17*x + 1
            sage: f.denominator()
            1
            
        Next we compute the denominator of a polynomial with rational coefficients.
            sage: Q = RationalField()
            sage: x = PolynomialRing(Q).gen()
            sage: f = Q('1/17')*x**19 - Q('2/3')*x + Q('1/3'); f
            1/17*x^19 - 2/3*x + 1/3
            sage: f.denominator()
            51
            
        Finally, we try to compute the denominator of a polynomial with
        coefficients in the real numbers, which is a ring whose elements
        do not have a denominator method.
            sage: R = RealField()
            sage: x = PolynomialRing(R).gen()
            sage: f = x + R('0.3'); f
            1.0000000000000000*x + 0.29999999999999999
            sage: f.denominator()
            Traceback (most recent call last):
            ...
            AttributeError: '_mpfr.RealNumber' object has no attribute 'denominator'
        """
        if self.degree() == -1:
            return 1
        R = self.base_ring()
        x = self.list()
        d = x[0].denominator()
        for y in x:
            d = d.lcm(y.denominator())
        return d
        
    def derivative(self):
        return self.polynomial([self[n]*n for n in xrange(1,self.degree()+1)])

    def integral(self):
        try:
            return self.polynomial([0] + [self[n]/(n+1) for n in xrange(0,self.degree()+1)])
        except TypeError:
            raise ArithmeticError, "coefficients of integral of %s cannot be coerced into the base ring"%self
            
        
    def dict(self):
        X = {}
        Y = self.list()
        for i in xrange(len(Y)):
            X[i] = Y[i]
        return X

    def factor(self):
        """
        Return polynomials f1, ..., fn and exponents e1, ..., en
        such that the gcd fo the coefficients of the fi is 1, and
        prod fi**ei is equal to a scalar multiple of self.

        INPUT:
            a polynomial

        OUTPUT:
            Factorization -- the factorization of self

        EXAMPLES:
            sage: x = PolynomialRing(RationalField()).gen()
            sage: f = (x**3 - 1)**2
            sage: f.factor()
            (x - 1)^2 * (x^2 + x + 1)^2
        """

        # PERFORMANCE NOTE:
        #     In many tests with SMALL degree PARI is substantially
        #     better than NTL.  (And magma is better yet.)  And the
        #     timing difference has nothing to do with moving Python
        #     data to NTL and back.
        #     For large degree ( > 1500) in the one test I tried, NTL was 
        #     *much* better than MAGMA, and far better than PARI.  So probably
        #     NTL's implementation is asymptotically better.  I could use
        #     PARI for smaller degree over other rings besides Z, and use
        #     NTL in general.
        
        R = self.parent().base_ring()
        if self.degree() <= 0:
            return []
        G = None
        import sage.rings.number_field
        if integer_mod_ring.is_IntegerModRing(R) or  isinstance(R, (integer_ring.IntegerRing,
                          rational_field.RationalField)):
            G = list(self._pari_().factor())
        elif isinstance(R, sage.rings.number_field.NumberField_generic):
            v = [x._pari_("a") for x in reversed(self.list())]
            f = pari(v).Pol()
            G = list(f.factor())
        if G is None:
            raise NotImplementedError
        return self._factor_pari_helper(G)

    def _factor_pari_helper(self, G):
        pols = G[0]
        exps = G[1]
        F = []
        R = self.parent()
        for i in xrange(len(pols)):
            F.append((R(pols[i]),int(exps[i])))
        return factorization.Factorization(F)

    def _lcm(self, other):
        """
        Let f and g be two polynomials.  Then this function
        returns the monic least common multiple of f and g.
        """
        f = self*other
        g = self.gcd(other)
        q = f//g  
        return ~(q[q.degree()])*q  # make monic  (~ is inverse in python)

    def is_monic(self):
        """
        Returns True if this polynomial is monic.  The zero
        polynomial is by definition not monic.
        """
        return not self.is_zero() and self[self.degree()] == 1

    def is_unit(self):
        if self.degree() > 0:
            return False
        return self[0].is_unit()
       
    def is_gen(self):
        return self._is_gen

    def is_zero(self):
        return self.degree() == -1

    def leading_coefficient(self):
        return self[self.degree()]

    def make_monic(self):
        """
        Return this polynomial divided by its leading coefficient.
        """
        if self.is_monic():
            return self
        a = ~self.leading_coefficient()        
        return self*self.parent()(a)
        
    def list(self):
        raise NotImplementedError
        
    def polynomial(self, *args, **kwds):
        return self.parent()(*args, **kwds)

    def _pari_(self):
        try:
            return self.__pari
        except AttributeError:
            v = list(reversed(self.list()))
            try:
                v = [x._pari_() for x in v]
            except AttributeError:
                pass
            self.__pari = pari(v).Pol(self.parent().name())
            return self.__pari
        
    def resultant(self, other, flag=0):
        raise NotImplementedError
            
        ## This should be switched to use NTL, which can apparently compute 
        ## resultants!
        ##        void XGCD(ZZ& r, ZZX& s, ZZX& t, const ZZX& a, const ZZX& b, 
        ##          long deterministic=0);
        ##// r = resultant of a and b; if r != 0, then computes s and t such
        ##// that: a*s + b*t = r; otherwise s and t not affected.  if
        ##// !deterministic, then resultant computation may use a randomized
        ##// strategy that errs with probability no more than 2^{-80}.
        #m = magma.Magma()
        #cmd = "R<%s> := PolynomialRing(RationalField()); "%self.parent().name() + \
        #      "Resultant(%s, %s);"%(self,other)
        #s = m.cmd(cmd)
        #i = s.find("\r")
        #return eval(s[:i])

    def reverse(self):
        v = list(self.list())
        v.reverse()
        return self.parent()(v)

    def roots(self):
        """
        Return all roots of this polynomial.

        EXAMPLES:
            sage: x = PolynomialRing(RationalField()).gen()
            sage: f = x**3 - 1
            sage: f.roots()
            [(1, 1)]
            sage: f = (x**3 - 1)**2
            sage: f.roots()
            [(1, 2)]
        """
        seq = []
        try:
            rts = self.factor()
        except NotImplementedError:
            raise NotImplementedError, "root finding for this polynomial not implemented"
        for fac in rts:                                                      
            g = fac[0]                                                       
            if g.degree() == 1:                                              
                seq.append((-g[0],fac[1]))                                   
        return seq      

    def valuation(self):
        r"""
        If $f = a_r x^r + a_{r+1}x^{r+1} + \cdots$, with $a_r$ nonzero,
        then the valuation of $f$ is $r$.  The valuation of the zero
        polynomial is $\infty$.
        """
        if self.is_zero():
            return infinity
        for i in xrange(self.degree()+1):
            if self[i] != 0:
                return i
        raise RuntimeError, "bug in computing valuation of polynomial"
            
    def name(self):
        return self.parent().name()

    def _xgcd(self, other):
        """
        Extended gcd of self and polynomial other.
        """
        # Algorithm 3.2.2 of Cohen, GTM 138
        R = self.parent()
        A = self
        B = other
        U = R(1)
        G = A
        V1 = R(0)
        V3 = B
        while V3.is_nonzero():
            Q, R = G.quo_rem(V3)
            T = U - V1*Q
            U = V1
            G = V3
            V1 = T
            V3 = R
        V = (G-A*U)//B
        return G, U, V

    def is_irreducible(self):
        F = self.factor()
        if len(F) > 1 or F[0][1] > 1:
            return False
        return True


    def truncate(self, n):
        r"""
        Replace this polynomial by $\sum a_m x^m$ where the sum is
        over $m < n$.  The resulting polynomial is equivalent to self
        modulo $x^n$.
        """
        return self.parent()(self[:int(n)])
        
    
class Polynomial_generic_dense(Polynomial):
    """
    A generic dense polynomial.
    """
    def __init__(self, parent, x=None, check=True, is_gen=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)
        if x == None:
            self.__coeffs = []
            return
        R = parent.base_ring()
#        if isinstance(x, Polynomial) and x.parent() == self.parent():
#            x = list(x.list())
        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                x = list(x.list())
            elif x.parent() == R:
                x = [x]
            else:
                x = [R(a) for a in x.list()]
                check = False
                #raise TypeError, "Cannot coerce %s into %s."%(x, parent)
        elif isinstance(x, dict):
            zero = R(0)
            n = max(x.keys())
            v = [zero for _ in xrange(n+1)]
            for i, z in x.iteritems():
                v[i] = z
            x = v
        elif isinstance(x, gen):
            x = [R(w) for w in reversed(x.Vec())]
            check = True
        elif not isinstance(x, list):
            x = [x]   # constant polynomials
        if check:
            self.__coeffs = [R(z) for z in x]
        else:
            self.__coeffs = x
        if check:
            self.__normalize()
        
    def __reduce__(self):
        return Polynomial_generic_dense, (self.parent(), self.list(), False)

    def __normalize(self):
        x = self.__coeffs
        n = len(x)-1
        while n>=0 and x[n] == 0:
            del x[n]
            n -= 1
        
    def __getitem__(self,n):
        if n < 0 or n >= len(self.__coeffs):
            return self.base_ring()(0)
        return self.__coeffs[n]

    def __getslice__(self, i, j):
        if i < 0:
            i = 0
        v = self.__coeffs[i:j]
        zero = self.base_ring()(0)
        for k in xrange(len(self.__coeffs),j):
            v.append(zero)
        return v

    def __setitem__(self, n, value):        
        if self._is_gen:
            raise ValueError, "the generator cannot be changed"
        n = int(n)
        value = self.base_ring()(value)
        if n >= 0 and n < len(self.__coeffs):
            self.__coeffs[n] = value
            if n == len(self.__coeffs) and value == 0:
                self.__normalize()  
        elif n < 0:
            raise IndexError, "polynomial coefficient index must be nonnegative"
        elif value != 0:
            zero = self.base_ring()(0)
            for _ in xrange(len(self.__coeffs), n):
                self.__coeffs.append(zero)
            self.__coeffs.append(value)
            
    def list(self):
        return self.__coeffs

    def degree(self):
        return len(self.__coeffs) - 1


class Polynomial_generic_sparse(Polynomial):
    """
    A generic sparse polynomial.
    """
    def __init__(self, parent, x=None, check=True, is_gen=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)
        if x == None:
            self.__coeffs = {}
            return
        R = parent.base_ring()
        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                x = dict(x.dict())
            elif x.parent() == R:
                x = {0:x}
            else:
                w = {}
                for n, c in x.dict().iteritems():
                    w[n] = R(c)
                #raise TypeError, "Cannot coerce %s into %s."%(x, parent)
        elif isinstance(x, list):
            y = {}
            for i in xrange(len(x)):
                if x[i] != 0:
                    y[i] = x[i]
            x = y
        elif not isinstance(x, dict):
            x = {0:x}   # constant polynomials
        elif isinstance(x, gen):
            x = [R(w) for w in reversed(x.Vec())]
            check = True
        if check:
            self.__coeffs = {}
            for i, z in x.iteritems():
                self.__coeffs[i] = R(z)
        else:
            self.__coeffs = x
        if check:
            self.__normalize()
        
    def __reduce__(self):
        return Polynomial_generic_sparse, (self.parent(), self.dict(), False)

    def __normalize(self):
        x = self.__coeffs
        zero = self.base_ring()(0)
        D = [n for n, z in x.iteritems() if z == 0]
        for n in D:
            del x[n]
        
    def __getitem__(self,n):
        if not self.__coeffs.has_key(n):
            return self.base_ring()(0)
        return self.__coeffs[n]

    def __getslice__(self, i, j):
        if i < 0:
            i = 0
        zero = self.base_ring()(0)
        v = [zero for _ in xrange(i,j)]
        x = self.__coeffs
        for k in set(x.keys()).intersection(set(xrange(i,j))):
            v[k] = x[k]
        return v

    def __setitem__(self, n, value):
        if self._is_gen:
            raise ValueError, "the generator cannot be changed"
        n = int(n)
        value = self.base_ring()(value)
        x = self.__coeffs
        if n < 0:
            raise IndexError, "polynomial coefficient index must be nonnegative"
        if value == 0:
            if x.has_key(n):
                del x[n]
        else:
            x[n] = value
            
    def list(self):
        zero = self.base_ring()(0)
        v = [zero for _ in xrange(self.degree()+1)]
        for n, x in self.__coeffs.iteritems():
            v[n] = x
        return v

    def _pari_(self):
        return self.__pari

    def degree(self):
        v = self.__coeffs.keys()
        if len(v) == 0:
            return -1
        return max(v)


class Polynomial_generic_field(Polynomial, euclidean_domain_element.EuclideanDomainElement):
    def __init__(self, parent, is_gen=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)
    
##     def quo_rem(self, other):
##         """
##         Returns a tuple (quotient, remainder) where
##             self = quotient*other + remainder.
##         """
##         #       idea: do with ints modulo several primes, and use CRT.
##         #             what is the bound on coefficients of quotient and remainder?
##         #             note -- can also just multiply out to see that results is correct.
##         if not isinstance(other,Polynomial):
##             return bin_op(self, other, operator.div)
##         if self.parent() != other.parent():
##             raise TypeError
            
##         if other.is_zero(): 
##             raise ZeroDivisionError
            
##         # This is algorithm 3.1.1 in Cohen GTM 138
##         A = self
##         B = other
##         R = A
##         Q = self.polynomial(0)
##         X = self.parent().gen()
##         while R.degree() >= B.degree():
##             S =  (R.leading_coefficient()/B.leading_coefficient()) * X**(R.degree()-B.degree())
##             Q += S
##             R -= S*B            
##         return (Q, R)

    def _gcd(self, other):
        """
        Return the GCD of self and other, as a monic polynomial.
        """
        g = euclidean_domain_element.EuclideanDomainElement._gcd(self, other)
        c = g.leading_coefficient()
        if c.is_unit():
            return (1/c)*g
        return g
        
        
class Polynomial_generic_sparse_field(Polynomial_generic_sparse, Polynomial_generic_field):
    def __init__(self, parent, x=None, check=True, is_gen = False):
        Polynomial_generic_sparse.__init__(self, parent, x, check, is_gen)

class Polynomial_generic_dense_field(Polynomial_generic_dense, Polynomial_generic_field):
    def __init__(self, parent, x=None, check=True, is_gen = False):
        Polynomial_generic_dense.__init__(self, parent, x, check, is_gen)


class Polynomial_rational_dense(Polynomial_generic_field):
    """
    A dense polynomial over the rational numbers.
    """
    def __init__(self, parent, x=None, check=True, is_gen=False, construct=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)

        if construct:
            self.__poly = x
            return

        self.__poly = pari([]).Pol()

        if x is None:
            return         # leave initialized to 0 polynomial.
        
        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                self.__poly = x.__poly.copy()
                return
            else:
                x = [QQ(a) for a in x.list()]
                check = False

        if isinstance(x, dict):
            zero = QQ(0)
            n = max(x.keys())
            v = [zero for _ in xrange(n+1)]
            for i, z in x.iteritems():
                v[i] = z
            x = v
            
        elif isinstance(x, gen):
            f = x.Pol()
            self.__poly = f
            assert self.__poly.type() == "t_POL"            
            return
            
        elif not isinstance(x, list):
            x = [x]   # constant polynomials

        if check:
            x = [QQ(z) for z in x]

        x.reverse()
        self.__poly = pari(x).Pol()
        assert self.__poly.type() == "t_POL"        

    def __repr__(self, name=None):
        if name is None:
            name = self.parent().name()
        return str(self.__poly).replace("x", name)

    def __reduce__(self):
        return Polynomial_rational_dense, \
               (self.parent(), self.list(), False, self._is_gen)
        
    def __getitem__(self, n):
        return QQ(self.__poly[n])

    def __getslice__(self, i, j):
        return [QQ(self.__poly[k]) for k in range(i,j)]

    def __pow__(self, n):
        n = int(n)
        if self.degree() <= 0:
            return self.parent()(self[0]**n)
        if n < 0:
            return (~self)**(-n)
        return Polynomial_rational_dense(self.parent(), self.__poly**n, construct=True)
     
    def _add(self, right):
        return Polynomial_rational_dense(self.parent(), 
                                         self.__poly + right.__poly, construct=True)

    def is_irreducible(self):
        try:
            return self.__poly.polisirreducible()
        except NotImplementedError:
            F = self.__poly.factor()
            if len(F) > 1 or F[0][1] > 1:
                return False
            return True

    def quo_rem(self, right):
        """
        Returns a tuple (quotient, remainder) where
            self = quotient*other + remainder.
        """
        if not isinstance(right, Polynomial_rational_dense):
            right = self.parent()(right)
        if right.parent() != self.parent():
            raise TypeError
        v = self.__poly.divrem(right.__poly)
        return Polynomial_rational_dense(self.parent(), v[0], construct=True), \
               Polynomial_rational_dense(self.parent(), v[1], construct=True)
        
                                   
    def _mul(self, right):
        """
        EXAMPLES:
            sage: x = PolynomialRing(QQ).gen()
            sage: (x - QQ('2/3'))*(x**2 - 8*x + 16)
            x^3 - 26/3*x^2 + 64/3*x - 32/3
        """
        return self.parent()(self.__poly * right.__poly, construct=True)        

    def _sub(self, right):
        return self.parent()(self.__poly - right.__poly, construct=True)
                               
    def __setitem__(self, n, value):
        if self._is_gen:
            raise ValueError, "the generator cannot be changed"
        n = int(n)
        if n < 0:
            raise IndexError, "n (=%s) must be >= 0"%n
        if n <= self.__poly.poldegree():
            self.__poly[n] = QQ(value)
        else:
            self.__poly = self.__poly + pari('(%s)*x^%s'%(QQ(value),n))
        if hasattr(self, "__list"):
            del self.__list
     
    def complex_roots(self, flag=0):
        """
        Returns the complex roots of this polynomial.
        INPUT:
            flag -- optional, and can be
                    0: (default), uses Schonhage's method modified by Gourdon,
                    1: uses a modified Newton method.
        OUTPUT:
            list of complex roots of this polynomial, counted with multiplicities.
            
        NOTE: Calls the pari function polroots.

        EXAMPLE:
        We compute the roots of the characteristic polynomial of some Salem numbers:
            sage: R = PolynomialRing(QQ); x = R.gen()
            sage: f = 1 - x**2 - x**3 - x**4 + x**6
            sage: f.complex_roots()[0]
            0.71363917353690087
        """
        R = self.__poly.polroots(flag)
        C = complex_field.CC
        return [C(a) for a in R]

    def copy(self):
        f = Polynomial_rational_dense(self.parent())
        f.__poly = self.__poly.copy()
        return f
        
    def degree(self):
        """
        Return the degree of this polynomial.  The zero polynomial
        has degree -1.
        """
        return max(self.__poly.poldegree(), -1)

    def discriminant(self):
        """
        EXAMPLES:
            sage: x = PolynomialRing(QQ).gen()        
            sage: f = x**3 + 3*x - 17
            sage: f.discriminant()
            -7911
        """
        return QQ(self.__poly.poldisc())
      
    def factor_mod(self, p):
        """
        Return the factorization of self modulo the prime p.

        INPUT:
            p -- prime

        OUTPUT:
            factorization of self reduced modulo p.
        """
        p = integer.Integer(p)
        if not p.is_prime():
            raise ValueError, "p (=%s) must be prime"%p
        G = self._pari_().factormod(p)
        R = sage.rings.polynomial_ring.PolynomialRing(finite_field.FiniteField(p), name=self.parent().name())
        return R(0)._factor_pari_helper(G)
        

    def factor_padic(self, p, prec=10):
        """
        Return p-adic factorization of self to given precision.

        INPUT:
            p -- prime
            prec -- integer; the precision

        OUTPUT:
            factorization of self reduced modulo p.
        """
        p = integer.Integer(p)
        if not p.is_prime():
            raise ValueError, "p (=%s) must be prime"%p
        prec = integer.Integer(prec)
        if prec <= 0:
            raise ValueError, "prec (=%s) must be positive"%prec
        G = self._pari_().factorpadic(p, prec)
        R = sage.rings.polynomial_ring.PolynomialRing(padic_field.pAdicField(p), name=self.parent().name())
        return R(0)._factor_pari_helper(G)

    def list(self):
        """
        EXAMPLES:
            sage: x = PolynomialRing(QQ).gen()
            sage: f = x**3 + 3*x - QQ('17/13')
            sage: f.list()
            [-17/13, 3, 0, 1]
        """
        return [QQ(x) for x in reversed(self.__poly.Vec())]

    def rescale(self, a):
        """
        Return f(a*X).
        """
        b = 1
        c = []
        for i in range(self.degree()+1):
            c.append(self[i]*b)
            b *= a
        return self.parent()(c)

    def resultant(self, other):
        """
        Returns the resultant of self and other, which must lie in the same
        polynomial ring.

        INPUT:
            other -- a polynomial
        OUTPUT:
            an element of the base ring of the polynomial ring

        NOTES:
            Implemented using pari's polresultant function.

        EXAMPLES:
            sage: x = PolynomialRing(RationalField()).gen()
            sage: f = x**3 + x + 1;  g = x**3 - x - 1
            sage: f.resultant(g)
            -8
        """
        if not isinstance(other, Polynomial):
            other = self.polynomial(other)
        if self.parent() != other.parent():
            raise TypeError
        return QQ(str(self.__poly.polresultant(other.__poly, 0)))

    def hensel_lift(self, p, e):
        """
        Assuming that self factors modulo $p$ into distinct factors,
        computes the Hensel lifts of these factors modulo $p^e$.  We
        assume that $p$ has integer coefficients.
        """
        p = integer.Integer(p)
        if not p.is_prime():
            raise ValueError, "p (=%s) must be prime"%p
        e = integer.Integer(e)
        if e < 1:
            raise ValueError, "e (=%s) must be at least 1"%e
        F = self.factor_mod(p)
        y = []
        for g, n in F:
            if n > 1:
                raise ArithmeticError, "The polynomial must be square free modulo p."
            y.append(g)
        H = self._pari_().polhensellift(y, p, e)
        R = integer_mod_ring.IntegerModRing(p**e)
        S = sage.rings.polynomial_ring.PolynomialRing(R, self.parent().name())
        return [S(eval(str(m.Vec().Polrev().Vec()))) for m in H]

class Polynomial_integer_dense(integral_domain_element.IntegralDomainElement, Polynomial):
    """
    A dense polynomial over the integers. 
    """
    def __init__(self, parent, x=None, check=True, is_gen=False, construct=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)
        if construct:
            if isinstance(x, ZZX_class):
                self.__poly = x
                return
            self.__poly = ZZX(x)
            return

        self.__poly = ZZX([])

        if x == None:
            return         # leave initialized to 0 polynomial.
        
        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                self.__poly = x.__poly.copy()
                return
            else:
                x = [ZZ(a) for a in x.list()]
                check = False

        if isinstance(x, dict):
            zero = ZZ(0)
            n = max(x.keys())
            v = [zero for _ in xrange(n+1)]
            for i, z in x.iteritems():
                v[i] = z
            x = v

        elif isinstance(x, ZZX_class):
            self.__poly = x.copy()
            return

        elif isinstance(x, gen):
            x = x.list()
            check = False

        elif isinstance(x, fraction_field_element.FractionFieldElement) and \
                 isinstance(x.numerator(), Polynomial_integer_dense):
            if x.denominator() == 1:
                x = x.numerator().__poly
                check = False

        elif not isinstance(x, list):
            x = [x]   # constant polynomials

        if check:
            x = [ZZ(z) for z in x]

        self.__poly = ZZX(x)

    def ntl_ZZX(self):
        """
        Return underlying NTL representation of this polynomial.
        Additional ``bonus'' functionality may be available through
        this function.
        """
        return self.__poly

    def __reduce__(self):
        return Polynomial_integer_dense, \
               (self.parent(), self.list(), False, self._is_gen)
        
    def __getitem__(self, n):
        return ZZ(self.__poly[n])

    def __getslice__(self, i, j):
        return [ZZ(self.__poly[k]) for k in range(i,j)]

    def __pow__(self, n):
        n = int(n)
        if self.degree() <= 0:
            return self.parent()(self[0]**n)
        if n < 0:
            return (~self)**(-n)
        return self.parent()(self.__poly**n, construct=True)
     
    def _add(self, right):
        return self.parent()(self.__poly + right.__poly, construct=True)

    def quo_rem(self, right):
        """
        Returns a tuple (quotient, remainder) where
            self = quotient*other + remainder.
        """
        if not isinstance(right, Polynomial_integer_dense):
            right = self.parent()(right)
        elif self.parent() != right.parent():
            raise TypeError
        v = self.__poly.quo_rem(right.__poly)
        return self.parent()(v[0], construct=True), \
               self.parent()(v[1], construct=True)
        
    def gcd(self, right):
        """
        Return the GCD of self and other.  The leading
        coefficient need not be 1.
        """
        if not isinstance(right, Polynomial_integer_dense):
            right = self.parent()(right)
        elif self.parent() != right.parent():
            raise TypeError
        g = self.__poly.gcd(right.__poly)
        return self.parent()(g, construct=True)

    def lcm(self, right):
        """
        Return the LCM of self and other, as a monic polynomial.
        """
        if not isinstance(right, Polynomial_integer_dense):
            right = self.parent()(right)
        elif self.parent() != right.parent():
            raise TypeError
        g = self.__poly.lcm(right.__poly)
        return self.parent()(g, construct=True)

    def xgcd(self, right):
        """
        Returns r,s,t such that r = s*self + t*other.
        """
        r, s, t = self.__poly.xgcd(right.__poly)
        return self.parent()(r, construct=True), self.parent()(s, construct=True), \
               self.parent()(t, construct=True)
        
                                   
    def _mul(self, right):
        """
        EXAMPLES:
            sage: x = PolynomialRing(ZZ).gen()
            sage: (x - 2)*(x**2 - 8*x + 16)
            x^3 - 10*x^2 + 32*x - 32
        """
        return self.parent()(self.__poly * right.__poly, construct=True)        

    def _sub(self, right):
        return self.parent()(self.__poly - right.__poly, construct=True)
                               
    def __setitem__(self, n, value):
        if self._is_gen:
            raise ValueError, "the generator cannot be changed"
        n = int(n)
        if n < 0:
            raise IndexError, "n (=%s) must be >= 0"%n
        self.__poly[n] = int(value)
     
    def complex_roots(self, flag=0):
        """
        Returns the complex roots of this polynomial.
        INPUT:
            flag -- optional, and can be
                    0: (default), uses Schonhage's method modified by Gourdon,
                    1: uses a modified Newton method.
        OUTPUT:
            list of complex roots of this polynomial, counted with multiplicities.
            
        NOTE: Calls the pari function polroots.

        EXAMPLE:
        We compute the roots of the characteristic polynomial of some Salem numbers:
            sage: R = PolynomialRing(ZZ); x = R.gen()
            sage: f = 1 - x**2 - x**3 - x**4 + x**6
            sage: f.complex_roots()[0]    # todo: known bug in PARI 2.2.10 !!
            0.71363917353690087
        """
        QQ = sage.rings.rational_field.RationalField()
        R = sage.rings.polynomial_ring.PolynomialRing(QQ)
        return R(self.list()).complex_roots()

    def copy(self):
        f = Polynomial_integer_dense(self.parent())
        f.__poly = self.__poly.copy()
        return f
        
    def degree(self):
        """
        Return the degree of this polynomial.  The zero polynomial
        has degree -1.
        """
        return max(self.__poly.degree(), -1)

    def discriminant(self):
        """
        EXAMPLES:
            sage: x = PolynomialRing(ZZ).gen()        
            sage: f = x**3 + 3*x - 17
            sage: f.discriminant()
            -7911
        """
        return ZZ(str(self.__poly.discriminant()))

    def _pari_(self):
        return pari(str(self.__poly).replace(' ',',')).Pol().polrecip()

    def factor_mod(self, p):
        """
        Return the factorization of self modulo the prime p.

        INPUT:
            p -- prime

        OUTPUT:
            factorization of self reduced modulo p.
        """
        p = integer.Integer(p)
        if not p.is_prime():
            raise ValueError, "p (=%s) must be prime"%p
        G = self._pari_().factormod(p)
        R = sage.rings.polynomial_ring.PolynomialRing(finite_field.FiniteField(p), name=self.parent().name())
        return R(0)._factor_pari_helper(G)
        

    def factor_padic(self, p, prec=10):
        """
        Return p-adic factorization of self to given precision.

        INPUT:
            p -- prime
            prec -- integer; the precision

        OUTPUT:
            factorization of self reduced modulo p.
        """
        p = integer.Integer(p)
        if not p.is_prime():
            raise ValueError, "p (=%s) must be prime"%p
        prec = integer.Integer(prec)
        if prec <= 0:
            raise ValueError, "prec (=%s) must be positive"%prec
        G = self._pari_().factorpadic(p, prec)
        R = sage.rings.polynomial_ring.PolynomialRing(padic_field.pAdicField(p), name=self.parent().name())
        return R(0)._factor_pari_helper(G)
      
    def list(self):
        """
        EXAMPLES:
            sage: x = PolynomialRing(ZZ).gen()
            sage: f = x**3 + 3*x - 17
            sage: f.list()
            [-17, 3, 0, 1]
        """
        return [ZZ(str(self.__poly[i])) for i in xrange(self.degree()+1)]

    def resultant(self, other):
        """
        Returns the resultant of self and other, which must lie in the same
        polynomial ring.

        INPUT:
            other -- a polynomial
        OUTPUT:
            an element of the base ring of the polynomial ring

        NOTES:
            Implemented using NTL's polresultant function.

        EXAMPLES:
            sage: x = PolynomialRing(ZZ).gen()
            sage: f = x**3 + x + 1;  g = x**3 - x - 1
            sage: f.resultant(g)
            -8
        """
        if not isinstance(other, Polynomial) or self.parent() != other.parent():
            other = self.polynomial(other)
        return ZZ(str(self.__poly.resultant(other.__poly, 0)))

    def set_directly(self, v):
        """
        Set the value of this polynomial directly from a vector or string.

        Polynomials over the integers are stored internally using NTL's ZZX
        class.  Use this function to set the value of this polynomial using
        the NTL constructor, which is potentially quicker.   The input v
        is either a vector of ints or a string of the form '[ n1 n2 n3 ... ]'
        where the ni are integers and there are no commas between them.
        The optimal input format is the string format, since that's what NTL uses.

        EXAMPLES:
            sage: R = PolynomialRing(ZZ)
            sage: R([1,2,3])
            3*x^2 + 2*x + 1
            sage: f = R(0)
            sage: f.set_directly([1,2,3])
            sage: f
            3*x^2 + 2*x + 1
            sage: f.set_directly('[1 2 3 4]')
            sage: f
            4*x^3 + 3*x^2 + 2*x + 1
        """
        if self._is_gen:
            raise TypeError, "Cannot change the value of the generator."
        self.__poly = ZZX(v)
        try:
            del self.__list
        except AttributeError:
            pass

        
    
class Polynomial_dense_mod_n(Polynomial):
    """
    A dense polynomial over the integers modulo n, where n is not prime. 
    """
    def __init__(self, parent, x=None, check=True, is_gen=False, construct=False):
        Polynomial.__init__(self, parent, is_gen=is_gen)

        if construct:
            if isinstance(x, ZZ_pX_class):
                self.__poly = x
                return
            parent._ntl_set_modulus()
            self.__poly = ZZ_pX(x)
            return

        self.__poly = ZZ_pX([])

        if x == None:
            return         # leave initialized to 0 polynomial.
        
        if isinstance(x, Polynomial):
            if x.parent() == self.parent():
                parent._ntl_set_modulus()
                self.__poly = x.__poly.copy()
                return
            else:
                x = [ZZ._coerce_(a) for a in x.list()]
                check = False

        if isinstance(x, dict):
            zero = ZZ(0)
            n = max(x.keys())
            v = [zero for _ in xrange(n+1)]
            for i, z in x.iteritems():
                v[i] = z
            x = v

        elif isinstance(x, ZZX_class):
            self.__poly = x.copy()
            return

        elif isinstance(x, gen) and x.type() == 't_POL':
            x = list(reversed(eval(str(x.lift().list()))))
            check = False

        elif isinstance(x, fraction_field_element.FractionFieldElement) and \
                 isinstance(x.numerator(), Polynomial_dense_mod_n):
            if x.denominator() == 1:
                x = x.numerator().__poly
                check = False

        elif not isinstance(x, list):
            x = [x]   # constant polynomials

        if check:
            x = [ZZ(z) for z in x]

        parent._ntl_set_modulus()
        self.__poly = ZZ_pX(x)

    def int_list(self):
        return eval(str(self.__poly).replace(' ',','))

    def _pari_(self):
        return pari(str(list(reversed(self.int_list())))).Pol() * \
               pari('Mod(1,%s)'%self.parent().base_ring().modulus())

    def ntl_ZZ_pX(self):
        """
        Return underlying NTL representation of this polynomial.
        Additional ``bonus'' functionality is available through this
        function.
        """
        return self.__poly

    def __reduce__(self):
        return Polynomial_dense_mod_n, \
               (self.parent(), self.list(), False, self._is_gen)
        
    def __getitem__(self, n):
        return self.base_ring()(self.__poly[n])

    def __getslice__(self, i, j):
        R = self.base_ring()
        return [R(self.__poly[k]) for k in range(i,j)]

    def __pow__(self, n):
        n = int(n)
        if self.degree() <= 0:
            return self.parent()(self[0]**n)
        if n < 0:
            return (~self)**(-n)
        return self.parent()(self.__poly**n, construct=True)
     
    def _add(self, right):
        return self.parent()(self.__poly + right.__poly, construct=True)

    def quo_rem(self, right):
        """
        Returns a tuple (quotient, remainder) where
            self = quotient*other + remainder.
        """
        if not isinstance(right, Polynomial_dense_mod_n):
            right = self.parent()(right)
        elif self.parent() != right.parent():
            raise TypeError
        v = self.__poly.quo_rem(right.__poly)
        P = self.parent()
        return P(v[0], construct=True), P(v[1], construct=True)
        
    def _mul(self, right):
        """
        EXAMPLES:
            sage: x = PolynomialRing(Integers(100)).gen()
            sage: (x - 2)*(x**2 - 8*x + 16)
            x^3 + 90*x^2 + 32*x + 68
        """
        return self.parent()(self.__poly * right.__poly, construct=True)

    def _sub(self, right):
        return self.parent()(self.__poly - right.__poly, construct=True)
                               
    def __setitem__(self, n, value):
        if self._is_gen:
            raise ValueError, "the generator cannot be changed"
        n = int(n)
        if n < 0:
            raise IndexError, "n (=%s) must be >= 0"%n
        self.parent()._ntl_set_modulus()
        self.__poly[n] = int(value)
     
    def copy(self):
        self.parent()._ntl_set_modulus()
        f = self.parent()()
        f.__poly = self.__poly.copy()
        return f
        
    def degree(self):
        """
        Return the degree of this polynomial.  The zero polynomial
        has degree -1.
        """
        return max(self.__poly.degree(), -1)

    def is_irreducible(self):
        return bool(self._pari_().polisirreducible())

    def list(self):
        """
        EXAMPLES:
            sage: x = PolynomialRing(Integers(100)).gen()
            sage: f = x**3 + 3*x - 17
            sage: f.list()
            [83, 3, 0, 1]
        """
        R = self.base_ring()
        return [R(x) for x in self.int_list()]

    def set_directly(self, v):
        r"""
        Set the value of this polynomial directly from a vector or string.

        Polynomials over the integers modulo n are stored internally
        using NTL's ZZ_pX class.  Use this function to set the value
        of this polynomial using the NTL constructor, which is
        potentially \emph{very} fast.  The input v is either a vector
        of ints or a string of the form '[ n1 n2 n3 ... ]' where the
        ni are integers and there are no commas between them.  The
        optimal input format is the string format, since that's what
        NTL uses by default. 

        EXAMPLES:
            sage: R = PolynomialRing(Integers(100))
            sage: R([1,-2,3])
            3*x^2 + 98*x + 1
            sage: f = R(0)
            sage: f.set_directly([1,-2,3])
            sage: f
            3*x^2 + 98*x + 1
            sage: f.set_directly('[1 -2 3 4]')
            sage: f
            4*x^3 + 3*x^2 + 98*x + 1
        """
        if self._is_gen:
            raise TypeError, "Cannot change the value of the generator."
        self.parent()._ntl_set_modulus()
        self.__poly = ZZ_pX(v)
        try:
            del self.__list
        except AttributeError:
            pass

class Polynomial_dense_mod_p(integral_domain_element.IntegralDomainElement, Polynomial_dense_mod_n):
    """
    A dense polynomial over the integers modulo p, where p is prime. 
    """
    def __reduce__(self):
        return Polynomial_dense_mod_p, \
               (self.parent(), self.list(), False, self._is_gen)

    def _gcd(self, right):
        """
        Return the GCD of self and other, as a monic polynomial.
        """
        if not isinstance(right, Polynomial_dense_mod_p):
            right = self.parent()(right)
        elif self.parent() != right.parent():
            raise TypeError
        g = self.ntl_ZZ_pX().gcd(right.ntl_ZZ_pX())
        return self.parent()(g, construct=True)
                                   
    def _xgcd(self, right):
        """
        Return r,s,t such that r = s*self + t*right.
        """
        r, s, t = self.ntl_ZZ_pX().xgcd(right.ntl_ZZ_pX())
        return self.parent()(r, construct=True), self.parent()(s, construct=True), \
               self.parent()(t, construct=True)
    
    def resultant(self, other):
        """
        Returns the resultant of self and other, which must lie in the same
        polynomial ring.

        INPUT:
            other -- a polynomial
        OUTPUT:
            an element of the base ring of the polynomial ring

        EXAMPLES:
            sage: x = PolynomialRing(GF(19)).gen()
            sage: f = x**3 + x + 1;  g = x**3 - x - 1
            sage: f.resultant(g)
            11
        """
        if not isinstance(other, Polynomial) or self.parent() != other.parent():
            other = self.polynomial(other)
        self.parent()._ntl_set_modulus()
        return self.base_ring()(str(self.ntl_ZZ_pX().resultant(other.ntl_ZZ_pX())))
        
    def discriminant(self):
        """
        EXAMPLES:
            sage: x = PolynomialRing(GF(19)).gen()        
            sage: f = x**3 + 3*x - 17
            sage: f.discriminant()
            12
        """
        self.parent()._ntl_set_modulus()
        return self.base_ring()(str(self.ntl_ZZ_pX().discriminant()))

    # PARI is way better than NTL for poly factor, and is called by default in the base class.
    #def factor(self, verbose=False):
    #    M = self.make_monic()
    #    self.parent()._ntl_set_modulus()
    #    F = [(self.parent()(f, construct=True), n) for f, n in M.ntl_ZZ_pX().factor(verbose)]
    #    return factorization.Factorization(F)

    

# doctests
if __name__ ==  '__main__':
    import doctest, sys
    
    # The doctests assume that the following names have been imported:
    from sage.rings.all import GF, Integers, IntegerRing, RationalField, RealField, PolynomialRing
    ZZ = IntegerRing()
    QQ = RationalField()
    
    # Run the doctests.
    doctest.testmod(sys.modules[__name__])
    
    

