"""
Univariate Polynomial Rings
"""


#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import random
import weakref

import commutative_ring
import ring
import ring_element
import field
import integral_domain
import principal_ideal_domain
import polynomial_element as polynomial
import rational_field
import integer_ring
import integer_mod_ring
from sage.libs.pari import pari
import sage.misc.defaults
import sage.misc.latex as latex

from sage.libs.ntl import ZZ as ntl_ZZ, set_modulus

_objsPolynomialRing = {}
def PolynomialRing(base_ring, name=None, sparse=False):
    global _objsPolynomialRing
    key = (base_ring, name, sparse)
    if _objsPolynomialRing.has_key(key):
        R = _objsPolynomialRing[key]()
        if R != None:
            return R

    if isinstance(name, int):
        raise TypeError, "name must be a string"
    if integer_mod_ring.is_IntegerModRing(base_ring) and not sparse:
        n = base_ring.modulus()
        if n.is_prime():
            R = PolynomialRing_dense_mod_p(base_ring, name)
        else:
            R = PolynomialRing_dense_mod_n(base_ring, name)
    elif isinstance(base_ring, field.Field):
        R = PolynomialRing_field(base_ring, name, sparse)
    elif isinstance(base_ring, integral_domain.IntegralDomain):
        R = PolynomialRing_integral_domain(base_ring, name, sparse)
    else:
        R = PolynomialRing_generic(base_ring, name, sparse)
        
    _objsPolynomialRing[key] = weakref.ref(R)
    return R

def is_PolynomialRing(x):
    return isinstance(x, PolynomialRing_generic)


class PolynomialRing_generic(commutative_ring.CommutativeRing, object):
    """
    Univariate polynomial ring over a commutative ring.
    """
    def __init__(self, base_ring, name=None, sparse=False):
        """
        EXAMPLES:
            sage: Q = RationalField()
            sage: R = PolynomialRing(Q)
            sage: x = R.gen()
            sage: R(-1) + R(1)
            0
            sage: (x - Q('2/3'))*(x**2 - 8*x + 16)
            x^3 - 26/3*x^2 + 64/3*x - 32/3
        """
        if not isinstance(base_ring, commutative_ring.CommutativeRing):
            raise TypeError, "Base ring (=%s) must be a commutative ring."%base_ring
        self.__base_ring = base_ring
        self.assign_names(name)
        self.__is_sparse = sparse
        ring.Ring.__init__(self)
        self.__set_polynomial_class()
        self.__generator = self([0,1], is_gen=True)
        self.__cyclopoly_cache = {}
        
    def __call__(self, *args, **kwds):
        return self.__polynomial_class(self, *args, **kwds)

    def _coerce_(self, x):
        if isinstance(x, polynomial.Polynomial) and x.parent() == self:
            return x
        if isinstance(x, ring_element.RingElement) and x.parent() == self.base_ring():
            return self([x])
        return self([self.base_ring()._coerce_(x)])

    def __cmp__(self, other):
        if not isinstance(other, PolynomialRing_generic):
            return -1
        if self.name() == other.name() and \
               self.__base_ring == other.__base_ring:
            return 0
        elif self.name() < other.name():
            return -1
        return +1
        
    def __repr__(self):
        s = "Univariate Polynomial Ring in %s over %s"%(
                self.name(), self.__base_ring)
        if self.is_sparse():
            s = "Sparse " + s
        return s

    def _latex_(self):
        return "%s[%s]"%(latex.latex(self.base_ring()), latex.latex(self.name()))
        
    def __set_polynomial_class(self, cls=None):
        if not (cls is None):
            self.__polynomial_class = cls
            return
        if isinstance(self.base_ring(), rational_field.RationalField) and not self.is_sparse():
            self.__polynomial_class = polynomial.Polynomial_rational_dense
        elif isinstance(self.base_ring(), integer_ring.IntegerRing) and not self.is_sparse():
            self.__polynomial_class = polynomial.Polynomial_integer_dense
        elif isinstance(self.base_ring(), field.Field):
            if self.__is_sparse:
                self.__polynomial_class = polynomial.Polynomial_generic_sparse_field
            else:
                self.__polynomial_class = polynomial.Polynomial_generic_dense_field
        elif self.__is_sparse:
            self.__polynomial_class = polynomial.Polynomial_generic_sparse
        else:
            self.__polynomial_class = polynomial.Polynomial_generic_dense
                
    def __reduce__(self):
        return sage.rings.polynomial_ring.PolynomialRing, \
               (self.__base_ring, self.name(), self.__is_sparse)
    def base_ring(self):
        return self.__base_ring
            
    def characteristic(self):
        return self.__base_ring.characteristic()
        
    def cyclotomic_polynomial(self, n):
        """
        The nth cyclotomic polynomial.

        EXAMPLES:
            sage: R = PolynomialRing(RationalField())
            sage: R.cyclotomic_polynomial(8)
            x^4 + 1
            sage: R.cyclotomic_polynomial(12)
            x^4 - x^2 + 1
            sage: S = PolynomialRing(FiniteField(7))
            sage: S.cyclotomic_polynomial(12)
            x^4 + 6*x^2 + 1
        """
        if n <= 0:
            raise ArithmeticError, "n=%s must be positive"%n
        f = pari.polcyclo(n)
        C = self.__polynomial_class 
        if C == polynomial.Polynomial_rational_dense:
            return self(f, construct=True)
        coeffs = str(f.Vec())
        if C == polynomial.Polynomial_integer_dense:
            return self(coeffs, construct=True)

        coeffs = eval(coeffs)
        return self(coeffs, check=True)

    def gen(self, n=0):
        """
        If this is R[x], return x.
        """
        if n != 0:
            raise IndexError, "Generator %s not defined."%n
        return self.__generator

    def parameter(self):
        return self.gen()

    def is_field(self):
        return False
        
    def is_sparse(self):
        return self.__is_sparse

    def krull_dimension(self):
        return self.base_ring().krull_dimension() + 1
        
    def ngens(self):
        return 1
        
    def quotient(self, f, name=None):
        import polynomial_quotient_ring
        return polynomial_quotient_ring.PolynomialQuotientRing(self, f, name)
    
    def random_element(self, degree, bound=0):
        """
        Return a random polynomial.
        
        INPUT:
            degree -- an int
            bound -- an int (default: 0, which tries to spread choice across ring, if implemented)

        OUTPUT:
            Polynomial -- A polynomial such that the coefficient of x^i,
            for i up to degree, are coercisions to the base ring of
            random integers between -bound and bound.
           
        """
        R = self.base_ring()
        return self([R.random_element(bound) for _ in xrange(degree+1)])


class PolynomialRing_integral_domain(PolynomialRing_generic, integral_domain.IntegralDomain):
    def __init__(self, base_ring, name="x", sparse=False):
        PolynomialRing_generic.__init__(self, base_ring, name, sparse)
        
class PolynomialRing_field(PolynomialRing_integral_domain, principal_ideal_domain.PrincipalIdealDomain):
    def __init__(self, base_ring, name="x", sparse=False):
        PolynomialRing_generic.__init__(self, base_ring, name, sparse)

class PolynomialRing_dense_mod_n(PolynomialRing_generic):
    def __init__(self, base_ring, name="x"):
        self.__modulus = ntl_ZZ(base_ring.modulus()) 
        PolynomialRing_generic.__init__(self, base_ring, name)

    def _ntl_set_modulus(self):
        set_modulus(self.__modulus)

    def __call__(self, *args, **kwds):
        set_modulus(self.__modulus)
        return polynomial.Polynomial_dense_mod_n(self, *args, **kwds)

class PolynomialRing_dense_mod_p(PolynomialRing_dense_mod_n, principal_ideal_domain.PrincipalIdealDomain):
    def __init__(self, base_ring, name="x"):
        self.__modulus = ntl_ZZ(base_ring.modulus())
        PolynomialRing_dense_mod_n.__init__(self, base_ring, name)

    def __call__(self, *args, **kwds):
        set_modulus(self.__modulus)
        return polynomial.Polynomial_dense_mod_p(self, *args, **kwds)
        

def polygen(base_ring, name="x"):
    if isinstance(name, list):
        return polygens(base_ring, name)
    return PolynomialRing(base_ring, name).gen()

    
# doctests
if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.all import FiniteField, RationalField
    doctest.testmod(sys.modules[__name__])
    
