"""
Element of a power series 
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

__doc_exclude=["infinity", "bin_op", "Polynomial"]

import operator

from infinity import infinity
import polynomial_element as polynomial
import power_series_ring
import sage.misc.misc as misc
import ring_element
import arith
import sage.misc.latex as latex
from coerce import bin_op
import rational_field, integer_ring
import sage.libs.pari as pari
import sage.misc.latex as latex


Polynomial = polynomial.Polynomial_generic_dense

class PowerSeries(ring_element.RingElement):
    def __init__(self, parent, prec, is_gen=False): 
        ring_element.RingElement.__init__(self, parent)
        self.__is_gen = is_gen
        if not (prec is infinity):
            prec = int(prec)
        self._prec = prec


    def __cmp__(self, other):
        r"""
        Comparison of self and other.
        
        We say two approximate power series are equal, if they agree
        for all coefficients up to the *minimum* of the precisions of
        each.  Thus, e.g., $f=1+q+O(q^2)$ is equal to $g=1+O(q)$.
        This is how PARI defines equality of power series, but not how
        MAGMA defines equality.  (MAGMA would declare f and g
        unequal.)  I side with PARI, because even if $g=1+q+O(q^2)$,
        we don't really know whether f equals g, since we don't know
        the coefficients of $q^2$.
        """
        if not isinstance(other, PowerSeries):
            try:
                other = self.parent()(other)
            except TypeError, ValueError:
                return -1
        if self.parent() != other.parent():
            return -1
        prec = self.common_prec(other)
        x = self.list()
        y = other.list()
        if prec != infinity:
            x = x[:prec]
            y = y[:prec]
        return misc.generic_cmp(x,y)

    def __call__(self, x):
        raise NotImplementedError

    def base_ring(self):
        return self.parent().base_ring()

    def list(self):
        raise NotImplementedError

    def prec(self):
        """
        The precision of $...+O(x^r)$ is by definition $r$.
        """
        return self._prec

    def __repr__(self):
        if self.is_zero():
            if self.prec() == infinity:
                return "0"
            else:
                return "O(%s^%s)"%(self.parent().name(),self.prec())
        s = " "
        v = self.list()
        m = len(v)
        X = self.parent().name()
        atomic_repr = self.parent().base_ring().is_atomic_repr()
        first = True
        for n in xrange(m):
            x = v[n]
            if x != 0:
                if not first:
                    s += " + "
                x = str(x)
                if not atomic_repr and n > 0 and (x[1:].find("+") != -1 or x[1:].find("-") != -1):
                    x = "(%s)"%x
                if n > 1:
                    var = "*%s^%s"%(X,n)
                elif n==1:
                    var = "*%s"%X
                else:
                    var = ""
                s += "%s%s"%(x,var)
                first = False

        if atomic_repr:
            s = s.replace(" + -", " - ")
        s = s.replace(" 1*"," ")
        s = s.replace(" -1*", " -")
        if self._prec == 0:
            bigoh = "O(1)"
        elif self._prec == 1:
            bigoh = "O(%s)"%self.parent().name()
        else:
            bigoh = "O(%s^%s)"%(self.parent().name(),self._prec)
        if self._prec != infinity:
            if s==" ":
                return bigoh
            s += " + %s"%bigoh
        return s[1:]

    def _latex_(self):
        if self.is_zero():
            if self.prec() == infinity:
                return "0"
            else:
                return "0 + \\cdots"
        s = " "
        v = self.list()
        m = len(v)
        X = latex.latex(self.parent().name())
        atomic_repr = self.parent().base_ring().is_atomic_repr()
        first = True
        for n in xrange(m):
            x = v[n]
            if x != 0:
                if not first:
                    s += " + "
                x = str(latex.latex(x))
                if not atomic_repr and n > 0 and (x[1:].find("+") != -1 or x[1:].find("-") != -1):
                    x = "\\left(%s\\right)"%x
                if n > 1:
                    var = "%s^{%s}"%(X,n)
                elif n==1:
                    var = "%s"%X
                else:
                    var = ""
                if n > 0:
                    s += "%s|%s"%(x,var)
                else:
                    s += str(x)
                first = False

        if atomic_repr:
            s = s.replace(" + -", " - ")
        print s
        s = s.replace(" -1|", " -")
        s = s.replace(" 1|"," ")
        s = s.replace("|","")
        if self._prec != infinity:
            if s==" ":
                return "0 + \\cdots"
            s += " + \\cdots"
        return s[1:]

    
    def trunc(self, prec=infinity):
        """
        Polynomial obtained from power series by truncating.
        """
        if prec == infinity:
            prec = self._prec
        a = self.list()
        v = [a[i] for i in range(min(prec, len(a)))]
        return self.parent()._poly_ring()(v)

    def add_bigoh(self, prec):
        r"""
        Returns the power series of precision at most prec got by
        adding $O(q^\text{prec})$ to f, where q is the variable.
        """
        if prec == infinity or prec >= self.prec():
            return self
        a = self.list()
        v = [a[i] for i in range(min(prec, len(a)))]        
        return self.parent()(v, prec)
        
    def __getitem__(self,n):
        if n<0:
            return 0
        c = self.list()
        if n >= len(c):
            if self._prec > n:
                return self.base_ring()(0)
            else:
                raise IndexError, "Coefficient %s not known."%n
        return c[n]
    
    def __getslice__(self, i, j):
        if j>self.prec():
            j = self.prec()
        return self.list()[int(i):int(j)]
    
    def __setitem__(self, n, value):
        raise NotImplementedError
        
    def common_prec(self, f):
        if self.prec() == infinity:
            return f.prec()
        elif f.prec() == infinity:
            return self.prec()
        return min(self.prec(), f.prec())

    def __add__(self):
        raise NotImplementedError

    def __radd__(self, left):
        return self.parent()(left) + self

    def __sub__(self, right):
        raise NotImplementedError

    def __mul__(self, right):
        if not isinstance(right,PowerSeries) or self.parent() != right.parent():
            return bin_op(self, right, operator.mul)
        if self.is_zero():
            return self
        if right.is_zero():
            return right
        sp = self.prec()
        rp = right.prec()
        if sp == infinity:
            if rp == infinity:
                prec = infinity
            else:
                prec = rp + self.valuation()
        else:  # sp != infinity
            if rp == infinity:
                prec = sp + right.valuation()
            else:
                prec = min(rp + self.valuation(), sp + right.valuation())
        # endif
        return self._mul(right, prec)

    def _mul(self, right, prec):
        raise NotImplementedError
    
    def __invert__(self):
        """
        Inverse of the power series, which we assume to have nonzero
        constant term so that the inverse is again a power series.
        """
        if self.valuation() > 0:
            u = ~self.unit_part()  # inverse of unit part
            R = self.parent().laurent_series_ring()
            return R(u, -self.valuation())
        a0 = self[0]
        try:
            b = [~a0]
        except ValueError, ZeroDivisionError:
            raise ZeroDivisionError, "leading coefficient must be a unit."
        #  By multiplying through we may assume that the leading coefficient
        #  of f=self is 1.   If f = 1 + a_1*q + a_2*q^2 + ..., then we have
        #  the following recursive formula for the coefficients b_n of the
        #  expansion of f^(-1):
        #        b_n = -b_0*(b_{n-1}*a_1 + b_{n-2}*a_2 + ... + b_0 a_n).
        a = self.list()
        for n in range(1,len(a)):
            b.append(-b[0]*sum([b[n-i]*a[i] for i in range(1,n+1)]))
        return self.parent()(b, prec=self.prec())
    
    def unit_part(self):
        """
        Suppose self factors as $q^n\cdot (a_0 + a_1 q + \cdots)$
        with $a_0$ nonzero.  Then this function returns $a_0 + a_1 q +
        \cdots $.
        """
        #assert not self.is_zero(), "Argument must be nonzero to have a unit part."
        if self.is_zero():
            return self.parent()(0, self.prec())
        n = self.valuation()
        v = self[int(n):]
        return self.parent()(v, self.prec()-n)
    
    def __div__(self, denom):
        if not isinstance(denom, PowerSeries) or self.parent() != denom.parent():
            return bin_op(self, denom, operator.div)

        u = denom.unit_part()
        inv = ~u  # inverse
        
        v = denom.valuation()
        if v > self.valuation():
            R = self.parent().laurent_series_ring()
            return R(self)/R(denom)
        
        # Algorithm: Cancel common factors of q from top and bottom,
        # then invert the denominator.  We do the cancellation first
        # because we can only invert a unit (and remain in the ring
        # of power series).
        
        u = denom.unit_part()
        if v > 0:
            coeffs = self[int(v):]
            num = self.parent()(coeffs, self.prec()-v)
        else:
            num = self
        return num*inv
    
    def __rdiv__(self, left):
        return self.parent()(left) / self
    
    def __pow__(self, right):
        right=int(right)
        if right < 0:
            return (~self)**(-right)
        if right == 0:
            return self.parent()(1)
        if self.is_zero():
            return self
        if self.valuation() == 1 and self.degree() == 1:   # alpha*q 
            v = [0 for _ in range(right)]
            v.append(self[1]**right)
            return self.parent()(v, prec=self.prec()+right-1)
        return arith.generic_power(self, right)
    
    def __mod__(self, other):
        if isinstance(other,(int,Integer,long)):
            try:
                return power_series_ring.PowerSeriesRing(Zmod(other))(self)
            except TypeError:
                raise ZeroDivisionError
        raise NotImplementedError, "Mod on power series ring not defined."
    
    def O(self, prec):
        r"""
        Return this series plus $O(x^\text{prec})$.  Does not change
        self.
        """
        if prec == infinity or prec >= self.prec():
            return self
        coeffs = self[:prec]
        return self.parent()(coeffs, prec)

    def exp(self, prec = infinity):
        if prec == infinity:
            if self.prec() == infinity:
                raise RuntimeError, "prec must be finite for exponentiation " + \
                      "of infinite-precision power series"
            prec = self.prec()
        if self.valuation() <= 0:
            raise ArithmeticError, "valuation must be positive"
        x = self.parent()(self.list(), prec)
        ans = self.parent()(1)
        k = self.parent().base_ring()
        for n in range(1,prec):
            ans += x * (k(1)/k(arith.factorial(n)))
            x *= self
        return self.parent()(ans, prec)

    def copy(self):
        raise NotImplementedError

    def V(self, n):
        """
        If $f = \sum a_m x^m$, then this function returns $\sum a_m x^{nm}$.
        """
        v = self.list()
        m = 0
        w = []
        zero = self.base_ring()(0)
        for i in range(len(v)*n):
            if i%n != 0:
                w.append(zero)
            else:
                w.append(v[m])
                m += 1
        return self.parent()(w, self.prec()*n)

    def valuation(self):
        prec = self.prec()
        v = self.list()
        if prec == infinity:
            if len(v) == 0:
                return infinity
        n = 0
        while n < len(v) and v[n] == 0:
            n += 1
        return n

    def degree(self):
        return len(self.list())-1

    def derivative(self):
        raise NotImplementedError

#def _normalize(v, prec):
#    n = len(v)-1
#    while n>=0 and (v[n] == 0 or (prec != infinity and n >= prec)):
#        del v[n]
#        n -= 1

class PowerSeries_generic_dense(PowerSeries):
    def __init__(self, parent, f=0, prec=infinity, check=True, is_gen=False):
        R = parent._poly_ring()
        if not (isinstance(f, polynomial.Polynomial) and f.parent() == R):
            f = R(f)
        self.__f = f
        if check and prec != infinity:
            self.__f = self.__f.truncate(prec)
        PowerSeries.__init__(self, parent, prec, is_gen)
    
    def __call__(self, x):
        return self.__f(x)

    def __setitem__(self, n, value):
        self.__f[n] = value
        self._prec = max(self._prec, n+1)

    def __neg__(self):
        return PowerSeries_generic_dense(self.parent(), -self.__f,
                                         self._prec, check=False)
        
    def __add__(self, right):
        if not isinstance(right,PowerSeries_generic_dense) or self.parent() != right.parent():
            return bin_op(self, right, operator.add)
        return PowerSeries_generic_dense(self.parent(), self.__f + right.__f, \
                                         self.common_prec(right), check=False)

    def __sub__(self, right):
        if not isinstance(right,PowerSeries_generic_dense) or self.parent() != right.parent():
            return bin_op(self, right, operator.sub)
        return PowerSeries_generic_dense(self.parent(), self.__f - right.__f, \
                                         self.common_prec(right), check=False)

    def _mul(self, right, prec):
        return PowerSeries_generic_dense(self.parent(), self.__f * right.__f, prec)
    

    def copy(self):
        return PowerSeries_generic_dense(self.parent(), self.__f, self.prec(), check=False)

    def list(self):
        return self.__f.list()

    def derivative(self):
        return PowerSeries_generic_dense(self.parent(), self.__f.derivative(),
                                         self.prec()-1, check=False)

    def integral(self):
        """
        The integral of this power series with 0 constant term.
        """
        return PowerSeries_generic_dense(self.parent(), self.__f.integral(),
                                         self.prec()+1, check=False)

    def laurent_series(self):
        return self.parent().laurent_series_ring()(self)

    def reversion(self):
        """
        Return the reversion of f, i.e., the series g such that
        g(f(x)) = x.

        EXAMPLES:
            sage: x = PowerSeriesRing(RationalField()).gen()
            sage: f = 2*x + 3*x**2 - x**4 + O(x**5)
            sage: g = f.reversion()
            sage: g
            1/2*x - 3/8*x^2 + 9/16*x^3 - 131/128*x^4 + O(x^5)
            sage: f(g)
            x + O(x^5)
            sage: g(f)
            x + O(x^5)
        """
        if not isinstance(self.parent().base_ring(), rational_field.RationalField):
            raise NotImplementedError
        if self.prec() == infinity:
            raise RuntimeError, "series must have finite precision for reversion."
        f = self._pari_()
        g = f.serreverse()
        return PowerSeries_generic_dense(self.parent(),[0]+ g.list(),self.prec())

    def _pari_(self):
        if not isinstance(self.parent().base_ring(),
                          (rational_field.RationalField, integer_ring.IntegerRing)):
            raise NotImplementedError
        if self.prec() == infinity:
            raise RuntimeError, "series precision must be finite for conversion to pari object."
        return pari.pari(str(self))
            
    
# doctests
if __name__ ==  '__main__':
    import doctest, sys
    from sage.rings.all import PowerSeriesRing, RationalField, O
    doctest.testmod(sys.modules[__name__])
    
