"""
Univariate Power Series Rings
"""

import weakref
import power_series
import polynomial_ring
import laurent_series_ring
import commutative_ring
import integral_domain
import field
import sage.structure.gens as gens
from infinity import infinity
import sage.misc.latex as latex

_objsPowerSeriesRing = {}
def PowerSeriesRing(base_ring, name=None):
    global _objsPowerSeriesRing
    key = (base_ring, name)
    if _objsPowerSeriesRing.has_key(key):
        x = _objsPowerSeriesRing[key]()
        if x != None: return x
    if isinstance(base_ring, field.Field):
        R = PowerSeriesRing_over_field(base_ring, name)
    elif isinstance(base_ring, integral_domain.IntegralDomain):
        R = PowerSeriesRing_domain(base_ring, name)
    elif isinstance(base_ring, commutative_ring.CommutativeRing):
        R = PowerSeriesRing_generic(base_ring, name)
    else:
        raise TypeError, "base_ring must be a commutative ring"
    _objsPowerSeriesRing[key] = weakref.ref(R)
    return R

def is_PowerSeriesRing(x):
    return isinstance(x, PowerSeriesRing_generic)

class PowerSeriesRing_generic(commutative_ring.CommutativeRing):
    def __init__(self, base_ring, name=None):
        self.__base_ring = base_ring
        self.__poly_ring = polynomial_ring.PolynomialRing(base_ring, name)
        self.__power_series_class = power_series.PowerSeries_generic_dense
        self.__generator = self.__power_series_class(self, [0,1], check=True, is_gen=True)
        self.assign_names(name)

    def __repr__(self):
        return "Power series ring in %s over %s"%(self.name(), self.base_ring())

    def _latex_(self):
        return "%s[[%s]]"%(latex.latex(self.base_ring()), latex.latex(self.name()))

    def __call__(self, f, prec=infinity):
        if isinstance(f, power_series.PowerSeries) and f.parent() == self and prec == f.prec():
            return f
        return self.__power_series_class(self, f, prec)

    def base_ring(self):
        return self.__base_ring

    def _poly_ring(self):
        return self.__poly_ring

    def gen(self, n=0):
        if n != 0:
            raise IndexError, "Generator %s not defined."%n
        return self.__generator

    def ngens(self):
        return 1

    def random_element(self, prec, bound=0):
        """
        Return a random power series.
        
        INPUT:
            prec -- an int
            bound -- an int (default: 0, which tries to spread choice across ring, if implemented)

        OUTPUT:
            Polynomial -- A polynomial such that the coefficient of x^i,
            for i up to degree, are coercisions to the base ring of
            random integers between -bound and bound.
        """
        return self(self.__poly_ring.random_element(prec, bound), prec)

    def __cmp__(self, other):
        if not isinstance(other, PowerSeriesRing_generic):
            return -1
        if self.base_ring() != other.base_ring():
            return -1
        if self.name() != other.name():
            return -1
        return 0

    def __contains__(self, x):
        if not isinstance(x, PowerSeries):
            return False
        return x.parent() == self

    def is_atomic_repr(self):
        return False

    def is_field(self):
        return False

    def is_finite(self):
        return False

    def characteristic(self):
        return self.base_ring().characteristic()

    def laurent_series_ring(self):
        """
        If this is the power series ring R[[t]], return the Laurent
        series ring R((t)).
        """
        try:
            return self.__laurent_series_ring
        except AttributeError:
            self.__laurent_series_ring = laurent_series_ring.LaurentSeriesRing(
                                                 self.base_ring(), self.name())
            return self.__laurent_series_ring            
    
class PowerSeriesRing_domain(integral_domain.IntegralDomain, PowerSeriesRing_generic):
    def __init__(self, base_ring, name=None):
        PowerSeriesRing_generic.__init__(self, base_ring, name)

    
class PowerSeriesRing_over_field(PowerSeriesRing_domain):
    def __init__(self, base_ring, name=None):
        PowerSeriesRing_generic.__init__(self, base_ring, name)

    def fraction_field(self):
        return self.laurent_series_ring()
    
