"""
Quotient ring element.

AUTHOR: William Stein
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

__doc_exclude = ["bin_op"]

import operator

import ring_element
import quotient_ring

from sage.rings.coerce import bin_op

class QuotientRingElement(ring_element.RingElement):
    def __init__(self, parent, rep):
        ring_element.RingElement.__init__(self, parent)
        self.__rep = rep

    def copy(self):
        return QutientRingElement(self.parent(), self.__rep)

    def lift(self):
        return self.__rep

    def is_zero(self):
        return self.__rep in self.parent().ideal()

    def is_unit(self):
        if self.__rep.is_unit():
            return True
        raise NotImplementedError
    
    def __repr__(self):
        s = ', '.join([str(x) for x in self.ring().ideal().gens()])
        return '[%s mod (%s)]'%(self.__rep, s)
    
    def __add__(self, right):
        if not isinstance(right, QuotientRingElement):
            return bin_op(self, right, operator.add)
        if self.parent() != right.parent():
            raise TypeError
        return QuotientRingElement(self.parent(), self.__rep + right.__rep)
    
    def __radd__(self, left):
        return self.parent()(left) + self
    
    def __sub__(self, right):
        if not isinstance(right, QuotientRingElement):
            return bin_op(self, right, operator.sub)
        if self.parent() != right.parent():
            raise TypeError
        return QuotientRingElement(self.parent(), self.__rep - right.__rep)
    
    def __rsub__(self, left):
        return self.parent()(left) - self        

    def __mul__(self, right):
        if not isinstance(right, QuotientRingElement):
            return bin_op(self, right, operator.mul)
        if self.parent() != right.parent():
            raise TypeError
        return QuotientRingElement(self.parent(), self.__rep * right.__rep)

    def __div__(self, right):
        if not isinstance(right, QuotientRingElement):
            return bin_op(self, right, operator.div)
        if self.parent() != right.parent():
            raise TypeError
        if not right.is_unit():
            raise ZeroDivisionError
        raise NotImplementedError        

    def __int__(self):
        return int(self.lift())

    def _integer_(self):
        try:
            return self.lift()._integer_()
        except AttributeError:
            raise NotImplementedError

    def _rational_(self):
        try:
            return self.lift()._rational_()
        except AttributeError:
            raise NotImplementedError

    def __long__(self):
        return int(self.lift())        

    def __rdiv__(self, left):
        return self.parent()(left)/self
    
    def __pow__(self, right):
        right = int(right)
        return QuotientRingElement(self.parent(), self.__rep**right)
    
    def __neg__(self):
        return QuotientRingElement(self.parent(), -self.__rep)
    
    def __pos__(self):
        return self
    
    def __invert__(self):
        raise NotImplementedError

    def __float__(self):
        return float(self.lift())
    
    def __cmp__(self, other):
        if not isinstance(other, QuotientRingElement):
            other = self.parent()(other)
        if self.__rep - other.__rep in self.parent().ideal():
            return 0
        return -1

if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
