"""
Abstract base class for rings
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import random
import sage.rings.coerce

cdef class Ring(_gens.Generators):
    """
    Generic ring class.
    """
    def __init__(self):
        pass
    
    def __call__(self, x):
        """
        Coerce x into the ring.
        """
        raise NotImplementedError

    def _coerce_(self, x):
        return self(x)

    def __contains__(self, x):
        r"""
        True if coercion of $x$ into self is possible.  Thus, e.g.,
        $2$ is in the integers and in $\Z/7\Z$ and the element $3$ of
        $\Z/7\Z$ {\em is} in $\Z$.  However, $2/3$ is not in $\Z$.
        Think of this as returning True if $x==y$ for some $y$ in
        self.
        
        EXAMPLES:
            sage: 2 in Integers(7)
            True
            sage: 2 in IntegerRing()
            True
            sage: Integers(7)(3) in IntegerRing()
            True
        """
        try:
            self(x)
        except TypeError:
            return False
        return True

    def __reduce__(self):
        """
        Needed for object persistence.
        """
        raise NotImplementedError, "object persistence not yet supported for this Ring"

    def category(self):
        """
        Return the category to which this ring belongs.
        """
        import sage.categories.all
        return sage.categories.all.Rings()

    def ideal(self, gens, coerce=True):
        """
        Return the ideal generated by gens.
        """
        import sage.rings.ideal
        return sage.rings.ideal.Ideal(self, gens, coerce=coerce)

    def principal_ideal(self, gen, coerce=True):
        """
        Return the principal ideal generated by gen.
        """
        return self.ideal([gen], coerce=coerce)

    def unit_ideal(self):
        """
        Return the unit ideal of this ring.
        """
        return self.ideal([self(1)], coerce=False)

    def zero_ideal(self):
        """
        Return the zero ideal of this ring.
        """
        return self.ideal([self(0)], coerce=False)

    def is_atomic_repr(self):
        """
        True if the elements have atomic string representations, in the sense
        that they print if they print at s, then -s means the negative of s.
        For example, integers are atomic but polynomials are not.
        """
        return False

    def is_commutative(self):
        """
        Return True if this ring is commutative.
        """
        raise NotImplementedError

    def is_field(self):
        """
        Return True if this ring is a field.
        """
        raise NotImplementedError

    def is_finite(self):
        """
        Return True if this ring is finite.
        """
        raise NotImplementedError

    def is_integral_domain(self):
        """
        Return True if this ring is an integral domain.
        """
        return NotImplementedError

    def is_ring(self):
        """
        Return True since self is a ring.
        """
        return True

    def is_noetherian(self):
        """
        Return True if this ring is Noetherian.
        """
        raise NotImplementedError

    def characteristic(self):
        """
        Return the characteristic of this ring.
        """
        raise NotImplementedError

    def order(self):
        """
        The number of elements of self.
        """
        raise NotImplementedError
    
    def __hash__(self):
        return hash(self.__repr__())

    def random_element(self, bound=None):
        """
        Return a random integer coerced into this ring, where the
        integer is chosen uniformly from the interval [-bound,bound].

        INPUT:
            bound -- int or None; (default: None, which defaults to 2.)
        """
        if bound is None:
            bound = 2
        return self(random.randrange(-bound, bound+1))


cdef class CommutativeRing(Ring):
    """
    Generic commutative ring.
    """
    def is_commutative(self):
        """
        Return True, since this ring is commutative.
        """
        return True
    
    def krull_dimension(self):
        """
        Return the Krull dimension if this commutative ring.

        The Krull dimension is the length of the longest ascending chain
        of prime ideals.
        """
        raise NotImplementedError

    def quotient(self, I):
        """
        Return the quotient of self by the ideal I of self.
        """
        import sage.rings.quotient_ring
        return sage.rings.quotient_ring.QuotientRing(self, I)
    

cdef class IntegralDomain(CommutativeRing):
    """
    Generic integral domain class.
    """
    def is_integral_domain(self):
        """
        Return True, since this ring is an integral domain.
        """
        return True
    
    def fraction_field(self):
        """
        Return the fraction field of self.
        """
        try:
            return self.__fraction_field
        except AttributeError:
            import sage.rings.fraction_field
            self.__fraction_field = sage.rings.fraction_field.FractionField_generic(self)
        return self.__fraction_field

    def is_field(self):
        """
        Return True if this ring is a field.
        """
        if self.is_finite():
            return True
        raise NotImplementedError, "unable to determine whether or not is a field."

cdef class NoetherianRing(CommutativeRing):
    """
    Generic Noetherian ring class.

    A Noetherian ring is a commutative ring in which every ideal is
    finitely generated.
    """
    def is_noetherian(self):
        """
        Return True since this ring is Noetherian.
        """
        return True

cdef class DedekindDomain(IntegralDomain):
    """
    Generic Dedekind domain class.

    A Dedekind domain is a Noetherian integral domain of Krull
    dimension one that is integrally closed in its field of fractions.
    """
    def krull_dimension(self):
        """
        Return 1 since Dedekind domains have Krull dimension 1.
        """
        return 1

    def is_integrally_closed(self):
        """
        Return True since Dedekind domains are integrally closed.
        """
        return True

    def integral_closure(self):
        """
        Return self since Dedekind domains are integrally closed.
        """
        return self

    def is_noetherian(self):
        """
        Return True since Dedekind domains are noetherian.
        """
        return True
    

cdef class PrincipalIdealDomain(IntegralDomain):
    """
    Generic principal ideal domain.
    """
    def class_group(self):
        """
        Return the trivial group, since the class group of a PID is trivial.
        """
        from sage.groups.abelian import AbelianGroup
        return AbelianGroup([])

    def gcd(self, x, y, coerce=True):
        """
        Return the greatest common divisor of x and y, as elements
        of self.
        """
        if coerce:
            x = self(x)
            y = self(y)
        return x.gcd(y)
    

cdef class EuclideanDomain(PrincipalIdealDomain):
    """
    Generic Euclidean domain class.
    """
    def parameter(self):
        """
        Return an element of degree 1.
        """
        raise NotImplementedError
        
def is_Field(x):
    """
    Return True if x is of class Field.
    """
    return isinstance(x, Field)

cdef class Field(PrincipalIdealDomain):
    """
    Generic field
    """
    def fraction_field(self):
        """
        Return the fraction field of self.
        """
        return self

    def divides(self, x, y, coerce=True):
        """
        Return True if x divides y in this field (usually True in a
        field!).  If coerce is True (the default), first coerce x and
        y into self.
        """
        if coerce:
            x = self(x)
            y = self(y)
        if x.is_zero():
            return y.is_zero()
        return True

    def ideal(self, gens, coerce=True):
        """
        Return the ideal generated by gens.
        """
        if coerce:
            gens = []
            for g in gens:
                gens.append(self(g))
        for x in gens:
            if not x.is_zero():
                return self.unit_ideal()
        return self.zero_ideal()

    def integral_closure(self):
        """
        Return this field, since fields are integrally closed in their
        fraction field.
        """
        return self

    def is_field(self):
        """
        Return True since this is a field.
        """
        return True

    def is_integrally_closed(self):
        """
        Return True since fields are integrally closed in their
        fraction field.
        """
        return True

    def is_noetherian(self):
        """
        Return True since fields are noetherian rings.
        """
        return True

    def krull_dimension(self):
        """
        Return the Krull dimension of this field, which is 0.
        """
        return 0

    def prime_subfield(self):
        """
        Return the prime subfield of self.

        EXAMPLES:
            sage: k = GF(9)
            sage: k.prime_subfield()
            Finite field of size 3
        """
        if self.characteristic() == 0:
            import sage.rings.rational_field
            return sage.rings.rational_field.RationalField()
        else:
            import sage.rings.finite_field
            return sage.rings.finite_field.FiniteField(self.characteristic())

cdef class Algebra(Ring):
    """
    Generic algebra
    """
    def __init__(self, base_ring):
        self.__base_ring = base_ring

    def base_ring(self):
        """
        Return the base ring of this algebra.
        """
        return self.__base_ring

    def characteristic(self):
        """
        Return the characteristic of this algebra, which is the same
        as the characteristic of its base ring.
        """
        return self.base_ring().characteristic()
    

cdef class CommutativeAlgebra(Algebra):
    """
    Generic algebra
    """
    def __init__(self, base_ring):
        Algebra.__init__(self, base_ring)

    def base_ring(self):
        """
        Return the base ring of this commutative algebra.
        """
        return self.__base_ring
    
    def is_commutative(self):
        """
        Return True since this algebra is commutative.
        """
        return True
