"""
Elliptic curves over finite fields
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

#import sage.schemes.curves.elliptic.ell_generic
import ell_generic
import sage.rings.ring as ring
import cremona
import sea
from sage.groups.all import AbelianGroup

class EllipticCurve_finite_field(ell_generic.EllipticCurve):
    """
    Elliptic curve over a finite field.
    """
    def __init__(self, x, y=None):
        if isinstance(x, list):
            ainvs = x
            field = ainvs[0].parent()
        else:
            field = x
            ainvs = y
        if not (isinstance(field, ring.Ring) and isinstance(ainvs,list)):
            raise TypeError
        
        ell_generic.EllipticCurve.__init__(
            self, [field(x) for x in ainvs])

    def _type(self):
        return "finite_field"

    def points(self):
        """
        All the points on this elliptic curve.

        TODO: currently very stupid implementation -- make faster using self.abelian_group()?
        """
        try:
            return self.__points
        except AttributeError: pass
        
        points = [self(0)]
        for x in self.base_field():
            for y in self.base_field():
                try:
                    points.append(self(x,y, check=True))
                except ArithmeticError:
                    pass
        self.__points = points
        return self.__points

    def cardinality(self, early_abort=False):
        """
        Return the number of points on this elliptic curve over this
        finite field.

        INPUT:
            early_abort -- bool (default: Falst); if True an early abort technique
                       is used and the computation is interrupted as soon
                       as a small divisor of the order is detected.

        Uses SEA (via sea.gp), as implemented in PARI by Christophe
        Doche and Sylvain Duquesne.
        """
        try:
            return self.__cardinality
        except AttributeError:
            pass
        if self.base_ring().degree() == 1:
            N = sea.ellsea(self.a_invariants(), self.base_ring().characteristic(), \
                           early_abort=early_abort)
        else:
            N = len(self.points())
        self.__cardinality = N
        return N

    def abelian_group(self):
        """
        Returns the abelian group structure of the group of points on
        this elliptic curve.

        OUTPUT:
            -- an abelian group
            -- tuple of images of each of the generators of the
               abelian group as points on this curve
        """
        try:
            return self.__abelian_group
        except AttributeError:
            pass
        if self.base_ring().degree() == 1:
            I, G = eval(cremona.ellzp(self.a_invariants(), self.base_ring().characteristic()))
            A = AbelianGroup(I)
            G = tuple(reversed([self(P) for P in G]))
        else:
            raise NotImplementedError
        self.__abelian_group = A, G
        return A, G

    def __getitem__(self, n):
        return self.points()[n]

if __name__ ==  '__main__':
    from sage.rings.all import *
    from constructor import EllipticCurve
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
