"""
Points on elliptic curves
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import ell_generic
import sage.rings.all as rings
import sage.rings.arith as arith
import sage.misc.misc as misc
from sage.modules.module_element import ModuleElement
oo = rings.infinity       # infinity
    
class Point(ModuleElement):
    """
    A point on an elliptic curve.

    EXAMPLES:
       sage: E = EllipticCurve('37a')

    TODO: This is a bug, since [1,0,0] is not on the curve.
       sage: E([1,0,0])  
       0
    """
    def __init__(self, curve, x=oo, y=oo, check=True):
        ModuleElement.__init__(self, curve)
        if not isinstance(curve, ell_generic.EllipticCurve):
            raise TypeError, "Argument 1 must be an elliptic curve."
        self.__curve = curve
        if y is oo and isinstance(x, (int, long, rings.Integer)) and x==0:
            self.__x, self.__y = oo, oo
            return
        K = curve.base_ring()
        if isinstance(x, (list,tuple)):
            if len(x) == 1:
                x, y = x[0], 1
            elif len(x) == 2:
                x, y = x[0], x[1]
            else:
                # TODO: fix:
                # E = EllipticCurve('37a')
                # E([1,0,0])  (but it's not on the curve!)
                # 0
                if x[2] == 0:
                    x = rings.infinity
                else:
                    try:
                        z = K(x[2])
                        x, y = K(x[0])/z, K(x[1])/z
                    except TypeError:
                        raise ArithmeticError, "Point %s is not on curve."%self
        if x is oo:
            self.__x, self.__y = oo, oo
            return
        try:
            x = K(x)
            y = K(y)
        except TypeError:
            raise ArithmeticError, "Point %s is not on curve."%((x,y),)
        self.__x, self.__y = x, y
        if check:
            if y*y + curve.a1()*x*y + curve.a3()*y != \
                   x**3 + curve.a2()*x**2 + curve.a4()*x + curve.a6():
                raise ArithmeticError, "Point %s is not on curve."%self

    def __repr__(self):
        if self.is_infinity():
            return "0"
##         if isinstance(self.base_ring(), rings.RationalField):
##             d = self.__x.denominator().lcm(self.__y.denominator())
##         else:
##             d = self.base_ring()(1)
        return "(%s, %s)"%(self.__x, self.__y)

    def __cmp__(self, other):
        if isinstance(other, (int, long, rings.Integer)) and other == 0:
            if self.is_zero():
                return 0
            else:
                return -1
        if not isinstance(other, Point):
            other = self.curve()(other)
        if self.__x < other.__x:
            return -1
        elif self.__x > other.__x:
            return 1
        elif self.__y < other.__y:
            return -1
        elif self.__y > other.__y:
            return 1
        return 0

    def base_ring(self):
        return self.__curve.base_ring()
    base_field = base_ring

    def order(self):
        """
        Return the order of this point on the elliptic curve.
        If the point has infinite order, returns 0.

        EXAMPLE:
            sage: E = EllipticCurve([0,0,1,-1,0])
            sage: P = E([0,0]); P
            (0, 0)
            sage: P.order()
            0

            sage: E = EllipticCurve([0,1])
            sage: P=E([-1,0])
            sage: P.order()
            2
        """
        if self.is_infinity():
            return 1
        E = self.__curve
        try:
            return int(E.pari_curve().ellorder([self.__x, self.__y]))
        except AttributeError:
            if E._type() == "finite_field":
                print "WARNING -- using naive point order finding over finite field!"
                # TODO: This is very very naive!!  -- should use baby-step giant step; maybe in mwrank
                #      note that this is *not* implemented in PARI!
                P = self
                n = 1
                while not P.is_infinity():
                    n += 1
                    P += self
                return n
            else:
                raise NotImplementedError
        

    def xy(self):
        return self.__x, self.__y

    def x(self):
        return self.__x

    def y(self):
        return self.__y

    def curve(self):
        return self.__curve

    def __getitem__(self, n):
        if n==0:
            return self.__x
        elif n==1:
            return self.__y
        else:
            raise IndexError, "Index must be 0 or 1"

    def is_infinity(self):
        return self.__x is rings.infinity

    def is_zero(self):
        return self.__x is rings.infinity
    
    def __add__(self, right):
        """Add self to right."""
        if self.curve() != right.curve():
            raise ArithmeticError, "Points must be on the same curve."

        # Use Prop 7.1.7 of Cohen "A Course in Computational Algebraic Number Theory"
        if self.is_infinity():
            return right
        if right.is_infinity():
            return self
        a1, a2, a3, a4, a6 = self.__curve.ainvs()
        x1, y1 = self.__x, self.__y
        x2, y2 = right.__x, right.__y
        if x1 == x2 and y1 == -y2 - a1*x2 - a3:
            return Point(self.__curve) # point at infinity
        if x1 == x2 and y1 == y2:
            m = (3*x1*x1 + 2*a2*x1 + a4 - a1*y1) / (2*y1 + a1*x1 + a3)
        else:
            m = (y1-y2)/(x1-x2)
        x3 = -x1 - x2 - a2 + m*(m+a1)
        y3 = -y1 - a3 - a1*x3 + m*(x1-x3)
        return Point(self.__curve, x3, y3, check=False)

    def __sub__(self, right):
        return self + (-right)

    def __neg__(self):
        if self.is_infinity():
            return self
        E, x, y = self.__curve, self.__x, self.__y
        return Point(self.__curve, x, -y - E.a1()*x - E.a3(), check=False)

    def __mul__(self, m):
        return self.__rmul__(m)

    def __rmul__(self, m):
        m = int(m)
        if m<0:
            return (-m)*(-self)
        if m==1:
            return self
        if m==0:
            return Point(self.__curve)
        power = Point(self.__curve)
        i = 0
        apow2 = self
        while ((m>>i) > 0):
            if((m>>i) & 1):
                power += apow2
            apow2 += apow2
            i += 1
        return power

    def height(self):
        """
        The height of the point.
        """
        return self.curve().pari_curve().ellheight([self[0], self[1]])

    def sigma(self, p):
        k = rings.Qp(p)
        if self.is_infinity():
            return k(0)
        sigma = self.__curve.sigma(p)
        return sigma(k(-self.__x/self.__y))


##     def padic_height(self, p):
##         """Returns the p-adic cyclotomic height of the point.

##         padic_height(p)

##         Input:
##            p: a prime number
##         Output:
##            p-adic cyclotomic height, which is a single p-adic number.

##         Algorithm:
##         We compute this height using the following formula, which
##         is valid for points that are in the intersection of the
##         identity component of the Neron model with the kernel of
##         reduction modulo $p$:
## $$
##      h(P) = 1/2 * sum_{ell!=p} sup(0,-ord_ell(x(P)))  + log_p(sigma_p(-x(P)/y(P)) / e),
## $$
##         where $P=(a/e^2, b/e^3)$ with $\gcd(a,e)=1$, and
##         where the first sum is over primes ell that don't equal $p$.
##         If $P$ isn't in the subgroup mentioned above, let $n$ be a positive
##         integer so that $nP$ is in that subgroup.  Then we return
##         $h(nP)/(n^2)$, which does not depend on the choice of $n$, and
##         is defined using the above formula.

##         """

##         if not arith.is_prime(p):
##             raise ArithmeticError, "p must be a prime number."
##         E = self.curve()
##         if E.conductor() % p == 0:
##             raise ArithmeticError, "p must be a prime of good reduction."

##         ap = E.ap(p)
##         if ap == 0:
##             raise ArithmeticError, "p must a prime of good ordinary reduction."
        
##         Np = p+1 - E.ap(p)
##         c = 1
##         for q, _ in arith.factor(E.conductor()):
##             c = arith.lcm(c, E.tamagawa_number(q))

##         n = Np*c
##         misc.verbose("Multiplying point by %s."%n)
##         R = n*self
            
##         x, y = R.xy()
##         sigma = E.padic_sigma(p)
##         Qp = rings.pAdicField(p)
##         print "x = ", x
##         print "-x/y = ", Qp(-x/y)
##         d = x.denominator()
##         e = d.sqrt()
##         if e*e != d:
##             raise ArithmeticError, "The denominator is not a perfect square!?"
##         t = Qp(-x/y)
##         return (sigma(t)/Qp(e)).log()



##         w = Qp(0)
##         for ell, ord in d.factor():
##             if ell != p:
##                 w += ord * Qp(ell).log()
##         # note that there is no 1/2, since we factored the square root.
##         t = Qp(-x/y)
##         print "sigm = ", sigma(t)
##         print "e = ", e
##         hP = w + (sigma(t)/e).log()
##         print "w = ", w
##         print "sigma(t) = ", sigma(t)
##         print "sigma(t)/e = ", sigma(t)/e
##         print "(sigma(t)/e).log() = ", (sigma(t)/e).log()
##         print "sum = ", w + (sigma(t)/e).log()

##         return hP/Qp(n**2)

if __name__ ==  '__main__':
    from sage.rings.all import *
    from constructor import EllipticCurve
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
